#include "global.h"
#include "cpu.h"
#include "crystal.h"
#include "apu.h"
#include "sound.h"
#include "mapper.h"

#include "apu_components.c"

/* (pseudo: it's part of the 2a03) APU

problems (unknown if accurate):
	- Greg Norman's Golf Power: sound buzzes (DMC)
	- Kyoro Chan Land: buzz in the background in-game
	- Mega Man 1: 3rd row of toms at the beginning of Fireman's stage music sound wrongly timed
	- Twin Cobra: sweeps sound wrong (when getting power-ups)
	- intros of Codemasters games have some crackling sound (noise?)
	
	- DMC: nestank: screen shaking sometimes, normal?
	
	- Super Mario Bros. 1: sound buzzes slightly at the start of some songs (main tune, invincible tune, death tune, ?) = normal! (tested on my PAL NES)
*/

#define NO_SHORT_NOISE	FALSE	/* no short looped noise (on 1st revision recalled Famicom) */


static struct {
	int* dac;
	int irqs;
	int dmc_irq_isenabled;
	int frame_irq_isenabled;
	int fs_mode;
	int fs_count;
	int fs_max;
	int mul_frame_irq;
	u8 fs_last;
} apu;

static ac_channel* apu_square0;
static ac_channel* apu_square1;
static ac_channel* apu_triangle;
static ac_channel* apu_noise;
static ac_channel* apu_dmc;


void __fastcall (*mapper_apu_clock)(int);
void apu_set_mapper_apu_clock(fp_mapper_apu_clock fun) { mapper_apu_clock=fun; }

void __fastcall apu_force_update(void)
{
	apu_square0_update();
	apu_square1_update();
	apu_triangle_update();
	apu_noise_update();
	apu_dmc_update();
}

void apu_set_buffer_counters(int buffer_counter)
{
	apu_square0->buffer_counter=apu_square1->buffer_counter=apu_triangle->buffer_counter=apu_noise->buffer_counter=apu_dmc->buffer_counter=buffer_counter;
	mapper_set_sound_buffer_counters(buffer_counter);
}

int apu_get_buffer_dmc_counter(void) { return apu_dmc->buffer_counter; }

void apu_kill_cycle(void)
{
	apu_square0->cycles-=crystal->ppu_cycle;
	apu_square1->cycles-=crystal->ppu_cycle;
	apu_triangle->cycles-=crystal->ppu_cycle;
	apu_noise->cycles-=crystal->ppu_cycle;
	apu_dmc->cycles-=crystal->ppu_cycle;
}


/* squares */
void __fastcall apu_write_square0_control(u8 data) { AC_ENVELOPE_REGISTER(apu_square0); AC_DUTY_REGISTER(apu_square0); } /* $4000 */
void __fastcall apu_write_square1_control(u8 data) { AC_ENVELOPE_REGISTER(apu_square1); AC_DUTY_REGISTER(apu_square1); } /* $4004 */
void __fastcall apu_write_square0_sweep(u8 data) { AC_SWEEP_REGISTER(apu_square0); } /* $4001 */
void __fastcall apu_write_square1_sweep(u8 data) { AC_SWEEP_REGISTER(apu_square1); } /* $4005 */
void __fastcall apu_write_square0_timer(u8 data) { AC_TIMER_LOW_REGISTER(apu_square0); AC_SWEEP_RECALC(apu_square0); } /* $4002 */
void __fastcall apu_write_square1_timer(u8 data) { AC_TIMER_LOW_REGISTER(apu_square1); AC_SWEEP_RECALC(apu_square1); } /* $4006 */
void __fastcall apu_write_square0_length(u8 data) { AC_LENGTH_COUNTER_REGISTER(apu_square0); AC_TIMER_HIGH_REGISTER(apu_square0); AC_DUTY_RESET(apu_square0); AC_SWEEP_RECALC(apu_square0); } /* $4003 */
void __fastcall apu_write_square1_length(u8 data) { AC_LENGTH_COUNTER_REGISTER(apu_square1); AC_TIMER_HIGH_REGISTER(apu_square1); AC_DUTY_RESET(apu_square1); AC_SWEEP_RECALC(apu_square1); } /* $4007 */

#define APU_SQUARE_UPDATE(x) \
	const int cpu_cycles=cpu_get_cycles(); \
	\
	for (;x->cycles>cpu_cycles;x->cycles-=crystal->rel_cycle) { \
		if (!--x->timer) { \
			x->timer=(x->timer_raw+1)<<1; \
			x->sequencer_counter=(x->sequencer_counter+1)&7; \
			x->sample=ac_duty_waveform[x->duty][x->sequencer_counter]; \
		} \
		if ((x->length_counter==0)|x->sweep_issilenced) x->buffer[x->buffer_counter++]=0; \
		else x->buffer[x->buffer_counter++]=x->volume*x->sample; \
	}

void __fastcall apu_square0_update(void) { APU_SQUARE_UPDATE(apu_square0) }
void __fastcall apu_square1_update(void) { APU_SQUARE_UPDATE(apu_square1) }


/* triangle */
void __fastcall apu_write_triangle_linear(u8 data) { AC_LINEAR_COUNTER_REGISTER(apu_triangle); } /* $4008 */
void __fastcall apu_write_triangle_timer(u8 data) { AC_TIMER_LOW_REGISTER(apu_triangle); } /* $400a */
void __fastcall apu_write_triangle_length(u8 data) { AC_LENGTH_COUNTER_REGISTER(apu_triangle); AC_TIMER_HIGH_REGISTER(apu_triangle); apu_triangle->linear_counter_halt=TRUE; } /* $400b */

void __fastcall apu_triangle_update(void)
{
	const int cpu_cycles=cpu_get_cycles();
	
	for (;apu_triangle->cycles>cpu_cycles;apu_triangle->cycles-=crystal->rel_cycle) {
		if (!--apu_triangle->timer) {
			apu_triangle->timer=apu_triangle->timer_raw+1;
			if (apu_triangle->timer_raw>3) {
				if ((apu_triangle->length_counter!=0)&(apu_triangle->linear_counter!=0)) {
					apu_triangle->sample=ac_triangle_waveform[apu_triangle->sequencer_counter++];
					apu_triangle->sequencer_counter&=0x1f;
				}
			}
		}
		apu_triangle->buffer[apu_triangle->buffer_counter++]=apu_triangle->sample;
	}
}


/* noise */
static int ac_noise_timer[0x10]; /* 4 bit lookup */
void apu_fill_noise_frequency_lookup(void)
{
	/* measured by blargg */
	const int ac_noise_timer_lut[2][16]={
		{4,8,16,32,64,96,128,160,202,254,380,508,762,1016,2034,4068},
		{4,7,14,30,60,88,118,148,188,236,354,472,708, 944,1890,3778}
	};
	
	int i;
	for (i=0;i<16;i++) ac_noise_timer[i]=ac_noise_timer_lut[crystal->mode][i];
}
void __fastcall apu_write_noise_envelope(u8 data) { AC_ENVELOPE_REGISTER(apu_noise); } /* $400c */
void __fastcall apu_write_noise_timer(u8 data) /* $400e */
{
	AC_TIMER_NOISE_REGISTER(apu_noise);
#if NO_SHORT_NOISE
	apu_noise->loop=1;
#endif
}
void __fastcall apu_write_noise_length(u8 data) { AC_LENGTH_COUNTER_REGISTER(apu_noise); } /* $400f */

void __fastcall apu_noise_update(void)
{
	const int cpu_cycles=cpu_get_cycles();
	
	for (;apu_noise->cycles>cpu_cycles;apu_noise->cycles-=crystal->rel_cycle) {
		if (!--apu_noise->timer) {
			apu_noise->timer=ac_noise_timer[apu_noise->timer_raw];
			apu_noise->random=apu_noise->random|((apu_noise->random&1)^(apu_noise->random>>apu_noise->loop&1))<<15;
			apu_noise->random>>=1;
			apu_noise->sample=apu_noise->random&1;
		}
		if (apu_noise->sample&(apu_noise->length_counter!=0)) apu_noise->buffer[apu_noise->buffer_counter++]=apu_noise->volume;
		else apu_noise->buffer[apu_noise->buffer_counter++]=0;
	}
}


/* DMC, different from the other channels */
static int apu_dmc_update_frequency_lookup[16];
void apu_fill_dmc_update_frequency_lookup(void)
{
	/* measured by blargg */
	const int dmc_update_frequency[2][16]={
		{428,380,340,320,286,254,226,214,190,160,142,128,106, 84, 72, 54},
		{398,354,316,298,276,236,210,198,176,148,132,118, 98, 78, 66, 50}
	};
	
	int i;
	for (i=0;i<16;i++) apu_dmc_update_frequency_lookup[i]=dmc_update_frequency[crystal->mode][i]*crystal->rel_cycle;
}

#define APU_DMC_REFILL() \
	if (apu_dmc->timer) { \
		apu_dmc->sequencer_isempty=FALSE; \
		apu_dmc->sequencer=cpu_dmc_dma_transfer((apu_dmc->address_counter&0x7fff)|0x8000); \
		apu_dmc->address_counter++; \
		if (!--apu_dmc->timer) { \
			if (apu_dmc->loop) { apu_dmc->timer=apu_dmc->timer_raw; apu_dmc->address_counter=apu_dmc->address; } \
			else if (apu.dmc_irq_isenabled) { apu.irqs|=BIT(7); cpu_set_interrupt(INTERRUPT_IRQ|IRQ_DMC|(cpu_test_apu_dmc_next_cycle()?0:INT_EDGE_IRQ_EX)); } /* works either way: is always edge when writing status register */ \
		} \
	}

/* $4010, DMC control
		7	enable DMC IRQs
		6	enable looping
		3,2,1,0	update frequency */
void __fastcall apu_write_dmc_control(u8 data)
{
	apu.dmc_irq_isenabled=data&BIT(7);
	if (apu.dmc_irq_isenabled==0) { apu.irqs&=~BIT(7); cpu_acknowledge_interrupt(~IRQ_DMC); }
	apu_dmc->loop=data&BIT(6);
	cpu_reset_apu_dmc_freq(apu_dmc_update_frequency_lookup[data&BIN8(00001111)]);
}

void __fastcall apu_write_dmc_dac(u8 data) { apu_dmc_update(); apu_dmc->sample=data&~BIT(7); } /* $4011, (force) 7 bit pcm */
void __fastcall apu_write_dmc_address(u8 data) { apu_dmc->address=data<<6|0xc000; } /* $4012, address */
void __fastcall apu_write_dmc_length(u8 data) { apu_dmc->timer_raw=data<<4|1; } /* $4013, length */

void __fastcall apu_run_dmc(void)
{
dmc_redo:
	if (apu_dmc->sequencer_counter) {
		
		if (!apu_dmc->sequencer_isempty) {
			if (apu_dmc->sequencer&1) {
				if (apu_dmc->sample<126) {
					apu_dmc_update();
					apu_dmc->sample++; apu_dmc->sample++;
				}
			}
			else {
				if (apu_dmc->sample>1) {
					apu_dmc_update();
					apu_dmc->sample--; apu_dmc->sample--;
				}
			}
			apu_dmc->sequencer>>=1;
		}
		
		if (!--apu_dmc->sequencer_counter) {
			apu_dmc->sequencer_counter=8;
			apu_dmc->sequencer_isempty=TRUE;
			APU_DMC_REFILL()
		}
	}
	if (cpu_set_apu_dmc_next_step()) goto dmc_redo;
}

void __fastcall apu_dmc_update(void)
{
	const int cpu_cycles=cpu_get_cycles();
	
	for (;apu_dmc->cycles>cpu_cycles;apu_dmc->cycles-=crystal->rel_cycle)
		apu_dmc->buffer[apu_dmc->buffer_counter++]=apu_dmc->sample;
}



/* status */
/* $4015, status register
read	7	DMC IRQ
		6	frame IRQ
		5	-
		4	DMC bytes remaining is > 0
		3	noise length counter is > 0
		2	triangle length counter is > 0
		1	square 1 length counter is > 0
		0	square 0 length counter is > 0 */
u8 __fastcall apu_read_status(u16 address)
{
	u8 status=apu.irqs|(apu_square0->length_counter!=0)|(apu_square1->length_counter!=0)<<1|(apu_triangle->length_counter!=0)<<2|(apu_noise->length_counter!=0)<<3|(apu_dmc->timer!=0)<<4;
	
	apu.irqs&=~BIT(6);
	cpu_acknowledge_interrupt(~IRQ_FRAME);
	
	return status;
}
/*write	4	enable DMC
		3	enable noise length counter
		2	enable triangle length counter
		1	enable square 1 length counter
		0	enable square 0 length counter */
void __fastcall apu_write_status(u8 data)
{
	AC_LENGTH_COUNTER_ENABLE_REGISTER(apu_square0,0) AC_LENGTH_COUNTER_ENABLE_REGISTER(apu_square1,1) AC_LENGTH_COUNTER_ENABLE_REGISTER(apu_triangle,2) AC_LENGTH_COUNTER_ENABLE_REGISTER(apu_noise,3)
	
	apu.irqs&=~BIT(7);
	cpu_acknowledge_interrupt(~IRQ_DMC);
	
	if (data&BIT(4)) {
		if (apu_dmc->timer==0) {
			apu_dmc->timer=apu_dmc->timer_raw; apu_dmc->address_counter=apu_dmc->address;
			if (apu_dmc->sequencer_isempty) APU_DMC_REFILL()
		}
	}
	else apu_dmc->timer=0;
}


/* frame sequencer */
static int apu_fs_update_frequency_lookup[2][5];
void apu_fill_fs_update_frequency_lookup(void)
{
	/* measured by blargg */
	const int fs_update_frequency[2][2][5]={
		{ {7456,7458,7457,7457,7458},{7458,7456,7458,7456,7454} },
		{ {8314,8312,8313,8313,8314},{8314,8314,8312,8314,8312} }
	};
	
	int i,j;
	for (i=0;i<2;i++) for (j=0;j<5;j++) apu_fs_update_frequency_lookup[i][j]=fs_update_frequency[crystal->mode][i][j]*crystal->rel_cycle;
}

/* $4017, frame sequencer
		7	mode (4 step:0 or 5 step:1)
		6	disable frame irqs */
void __fastcall apu_write_fs(u8 data)
{
	apu.mul_frame_irq=0;
	apu.fs_last=data;
	apu.frame_irq_isenabled=(data&BIN8(11000000))==0;
	if (data&BIT(6)) { apu.irqs&=~BIT(6); cpu_acknowledge_interrupt(~IRQ_FRAME); }
	
	apu.fs_mode=data>>7;
	apu.fs_max=apu.fs_mode+3; /* 3 or 4 */
	apu.fs_count=apu.fs_max<<1&4; /* 3?:4 4?:0 */
	cpu_reset_apu_fs_next_step();
}
void apu_rewrite_fs(void)
{
	apu_write_fs(apu.fs_last);
	
	/* not written exactly before the reset interrupt, it seems. affects blargg apu test 09.reset_timing, Mezase Pachi Pro - Pachio Kun (lockup after you wait a while) (both currently OK) */
	cpu_set_apu_fs_next_step(-(crystal->cycle*4));
}

void __fastcall apu_run_fs(void)
{
	if (apu.fs_count<4) {
		int apu_odd_clock=(apu.fs_count+apu.fs_mode)&1;
		
		#define _FS_IRQ(x) \
			if (apu.frame_irq_isenabled) { \
				apu.irqs|=BIT(6); \
				cpu_set_interrupt(INTERRUPT_IRQ|IRQ_FRAME|x); \
			}
		
		#define _FS_CLOCK() \
			(*mapper_apu_clock)(apu_odd_clock); \
			\
			/* clock envelopes and triangle linear counter */ \
			AC_ENVELOPE_CLOCK(apu_square0); \
			AC_ENVELOPE_CLOCK(apu_square1); \
			AC_ENVELOPE_CLOCK(apu_noise); \
			AC_LINEAR_COUNTER_CLOCK(apu_triangle); \
			\
			if (apu_odd_clock) { \
				/* clock length counters and sweeps */ \
				AC_LENGTH_COUNTER_CLOCK(apu_square0); \
				AC_LENGTH_COUNTER_CLOCK(apu_square1); \
				AC_LENGTH_COUNTER_CLOCK(apu_triangle); \
				AC_LENGTH_COUNTER_CLOCK(apu_noise); \
				AC_SWEEP_CLOCK(apu_square0); AC_SWEEP_CLOCK(apu_square1); \
			}
		
		#define _FS_DONE() \
			cpu_set_apu_fs_next_step(apu_fs_update_frequency_lookup[apu.fs_mode][apu.fs_count++]); \
			if (apu.fs_count>apu.fs_max) apu.fs_count=0
		
		/* 3 irqs in a row */
		if ((apu.fs_mode==0)&(apu.fs_count==3))
			for (;;) {
				switch (++apu.mul_frame_irq) {
					case 2: _FS_CLOCK() break;
					case 3:
						if (cpu_test_apu_fs_next_step(crystal->rel_cycle)) { _FS_IRQ(0) }
						else { _FS_IRQ(INT_EDGE_IRQ_EX) }
						_FS_DONE(); return;
					default: break;
				}
				
				if (cpu_set_apu_fs_next_step(crystal->rel_cycle)) { _FS_IRQ(0) }
				else { _FS_IRQ(INT_EDGE_IRQ_EX) return; }
			}
		
		/* normal clocking */
		_FS_CLOCK()
	}
	
	_FS_DONE();
}



void apu_new_frame(void)
{
	apu_square0->cycles=apu_square1->cycles=apu_triangle->cycles=apu_noise->cycles=apu_dmc->cycles=cpu_get_cycles();
	apu.mul_frame_irq=0;
}


void apu_end_frame(void)
{
	int i;
	apu_force_update();
	
	/* fill dac */
	i=apu_dmc->buffer_counter;
	while (i--)
		apu.dac[i]=ac_square_volume_lookup[apu_square0->buffer[i]+apu_square1->buffer[i]]+ac_tnd_volume_lookup[3*apu_triangle->buffer[i]+(apu_noise->buffer[i]<<1)+apu_dmc->buffer[i]];
}

void apu_init(void)
{
	ac_fill_volume_lookup();
	
	memset(&apu,0,sizeof(apu));
	
	AC_CREATE_CHANNEL(apu_square0);
	AC_CREATE_CHANNEL(apu_square1);
	AC_CREATE_CHANNEL(apu_triangle);
	AC_CREATE_CHANNEL(apu_noise);
	AC_CREATE_CHANNEL(apu_dmc);
	
	apu_square0->buffer=sound_create_channel();
	apu_square1->buffer=sound_create_channel();
	apu_triangle->buffer=sound_create_channel();
	apu_noise->buffer=sound_create_channel();
	apu_dmc->buffer=sound_create_channel();
	
	apu_fill_noise_frequency_lookup();
	
	apu_square0->timer=apu_square1->timer=2;
	apu_square0->sweep_sub=1;
	apu_triangle->timer=1;
	apu_triangle->sequencer_counter=1;
	apu_triangle->sample=ac_triangle_waveform[apu_triangle->sequencer_counter];
	apu_noise->timer=ac_noise_timer[0];
	apu_noise->loop=1;
	apu_dmc->sequencer_counter=8;
	apu_dmc->timer_raw=1;
	
	apu.dac=sound_create_dac();
	
	apu_fill_fs_update_frequency_lookup();
	apu_fill_dmc_update_frequency_lookup();
	apu_write_dmc_control(0xf);
	
	apu_write_status(0);
	
	apu_write_fs(0);
/*	confirmed behaviour: 0
	
	problems with 0:
	- Colour 2001 Streetfighter II (Unl): won't work
	- Stars SE: locks up at boot (fix: change CLI to SEI at init: iNES file offset $17: change $58 to $78)
	
	problems with $40:
	dunno
*/
	LOG(LOG_VERBOSE,"APU initialised\n");
}

void apu_clean(void)
{
	memset(&apu,0,sizeof(apu));
	
	AC_CLEAN_CHANNEL(apu_square0);
	AC_CLEAN_CHANNEL(apu_square1);
	AC_CLEAN_CHANNEL(apu_triangle);
	AC_CLEAN_CHANNEL(apu_noise);
	AC_CLEAN_CHANNEL(apu_dmc);
	
	LOG(LOG_VERBOSE,"APU cleaned\n");
}
