/*
	common NES APU components, most are used in more than 1 channel
	
	Famicom MMC5 makes use of some too
	
	
	x=struct/function prefix
*/

#include "global.h"


#define HACK_TRIANGLE_SAWTOOTH		FALSE	/* convert triangle wave into a sawtooth wave */


typedef struct {
	int cycles;
	int* buffer;
	int buffer_counter;
	int sample;
	
	int duty;				/* squares only */
	int sequencer_counter;
	int sequencer;			/* dmc only */
	int sequencer_isempty;	/* dmc only */
	
	int timer;
	int timer_raw;
	
	int length_counter_isenabled;
	int length_counter_halt;
	int length_counter;
	
	int envelope;
	int envelope_loop;
	int envelope_start;
	int envelope_constant;
	int envelope_divider;
	int envelope_counter;
	int volume;
	
	int linear_counter;		/* all linear: triangle only */
	int linear_counter_period;
	int linear_counter_halt;
	
	int loop;
	
	int random;				/* noise only */
	
	int sweep_start;		/* all sweeps: squares only */
	int sweep_divider;
	int sweep_period;
	int sweep_isenabled;
	int sweep_negate;
	int sweep_shift;
	int sweep_issilenced;
	int sweep_sub;
	u16 sweep_shiftresult;
	
	u16 address;			/* dmc only */
	u16 address_counter;	/* dmc only */
} ac_channel;


#define AC_CREATE_CHANNEL(x) \
	MEM_CREATE(x,sizeof(ac_channel)); \
	/* set default values */ \
	x->sequencer_isempty=TRUE; \
	x->address=0xc000; \
	x->envelope_divider=1; \
	x->sweep_divider=1; \
	x->sweep_period=1; \
	x->random=1
#define AC_CLEAN_CHANNEL(x) MEM_CLEAN(x)


/********/

/* length counter */
static const int ac_length_counter[0x20]={10,254,20,2,40,4,80,6,160,8,60,10,14,12,26,14,12,16,24,18,48,20,96,22,192,24,72,26,16,28,32,30};

/* register:
	square0:	$4003
	square1:	$4007
	triangle:	$400b
	noise:		$400f
	
	bits:		7-3
	
	also starts envelope (if available) */
#define AC_LENGTH_COUNTER_REGISTER(x) \
	x##_update(); \
	if (cpu_update_apu_fs_onclock()) { if (x->length_counter_isenabled&&x->length_counter==0) x->length_counter=ac_length_counter[data>>3]; } \
	else { if (x->length_counter_isenabled) x->length_counter=ac_length_counter[data>>3]; } \
	x->envelope_start=TRUE
/*	all:		$4015 (see underneath at statusregister) */
#define AC_LENGTH_COUNTER_ENABLE_REGISTER(x,y) x->length_counter_isenabled=data>>y&1; if (!x->length_counter_isenabled) { x##_update(); x->length_counter=0; }

/* clock */
#define AC_LENGTH_COUNTER_CLOCK(x) \
	/* update has already been done */ \
	x->length_counter-=((x->length_counter_halt==0)&(x->length_counter!=0))



/* linear counter */
/* register:
	triangle:	$4008
	
	bits:
	7			control/length counter halt
	6-0			period */
#define AC_LINEAR_COUNTER_REGISTER(x) \
	x##_update(); \
	cpu_update_apu_fs_onclock(); \
	x->loop=x->length_counter_halt=data>>7; \
	x->linear_counter_period=data&~BIT(7)

/* clock */
#define AC_LINEAR_COUNTER_CLOCK(x) \
	x##_update(); \
	if (x->linear_counter_halt) x->linear_counter=x->linear_counter_period; \
	else x->linear_counter-=(x->linear_counter!=0); \
	if (!x->loop) x->linear_counter_halt=FALSE



/* timer */
/* 11 bit */
/* register:	low		high
	square0:	$4002	$4003
	square1:	$4006	$4007
	triangle:	$400a	$400b
	
	bits:		all	0,1,2 */
#define AC_TIMER_LOW_REGISTER(x) x##_update(); x->timer_raw=(x->timer_raw&0xff00)|data
#define AC_TIMER_HIGH_REGISTER(x) \
	/* update is done in length counter register */ \
	x->timer_raw=(x->timer_raw&0xff)|(data<<8&0x700)

/* register:
	noise:		$400e
	
	bits:
	7			loop
	6-0			period */
#define AC_TIMER_NOISE_REGISTER(x) x##_update(); \
	x->timer_raw=data&0xf; \
	if (data&BIT(7)) x->loop=6; \
	else x->loop=1



/* envelope */
#define AC_ENVELOPE_UPDATE_VOLUME(x) if (x->envelope_constant==0) x->volume=x->envelope_counter

/* register:
	square0:	$4000
	square1:	$4004
	noise:		$400c
	
	bits:
	5			envelope loop/length counter halt
	4			constant volume
	3,2,1,0		envelope volume/period */
#define AC_ENVELOPE_REGISTER(x) \
	x##_update(); \
	cpu_update_apu_fs_onclock(); \
	x->envelope_loop=x->length_counter_halt=data&BIT(5); \
	x->envelope_constant=data&BIT(4); \
	x->envelope=data&0xf; \
	AC_ENVELOPE_UPDATE_VOLUME(x); \
	else x->volume=x->envelope

/* clock */
#define AC_ENVELOPE_CLOCK(x) \
	x##_update(); \
	if (x->envelope_start) { \
		x->envelope_start=FALSE; \
		x->envelope_divider=x->envelope+1; \
		x->envelope_counter=0xf; \
		AC_ENVELOPE_UPDATE_VOLUME(x); \
	} \
	else { \
		if (!--x->envelope_divider) { \
			x->envelope_divider=x->envelope+1; \
			if (x->envelope_counter) x->envelope_counter--; \
			else if (x->envelope_loop) x->envelope_counter=0xf; \
			AC_ENVELOPE_UPDATE_VOLUME(x); \
		} \
	}



/* duty/waveform */

#if HACK_TRIANGLE_SAWTOOTH
static const int ac_triangle_waveform[0x20]={15,15,14,14,13,13,12,12,11,11,10,10,9,9,8,8,7,7,6,6,5,5,4,4,3,3,2,2,1,1,0,0}; /* saw */
#else
static const int ac_triangle_waveform[0x20]={15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15}; /* triangle */
#endif
static const int ac_duty_waveform[4][8]={
	{ 0,1,0,0,0,0,0,0 },	/* _-______ */
	{ 0,1,1,0,0,0,0,0 },	/* _--_____ */
	{ 0,1,1,1,1,0,0,0 },	/* _----___ */
	{ 1,0,0,1,1,1,1,1 }		/* -__----- */
};
#define AC_DUTY_RESET(x) x->sequencer_counter=0

/* register:
	square0:	$4000
	square1:	$4004
	
	bits:		7,6 */
#define AC_DUTY_REGISTER(x) \
	/* update is done in envelope register */ \
	x->duty=data>>6;



/* sweep */
#define AC_SWEEP_RECALC(x) \
	x->sweep_shiftresult=x->timer_raw>>x->sweep_shift; \
	if (x->sweep_negate) { x->sweep_shiftresult=x->timer_raw-x->sweep_shiftresult-x->sweep_sub; x->sweep_issilenced=x->timer_raw<8; } \
	else { x->sweep_shiftresult+=x->timer_raw; x->sweep_issilenced=(x->sweep_shiftresult>0x7ff)|(x->timer_raw<8); }

/* register:
	square0:	$4001
	square1:	$4005
	
	bits:
	7			enable
	6,5,4		period
	3			negate
	2,1,0		shift */
#define AC_SWEEP_REGISTER(x) \
	x##_update(); \
	x->sweep_start=TRUE; \
	x->sweep_period=(data>>4&7)+1; \
	x->sweep_negate=data&BIT(3); \
	x->sweep_shift=data&7; \
	x->sweep_isenabled=data>>7&(x->sweep_shift!=0); \
	AC_SWEEP_RECALC(x)

/* clock */
#define AC_SWEEP_CLOCK(x) \
	if (!--x->sweep_divider) { \
		x->sweep_divider=x->sweep_period; \
		if (~x->sweep_issilenced&x->sweep_isenabled) { \
			x->timer_raw=x->sweep_shiftresult; \
			AC_SWEEP_RECALC(x); \
		} \
	} \
	if (x->sweep_start) { x->sweep_start=FALSE; x->sweep_divider=x->sweep_period; }



/* volume lookups */
static int ac_square_volume_lookup[31];
static int ac_tnd_volume_lookup[203];

static void ac_fill_volume_lookup(void)
{
	int i;
	
	/* fill volume lookup tables, estimated by blargg */
	for (i=0;i<31;i++) ac_square_volume_lookup[i]=(95.52/(8128.0/i+100))*15000;
	for (i=0;i<203;i++) ac_tnd_volume_lookup[i]=(163.67/(24329.0/i+100))*15000;
}
