#include "global.h"
#include "crystal.h"
#include "cpu_memorymap.h"
#include "cpu.h"
#include "ppu.h"
#include "mapper.h"
#include "cartridge.h"
#include "apu.h"
#include "draw.h"

/*
NES RP2a03x (or 2a07 for PAL?) (minus sound, sound in apu*) = 6502 minus decimal maths

Games/programs highly sensitive to CPU timing (NMI, IRQ, cycle deduction), these fall into the category 'fix something --> break something else':
currently working correctly (so the described problems do not happen) (a lot more, but these are the ones I use(d) for testing):
	- Battletoads: lock up at level 2
	- Battletoads & Double Dragon: lock up at level 1
	- Cobra Triangle: lock up at or before the start of gameplay
	- Downtown Special (Kunio-Kun): garbage/flickering scanline above the statusbar
	- Fire Hawk: lock up in-game, or wrong cutoff anywhere
	- Gun Sight/Laser Invasion: screen shaking anywhere
	- Ironsword (W&W2): lock up at titlescreen (related to frame irq/vblank)
	- Kamen Rider Club: lock up at boot (related to cold boot $2006 disabled)
	- Kick Master: screen shaking at title, and at in-game statusbar
	- Marble Madness: garbage to the left of the textbox at the start of each level
	- nestank: lock up or wrong cutoff
	- Pin Bot: screen shaking after you die
	- Time Zone: screen shaking in the 3rd level (or 1st demo)
	- Uchuu Keibitai SDF: flickering or garbage in the intro
	- Wizards & Warriors 3: garbage in the intro, after the ghost shoots the wizard
	
	- test roms by blargg (lots)
	- pputime2 (by Quietust/Kevin Horton): block moves left or right
	- read2004 (by Quietust): wrong values
	- scanline (2 versions, by Quietust): wrong cutoff
currently not working correctly:
	- nothing that I know of

Other problems:
	- Dragon Fighter: demo mode messes up (normal?)
	- Metal Fighter (Wei Lai Xiao Zi) (Joy Van): screen shaking on the title screen (normal/bad dump? USA version works)
	- (some bad.. hacks/old PD stuff/translated roms/dumps, but that's normal, they wouldn't work on a real NES either), eg:
		- Halloween Demo (PD): logo jumps when blood scrolls down
		- Mouser (Beta 1.0) (PD): garbage (a1: music, black screen)
		- PasoFami Demo 2 (TwinBee) (PD): garbage/black (but is 'playable')
		- Pong by Paul Talbot (PD): locks up


cpu bugs that are in the real 6502 (all 'working'):
	- Return address pushed on the stack by JSR is one less than actual next instruction. RTS increments PC after popping. RTI doesn't.
	- The status bits pushed on the stack by PHP have the breakpoint bit set.
	- The D flag is not defined after reset.
	- The D flag is not cleared by interrupts.
	- An indirect JMP (xxFF) will fail because the MSB will be fetched from address xx00 instead of page xx+1.
	- If an interrupt occurs on a BRK instruction, the breakpoint is ignored.
*/

#define FLAG_B 0x10

static struct {
	/* p*=padding for 32-bit alignment */
	
	u8 A,p0[3];		/* accumulator */
	u8 X,p1[3];		/* index register x */
	u8 Y,p2[3];		/* index register y */
	u16 PC,p3[1];	/* program counter (HIGHBYTE PCH+LOWBYTE PCL) */
	u8 S,p4[3];		/* stack pointer ([(u16)HIGHBYTE 1+]LOWBYTE S) */
	
	/* status register */
	u16 P_NZ,p5[1];	/* negative flag: lowbyte=$80 or highbyte=$80. zero flag: lowbyte=0 */
	u8 P_V,p6[3];	/* overflow flag: x or 0 */
	/*B*/			/* brk flag: technically inexistant, the only way you're able to read/write it yourself, is when it's on the stack */
	u8 P_D,p7[3];	/* decimal flag: bit: unused in maths on the NES */
	u8 P_I,p8[3];	/* irq flag: bit */
	u16 P_C,p9[1];	/* carry flag: highbyte=1, don't allow bits above */
	
	int irq_pend;
	int irq_type;
	int irq_edge;
	int interrupt;
	int nmi;
	
	int halt;
	int illegal_count;
	int cycles;
	int cycles_start;
	u32 cycles_full;
	int apu_fs_next_step;
	int apu_dmc_next_step;
	int apu_dmc_frequency;
	
	int sprite_dma_address;
	int sprite_dma_count;
	int sprite_dma_nf;
} cpu;

static __inline__ void update_sync(void);
void illegal_nothing(u8);
void illegal_show(u8);
void (*_illegal)(u8); /* fp, or compiler will inline */
void __fastcall (*mapper_irq)(void);

/* opcode mnemonics */
static const char* c_op_mnemonic[]= { "BRK","ORA","HLT","SLO","NOP","ASL","PHP","AAC","BPL","CLC","JSR","AND","RLA","BIT","ROL","PLP","BMI","SEC","RTI","EOR","SRE","LSR","PHA","ASR","JMP","BVC","CLI","RTS","ADC","RRA","ROR","PLA","ARR","BVS","SEI","STA","AAX","STY","STX","DEY","TXA","XAA","BCC","AXA","TYA","TXS","XAS","SYA","SXA","LDY","LDA","LDX","LAX","TAY","TAX","ATX","BCS","CLV","TSX","LAR","CPY","CMP","DCP","DEC","INY","DEX","AXS","BNE","CLD","CPX","SBC","ISC","INC","INX","BEQ","SED" };
enum {                               _BRK=0,_ORA, _HLT, _SLO, _NOP, _ASL, _PHP, _AAC, _BPL, _CLC, _JSR, _AND, _RLA, _BIT, _ROL, _PLP, _BMI, _SEC, _RTI, _EOR, _SRE, _LSR, _PHA, _ASR, _JMP, _BVC, _CLI, _RTS, _ADC, _RRA, _ROR, _PLA, _ARR, _BVS, _SEI, _STA, _AAX, _STY, _STX, _DEY, _TXA, _XAA, _BCC, _AXA, _TYA, _TXS, _XAS, _SYA, _SXA, _LDY, _LDA, _LDX, _LAX, _TAY, _TAX, _ATX, _BCS, _CLV, _TSX, _LAR, _CPY, _CMP, _DCP, _DEC, _INY, _DEX, _AXS, _BNE, _CLD, _CPX, _SBC, _ISC, _INC, _INX, _BEQ, _SED  }; /* for readability */
static const u8 op_mnemonic[0x100]=
{ /*  0     1     2     3     4     5     6     7     8     9     A     B     C     D     E     F
0 */ _BRK, _ORA, _HLT, _SLO, _NOP, _ORA, _ASL, _SLO, _PHP, _ORA, _ASL, _AAC, _NOP, _ORA, _ASL, _SLO,  /* 0
1 */ _BPL, _ORA, _HLT, _SLO, _NOP, _ORA, _ASL, _SLO, _CLC, _ORA, _NOP, _SLO, _NOP, _ORA, _ASL, _SLO,  /* 1
2 */ _JSR, _AND, _HLT, _RLA, _BIT, _AND, _ROL, _RLA, _PLP, _AND, _ROL, _AAC, _BIT, _AND, _ROL, _RLA,  /* 2
3 */ _BMI, _AND, _HLT, _RLA, _NOP, _AND, _ROL, _RLA, _SEC, _AND, _NOP, _RLA, _NOP, _AND, _ROL, _RLA,  /* 3
4 */ _RTI, _EOR, _HLT, _SRE, _NOP, _EOR, _LSR, _SRE, _PHA, _EOR, _LSR, _ASR, _JMP, _EOR, _LSR, _SRE,  /* 4
5 */ _BVC, _EOR, _HLT, _SRE, _NOP, _EOR, _LSR, _SRE, _CLI, _EOR, _NOP, _SRE, _NOP, _EOR, _LSR, _SRE,  /* 5
6 */ _RTS, _ADC, _HLT, _RRA, _NOP, _ADC, _ROR, _RRA, _PLA, _ADC, _ROR, _ARR, _JMP, _ADC, _ROR, _RRA,  /* 6
7 */ _BVS, _ADC, _HLT, _RRA, _NOP, _ADC, _ROR, _RRA, _SEI, _ADC, _NOP, _RRA, _NOP, _ADC, _ROR, _RRA,  /* 7
8 */ _NOP, _STA, _NOP, _AAX, _STY, _STA, _STX, _AAX, _DEY, _NOP, _TXA, _XAA, _STY, _STA, _STX, _AAX,  /* 8
9 */ _BCC, _STA, _HLT, _AXA, _STY, _STA, _STX, _AAX, _TYA, _STA, _TXS, _XAS, _SYA, _STA, _SXA, _AXA,  /* 9
A */ _LDY, _LDA, _LDX, _LAX, _LDY, _LDA, _LDX, _LAX, _TAY, _LDA, _TAX, _ATX, _LDY, _LDA, _LDX, _LAX,  /* A
B */ _BCS, _LDA, _HLT, _LAX, _LDY, _LDA, _LDX, _LAX, _CLV, _LDA, _TSX, _LAR, _LDY, _LDA, _LDX, _LAX,  /* B
C */ _CPY, _CMP, _NOP, _DCP, _CPY, _CMP, _DEC, _DCP, _INY, _CMP, _DEX, _AXS, _CPY, _CMP, _DEC, _DCP,  /* C
D */ _BNE, _CMP, _HLT, _DCP, _NOP, _CMP, _DEC, _DCP, _CLD, _CMP, _NOP, _DCP, _NOP, _CMP, _DEC, _DCP,  /* D
E */ _CPX, _SBC, _NOP, _ISC, _CPX, _SBC, _INC, _ISC, _INX, _SBC, _NOP, _SBC, _CPX, _SBC, _INC, _ISC,  /* E
F */ _BEQ, _SBC, _HLT, _ISC, _NOP, _SBC, _INC, _ISC, _SED, _SBC, _NOP, _ISC, _NOP, _SBC, _INC, _ISC   /* F
      0     1     2     3     4     5     6     7     8     9     A     B     C     D     E     F     */
};

/* opcode addressing modes */
#if DEBUG_CPU
static const char* addressing_mode_names[]= { "Accumulator", "Immediate", "Absolute", "Zero Page", "Indexed Zero Page (X)", "Indexed Zero Page (Y)", "Indexed Absolute (X)", "Indexed Absolute (Y)", "Implied", "Relative", "Indexed Indirect (X)", "Indirectly Indexed (Y)", "Indirect" };
#endif
static const char* addressing_mode_names_short[]={
		"Acc", "Imm", "Abs", "ZP", "ZP,X", "ZP,Y", "Abs,X", "Abs,Y", "Imp", "Rel", "Ind,X", "Ind,Y", "Ind" };
enum {  _ACC=0,_IMM,  _ABS,  _ZP,  _ZPX,   _ZPY,   _ABSX,   _ABSY,   _IMP,  _REL,  _INDX,   _INDY,   _INDR }; /* for readability */
static const u8 op_addressing_mode[0x100]=
{ /*  0     1     2     3     4     5     6     7     8     9     A     B     C     D     E     F
0 */ _IMP, _INDX,_IMP, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _ACC, _IMM, _ABS, _ABS, _ABS, _ABS,  /* 0
1 */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPX, _ZPX, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSX,_ABSX, /* 1
2 */ _ABS, _INDX,_IMP, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _ACC, _IMM, _ABS, _ABS, _ABS, _ABS,  /* 2
3 */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPX, _ZPX, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSX,_ABSX, /* 3
4 */ _IMP, _INDX,_IMP, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _ACC, _IMM, _ABS, _ABS, _ABS, _ABS,  /* 4
5 */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPX, _ZPX, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSX,_ABSX, /* 5
6 */ _IMP, _INDX,_IMP, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _ACC, _IMM, _INDR,_ABS, _ABS, _ABS,  /* 6
7 */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPX, _ZPX, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSX,_ABSX, /* 7
8 */ _IMM, _INDX,_IMM, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _IMP, _IMM, _ABS, _ABS, _ABS, _ABS,  /* 8
9 */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPY, _ZPY, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSY,_ABSY, /* 9
A */ _IMM, _INDX,_IMM, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _IMP, _IMM, _ABS, _ABS, _ABS, _ABS,  /* A
B */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPY, _ZPY, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSY,_ABSY, /* B
C */ _IMM, _INDX,_IMM, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _IMP, _IMM, _ABS, _ABS, _ABS, _ABS,  /* C
D */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPX, _ZPX, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSX,_ABSX, /* D
E */ _IMM, _INDX,_IMM, _INDX,_ZP,  _ZP,  _ZP,  _ZP,  _IMP, _IMM, _IMP, _IMM, _ABS, _ABS, _ABS, _ABS,  /* E
F */ _REL, _INDY,_IMP, _INDY,_ZPX, _ZPX, _ZPX, _ZPX, _IMP, _ABSY,_IMP, _ABSY,_ABSX,_ABSX,_ABSX,_ABSX  /* F
      0     1     2     3     4     5     6     7     8     9     A     B     C     D     E     F     */
};

/* true(1): official/documented, false: illegal/undocumented */
static const u8 op_official[0x100]=
{ /* 0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
0 */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 0, /* 0
1 */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, /* 1
2 */ 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, /* 2
3 */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, /* 3
4 */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, /* 4
5 */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, /* 5
6 */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, /* 6
7 */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, /* 7
8 */ 0, 1, 0, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, /* 8
9 */ 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, /* 9
A */ 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, /* A
B */ 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, /* B
C */ 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, /* C
D */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, /* D
E */ 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, /* E
F */ 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0  /* F
     0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  */
};

/* number of cycles to burn for each opcode */
static int op_cycles_lookup[0x100]; /* opcode cycles multiplied by ntsc/pal cycle */
static const u8 op_cycles[0x100]=
/* 00-BRK(0) is 7, cycles are deducted in the interrupt handler instead */
{ /* 0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
0 */ 0, 6, 0, 8, 3, 3, 5, 5, 3, 2, 2, 2, 4, 4, 6, 6, /* 0
1 */ 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 2, 7, 4, 4, 7, 7, /* 1
2 */ 6, 6, 0, 8, 3, 3, 5, 5, 4, 2, 2, 2, 4, 4, 6, 6, /* 2
3 */ 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 2, 7, 4, 4, 7, 7, /* 3
4 */ 6, 6, 0, 8, 3, 3, 5, 5, 3, 2, 2, 2, 3, 4, 6, 6, /* 4
5 */ 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 2, 7, 4, 4, 7, 7, /* 5
6 */ 6, 6, 0, 8, 3, 3, 5, 5, 4, 2, 2, 2, 5, 4, 6, 6, /* 6
7 */ 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 2, 7, 4, 4, 7, 7, /* 7
8 */ 2, 6, 2, 6, 3, 3, 3, 3, 2, 2, 2, 2, 4, 4, 4, 4, /* 8
9 */ 2, 6, 0, 6, 4, 4, 4, 4, 2, 5, 2, 5, 5, 5, 5, 5, /* 9
A */ 2, 6, 2, 6, 3, 3, 3, 3, 2, 2, 2, 2, 4, 4, 4, 4, /* A
B */ 2, 5, 0, 5, 4, 4, 4, 4, 2, 4, 2, 4, 4, 4, 4, 4, /* B
C */ 2, 6, 2, 8, 3, 3, 5, 5, 2, 2, 2, 2, 4, 4, 6, 6, /* C
D */ 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 2, 7, 4, 4, 7, 7, /* D
E */ 2, 6, 2, 8, 3, 3, 5, 5, 2, 2, 2, 2, 4, 4, 6, 6, /* E
F */ 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 2, 7, 4, 4, 7, 7  /* F
     0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  */
};

/* extra cycles to burn on page overflow: only on addressing modes ABS_X, ABS_Y, IND_Y (and REL). can be calculated with masks, but this is faster */
#define v 0x100
static const u16 op_extracycles[0x100]=
{ /* 0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
0 */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0
1 */ 0, v, 0, 0, 0, 0, 0, 0, 0, v, 0, 0, v, v, 0, 0, /* 1
2 */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 2
3 */ 0, v, 0, 0, 0, 0, 0, 0, 0, v, 0, 0, v, v, 0, 0, /* 3
4 */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 4
5 */ 0, v, 0, 0, 0, 0, 0, 0, 0, v, 0, 0, v, v, 0, 0, /* 5
6 */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 6
7 */ 0, v, 0, 0, 0, 0, 0, 0, 0, v, 0, 0, v, v, 0, 0, /* 7
8 */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 8
9 */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 9
A */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* A
B */ v, v, v, v, v, v, v, v, v, v, v, v, v, v, v, v, /* B
C */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* C
D */ 0, v, 0, 0, 0, 0, 0, 0, 0, v, 0, 0, v, v, 0, 0, /* D
E */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* E
F */ 0, v, 0, 0, 0, 0, 0, 0, 0, v, 0, 0, v, v, 0, 0  /* F
     0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  */
};
#undef v


/* interrupt handler: function declarations, defines, fpt define, functionpointertable */
static void __fastcall int_none(void); static void __fastcall int_brk(void); static void __fastcall int_irq(void); static void __fastcall int_nmi(void); static void __fastcall int_res(void); static void __fastcall int_mul(void);
typedef void __fastcall (*fpt_handle_interrupt)(void);
static const fpt_handle_interrupt handle_interrupt[0x10]= { int_none, int_brk, int_irq, int_mul, int_nmi, int_mul, int_mul, int_mul, int_res, int_mul, int_mul, int_mul, int_mul, int_mul, int_mul, int_mul };
#define IRQ_PEND(x) if (tempb&cpu.irq_pend&!cpu.P_I) if (!(cpu.interrupt&INTERRUPT_IRQ)) { cpu.interrupt|=INTERRUPT_IRQ; cpu.irq_type|=x; }



/* shortcuts */

/* memory handler */
typedef u8 __fastcall(*fpt_cpu_read_memorymap)(u16);
typedef void __fastcall(*fpt_cpu_write_memorymap)(u16,u8);
static fpt_cpu_read_memorymap cpu_read_memorymap[9]; /* last one in case of overflow */
static fpt_cpu_write_memorymap cpu_write_memorymap[9]; /* last one in case of overflow */

#define READ_BYTE(x)		cpu_read_memorymap[(x)>>13](x)
#define READ_BYTE_RAM(x)	cpu_read_ram(x)
#define READ_BYTE_MAP(x,y)	cpu_read_memorymap[x](y)
#define READ_WORD(x)		READ_BYTE(x)|READ_BYTE((x)+1)<<8
#define WRITE_BYTE(x,y)		cpu_write_memorymap[(x)>>13](x,y)
#define WRITE_BYTE_RAM(x,y)	cpu_write_ram(x,y)
#define WRITE_BYTE_MAP(x,y,z) cpu_write_memorymap[x](y,z)
#define DATA				READ_BYTE(address)

/* branch */
#define BRANCH()	address=cpu.PC+(s8)DATA; cpu.cycles-=crystal->cycle<<((cpu.PC^address)>>8&1); cpu.PC=address /* shift left with result (1 if other mempage, 0 if not)=2 or 1 cycle */

/* stack */
#define PUSH(x)		WRITE_BYTE_RAM(0x100|cpu.S--,x)
#define PULL()		READ_BYTE_RAM(++cpu.S|0x100)

#define PULL_PC()	tempb=PULL(); cpu.PC=PULL(); cpu.PC=cpu.PC<<8|tempb /* pull twice, store in program counter */
#define PUSH_PC()	PUSH(cpu.PC>>8); PUSH(cpu.PC&0xff) /* push program counter */

/* push/pull status register NV1BDIZC (gotta convert every var to bits. tough job, but hardly used, so the var method of flags is faster) */
/* 6502 flags optimisation idea from blargg */
#define PUSH_STATUS(b)	PUSH(((cpu.P_NZ>>8|cpu.P_NZ)&0x80)|((cpu.P_NZ&0xff)==0)<<1|(cpu.P_V!=0)<<6|cpu.P_D<<3|cpu.P_I<<2|cpu.P_C>>8|0x20|b)
#define PULL_STATUS()	cpu.P_NZ=PULL(); cpu.P_V=cpu.P_NZ&0x40; cpu.P_D=cpu.P_NZ>>3&1; cpu.P_I=cpu.P_NZ>>2&1; cpu.P_C=cpu.P_NZ<<8&0x100; cpu.P_NZ=(~cpu.P_NZ&2)|cpu.P_NZ<<8


/* opcodes */

/* (with RMW garbage store on ASL,DEC,INC,LSR,ROL,ROR) */

#define OP_ADC()	tempb=DATA; cpu.P_C=(cpu.P_C>>8)+cpu.A+tempb; cpu.P_V=(cpu.A^cpu.P_C)&(tempb^cpu.P_C)&0x80; cpu.A=cpu.P_NZ=cpu.P_C
#define OP_AND()	cpu.P_NZ=cpu.A&=DATA
#define OP_ASL()	tempb=DATA; WRITE_BYTE(address,tempb); cpu.P_NZ=cpu.P_C=tempb<<1; WRITE_BYTE(address,cpu.P_C)
#define OP_ASLa()	cpu.A=cpu.P_NZ=cpu.P_C=cpu.A<<1;
#define OP_BCC()	if (!(cpu.P_C&0x100)) { BRANCH(); }
#define OP_BCS()	if (cpu.P_C&0x100) { BRANCH(); }
#define OP_BEQ()	if (!(cpu.P_NZ&0xff)) { BRANCH(); }
#define OP_BIT()	tempb=DATA; cpu.P_NZ=tempb<<8|(cpu.A&tempb)>>1|(cpu.A&tempb&0x7f); cpu.P_V=tempb&0x40;
#define OP_BMI()	if (cpu.P_NZ&0x8080) { BRANCH(); }
#define OP_BNE()	if (cpu.P_NZ&0xff) { BRANCH(); }
#define OP_BPL()	if (!(cpu.P_NZ&0x8080)) { BRANCH(); }
#define OP_BRK()	cpu.PC++; cpu.interrupt|=INTERRUPT_BRK
#define OP_BVC()	if (cpu.P_V==FALSE) { BRANCH(); }
#define OP_BVS()	if (cpu.P_V) { BRANCH(); }
#define OP_CLC()	cpu.P_C=FALSE
#define OP_CLD()	cpu.P_D=FALSE
#define OP_CLI()	tempb=cpu.P_I; cpu.P_I=FALSE; IRQ_PEND(INT_EDGE_IRQ)
#define OP_CLV()	cpu.P_V=FALSE
#define OP_CMP()	cpu.P_C=cpu.P_NZ=cpu.A+0x100-DATA
#define OP_CPX()	cpu.P_C=cpu.P_NZ=cpu.X+0x100-DATA
#define OP_CPY()	cpu.P_C=cpu.P_NZ=cpu.Y+0x100-DATA
#define OP_DEC()	tempb=DATA; WRITE_BYTE(address,tempb); cpu.P_NZ=tempb+0xff; WRITE_BYTE(address,cpu.P_NZ)
#define OP_DEX()	cpu.P_NZ=cpu.X+=0xff
#define OP_DEY()	cpu.P_NZ=cpu.Y+=0xff
#define OP_EOR()	cpu.P_NZ=cpu.A^=DATA
#define OP_INC()	tempb=DATA; WRITE_BYTE(address,tempb); cpu.P_NZ=tempb+1; WRITE_BYTE(address,cpu.P_NZ)
#define OP_INX()	cpu.P_NZ=++cpu.X
#define OP_INY()	cpu.P_NZ=++cpu.Y
#define OP_JMP()	cpu.PC=address
#define OP_JSR()	cpu.PC--; PUSH_PC(); cpu.PC=address
#define OP_LDA()	cpu.P_NZ=cpu.A=DATA
#define OP_LDX()	cpu.P_NZ=cpu.X=DATA
#define OP_LDY()	cpu.P_NZ=cpu.Y=DATA
#define OP_LSR()	tempb=DATA; WRITE_BYTE(address,tempb); cpu.P_NZ=tempb>>1; cpu.P_C=tempb<<8&0x100; WRITE_BYTE(address,cpu.P_NZ)
#define OP_LSRa()	cpu.P_NZ=cpu.A>>1; cpu.P_C=cpu.A<<8&0x100; cpu.A=cpu.P_NZ /* same as above, but accumulator */
/* OP_NOP: nothing */
#define OP_ORA()	cpu.P_NZ=cpu.A|=DATA
#define OP_PHA()	PUSH(cpu.A)
#define OP_PHP()	PUSH_STATUS(FLAG_B)
#define OP_PLA()	cpu.P_NZ=cpu.A=PULL()
#define OP_PLP()	tempb=cpu.P_I; PULL_STATUS(); IRQ_PEND(INT_EDGE_IRQ)
#define OP_ROL()	tempb=DATA; WRITE_BYTE(address,tempb); cpu.P_NZ=cpu.P_C=tempb<<1|cpu.P_C>>8; WRITE_BYTE(address,cpu.P_NZ&0xff)
#define OP_ROLa()	cpu.A=cpu.P_NZ=cpu.P_C=cpu.A<<1|cpu.P_C>>8 /* same as above, but accumulator */
#define OP_ROR()	tempb=DATA; WRITE_BYTE(address,tempb); cpu.P_NZ=tempb>>1|(cpu.P_C>>1&0x80); cpu.P_C=tempb<<8&0x100; WRITE_BYTE(address,cpu.P_NZ)
#define OP_RORa()	cpu.P_NZ=cpu.A>>1|(cpu.P_C>>1&0x80); cpu.P_C=cpu.A<<8&0x100; cpu.A=cpu.P_NZ /* same as above, but accumulator */
#define OP_RTI()	tempb=cpu.P_I; PULL_STATUS(); IRQ_PEND(FALSE); PULL_PC(); cpu.irq_edge=FALSE
#define OP_RTS()	PULL_PC(); cpu.PC++
#define OP_SBC()	tempb=~DATA; cpu.P_C=(cpu.P_C>>8)+cpu.A+tempb; cpu.P_V=(cpu.A^cpu.P_C)&(tempb^cpu.P_C)&0x80; cpu.A=cpu.P_NZ=cpu.P_C
#define OP_SEC()	cpu.P_C=0x100
#define OP_SED()	cpu.P_D=TRUE
#define OP_SEI()	cpu.P_I=TRUE
#define OP_STA()	WRITE_BYTE(address,cpu.A)
#define OP_STX()	WRITE_BYTE(address,cpu.X)
#define OP_STY()	WRITE_BYTE(address,cpu.Y)
#define OP_TAX()	cpu.P_NZ=cpu.X=cpu.A
#define OP_TAY()	cpu.P_NZ=cpu.Y=cpu.A
#define OP_TSX()	cpu.P_NZ=cpu.X=cpu.S
#define OP_TXA()	cpu.P_NZ=cpu.A=cpu.X
#define OP_TXS()	cpu.S=cpu.X
#define OP_TYA()	cpu.P_NZ=cpu.A=cpu.Y


/*
unofficial opcodes, eg. opcodes ending with %11 (3,7,b,f), usually instructions in parallel, eg. 07=05+06.

unofficial 'official':
	NOP (though with more possible addressing modes instead of implied):
		1 byte IMP: $1a, $3a, $5a, $7a, $da, $fa
		2 bytes ZP(/x/y) or IMM: $04, $14, $34, $44, $54, $64, $74, $80, $82, $89, $c2, $d4, $e2, $f4
		3 bytes ABS(/x): $0c, $1c, $3c, $5c, $7c, $dc, $fc
	SBC IMM: $eb (same as $e9)

some examples using unofficial opcodes:
- Esper Bouken Tai: $D4: ZP,X NOP
- Ninja Jajamaru Kun: $04: ZP NOP
- Radia Senki - Reimei Hen: $DA/$FA: IMP NOP
- Simpsons - Bart vs World: $0B: IMM AAC (bad dump?)
- Super Cars: $B3: IND_Y LAX
- (Kevin Horton's nestest)
- (blargg's cpu.nes test)

*/

#define OP_AAC()	OP_AND(); cpu.P_C=cpu.P_NZ<<1
#define OP_AAX()	WRITE_BYTE(address,cpu.A&cpu.X)
#define OP_ARR()	OP_AND(); OP_RORa(); cpu.P_C=cpu.A<<2&0x100; cpu.P_V=cpu.P_C>>1^(cpu.A<<2&0x80)
#define OP_ASR()	OP_AND(); OP_LSRa()
#define OP_ATX()	cpu.A|=0xff; OP_AND(); OP_TAX() /* OR mask may differ on different NESes */
#define OP_AXA()	WRITE_BYTE(address,cpu.A&cpu.X&7)
#define OP_AXS()	cpu.X=cpu.P_C=cpu.P_NZ=(DATA^0xff)+(cpu.A&cpu.X)+1
#define OP_DCP()	OP_DEC(); OP_CMP()
#define OP_HLT()	cpu.halt=TRUE; return 1
#define OP_ISC()	OP_INC(); OP_SBC()
#define OP_LAR()	OP_TSX(); OP_TXA(); OP_AND(); OP_TAX(); OP_TXS()
#define OP_LAX()	OP_LDA(); OP_LDX()
#define OP_RLA()	OP_ROL(); OP_AND()
#define OP_RRA()	OP_ROR(); OP_ADC()
#define OP_SLO()	OP_ASL(); OP_ORA()
#define OP_SRE()	OP_LSR(); OP_EOR()
#define OP_SXA()	if ((tempw^address)&0x100) address&=0xff; WRITE_BYTE(address,cpu.X&((tempw>>8)+1)) /* not sure about page boundary cross behaviour (same for SYA and XAS) */
#define OP_SYA()	if ((tempw^address)&0x100) address&=0xff; WRITE_BYTE(address,cpu.Y&((tempw>>8)+1))
#define OP_XAA()	OP_TXA(); OP_AND()
#define OP_XAS()	if ((tempw^address)&0x100) address&=0xff; cpu.S=cpu.A&cpu.Y; WRITE_BYTE(address,cpu.S&((tempw>>8)+1))




/* addressing modes */

#define ADD_EXTRA_CYCLE() if ((tempw^address)&op_extracycles[opcode]) cpu.cycles-=crystal->cycle /* 1 cycle penalty on page boundary cross */
#define ADD_DUMMY_READ(x) if (((tempw^address)&op_extracycles[opcode])|(op_extracycles[opcode]^0x100)) tempb=READ_BYTE((tempw&0xff00)|((tempw+(x))&0xff))

/* ADD_ACCUM: nothing */
#define ADD_IMM()	address=cpu.PC++
#define ADD_ABS()	address=READ_WORD(cpu.PC); cpu.PC++; cpu.PC++
#define ADD_ZP()	address=READ_BYTE(cpu.PC); cpu.PC++
#define ADD_ZP_X()	address=(READ_BYTE(cpu.PC)+cpu.X)&0xff; cpu.PC++
#define ADD_ZP_Y()	address=(READ_BYTE(cpu.PC)+cpu.Y)&0xff; cpu.PC++
#define ADD_ABS_X()	tempw=READ_WORD(cpu.PC); address=tempw+cpu.X; ADD_DUMMY_READ(cpu.X); ADD_EXTRA_CYCLE(); cpu.PC++; cpu.PC++
#define ADD_ABS_Y()	tempw=READ_WORD(cpu.PC); address=tempw+cpu.Y; ADD_EXTRA_CYCLE(); cpu.PC++; cpu.PC++
/* ADD_IMP: nothing */
#define ADD_REL()	address=cpu.PC++
#define ADD_IND_X()	tempb=READ_BYTE(cpu.PC)+cpu.X; cpu.PC++; address=READ_BYTE_RAM((tempb+1)&0xff)<<8|READ_BYTE_RAM(tempb)
#define ADD_IND_Y()	tempb=READ_BYTE(cpu.PC); tempw=READ_BYTE_RAM((tempb+1)&0xff)<<8|READ_BYTE_RAM(tempb); address=tempw+cpu.Y; ADD_DUMMY_READ(cpu.Y); ADD_EXTRA_CYCLE(); cpu.PC++
#define ADD_INDIR()	tempw=READ_WORD(cpu.PC); address=READ_BYTE((tempw&0xff00)|((tempw+1)&0xff))<<8|READ_BYTE(tempw) /* there's a bug on the 6502 that it loads xx00 after xxff */




/* interrupt handlers */

/* irq/brk shared, with possible NMI in the middle */
#define _INT(b) \
	PUSH_PC(); PUSH_STATUS(b); \
	cpu.P_I=TRUE; \
	\
	/* 5 cycles */ \
	cpu.cycles-=5*crystal->cycle; \
	\
	if ((cpu.cycles<=crystal->vblank_trigger&&ppu_get_cycles()>crystal->vblank_trigger)||cpu.nmi) { \
		ppu_force_update(); \
		\
		/* nmi? interrupt is ignored for now */ \
		cpu.PC=READ_WORD(cpu.nmi?0xfffa:0xfffe); \
		cpu.cycles-=2*crystal->cycle; \
	} \
	else { \
		cpu.cycles-=2*crystal->cycle; \
		\
		if ((cpu.cycles<=crystal->vblank_trigger&&ppu_get_cycles()>crystal->vblank_trigger)||cpu.nmi) { \
			ppu_force_update(); \
			\
			/* nmi? interrupt with nmi one instruction after int */ \
			if (cpu.nmi) cpu.interrupt|=INTERRUPT_NMI; \
		} \
		\
		cpu.PC=READ_WORD(0xfffe); \
	} \
	\
	cpu.nmi=0

/* no interrupt, shouldn't be called */
static void __fastcall int_none(void) { return; }
/* software interrupt (brk) handler */
static void __fastcall int_brk(void)
{
	cpu.interrupt&=~INTERRUPT_BRK;
	
	_INT(FLAG_B);
	
	#if DEBUG_CPU
	LOG(LOG_CPU,"_INTERRUPT BRK\n");
	#endif
}
/* interrupt request handler */
static void __fastcall int_irq(void)
{
	if (cpu.irq_type&INT_EDGE_IRQ) { cpu.irq_type&=~INT_EDGE_IRQ; cpu.irq_edge=!cpu.P_I; }
	else {
		cpu.interrupt&=~INTERRUPT_IRQ;
		if ((!cpu.P_I)|cpu.irq_edge) {
			cpu.irq_edge=FALSE;
			
			_INT(0);
			
			#if DEBUG_CPU
			LOG(LOG_CPU,"_INTERRUPT IRQ\n");
			#endif
		}
	}
}
/* non maskable interrupt handler */
static void __fastcall int_nmi(void)
{
	if (cpu.irq_type&INT_EDGE_NMI) { cpu.irq_type&=~INT_EDGE_NMI; }
	else {
		cpu.interrupt&=~INTERRUPT_NMI;
		if (!ppu_get_vblank_nmi_disallowed()) {
			PUSH_PC(); PUSH_STATUS(FALSE);
			cpu.P_I=TRUE; cpu.PC=READ_WORD(0xfffa); cpu.cycles-=7*crystal->cycle;
			
			#if DEBUG_CPU
			LOG(LOG_CPU,"_INTERRUPT NMI\n");
			#endif
		}
	}
}
/* reset handler */
static void __fastcall int_res(void)
{
	cpu.interrupt=cpu.irq_type=cpu.irq_edge=FALSE;
	cpu.S-=3; cpu.P_I=TRUE; cpu.PC=READ_WORD(0xfffc); cpu.cycles-=7*crystal->cycle;
	
	#if DEBUG_CPU
	LOG(LOG_CPU,"_INTERRUPT RES\n");
	#endif
}
/* multiple interrupts: choose one */
static void __fastcall int_mul(void)
{
	if (cpu.interrupt&INTERRUPT_RES) int_res();
	else if (cpu.interrupt&INTERRUPT_NMI) {
		int n=((cpu.irq_type&INT_EDGE_NMI)==0)&ppu_get_vblank_nmi_disallowed();
		if (n) cpu.interrupt&=~INTERRUPT_NMI;
		n|=(cpu.irq_type&INT_EDGE_NMI);
		
		if (cpu.interrupt&INTERRUPT_IRQ) {
			if (n) {
				cpu.irq_type&=~INT_EDGE_NMI;
				if (cpu.interrupt&INTERRUPT_BRK) {
					if (cpu.P_I|(cpu.irq_type&INT_EDGE_IRQ)) { cpu.irq_type&=~INT_EDGE_IRQ; int_brk(); }
					else int_irq();
				}
				else int_irq();
			}
			else {
				cpu.interrupt&=~(INTERRUPT_IRQ|INTERRUPT_BRK); /* suppress */
				cpu.irq_type&=~INT_EDGE_IRQ; cpu.irq_edge=FALSE;
				int_nmi();
			}
		}
		else if (cpu.interrupt&INTERRUPT_BRK) {
			if (n) { cpu.irq_type&=~INT_EDGE_NMI; int_brk(); }
			else {
				cpu.interrupt&=~INTERRUPT_BRK;
				int_nmi();
				
				/* set B flag */
				WRITE_BYTE_RAM(((cpu.S+1)&0xff)|0x100,READ_BYTE_RAM(((cpu.S+1)&0xff)|0x100)|FLAG_B);
			}
		}
		else int_nmi();
	}
	else if (cpu.interrupt&INTERRUPT_IRQ) {
		if (cpu.interrupt&INTERRUPT_BRK) {
			if (cpu.P_I|(cpu.irq_type&INT_EDGE_IRQ)) { cpu.irq_type&=~INT_EDGE_IRQ; int_brk(); }
			else { cpu.interrupt&=~INTERRUPT_BRK; int_irq(); }
		}
		else int_irq();
	}
}



/**************************/

static __inline__ void update_sync(void)
{
	if (cpu.cycles<cpu.apu_dmc_next_step) apu_run_dmc();
	if (cpu.cycles<cpu.apu_fs_next_step) apu_run_fs();
	(*mapper_irq)();
}

void cpu_set_fpt(int type)
{
	const fpt_cpu_read_memorymap cpu_read_memorymap_std[8]={ cpu_read_ram, cpu_read_ppu, cpu_read_io, cpu_read_wram, cpu_read_prgbank1a, cpu_read_prgbank1b, cpu_read_prgbank2a, cpu_read_prgbank2b };
	const fpt_cpu_write_memorymap cpu_write_memorymap_std[8]={ cpu_write_ram, cpu_write_ppu, cpu_write_io, cpu_write_wram, cpu_write_prgbank1a, cpu_write_prgbank1b, cpu_write_prgbank2a, cpu_write_prgbank2b };
	
	if (type) {
		switch (type) {
			case CPU_FPT_WRAM_READ:		cpu_read_memorymap[3]=cpu_read_wram_override;		break;
			case CPU_FPT_WRAM_WRITE:	cpu_write_memorymap[3]=cpu_write_wram_override;		break;
			case CPU_FPT_PRG1A_READ:	cpu_read_memorymap[4]=cpu_read_prgbank1a_override;	break;
			case CPU_FPT_PRG1B_READ:	cpu_read_memorymap[5]=cpu_read_prgbank1b_override;	break;
			case CPU_FPT_PRG2A_READ:	cpu_read_memorymap[6]=cpu_read_prgbank2a_override;	break;
			case CPU_FPT_PRG2B_READ:	cpu_read_memorymap[7]=cpu_read_prgbank2b_override;	break;
			
			default: break;
		}
	}
	else {
		int i;
		
		for (i=0;i<9;i++) {
			cpu_read_memorymap[i]=cpu_read_memorymap_std[i&7];
			cpu_write_memorymap[i]=cpu_write_memorymap_std[i&7];
		}
	}
}

void illegal_nothing(u8 opcode) { return; }
void illegal_show(u8 opcode)
{
	if (!op_official[opcode]) {
		LOG(LOG_MISC|LOG_WARNING,"illegal opcode at $%04X:$%02X(%s %s)",(cpu.PC-1)&0xffff,opcode,addressing_mode_names_short[op_addressing_mode[opcode]],c_op_mnemonic[op_mnemonic[opcode]]);
		if (!--cpu.illegal_count) {
			_illegal=illegal_nothing;
			LOG(LOG_MISC|LOG_WARNING,", etc.");
		}
		LOG(LOG_MISC|LOG_WARNING,"\n");
	}
}

int __fastcall cpu_update_apu_fs_onclock(void)
{
	if (cpu.cycles==cpu.apu_fs_next_step) {
		apu_run_fs();
		return TRUE;
	}
	return FALSE;
}

int __fastcall cpu_set_apu_fs_next_step(int freq) { cpu.apu_fs_next_step-=freq; return cpu.cycles<cpu.apu_fs_next_step; }
int __fastcall cpu_test_apu_fs_next_step(int freq) { return cpu.cycles<(cpu.apu_fs_next_step-freq); }
void __fastcall cpu_reset_apu_fs_next_step(void) { cpu.apu_fs_next_step=cpu.cycles-(crystal->rel_cycle+(((((cpu.cycles_full+(cpu.cycles_start-cpu.cycles))-((cpu.cycles_full+(cpu.cycles_start-cpu.cycles))%crystal->cycle))/crystal->cycle)&1)*crystal->rel_cycle)); }
int __fastcall cpu_set_apu_dmc_next_step(void) { cpu.apu_dmc_next_step-=cpu.apu_dmc_frequency; return cpu.cycles<cpu.apu_dmc_next_step; }
int __fastcall cpu_test_apu_dmc_next_cycle(void) { return cpu.cycles<(cpu.apu_dmc_next_step-crystal->rel_cycle); }
void __fastcall cpu_reset_apu_dmc_freq(int freq) { cpu.apu_dmc_frequency=freq; }
void cpu_set_mapper_irq(fp_mapper_irq fun) { if (fun==NULL) fun=mapper_irq_nothing; mapper_irq=fun; }
int* cpu_get_cycles_ptr(void) { return &cpu.cycles; }
int __fastcall cpu_get_cycles(void) { return cpu.cycles; }
int __fastcall cpu_get_rel_cycles(void) { return cpu.cycles-cpu.cycles%crystal->rel_cycle; }
void cpu_adjust_rel_cycles(void) { cpu.cycles+=(cpu.cycles%crystal->ppu_cycle); }
int cpu_get_halt(void) { return cpu.halt; }
u16 cpu_get_pc(void) { return cpu.PC; }
void cpu_set_overflow(void) { cpu.P_V=TRUE; }
void cpu_set_nmi(int n) { cpu.nmi=n; }

void cpu_do_vblank_nmi(void)
{
	if (cpu.halt) return;
	
	if (cpu.nmi) {
		cpu.interrupt|=INTERRUPT_NMI;
		if ((cpu.cycles+crystal->ppu_cycle)<crystal->vblank_trigger) handle_interrupt[cpu.interrupt](); /* edge */
	}
}

void __fastcall cpu_set_interrupt(int i)
{
	if (i&INTERRUPT_IRQ) {
		cpu.irq_pend=TRUE;
		cpu.irq_type&=~INT_EDGE_IRQ;
		if (!(cpu.interrupt&INTERRUPT_IRQ)) cpu.irq_type|=(i>>8);
	}
	else cpu.irq_type|=(i>>8);
	
	cpu.interrupt|=(i&BIN8(00001111));
	cpu.irq_type|=(i&BIN8(11110000));
	
	#if DEBUG_CPU
	LOG(LOG_CPU,"_INTERRUPT request: ");
	switch (i&BIN8(11111100)) {
		case INTERRUPT_NMI:	LOG(LOG_CPU,"NMI");		break;
		case INTERRUPT_RES:	LOG(LOG_CPU,"reset");	break;
		case IRQ_FRAME:		LOG(LOG_CPU,"frame");	break;
		case IRQ_DMC:		LOG(LOG_CPU,"DMC");		break;
		case IRQ_MAPPER:	LOG(LOG_CPU,"ext1");	break;
		case IRQ_MAPPER2:	LOG(LOG_CPU,"ext2");	break;
		
		default:			LOG(LOG_CPU,"?");		break;
	}
	LOG(LOG_CPU,"\n");
	#endif
}

void __fastcall cpu_acknowledge_interrupt(int i)
{
	cpu.irq_type&=i;
	if (!(cpu.irq_type&~(INT_EDGE_NMI|INT_EDGE_IRQ))) cpu.irq_pend=FALSE;
	
	#if DEBUG_CPU
	LOG(LOG_CPU,"_INTERRUPT acknowledge: ");
	switch (~i) {
		case IRQ_FRAME:		LOG(LOG_CPU,"frame");	break;
		case IRQ_DMC:		LOG(LOG_CPU,"DMC");		break;
		case IRQ_MAPPER:	LOG(LOG_CPU,"ext1");	break;
		case IRQ_MAPPER2:	LOG(LOG_CPU,"ext2");	break;
		
		default:			LOG(LOG_CPU,"?");		break;
	}
	LOG(LOG_CPU,"\n");
	#endif
}

void cpu_fill_op_cycles_lookup(void)
{
	int i;
	for(i=0;i<0x100;i++) op_cycles_lookup[i]=op_cycles[i]*crystal->cycle;
}

void __fastcall cpu_sprite_dma_transfer(u8 addressbyte)
{
	int i=0x100;
	const int map=addressbyte>>5;
	int address=addressbyte<<8; /* no need for u16, it won't overflow */
	const int cc=crystal->cycle<<1;
	
	cpu.cycles-=crystal->cycle;
	
	while (i--) {
		ppu_write_sprite_memory_data(READ_BYTE_MAP(map,address++));
		if ((cpu.cycles-=cc)<0) {
			cpu.sprite_dma_nf=TRUE;
			cpu.sprite_dma_count=i;
			cpu.sprite_dma_address=address;
			break;
		}
	}
	
	update_sync();
}

void __fastcall cpu_sprite_dma_transfer_continue(void)
{
	if (cpu.sprite_dma_nf) {
		const int map=cpu.sprite_dma_address>>13;
		const int cc=crystal->cycle<<1;
		
		while (cpu.sprite_dma_count--) {
			ppu_write_sprite_memory_data(READ_BYTE_MAP(map,cpu.sprite_dma_address++));
			cpu.cycles-=cc;
		}
		cpu.sprite_dma_nf=FALSE;
		
		update_sync();
	}
}

u8 __fastcall cpu_dmc_dma_transfer(u16 address)
{
	cpu.cycles-=(crystal->cycle<<2);
	return READ_BYTE(address);
}

void cpu_kill_cycle(void)
{
	cpu.cycles-=crystal->ppu_cycle;
	cpu.cycles_start-=crystal->ppu_cycle;
	cpu.apu_dmc_next_step-=crystal->ppu_cycle;
	cpu.apu_fs_next_step-=crystal->ppu_cycle;
	apu_kill_cycle();
	mapper_kill_cycle();
	
	update_sync();
}

void cpu_reset(int hard)
{
	char oc[10];
	
	cpu.interrupt|=INTERRUPT_RES;
	handle_interrupt[cpu.interrupt]();
	
	sprintf(oc," (%d%%)",crystal->overclock);
	LOG(LOG_MISC|LOG_COLOUR(LC_YELLOW)|LOG_TYPE(LT_RESET),"%s NES%s reset            ",crystal->mode?"PAL":"NTSC",crystal->overclock==100?"":oc);
}

void cpu_sync_reset_time(void)
{
	if (cpu.cycles>(crystal->frame-356*crystal->ppu_cycle)) cpu.cycles=crystal->frame-356*crystal->ppu_cycle;
}

void cpu_init(void)
{
	memset(&cpu,0,sizeof(cpu));
	cpu_fill_op_cycles_lookup();
	cpu.P_NZ=1;
	
	#if !CPU_EMPTY_RAM
	WRITE_BYTE(0x0008,0xf7);
	WRITE_BYTE(0x0009,0xef);
	WRITE_BYTE(0x000a,0xdf);
	WRITE_BYTE(0x000f,0xbf);
	#endif
	
	/* first frame start */
	/* NTSC: at the end of the first scanline (that is, after vblank) */
	/* sensitive games:
		Chessmaster (skips title screen) (OK)
		Cobra Triangle (locks up in-game) (OK)
		Ironsword (locks up at title screen) (OK)
		RAF World (small vertical shake at title screen) (OK)
	
	some PD stuff not working properly, but better when starting from vblank (it's probably correct they don't work):
		Duel, The by Bokudono
		Jumpy Demo
		maybe more..
	*/
	/* Famicom's startup timing is random (tested on Famicom), emulating that could cause problems with eg. movies, so I won't */
	if (!crystal->mode) cpu.cycles=cpu.apu_fs_next_step=cpu.apu_dmc_next_step=-(356*crystal->ppu_cycle);
	
	/* PAL: 106 or 107 (at the end of the first scanline) consistent $1682D on Disch's startup test ROM, but Pacman won't boot like that */
	else cpu.cycles=cpu.apu_fs_next_step=cpu.apu_dmc_next_step=-(31143*crystal->rel_cycle); /* closest I can get to $1682D without Pacman/Kojiki breaking */
	/* sensitive games:
		Pacman (won't boot) (OK)
		Ironsword (same as NTSC) (OK) (not really a problem with PAL FS timing)
		Cobra Triangle (same as NTSC) (OK) (not really a problem with PAL FS timing)
	
	demos (tested on NES):
		Kojiki by wAMMA
	*/
	
	/* settings */
	cpu.illegal_count=draw_get_cpu_illegal_count_set();
	
	if (cpu.illegal_count) _illegal=illegal_show;
	else _illegal=illegal_nothing;
	
	LOG(LOG_VERBOSE,"CPU initialised\n");
}

void cpu_clean(void)
{
	memset(&cpu,0,sizeof(cpu));
	
	LOG(LOG_VERBOSE,"CPU cleaned\n");
}

void cpu_new_frame(void)
{
	cpu.cycles_full+=(cpu.cycles_start-cpu.cycles);
	cpu.cycles+=crystal->frame;
	cpu.cycles_start=cpu.cycles;
	
	cpu.apu_fs_next_step+=crystal->frame;
	cpu.apu_dmc_next_step+=crystal->frame;
	
	#if DEBUG_CPU
	LOG(LOG_CPU,"\n\tfc:%d\n",crystal->fc);
	#endif
}

int cpu_execute(const int border)
{
	#define _ILL (*_illegal)(opcode)
	
	#if DEBUG_CPU
	u16 temp_pc;
	char in_int[3];
	#endif
	
	u16 tempw;
	u8 tempb;
	u8 opcode;
	u16 address;
	
	if (cpu.halt) return 1;
	
 while (cpu.cycles>border) {
	
	opcode=READ_BYTE(cpu.PC);
	cpu.cycles-=op_cycles_lookup[opcode];
	
	update_sync();
	
	#if DEBUG_PPU
	ppu_force_update();
	#endif
	
	#if DEBUG_CPU
	address=0;
	temp_pc=cpu.PC;
	#endif
	
	cpu.PC++;
	
 switch (opcode) {
/*row	#opcode   =ill?	addr mode	opcode */
/*0*/
	case 0x00:		/*ADD_IMP*/		OP_BRK();	break;
	case 0x01:		ADD_IND_X();	OP_ORA();	break;
	case 0x02:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x03:_ILL;	ADD_IND_X();	OP_SLO();	break;
	case 0x04:_ILL;	ADD_ZP();		/*OP_NOP*/	break;
	case 0x05:		ADD_ZP();		OP_ORA();	break;
	case 0x06:		ADD_ZP();		OP_ASL();	break;
	case 0x07:_ILL;	ADD_ZP();		OP_SLO();	break;
	case 0x08:		/*ADD_IMP*/		OP_PHP();	break;
	case 0x09:		ADD_IMM();		OP_ORA();	break;
	case 0x0a:		/*ADD_ACCUM*/	OP_ASLa();	break;
	case 0x0b:_ILL;	ADD_IMM();		OP_AAC();	break;
	case 0x0c:_ILL;	ADD_ABS();		/*OP_NOP*/	break;
	case 0x0d:		ADD_ABS();		OP_ORA();	break;
	case 0x0e:		ADD_ABS();		OP_ASL();	break;
	case 0x0f:_ILL;	ADD_ABS();		OP_SLO();	break;
/*1*/
	case 0x10:		ADD_REL();		OP_BPL()	break;
	case 0x11:		ADD_IND_Y();	OP_ORA();	break;
	case 0x12:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x13:_ILL;	ADD_IND_Y();	OP_SLO();	break;
	case 0x14:_ILL;	ADD_ZP_X();		/*OP_NOP*/	break;
	case 0x15:		ADD_ZP_X();		OP_ORA();	break;
	case 0x16:		ADD_ZP_X();		OP_ASL();	break;
	case 0x17:_ILL;	ADD_ZP_X();		OP_SLO();	break;
	case 0x18:		/*ADD_IMP*/		OP_CLC();	break;
	case 0x19:		ADD_ABS_Y();	OP_ORA();	break;
	case 0x1a:_ILL;	/*ADD_IMP*/		/*OP_NOP*/	break;
	case 0x1b:_ILL;	ADD_ABS_Y();	OP_SLO();	break;
	case 0x1c:_ILL;	ADD_ABS_X();	/*OP_NOP*/	break;
	case 0x1d:		ADD_ABS_X();	OP_ORA();	break;
	case 0x1e:		ADD_ABS_X();	OP_ASL();	break;
	case 0x1f:_ILL;	ADD_ABS_X();	OP_SLO();	break;
/*2*/
	case 0x20:		ADD_ABS();		OP_JSR();	break;
	case 0x21:		ADD_IND_X();	OP_AND();	break;
	case 0x22:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x23:_ILL;	ADD_IND_X();	OP_RLA();	break;
	case 0x24:		ADD_ZP();		OP_BIT();	break;
	case 0x25:		ADD_ZP();		OP_AND();	break;
	case 0x26:		ADD_ZP();		OP_ROL();	break;
	case 0x27:_ILL;	ADD_ZP();		OP_RLA();	break;
	case 0x28:		/*ADD_IMP*/		OP_PLP();	break;
	case 0x29:		ADD_IMM();		OP_AND();	break;
	case 0x2a:		/*ADD_ACCUM*/	OP_ROLa();	break;
	case 0x2b:_ILL;	ADD_IMM();		OP_AAC();	break;
	case 0x2c:		ADD_ABS();		OP_BIT();	break;
	case 0x2d:		ADD_ABS();		OP_AND();	break;
	case 0x2e:		ADD_ABS();		OP_ROL();	break;
	case 0x2f:_ILL;	ADD_ABS();		OP_RLA();	break;
/*3*/
	case 0x30:		ADD_REL();		OP_BMI()	break;
	case 0x31:		ADD_IND_Y();	OP_AND();	break;
	case 0x32:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x33:_ILL;	ADD_IND_Y();	OP_RLA();	break;
	case 0x34:_ILL;	ADD_ZP_X();		/*OP_NOP*/	break;
	case 0x35:		ADD_ZP_X();		OP_AND();	break;
	case 0x36:		ADD_ZP_X();		OP_ROL();	break;
	case 0x37:_ILL;	ADD_ZP_X();		OP_RLA();	break;
	case 0x38:		/*ADD_IMP*/		OP_SEC();	break;
	case 0x39:		ADD_ABS_Y();	OP_AND();	break;
	case 0x3a:_ILL;	/*ADD_IMP*/		/*OP_NOP*/	break;
	case 0x3b:_ILL;	ADD_ABS_Y();	OP_RLA();	break;
	case 0x3c:_ILL;	ADD_ABS_X();	/*OP_NOP*/	break;
	case 0x3d:		ADD_ABS_X();	OP_AND();	break;
	case 0x3e:		ADD_ABS_X();	OP_ROL();	break;
	case 0x3f:_ILL;	ADD_ABS_X();	OP_RLA();	break;
/*4*/
	case 0x40:		/*ADD_IMP*/		OP_RTI();	break;
	case 0x41:		ADD_IND_X();	OP_EOR();	break;
	case 0x42:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x43:_ILL;	ADD_IND_X();	OP_SRE();	break;
	case 0x44:_ILL;	ADD_ZP();		/*OP_NOP*/	break;
	case 0x45:		ADD_ZP();		OP_EOR();	break;
	case 0x46:		ADD_ZP();		OP_LSR();	break;
	case 0x47:_ILL;	ADD_ZP();		OP_SRE();	break;
	case 0x48:		/*ADD_IMP*/		OP_PHA();	break;
	case 0x49:		ADD_IMM();		OP_EOR();	break;
	case 0x4a:		/*ADD_ACCUM*/	OP_LSRa();	break;
	case 0x4b:_ILL;	ADD_IMM();		OP_ASR();	break;
	case 0x4c:		ADD_ABS();		OP_JMP();	break;
	case 0x4d:		ADD_ABS();		OP_EOR();	break;
	case 0x4e:		ADD_ABS();		OP_LSR();	break;
	case 0x4f:_ILL;	ADD_ABS();		OP_SRE();	break;
/*5*/
	case 0x50:		ADD_REL();		OP_BVC()	break;
	case 0x51:		ADD_IND_Y();	OP_EOR();	break;
	case 0x52:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x53:_ILL;	ADD_IND_Y();	OP_SRE();	break;
	case 0x54:_ILL;	ADD_ZP_X();		/*OP_NOP*/	break;
	case 0x55:		ADD_ZP_X();		OP_EOR();	break;
	case 0x56:		ADD_ZP_X();		OP_LSR();	break;
	case 0x57:_ILL;	ADD_ZP_X();		OP_SRE();	break;
	case 0x58:		/*ADD_IMP*/		OP_CLI();	break;
	case 0x59:		ADD_ABS_Y();	OP_EOR();	break;
	case 0x5a:_ILL;	/*ADD_IMP*/		/*OP_NOP*/	break;
	case 0x5b:_ILL;	ADD_ABS_Y();	OP_SRE();	break;
	case 0x5c:_ILL;	ADD_ABS_X();	/*OP_NOP*/	break;
	case 0x5d:		ADD_ABS_X();	OP_EOR();	break;
	case 0x5e:		ADD_ABS_X();	OP_LSR();	break;
	case 0x5f:_ILL;	ADD_ABS_X();	OP_SRE();	break;
/*6*/
	case 0x60:		/*ADD_IMP*/		OP_RTS();	break;
	case 0x61:		ADD_IND_X();	OP_ADC();	break;
	case 0x62:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x63:_ILL;	ADD_IND_X();	OP_RRA();	break;
	case 0x64:_ILL;	ADD_ZP();		/*OP_NOP*/	break;
	case 0x65:		ADD_ZP();		OP_ADC();	break;
	case 0x66:		ADD_ZP();		OP_ROR();	break;
	case 0x67:_ILL;	ADD_ZP();		OP_RRA();	break;
	case 0x68:		/*ADD_IMP*/		OP_PLA();	break;
	case 0x69:		ADD_IMM();		OP_ADC();	break;
	case 0x6a:		/*ADD_ACCUM*/	OP_RORa();	break;
	case 0x6b:_ILL;	ADD_IMM();		OP_ARR();	break;
	case 0x6c:		ADD_INDIR();	OP_JMP();	break;
	case 0x6d:		ADD_ABS();		OP_ADC();	break;
	case 0x6e:		ADD_ABS();		OP_ROR();	break;
	case 0x6f:_ILL;	ADD_ABS();		OP_RRA();	break;
/*7*/
	case 0x70:		ADD_REL();		OP_BVS()	break;
	case 0x71:		ADD_IND_Y();	OP_ADC();	break;
	case 0x72:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x73:_ILL;	ADD_IND_Y();	OP_RRA();	break;
	case 0x74:_ILL;	ADD_ZP_X();		/*OP_NOP*/	break;
	case 0x75:		ADD_ZP_X();		OP_ADC();	break;
	case 0x76:		ADD_ZP_X();		OP_ROR();	break;
	case 0x77:_ILL;	ADD_ZP_X();		OP_RRA();	break;
	case 0x78:		/*ADD_IMP*/		OP_SEI();	break;
	case 0x79:		ADD_ABS_Y();	OP_ADC();	break;
	case 0x7a:_ILL;	/*ADD_IMP*/		/*OP_NOP*/	break;
	case 0x7b:_ILL;	ADD_ABS_Y();	OP_RRA();	break;
	case 0x7c:_ILL;	ADD_ABS_X();	/*OP_NOP*/	break;
	case 0x7d:		ADD_ABS_X();	OP_ADC();	break;
	case 0x7e:		ADD_ABS_X();	OP_ROR();	break;
	case 0x7f:_ILL;	ADD_ABS_X();	OP_RRA();	break;
/*8*/
	case 0x80:_ILL;	ADD_IMM();		/*OP_NOP*/	break;
	case 0x81:		ADD_IND_X();	OP_STA();	break;
	case 0x82:_ILL;	ADD_IMM();		/*OP_NOP*/	break;
	case 0x83:_ILL;	ADD_IND_X();	OP_AAX();	break;
	case 0x84:		ADD_ZP();		OP_STY();	break;
	case 0x85:		ADD_ZP();		OP_STA();	break;
	case 0x86:		ADD_ZP();		OP_STX();	break;
	case 0x87:_ILL;	ADD_ZP();		OP_AAX();	break;
	case 0x88:		/*ADD_IMP*/		OP_DEY();	break;
	case 0x89:_ILL;	ADD_IMM();		/*OP_NOP*/	break;
	case 0x8a:		/*ADD_IMP*/		OP_TXA();	break;
	case 0x8b:_ILL;	ADD_IMM();		OP_XAA();	break;
	case 0x8c:		ADD_ABS();		OP_STY();	break;
	case 0x8d:		ADD_ABS();		OP_STA();	break;
	case 0x8e:		ADD_ABS();		OP_STX();	break;
	case 0x8f:_ILL;	ADD_ABS();		OP_AAX();	break;
/*9*/
	case 0x90:		ADD_REL();		OP_BCC()	break;
	case 0x91:		ADD_IND_Y();	OP_STA();	break;
	case 0x92:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0x93:_ILL;	ADD_IND_Y();	OP_AXA();	break;
	case 0x94:		ADD_ZP_X();		OP_STY();	break;
	case 0x95:		ADD_ZP_X();		OP_STA();	break;
	case 0x96:		ADD_ZP_Y();		OP_STX();	break;
	case 0x97:_ILL;	ADD_ZP_Y();		OP_AAX();	break;
	case 0x98:		/*ADD_IMP*/		OP_TYA();	break;
	case 0x99:		ADD_ABS_Y();	OP_STA();	break;
	case 0x9a:		/*ADD_IMP*/		OP_TXS();	break;
	case 0x9b:_ILL;	ADD_ABS_Y();	OP_XAS();	break;
	case 0x9c:_ILL;	ADD_ABS_X();	OP_SYA();	break;
	case 0x9d:		ADD_ABS_X();	OP_STA();	break;
	case 0x9e:_ILL;	ADD_ABS_Y();	OP_SXA();	break;
	case 0x9f:_ILL;	ADD_ABS_Y();	OP_AXA();	break;
/*a*/
	case 0xa0:		ADD_IMM();		OP_LDY();	break;
	case 0xa1:		ADD_IND_X();	OP_LDA();	break;
	case 0xa2:		ADD_IMM();		OP_LDX();	break;
	case 0xa3:_ILL;	ADD_IND_X();	OP_LAX();	break;
	case 0xa4:		ADD_ZP();		OP_LDY();	break;
	case 0xa5:		ADD_ZP();		OP_LDA();	break;
	case 0xa6:		ADD_ZP();		OP_LDX();	break;
	case 0xa7:_ILL;	ADD_ZP();		OP_LAX();	break;
	case 0xa8:		/*ADD_IMP*/		OP_TAY();	break;
	case 0xa9:		ADD_IMM();		OP_LDA();	break;
	case 0xaa:		/*ADD_IMP*/		OP_TAX();	break;
	case 0xab:_ILL;	ADD_IMM();		OP_ATX();	break;
	case 0xac:		ADD_ABS();		OP_LDY();	break;
	case 0xad:		ADD_ABS();		OP_LDA();	break;
	case 0xae:		ADD_ABS();		OP_LDX();	break;
	case 0xaf:_ILL;	ADD_ABS();		OP_LAX();	break;
/*b*/
	case 0xb0:		ADD_REL();		OP_BCS()	break;
	case 0xb1:		ADD_IND_Y();	OP_LDA();	break;
	case 0xb2:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0xb3:_ILL;	ADD_IND_Y();	OP_LAX();	break;
	case 0xb4:		ADD_ZP_X();		OP_LDY();	break;
	case 0xb5:		ADD_ZP_X();		OP_LDA();	break;
	case 0xb6:		ADD_ZP_Y();		OP_LDX();	break;
	case 0xb7:_ILL;	ADD_ZP_Y();		OP_LAX();	break;
	case 0xb8:		/*ADD_IMP*/		OP_CLV();	break;
	case 0xb9:		ADD_ABS_Y();	OP_LDA();	break;
	case 0xba:		/*ADD_IMP*/		OP_TSX();	break;
	case 0xbb:_ILL;	ADD_ABS_Y();	OP_LAR();	break;
	case 0xbc:		ADD_ABS_X();	OP_LDY();	break;
	case 0xbd:		ADD_ABS_X();	OP_LDA();	break;
	case 0xbe:		ADD_ABS_Y();	OP_LDX();	break;
	case 0xbf:_ILL;	ADD_ABS_Y();	OP_LAX();	break;
/*c*/
	case 0xc0:		ADD_IMM();		OP_CPY();	break;
	case 0xc1:		ADD_IND_X();	OP_CMP();	break;
	case 0xc2:_ILL;	ADD_IMM();		/*OP_NOP*/	break;
	case 0xc3:_ILL;	ADD_IND_X();	OP_DCP();	break;
	case 0xc4:		ADD_ZP();		OP_CPY();	break;
	case 0xc5:		ADD_ZP();		OP_CMP();	break;
	case 0xc6:		ADD_ZP();		OP_DEC();	break;
	case 0xc7:_ILL;	ADD_ZP();		OP_DCP();	break;
	case 0xc8:		/*ADD_IMP*/		OP_INY();	break;
	case 0xc9:		ADD_IMM();		OP_CMP();	break;
	case 0xca:		/*ADD_IMP*/		OP_DEX();	break;
	case 0xcb:_ILL;	ADD_IMM();		OP_AXS();	break;
	case 0xcc:		ADD_ABS();		OP_CPY();	break;
	case 0xcd:		ADD_ABS();		OP_CMP();	break;
	case 0xce:		ADD_ABS();		OP_DEC();	break;
	case 0xcf:_ILL;	ADD_ABS();		OP_DCP();	break;
/*d*/
	case 0xd0:		ADD_REL();		OP_BNE()	break;
	case 0xd1:		ADD_IND_Y();	OP_CMP();	break;
	case 0xd2:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0xd3:_ILL;	ADD_IND_Y();	OP_DCP();	break;
	case 0xd4:_ILL;	ADD_ZP_X();		/*OP_NOP*/	break;
	case 0xd5:		ADD_ZP_X();		OP_CMP();	break;
	case 0xd6:		ADD_ZP_X();		OP_DEC();	break;
	case 0xd7:_ILL;	ADD_ZP_X();		OP_DCP();	break;
	case 0xd8:		/*ADD_IMP*/		OP_CLD();	break;
	case 0xd9:		ADD_ABS_Y();	OP_CMP();	break;
	case 0xda:_ILL;	/*ADD_IMP*/		/*OP_NOP*/	break;
	case 0xdb:_ILL;	ADD_ABS_Y();	OP_DCP();	break;
	case 0xdc:_ILL;	ADD_ABS_X();	/*OP_NOP*/	break;
	case 0xdd:		ADD_ABS_X();	OP_CMP();	break;
	case 0xde:		ADD_ABS_X();	OP_DEC();	break;
	case 0xdf:_ILL;	ADD_ABS_X();	OP_DCP();	break;
/*e*/
	case 0xe0:		ADD_IMM();		OP_CPX();	break;
	case 0xe1:		ADD_IND_X();	OP_SBC();	break;
	case 0xe2:_ILL;	ADD_IMM();		/*OP_NOP*/	break;
	case 0xe3:_ILL;	ADD_IND_X();	OP_ISC();	break;
	case 0xe4:		ADD_ZP();		OP_CPX();	break;
	case 0xe5:		ADD_ZP();		OP_SBC();	break;
	case 0xe6:		ADD_ZP();		OP_INC();	break;
	case 0xe7:_ILL;	ADD_ZP();		OP_ISC();	break;
	case 0xe8:		/*ADD_IMP*/		OP_INX();	break;
	case 0xe9:		ADD_IMM();		OP_SBC();	break;
	case 0xea:		/*ADD_IMP*/		/*OP_NOP*/	break;
	case 0xeb:_ILL;	ADD_IMM();		OP_SBC();	break;
	case 0xec:		ADD_ABS();		OP_CPX();	break;
	case 0xed:		ADD_ABS();		OP_SBC();	break;
	case 0xee:		ADD_ABS();		OP_INC();	break;
	case 0xef:_ILL;	ADD_ABS();		OP_ISC();	break;
/*f*/
	case 0xf0:		ADD_REL();		OP_BEQ()	break;
	case 0xf1:		ADD_IND_Y();	OP_SBC();	break;
	case 0xf2:_ILL;	/*ADD_IMP*/		OP_HLT();	break;
	case 0xf3:_ILL;	ADD_IND_Y();	OP_ISC();	break;
	case 0xf4:_ILL;	ADD_ZP_X();		/*OP_NOP*/	break;
	case 0xf5:		ADD_ZP_X();		OP_SBC();	break;
	case 0xf6:		ADD_ZP_X();		OP_INC();	break;
	case 0xf7:_ILL;	ADD_ZP_X();		OP_ISC();	break;
	case 0xf8:		/*ADD_IMP*/		OP_SED();	break;
	case 0xf9:		ADD_ABS_Y();	OP_SBC();	break;
	case 0xfa:_ILL;	/*ADD_IMP*/		/*OP_NOP*/	break;
	case 0xfb:_ILL;	ADD_ABS_Y();	OP_ISC();	break;
	case 0xfc:_ILL;	ADD_ABS_X();	/*OP_NOP*/	break;
	case 0xfd:		ADD_ABS_X();	OP_SBC();	break;
	case 0xfe:		ADD_ABS_X();	OP_INC();	break;
	case 0xff:_ILL;	ADD_ABS_X();	OP_ISC();	break;
	
} /* switch */
	
	#if DEBUG_CPU
	sprintf(in_int,"%s%s",cpu.interrupt?"i":"",cpu.irq_pend?"p":"");
	#endif
	
	if (cpu.interrupt) handle_interrupt[cpu.interrupt]();
	
	#if DEBUG_CPU
	LOG(LOG_CPU,"%s%s%2X %s %20s\tPC:%4X A:%2X X:%2X Y:%2X S:%2X P:%d%d%dB%d%d%d%d ad:%4X\n", \
	op_official[opcode]?"":"!",in_int,opcode,c_op_mnemonic[op_mnemonic[opcode]],addressing_mode_names[op_addressing_mode[opcode]], \
	temp_pc,cpu.A,cpu.X,cpu.Y,cpu.S,((cpu.P_NZ>>8|cpu.P_NZ)&0x80)>>7,cpu.P_V!=0,1,cpu.P_D,cpu.P_I, \
	(cpu.P_NZ&0xff)==0,cpu.P_C>>8,address);
	#endif
	
} /* main loop */
	
	return 0;
}

void cpu_execute_halted(const int border)
{
	while (cpu.cycles>border) {
		update_sync();
		
		#if DEBUG_PPU
		ppu_force_update();
		#endif
		
		/* nothing happens, except addressbus incrementing every cycle */
		
		cpu.cycles-=crystal->cycle;
	}
}
