#include "global.h"
#include "ppu.h"
#include "apu.h"
#include "cont.h"
#include "cpu.h"
#include "cpu_memorymap.h"
#include "cartridge.h"
#include "mapper.h"

/* memory map
	
	$10000
	$c000		PRG bank 2									cpu_read/write_prgbank2a/b
	$8000		PRG bank 1									cpu_read/write_prgbank1a/b
	$6000		WRAM/SRAM(battery)							cpu_read/write_wram
	$4000		sound/general IO, 32 registers.				cpu_read/write_io
	$2000		PPU IO, 8 registers (mirrored 1024 times)	cpu_read/write_ppu
	$0000		RAM, 4kB, mirrored 4 times					cpu_read/write_ram
*/

static u8 cpu_ram[0x800];

/* standard IO handlers */
static fp_cpu_read_io2000 cpu_read_io2000[8]; /* PPU IO (read) */
static fp_cpu_write_io2000 cpu_write_io2000[8]; /* PPU IO (write) */
static fp_cpu_read_io4000 cpu_read_io4000[0x20]; /* sound/general IO (read) */
static fp_cpu_write_io4000 cpu_write_io4000[0x20]; /* sound/general IO (write) */

u8 __fastcall cpu_read_io_nothing(u16 address) { return address>>8; }
void __fastcall cpu_write_io_nothing(u8 data) { return; }

void cpu_set_read_io2000(u16 address,fp_cpu_read_io2000 fun) { if (fun==NULL) fun=ppu_read_openbus; cpu_read_io2000[address&7]=fun; }
void cpu_set_write_io2000(u16 address,fp_cpu_write_io2000 fun) { if (fun==NULL) fun=cpu_write_io_nothing; cpu_write_io2000[address&7]=fun; }
void cpu_set_read_io4000(u16 address,fp_cpu_read_io4000 fun) { if (fun==NULL) fun=cpu_read_io_nothing; cpu_read_io4000[address&0x1f]=fun; }
void cpu_set_write_io4000(u16 address,fp_cpu_write_io4000 fun) { if (fun==NULL) fun=cpu_write_io_nothing; cpu_write_io4000[address&0x1f]=fun; }
fp_cpu_write_io2000 __fastcall cpu_get_write_io2000(u16 address) { return cpu_write_io2000[address&7]; }

void cpu_set_io2000_std(void)
{
	u16 address;
	
	/* read */
	for (address=0x2000;address<0x2008;address++) cpu_set_read_io2000(address,NULL);
	cpu_set_read_io2000(0x2002,ppu_read_status);
	cpu_set_read_io2000(0x2004,ppu_read_sprite_memory_data);
	cpu_set_read_io2000(0x2007,ppu_read_memory_data);
	
	/* write */
	for (address=0x2000;address<0x2008;address++) cpu_set_write_io2000(address,NULL);
	cpu_set_write_io2000(0x2000,ppu_write_control1);
	cpu_set_write_io2000(0x2001,ppu_write_control2);
	cpu_set_write_io2000(0x2003,ppu_write_sprite_memory_address);
	cpu_set_write_io2000(0x2004,ppu_write_sprite_memory_data);
	cpu_set_write_io2000(0x2005,ppu_write_scroll);
	cpu_set_write_io2000(0x2006,ppu_write_memory_address);
	cpu_set_write_io2000(0x2007,ppu_write_memory_data);
}
void cpu_set_io4000_std(void)
{
	u16 address;

	/* read */
	for (address=0x4000;address<0x4020;address++) cpu_set_read_io4000(address,NULL);
	cpu_set_read_io4000(0x4015,apu_read_status);
	cpu_set_read_io4000(0x4016,cont_read_port);
	cpu_set_read_io4000(0x4017,cont_read_port);
	
	/* write */
	for (address=0x4000;address<0x4020;address++) cpu_set_write_io4000(address,NULL);
	
	cpu_set_write_io4000(0x4000,apu_write_square0_control);
	cpu_set_write_io4000(0x4001,apu_write_square0_sweep);
	cpu_set_write_io4000(0x4002,apu_write_square0_timer);
	cpu_set_write_io4000(0x4003,apu_write_square0_length);

	cpu_set_write_io4000(0x4004,apu_write_square1_control);
	cpu_set_write_io4000(0x4005,apu_write_square1_sweep);
	cpu_set_write_io4000(0x4006,apu_write_square1_timer);
	cpu_set_write_io4000(0x4007,apu_write_square1_length);
	
	cpu_set_write_io4000(0x4008,apu_write_triangle_linear);
	cpu_set_write_io4000(0x400a,apu_write_triangle_timer);
	cpu_set_write_io4000(0x400b,apu_write_triangle_length);
	
	cpu_set_write_io4000(0x400c,apu_write_noise_envelope);
	cpu_set_write_io4000(0x400e,apu_write_noise_timer);
	cpu_set_write_io4000(0x400f,apu_write_noise_length);
	
	cpu_set_write_io4000(0x4010,apu_write_dmc_control);
	cpu_set_write_io4000(0x4011,apu_write_dmc_dac);
	cpu_set_write_io4000(0x4012,apu_write_dmc_address);
	cpu_set_write_io4000(0x4013,apu_write_dmc_length);
	
	cpu_set_write_io4000(0x4014,cpu_sprite_dma_transfer);
	cpu_set_write_io4000(0x4015,apu_write_status);
	cpu_set_write_io4000(0x4016,cont_write_port);
	cpu_set_write_io4000(0x4017,apu_write_fs);
}

/* mapper IO handlers */
static fp_cpu_read_io4020 cpu_read_io4020;		/* mapper IO at $4020-$5fff (read) */
static fp_cpu_write_io4020 cpu_write_io4020;	/* mapper IO at $4020-$5fff (write) */
static fp_cpu_read_io6000 cpu_read_io6000;		/* mapper IO at $6000-$7fff (read) */
static fp_cpu_write_io6000 cpu_write_io6000;	/* mapper IO at $6000-$7fff (write) */
static fp_cpu_read_io8000 cpu_read_io8000[4];	/* mapper IO at $8000-$ffff (read), in chunks of $2000 */
static fp_cpu_write_io8000 cpu_write_io8000[4];	/* mapper IO at $8000-$ffff (write), in chunks of $2000 */

void cpu_set_read_io4020(fp_cpu_read_io4020 fun) { cpu_read_io4020=fun; }
void cpu_set_write_io4020(fp_cpu_write_io4020 fun) { cpu_write_io4020=fun; }
void cpu_set_read_io6000(fp_cpu_read_io6000 fun)
{
	if (fun!=MAPIO_R_FUN(nothing)) {
		cpu_read_io6000=fun;
		cpu_set_fpt(CPU_FPT_WRAM_READ);
	}
}
void cpu_set_write_io6000(fp_cpu_write_io6000 fun)
{
	if (fun!=MAPIO_FUN(nothing)) {
		cpu_write_io6000=fun;
		cpu_set_fpt(CPU_FPT_WRAM_WRITE);
	}
}
void cpu_set_read_io8000(fp_cpu_read_io8000 fun1,fp_cpu_read_io8000 fun2,fp_cpu_read_io8000 fun3,fp_cpu_read_io8000 fun4)
{
	if (fun1!=MAPIO_R_FUN(nothing)) { cpu_read_io8000[0]=fun1; cpu_set_fpt(CPU_FPT_PRG1A_READ); }
	if (fun2!=MAPIO_R_FUN(nothing)) { cpu_read_io8000[1]=fun2; cpu_set_fpt(CPU_FPT_PRG1B_READ); }
	if (fun3!=MAPIO_R_FUN(nothing)) { cpu_read_io8000[2]=fun3; cpu_set_fpt(CPU_FPT_PRG2A_READ); }
	if (fun4!=MAPIO_R_FUN(nothing)) { cpu_read_io8000[3]=fun4; cpu_set_fpt(CPU_FPT_PRG2B_READ); }
}
void cpu_set_write_io8000(fp_cpu_write_io8000 fun1,fp_cpu_write_io8000 fun2,fp_cpu_write_io8000 fun3,fp_cpu_write_io8000 fun4)
{
	cpu_write_io8000[0]=fun1; cpu_write_io8000[1]=fun2; cpu_write_io8000[2]=fun3; cpu_write_io8000[3]=fun4;
}


void cpu_memmap_init(void)
{
	cpu_memmap_fill_ram(0xff);
/*	confirmed behaviour: $ff. CopyNES, and probably old FFE devices too, empty RAM at start
	(some roms could have been extracted from x-in-1 pirate carts, that emptied ram at init)
	
	problems with 0:
	- Final Fantasy 1: Disch: problem with 'random' encounters (offset differs by 1: 0-ff=1)
	- Minna no Taabou no Nakayoshi Daisakusen: locks up at boot
	
	problems with $ff:
	- pputime2: block moves to the right
	- Sayoonara: no sound ([a1] works fine)
	- Scroll Demo (PD): won't work
	- smwstomp: no sound
	- Sokoban (PD): crashes
	- Stars Demo (not SE): crashes
	- Super Mario Bros (several hacked versions): start at -1 world .. (A+start to start at the correct world)
*/
	cpu_set_io2000_std();
	cpu_set_io4000_std();
	cpu_set_fpt(CPU_FPT_NORMAL);
	
	LOG(LOG_VERBOSE,"CPU memorymap initialised\n");
}

void cpu_memmap_fill_ram(u8 b)
{
	memset(cpu_ram,b,0x800);
}

void cpu_memmap_clean(void)
{
	LOG(LOG_VERBOSE,"CPU memorymap cleaned\n");
}


/* ram */
u8 __fastcall cpu_read_ram(u16 address)
{
	#if DEBUG_CPU_MMAP
	LOG(LOG_CPU_MMAP,"cpu ramread ad: %x data: %x\n",address&0x7ff,cpu_ram[address&0x7ff]);
	#endif
	
	return cpu_ram[address&0x7ff];
}

void __fastcall cpu_write_ram(u16 address,u8 data)
{
	cpu_ram[address&0x7ff]=data;
	
	#if DEBUG_CPU_MMAP
	LOG(LOG_CPU_MMAP,"cpu ramwrite ad: %x data: %x\n",address&0x7ff, data);
	#endif
}

/* ppu */
u8 __fastcall cpu_read_ppu(u16 address) { return cpu_read_io2000[address&7](address); }
void __fastcall cpu_write_ppu(u16 address,u8 data) { cpu_write_io2000[address&7](data); }

/* io */
u8 __fastcall cpu_read_io(u16 address)
{
	if (address<0x4020) return cpu_read_io4000[address&0x1f](address);
	else return cpu_read_io4020(address);
}

void __fastcall cpu_write_io(u16 address,u8 data)
{
	if (address<0x4020) cpu_write_io4000[address&0x1f](data);
	else cpu_write_io4020(address,0,data);
}

/* wram */
u8 __fastcall cpu_read_wram(u16 address) { if (cartridge->wram_isenabled) return cartridge->cur_wrambank[address&0x1fff]; else return address>>8; }
u8 __fastcall cpu_read_wram_override(u16 address) { return cpu_read_io6000(address); }
void __fastcall cpu_write_wram(u16 address,u8 data) { if (cartridge->wram_isenabled&cartridge->wram_writeenable) cartridge->cur_wrambank[address&0x1fff]=data; }
void __fastcall cpu_write_wram_override(u16 address,u8 data) { cpu_write_io6000(address,cartridge->cur_wrambank[address&0x1fff],data); }

/* --- prg/rom --- */
#define PRGREAD(x) cartridge->cur_prgbank[x][address&0x1fff]
#define LOG_PRGREAD(x) LOG(LOG_CPU_MMAP,"cpu prgbank%dread ad: %x(%x) data: %x\n",x,address,address&0x1fff,PRGREAD(x))
#define PRGWRITE(x) cpu_write_io8000[x](address,PRGREAD(x),data)
#define PRGREAD_OVER(x) return cpu_read_io8000[x](address)

/* read */
u8 __fastcall cpu_read_prgbank1a(u16 address) /* $8000 - $9fff */
{
	#if DEBUG_CPU_MMAP
	LOG_PRGREAD(0);
	#endif
	
	return PRGREAD(0);
}
u8 __fastcall cpu_read_prgbank1b(u16 address) /* $a000 - $bfff */
{
	#if DEBUG_CPU_MMAP
	LOG_PRGREAD(1);
	#endif
	
	return PRGREAD(1);
}
u8 __fastcall cpu_read_prgbank2a(u16 address) /* $c000 - $dfff */
{
	#if DEBUG_CPU_MMAP
	LOG_PRGREAD(2);
	#endif
	
	return PRGREAD(2);
}
u8 __fastcall cpu_read_prgbank2b(u16 address) /* $e000 - $ffff */
{
	#if DEBUG_CPU_MMAP
	LOG_PRGREAD(3);
	#endif
	
	return PRGREAD(3);
}
u8 __fastcall cpu_read_prgbank1a_override(u16 address) { PRGREAD_OVER(0); } /* $8000 - $9fff */
u8 __fastcall cpu_read_prgbank1b_override(u16 address) { PRGREAD_OVER(1); } /* $a000 - $bfff */
u8 __fastcall cpu_read_prgbank2a_override(u16 address) { PRGREAD_OVER(2); } /* $c000 - $dfff */
u8 __fastcall cpu_read_prgbank2b_override(u16 address) { PRGREAD_OVER(3); } /* $e000 - $ffff */

/* write (mapper io usually) */
void __fastcall cpu_write_prgbank1a(u16 address,u8 data) { PRGWRITE(0); } /* $8000 - $9fff */
void __fastcall cpu_write_prgbank1b(u16 address,u8 data) { PRGWRITE(1); } /* $a000 - $bfff */
void __fastcall cpu_write_prgbank2a(u16 address,u8 data) { PRGWRITE(2); } /* $c000 - $dfff */
void __fastcall cpu_write_prgbank2b(u16 address,u8 data) { PRGWRITE(3); } /* $e000 - $ffff */
