#include "global.h"
#include "crystal.h"
#include "draw.h"
#include "main.h"
#include "resource.h"
#include "settings.h"
#include "cpu.h"
#include "nes.h"
#include "input.h"

/*
NES NTSC crystal runs at 236250000/11=~21477272.72Hz
NES NTSC PPU runs at crystal/4=~5369318.18Hz
NES NTSC CPU runs at crystal/12=~1789772.72Hz
NES PAL crystal runs at 26601712.5Hz
NES PAL CPU runs at crystal/16=1662607.03125Hz
NES PAL PPU runs at crystal/5=5320342.5Hz

In NTSC 3 PPU cycles take 1 CPU cycle, in PAL 3.2 PPU cycles take 1 CPU cycle.
One scanline is 256 pixel cycles and 85 hblank cycles, that makes 341 PPU cycles.
There are 262 scanlines per frame for NTSC, and 312 for PAL.

The NES does not use the default NTSC(60/1.001=~59.94Hz) or PAL(50Hz) refresh rate.
NTSC refresh rate: ((236250000/11)/12)/((262*341)/3)=~60.098Hz (not counting odd frames)
PAL refresh rate: (26601712.5/16)/((312*341)/3.2)=~50.007Hz

timing, with sound: create 1 sample every cpuspeed/44100 cycle,
NTSC: 40+45/77 (~40.584416)
PAL: 37+13187/18816 (~37.700839)
using base factors (5,15,16) to get accurate PAL/NTSC timing, idea from Disch

exact 60hz: 40+2281/4410 (~40.51723) (taking even/odd frames into account, [~samples(cycles) per frame]/(44100/60))
exact 50hz: 37+409/588 (~37.695578) ([spf]/(44100/50))
~60.01hz (my monitor :P ): 40+3244/6191
*/

#define PPU_CYCLE				5
#define PPU_SCANLINE			(PPU_CYCLE*256)
#define PPU_HBLANK				(PPU_CYCLE*85)
#define PPU_FULL_SCANLINE		(PPU_SCANLINE+PPU_HBLANK)

#define NTSC_LINES				262
#define NTSC_CYCLE				(PPU_CYCLE*3)
#define NTSC_FRAME				(PPU_FULL_SCANLINE*NTSC_LINES)
#define NTSC_VBLANK_TRIGGER		(NTSC_FRAME-(PPU_FULL_SCANLINE*242))

#define PAL_LINES				312
#define PAL_CYCLE				16 /*(PPU_CYCLE*3.2)*/
#define PAL_FRAME				(PPU_FULL_SCANLINE*PAL_LINES)
#define PAL_VBLANK_TRIGGER		(PAL_FRAME-(PPU_FULL_SCANLINE*242))

#define WIN_SOUND_NTSC_SLICE_1	40
#define WIN_SOUND_NTSC_SLICE_2	41
#define WIN_SOUND_NTSC_BASE		77
#define WIN_SOUND_NTSC_SUB		45
#define WIN_SOUND_NTSC_CSPF		734

#define WIN_SOUND_NTSCs_SLICE_1	40
#define WIN_SOUND_NTSCs_SLICE_2	41
#if 0
#define WIN_SOUND_NTSCs_BASE	6191
#define WIN_SOUND_NTSCs_SUB		3244
#else
#define WIN_SOUND_NTSCs_BASE	4410
#define WIN_SOUND_NTSCs_SUB		2281
#endif
#define WIN_SOUND_NTSCs_CSPF	735

#define WIN_SOUND_PAL_SLICE_1	37
#define WIN_SOUND_PAL_SLICE_2	38
#define WIN_SOUND_PAL_BASE		18816
#define WIN_SOUND_PAL_SUB		13187
#define WIN_SOUND_PAL_CSPF		882

#define WIN_SOUND_PALs_SLICE_1	37
#define WIN_SOUND_PALs_SLICE_2	38
#define WIN_SOUND_PALs_BASE		588
#define WIN_SOUND_PALs_SUB		409
#define WIN_SOUND_PALs_CSPF		882

#define OC_UPSCALE				128

void crystal_init(void)
{
	MEM_CREATE(crystal,sizeof(Crystal));
	
	crystal->ppu_cycle=PPU_CYCLE*OC_UPSCALE;
	crystal->ppu_scanline=PPU_SCANLINE*OC_UPSCALE;
	crystal->hblank=PPU_HBLANK*OC_UPSCALE;
	crystal->full_scanline=PPU_FULL_SCANLINE*OC_UPSCALE;
	crystal->max_cycles=PAL_FRAME/PAL_CYCLE; /* used to calculate sound buffer size */
	
	LOG(LOG_VERBOSE,"crystal initialised\n");
}

void crystal_settings_load(void)
{
	int i;
	
	crystal->overclock=100;
	if (SETTINGS_GET_INT(settings_info(SETTINGS_CPUSPEED),&i)) { CLAMP(i,25,400); crystal->overclock=i; }
}

void crystal_clean(void)
{
	LOG(LOG_VERBOSE,"crystal cleaned\n");
	
	MEM_CLEAN(crystal);
}

void crystal_set_cpuspeed(int p)
{
	float oc;
	
	if (p<25) p=25; if (p>400) p=400;
	if (p!=crystal->overclock) {
		crystal->overclock=p;
		LOG(LOG_MISC|LOG_COLOUR(LC_GREEN)|LOG_TYPE(LT_CPUSPEED),"%d%% 2A03 clock speed   ",p);
		if ((p%50)==0) input_cpuspeed_ticks_stall();
	}
	
	oc=crystal->overclock;
	oc=(float)OC_UPSCALE*(100.0/oc);
	
	crystal->cycle=(crystal->mode?PAL_CYCLE:NTSC_CYCLE)*oc;
	
	if (nes_is_running()) {
		cpu_fill_op_cycles_lookup();
		cpu_adjust_rel_cycles();
	}
}

void crystal_set_mode(int pal)
{
	int i,t=0;
	char c[STRING_SIZE]={0};
	
	if (draw_get_vidformat()!=DRAW_VIDFORMAT_AUTO) pal=draw_get_vidformat();
	crystal->mode=pal;
	LOG(LOG_VERBOSE,"crystal mode set to %s\n",pal?"PAL":"NTSC");
	
	crystal_set_cpuspeed(crystal->overclock);
	crystal->rel_cycle=(pal?PAL_CYCLE:NTSC_CYCLE)*OC_UPSCALE;
	crystal->lines=pal?PAL_LINES:NTSC_LINES;
	crystal->frame=(pal?PAL_FRAME:NTSC_FRAME)*OC_UPSCALE;
	crystal->vblank_trigger=(pal?PAL_VBLANK_TRIGGER:NTSC_VBLANK_TRIGGER)*OC_UPSCALE;
	crystal->odd_frame=pal?FALSE:FALSE; /* initial frame is odd(1) */
	
	if (draw_get_5060_auto()) {
		int mf=draw_get_refreshrate();
		if (mf>0) { t=mf%(pal?50:60); t=t==0; }
	}
	
	if (t|draw_get_5060()) {
		crystal->slice[0]=pal?WIN_SOUND_PALs_SLICE_1:WIN_SOUND_NTSCs_SLICE_1;
		crystal->slice[1]=pal?WIN_SOUND_PALs_SLICE_2:WIN_SOUND_NTSCs_SLICE_2;
		crystal->base=pal?WIN_SOUND_PALs_BASE:WIN_SOUND_NTSCs_BASE;
		crystal->sub=pal?WIN_SOUND_PALs_SUB:WIN_SOUND_NTSCs_SUB;
		crystal->cspf=pal?WIN_SOUND_PALs_CSPF:WIN_SOUND_NTSCs_CSPF;
	}
	else {
		crystal->slice[0]=pal?WIN_SOUND_PAL_SLICE_1:WIN_SOUND_NTSC_SLICE_1;
		crystal->slice[1]=pal?WIN_SOUND_PAL_SLICE_2:WIN_SOUND_NTSC_SLICE_2;
		crystal->base=pal?WIN_SOUND_PAL_BASE:WIN_SOUND_NTSC_BASE;
		crystal->sub=pal?WIN_SOUND_PAL_SUB:WIN_SOUND_NTSC_SUB;
		crystal->cspf=pal?WIN_SOUND_PAL_CSPF:WIN_SOUND_NTSC_CSPF;
	}
	
	for (i=0;i<313;i++) crystal->hc[i]=crystal->sc[i]=-crystal->frame;
	for (i=0;i<=crystal->lines;i++) {
		crystal->hc[i]=crystal->frame-(i*crystal->full_scanline)-crystal->ppu_scanline;
		crystal->sc[i]=crystal->frame-(i*crystal->full_scanline);
	}
	
	/* update menu */
	main_menu_caption(MAIN_MENUCAPTION_GET,IDM_VIDEOFORMAT,c);
	if (crystal->mode) {  c[11]='N'; c[12]='T'; c[13]='S'; c[14]='C'; }
	else { c[11]='P'; c[12]='A'; c[13]='L'; c[14]=' '; }
	main_menu_caption(MAIN_MENUCAPTION_PUT,IDM_VIDEOFORMAT,c);
}

void crystal_new_frame(void)
{
	crystal->odd_frame^=!crystal->mode;
	crystal->fc++;
	crystal->fc_h+=(crystal->fc==0);
}
