#define DIRECTINPUT_VERSION 0x500
#include <dinput.h>

#include "global.h"
#include "main.h"
#include "resource.h"
#include "input.h"
#include "nes.h"
#include "cont.h"
#include "cpu.h"
#include "crystal.h"
#include "draw.h"
#include "sound.h"
#include "settings.h"
#include "version.h"


#define INPUT_T_MAX				(CONT_T_MAX+INPUT_T_END)
#define INPUT_DA_EM				CONT_DA_MAX

#define INPUT_TT_FOREGROUND		(input.num_triggers-4)
#define INPUT_TA_FOREGROUND		(input.num_axes-4)
#define INPUT_TT_ACTIVE			(input.num_triggers-3)
#define INPUT_TA_ACTIVE			(input.num_axes-3)
#define INPUT_TT_NOTHING		(input.num_triggers-2)
#define INPUT_TA_NOTHING		(input.num_axes-2)

#define T_KEYBOARD_SIZE			256	/* number of keyboard triggers */
#define A_KEYBOARD_SIZE			0	/* axes, etc */
#define T_MOUSE_SIZE			10
#define A_MOUSE_SIZE			3
#define T_JOYSTICK_SIZE			64
#define A_JOYSTICK_SIZE			8

#define T_KEYBOARD_A_ALT		(T_KEYBOARD_SIZE)
#define T_KEYBOARD_A_CTRL		(T_KEYBOARD_SIZE+1)
#define T_KEYBOARD_A_SHIFT		(T_KEYBOARD_SIZE+2)
#define T_KEYBOARD_A_ENTER		(T_KEYBOARD_SIZE+3)
#define T_KEYBOARD_A_CIRCUMFLEX	(T_KEYBOARD_SIZE+4)
#define T_KEYBOARD_SIZE_TOTAL	(T_KEYBOARD_SIZE+5)

static struct {
	LPDIRECTINPUT dinput;
	int semaphore;
	int background;
	int updfreq;
	u32 updframe;
	DWORD updticks;
	int updcycles;
	int vblank_updated;
	int menukey;
	int skey_set[CONT_DA_MAX+2]; /* -2:emulator shortcuts, -1: empty */
	
	int* trigger_rule[INPUT_T_MAX];
	int* trigger; /* $80 or 0 */
	int* trigger_old;
	char** trigger_info;
	int num_triggers;
	
	long* axis;
	long* axis_old;
	char** axis_info;
	int num_axes;
	
	/* keyboard */
	int num_keyboards;
	
	/* mouse */
	int num_mouses; /* .. I mean mice */
	DWORD mouse_cursor_ticks;
	int mouse_cursor_dc;
	int mouse_cursor_visible;
	int doubleclick;
	DWORD doubleclick_ticks;
	
	/* joystick */
	int num_joysticks;
	int joystick_isenabled;
	
	/* shortcuts */
	int* emshort;
	int* emshort_old;
	DWORD cpuspeed_ticks;
	DWORD cpuspeed_ticks_next;
	DWORD cartswitch_ticks;
	DWORD cartswitch_ticks_next;
	
} input;

typedef struct input_device {
	LPDIRECTINPUTDEVICE pd;
	char* name;
	struct input_device* next;
	struct input_device* prev;
} input_device;

static input_device* input_device_keyboard_cursor;
static input_device* input_device_mouse_cursor;
static input_device* input_device_joystick_cursor;
static input_device* input_device_keyboard_begin;
static input_device* input_device_mouse_begin;
static input_device* input_device_joystick_begin;


static void enumerate_devices(void);
static void input_fill_info(void);

static BOOL CALLBACK keyboard_create(LPCDIDEVICEINSTANCE,LPVOID);
static void keyboard_clean(void);
static void keyboard_update(void);

static BOOL CALLBACK mouse_create(LPCDIDEVICEINSTANCE,LPVOID);
static void mouse_clean(void);
static void mouse_update(void);

static BOOL CALLBACK joystick_create(LPCDIDEVICEINSTANCE,LPVOID);
static void joystick_clean(void);
static void joystick_update(void);


/* trigger/axis info */

static void info_string(int type,int n,const char* c)
{
	const int sc=strlen(c)+1;
	const int trigger_keyboard_size=T_KEYBOARD_SIZE_TOTAL;
	const int trigger_keyboard_start=0;
	const int trigger_mouse_size=T_MOUSE_SIZE;
	const int trigger_mouse_start=input.num_keyboards*trigger_keyboard_size;
	const int axis_mouse_size=A_MOUSE_SIZE;
	const int axis_mouse_start=0;
	const int trigger_joystick_size=T_JOYSTICK_SIZE;
	const int trigger_joystick_start=trigger_mouse_start+(input.num_mouses*trigger_mouse_size);
	const int axis_joystick_size=A_JOYSTICK_SIZE;
	const int axis_joystick_start=input.num_mouses*axis_mouse_size;
	int i;
	
	/* create info string */
	#define INFO_CREATE(dev,str,ta) \
		MEM_CREATE(input.ta##_info[n+ta##_##dev##_start+(i*ta##_##dev##_size)],4+sc); \
		sprintf(input.ta##_info[n+ta##_##dev##_start+(i*ta##_##dev##_size)],"%s%d %s",str,i+1,c)
	#define INFO_LOOP(dev,str,ta) \
		i=input.num_##dev##s; \
		while (i--) { INFO_CREATE(dev,str,ta); } \
		break
	
	switch (type) {
		/* keyboard_trigger_string */
		case 0:
			i=input.num_keyboards;
			while (i--) {
				if (input.trigger_info[n+(i*trigger_keyboard_size)]==NULL) {
					if (i) { INFO_CREATE(keyboard,"K",trigger); }
					else {
						MEM_CREATE(input.trigger_info[n],sc);
						sprintf(input.trigger_info[n],"%s",c);
					}
				}
			}
			break;
		
		case 1: INFO_LOOP(mouse,"M",trigger); /* mouse_trigger_string */
		case 2: INFO_LOOP(mouse,"M",axis); /* mouse_axis_string */
		case 3: INFO_LOOP(joystick,"J",trigger); /* joystick_trigger_string */
		case 4: INFO_LOOP(joystick,"J",axis); /* joystick_axis_string */
		
		default: break;
	}
}

/* keyboard */
static void keyboard_fill_info(void)
{
	int ku;
	char kc[12];
	
	/* std */
	info_string(0,DIK_A,			"A");
	info_string(0,DIK_B,			"B");
	info_string(0,DIK_C,			"C");
	info_string(0,DIK_D,			"D");
	info_string(0,DIK_E,			"E");
	info_string(0,DIK_F,			"F");
	info_string(0,DIK_G,			"G");
	info_string(0,DIK_H,			"H");
	info_string(0,DIK_I,			"I");
	info_string(0,DIK_J,			"J");
	info_string(0,DIK_K,			"K");
	info_string(0,DIK_L,			"L");
	info_string(0,DIK_M,			"M");
	info_string(0,DIK_N,			"N");
	info_string(0,DIK_O,			"O");
	info_string(0,DIK_P,			"P");
	info_string(0,DIK_Q,			"Q");
	info_string(0,DIK_R,			"R");
	info_string(0,DIK_S,			"S");
	info_string(0,DIK_T,			"T");
	info_string(0,DIK_U,			"U");
	info_string(0,DIK_V,			"V");
	info_string(0,DIK_W,			"W");
	info_string(0,DIK_X,			"X");
	info_string(0,DIK_Y,			"Y");
	info_string(0,DIK_Z,			"Z");
	
	info_string(0,DIK_0,			"0");
	info_string(0,DIK_1,			"1");
	info_string(0,DIK_2,			"2");
	info_string(0,DIK_3,			"3");
	info_string(0,DIK_4,			"4");
	info_string(0,DIK_5,			"5");
	info_string(0,DIK_6,			"6");
	info_string(0,DIK_7,			"7");
	info_string(0,DIK_8,			"8");
	info_string(0,DIK_9,			"9");
	
	info_string(0,DIK_ESCAPE,		"Escape");
	info_string(0,DIK_F1,			"F1");
	info_string(0,DIK_F2,			"F2");
	info_string(0,DIK_F3,			"F3");
	info_string(0,DIK_F4,			"F4");
	info_string(0,DIK_F5,			"F5");
	info_string(0,DIK_F6,			"F6");
	info_string(0,DIK_F7,			"F7");
	info_string(0,DIK_F8,			"F8");
	info_string(0,DIK_F9,			"F9");
	info_string(0,DIK_F10,			"F10");
	info_string(0,DIK_F11,			"F11");
	info_string(0,DIK_F12,			"F12");
	
	info_string(0,DIK_GRAVE,		"`");
	info_string(0,DIK_MINUS,		"-");
	info_string(0,DIK_EQUALS,		"=");
	info_string(0,DIK_BACK,			"Backspace");
	
	info_string(0,DIK_TAB,			"Tab");
	info_string(0,DIK_LBRACKET,		"[");
	info_string(0,DIK_RBRACKET,		"]");
	info_string(0,DIK_BACKSLASH,	"\\");
	
	info_string(0,DIK_CAPITAL,		"Caps Lock");
	info_string(0,DIK_SEMICOLON,	";");
	info_string(0,DIK_APOSTROPHE,	"'");
	info_string(0,DIK_RETURN,		"Main Enter");
	
	info_string(0,DIK_LSHIFT,		"lShift");
	info_string(0,DIK_COMMA,		",");
	info_string(0,DIK_PERIOD,		".");
	info_string(0,DIK_SLASH,		"/");
	info_string(0,DIK_RSHIFT,		"rShift");
	
	info_string(0,DIK_LCONTROL,		"lCtrl");
	info_string(0,DIK_LWIN,			"lWindows");
	info_string(0,DIK_LMENU,		"lAlt");
	info_string(0,DIK_SPACE,		"Space");
	info_string(0,DIK_RMENU,		"rAlt");
	info_string(0,DIK_RWIN,			"rWindows");
	info_string(0,DIK_APPS,			"AppMenu");
	info_string(0,DIK_RCONTROL,		"rCtrl");
	
	info_string(0,DIK_SYSRQ,		"PrtScr"); /* Print Screen */
	info_string(0,DIK_SCROLL,		"Scroll Lock");
	info_string(0,DIK_PAUSE,		"Pause"); /* single pulse, can't seem to get held state */
	
	info_string(0,DIK_INSERT,		"Insert");
	info_string(0,DIK_DELETE,		"Delete");
	info_string(0,DIK_HOME,			"Home");
	info_string(0,DIK_END,			"End");
	info_string(0,DIK_PRIOR,		"Page Up");
	info_string(0,DIK_NEXT,			"Page Down");
	
	info_string(0,DIK_UP,			"Up");
	info_string(0,DIK_DOWN,			"Down");
	info_string(0,DIK_LEFT,			"Left");
	info_string(0,DIK_RIGHT,		"Right");
	
	info_string(0,DIK_NUMPAD0,		"Numpad 0");
	info_string(0,DIK_NUMPAD1,		"Numpad 1");
	info_string(0,DIK_NUMPAD2,		"Numpad 2");
	info_string(0,DIK_NUMPAD3,		"Numpad 3");
	info_string(0,DIK_NUMPAD4,		"Numpad 4");
	info_string(0,DIK_NUMPAD5,		"Numpad 5");
	info_string(0,DIK_NUMPAD6,		"Numpad 6");
	info_string(0,DIK_NUMPAD7,		"Numpad 7");
	info_string(0,DIK_NUMPAD8,		"Numpad 8");
	info_string(0,DIK_NUMPAD9,		"Numpad 9");
	
	info_string(0,DIK_NUMLOCK,		"Num Lock");
	info_string(0,DIK_DIVIDE,		"Numpad /");
	info_string(0,DIK_MULTIPLY,		"Numpad *");
	info_string(0,DIK_SUBTRACT,		"Numpad -");
	info_string(0,DIK_ADD,			"Numpad +");
	info_string(0,DIK_DECIMAL,		"Numpad .");
	info_string(0,DIK_NUMPADENTER,	"Numpad Enter");
	
	info_string(0,DIK_POWER,		"Power");
	info_string(0,DIK_SLEEP,		"Sleep");
	info_string(0,DIK_WAKE,			"Wake");
	
	/* non-standard keyboards */
	info_string(0,DIK_OEM_102,		"<"); /* or \ */
	info_string(0,DIK_ABNT_C1,		"/ [Brazil]");
	info_string(0,DIK_ABNT_C2,		"Numpad . [Brazil]");
	info_string(0,DIK_F13,			"F13");
	info_string(0,DIK_F14,			"F14");
	info_string(0,DIK_F15,			"F15");
	info_string(0,DIK_YEN,			"Yen");
	info_string(0,DIK_AT,			"@");
	info_string(0,DIK_COLON,		":");
	info_string(0,DIK_UNDERLINE,	"_");
	info_string(0,DIK_STOP,			"Stop");
	info_string(0,DIK_AX,			"AX");
	info_string(0,DIK_UNLABELED,	"J3100");
	info_string(0,DIK_KANA,			"Kana");
	info_string(0,DIK_KANJI,		"Kanji");
	info_string(0,DIK_CONVERT,		"Convert");
	info_string(0,DIK_NOCONVERT,	"No Convert");
	info_string(0,DIK_NUMPADEQUALS,	"Numpad =");
	info_string(0,DIK_NUMPADCOMMA,	"Numpad ,");
	
	/* multimedia (Windows) keys */
	info_string(0,DIK_MYCOMPUTER,	"My Computer");
	info_string(0,DIK_MAIL,			"Mail");
	info_string(0,DIK_CALCULATOR,	"Calculator");
	info_string(0,DIK_PREVTRACK,	"Previous Track"); /* or ^ on Japanese keyboards; DIK_CIRCUMFLEX */
	info_string(0,DIK_NEXTTRACK,	"Next Track");
	info_string(0,DIK_MEDIASELECT,	"Media Select");
	info_string(0,DIK_MEDIASTOP,	"Media Stop");
	info_string(0,DIK_PLAYPAUSE,	"Play");
	info_string(0,DIK_MUTE,			"Mute");
	info_string(0,DIK_VOLUMEDOWN,	"Volume -");
	info_string(0,DIK_VOLUMEUP,		"Volume +");
	info_string(0,DIK_WEBHOME,		"Web Home");
	info_string(0,DIK_WEBSTOP,		"Web Stop");
	info_string(0,DIK_WEBSEARCH,	"Search");
	info_string(0,DIK_WEBFAVORITES,	"Favourites");
	info_string(0,DIK_WEBBACK,		"Back");
	info_string(0,DIK_WEBFORWARD,	"Forward");
	info_string(0,DIK_WEBREFRESH,	"Refresh");
	
	/* custom double keys */
	info_string(0,T_KEYBOARD_A_ALT,"Alt");
	info_string(0,T_KEYBOARD_A_CTRL,"Ctrl");
	info_string(0,T_KEYBOARD_A_SHIFT,"Shift");
	info_string(0,T_KEYBOARD_A_ENTER,"Enter");
	info_string(0,T_KEYBOARD_A_CIRCUMFLEX,"^");
	
	/* the rest is unknown */
	for (ku=0;ku<T_KEYBOARD_SIZE;ku++) {
		sprintf(kc,"Unknown $%02X",ku); /* 12 */
		info_string(0,ku,kc);
	}
}

/* mouse */
static void mouse_fill_info(void)
{
	int i,j;
	char c[3];
	
	/* triggers */
	i=0;
	info_string(1,i++,	"X-");
	info_string(1,i++,	"X+");
	info_string(1,i++,	"Y-");
	info_string(1,i++,	"Y+");
	info_string(1,i++,	"Z-");
	info_string(1,i++,	"Z+");
	
	for (j=1;j<5;j++) {
		sprintf(c,"B%d",j); /* 3 */
		info_string(1,i++,c);
	}
	
	/* axes */
	i=0;
	info_string(2,i++,	"X");
	info_string(2,i++,	"Y");
	info_string(2,i++,	"Z");
}

/* joystick */
static void joystick_fill_info(void)
{
	int i,j;
	char c[8];
	
	/* triggers */
	i=0;
	info_string(3,i++,	"X-");
	info_string(3,i++,	"X+");
	info_string(3,i++,	"Y-");
	info_string(3,i++,	"Y+");
	info_string(3,i++,	"Z-");
	info_string(3,i++,	"Z+");
	info_string(3,i++,	"RX-");
	info_string(3,i++,	"RX+");
	info_string(3,i++,	"RY-");
	info_string(3,i++,	"RY+");
	info_string(3,i++,	"RZ-");
	info_string(3,i++,	"RZ+");
	info_string(3,i++,	"U-");
	info_string(3,i++,	"U+");
	info_string(3,i++,	"V-");
	info_string(3,i++,	"V+");
	
	for (j=1;j<5;j++) {
		sprintf(c,"POV%d X-",j); info_string(3,i++,c); /* 8 */
		sprintf(c,"POV%d X+",j); info_string(3,i++,c);
		sprintf(c,"POV%d Y-",j); info_string(3,i++,c);
		sprintf(c,"POV%d Y+",j); info_string(3,i++,c);
	}
	
	for (j=1;j<33;j++) {
		sprintf(c,"B%d",j);
		info_string(3,i++,c);
	}
	
	/* axes */
	i=0;
	info_string(4,i++,	"X");
	info_string(4,i++,	"Y");
	info_string(4,i++,	"Z");
	info_string(4,i++,	"RX");
	info_string(4,i++,	"RY");
	info_string(4,i++,	"RZ");
	info_string(4,i++,	"U");
	info_string(4,i++,	"V");
}

static void input_fill_info(void)
{
	keyboard_fill_info();
	mouse_fill_info();
	joystick_fill_info();
	
	MEM_CREATE(input.trigger_info[INPUT_TT_FOREGROUND],8);	sprintf(input.trigger_info[INPUT_TT_FOREGROUND],"IsFocus");
	MEM_CREATE(input.axis_info[INPUT_TA_FOREGROUND],8);		sprintf(input.axis_info[INPUT_TA_FOREGROUND],"IsFocus");
	MEM_CREATE(input.trigger_info[INPUT_TT_ACTIVE],9);		sprintf(input.trigger_info[INPUT_TT_ACTIVE],"IsActive");
	MEM_CREATE(input.axis_info[INPUT_TA_ACTIVE],9);			sprintf(input.axis_info[INPUT_TA_ACTIVE],"IsActive");
	MEM_CREATE(input.trigger_info[INPUT_TT_NOTHING],8);		sprintf(input.trigger_info[INPUT_TT_NOTHING],"nothing");
	MEM_CREATE(input.axis_info[INPUT_TA_NOTHING],8);		sprintf(input.axis_info[INPUT_TA_NOTHING],"nothing");
}

/* trigger rules, same numbers as in cont.h/c */

/* shortcut names */
static const char* input_t_info[INPUT_T_END]={
	"hard reset",		/* hard reset (reload) */
	"soft reset",
	"open  ",			/* spaces to prevent conflict with "open filetype index" */
	"exit",
	"full screen",
	"pause",
	"switch video format",
	"decrease 2A03 clock",
	"increase 2A03 clock",
	"cartridge switch up",
	"cartridge switch down",
	"help",
	"close",
	"disable shortcuts"
};

const char* input_get_trigger_info(u32 i) { if (i>=INPUT_T_END) return NULL; else return input_t_info[i]; }

/* default */
static const char* input_t_default[INPUT_T_MAX]={
	/* famicom mic */
	"sound()",
	
	/* joypad 1 */
	"!Shift & X | Shift & rapid(4,X) | J1 B2",
	"!Shift & Z | Shift & rapid(4,Z) | J1 B1",
	"C | J1 B3",
	"!Ctrl & V | J1 B4", /* prevent conflict with video format change (only here because it's most used) */
	"Up | J1 Y-",
	"Down & !Up | J1 Y+",
	"Left | J1 X-",
	"Right & !Left | J1 X+",
	
	/* joypad 2 */
	"!Shift & S | Shift & rapid(4,S) | J2 B2",
	"!Shift & A | Shift & rapid(4,A) | J2 B1",
	"D | J2 B3",
	"F | J2 B4",
	"I | J2 Y-",
	"K & !I | J2 Y+",
	"J | J2 X-",
	"L & !J | J2 X+",
	
	/* joypad 3 */
	"!Shift & W | Shift & rapid(4,W) | J3 B2",
	"!Shift & Q | Shift & rapid(4,Q) | J3 B1",
	"E | J3 B3",
	"R | J3 B4",
	"6 | J3 Y-",
	"Y & !6 | J3 Y+",
	"T | J3 X-",
	"U & !T | J3 X+",
	
	/* joypad 4 */
	"!Shift & 2 | Shift & rapid(4,2) | J4 B2",
	"!Shift & 1 | Shift & rapid(4,1) | J4 B1",
	"3 | J4 B3",
	"4 | J4 B4",
	"0 | J4 Y-",
	"P & !0 | J4 Y+",
	"O | J4 X-",
	"[ & !O | J4 X+",
	
	/* powerpad */
	"T | [",
	"Y | P",
	"U | O",
	"I",
	"G | '",
	"H | ;",
	"J | L",
	"K",
	"B | rShift",
	"N | /",
	"M | .",
	",",
	
	/* hypershot */
	"B",
	"N",
	"G",
	"H",
	
	/* keyboard, conflict with emulator shortcuts */
	"0",
	"1",
	"2",
	"3",
	"4",
	"5",
	"6",
	"7",
	"8",
	"9",
	"A",
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	"M",
	"N",
	"O",
	"P",
	"Q",
	"R",
	"S",
	"T",
	"U",
	"V",
	"W",
	"X",
	"Y",
	"Z",
	"F1",
	"F2",
	"F3",
	"F4",
	"F5",
	"F6",
	"F7",
	"F8",
	"-",
	"= | ^",
	"\\ | Yen",
	"`",
	"[ | @",
	"]",
	";",
	"' | :",
	",",
	".",
	"/",
	"rCtrl | Delete | _", /* rctrl doesn't exist on some keyboards */
	"Space",
	"Escape",
	"Enter",
	"Home",
	"Insert",
	"Backspace", /* del */
	"Up",
	"Down",
	"Left",
	"Right",
	"End | Stop",
	"lShift",
	"rShift",
	"lCtrl",
	"lAlt", /* grph */
	"rAlt | Kana", /* not sure though if ralt exists on all keyboards */
	
	/* emulator shortcuts */
	"Ctrl & R",					/* hard reset (reload) */
	"Shift & Ctrl & R",			/* soft reset */
	"Ctrl & O",					/* open game */
	"Alt & F4",					/* exit, though alt+f4 is standard */
	"Alt & Enter | dblclick()",	/* full screen */
	"Pause | Ctrl & P | !IsActive",		/* pause */
	"Ctrl & V",					/* video format */
	"Ctrl & -",					/* decrease 2A03 clock */
	"Ctrl & = | Ctrl & ^",		/* (+) increase 2A03 clock */
	"Ctrl & Page Up",			/* cart switch up */
	"Ctrl & Page Down",			/* cart switch down */
	"F1",						/* help */
	"nothing",					/* close game */
	"!IsFocus | AppMenu"		/* disable shortcuts */
};
/* set */
static char* input_t_set[INPUT_T_MAX];




/* init */
/* std */

static void input_menucaption(int idm,int idi)
{
	int i,j;
	char c[STRING_SIZE]={0};
	idi+=(INPUT_T_MAX-INPUT_T_END);
	
	main_menu_caption(MAIN_MENUCAPTION_GET,idm,c);
	i=strlen(c);
	
	if (stricmp(input_t_set[idi],"nothing")!=0) {
		c[i++]='\t';
		
		for (j=0;j<strlen(input_t_set[idi]);j++) {
			/* cap at | */
			if (input_t_set[idi][j]=='|') {
				if (i&&c[i-1]==' ') c[i-1]=0;
				break;
			}
			
			/* convert & to + (or menu will interpret it as shortcut) */
			if (input_t_set[idi][j]=='&') {
				if (i&&c[i-1]==' ') c[i-1]='+';
				else c[i++]='+';
				if (input_t_set[idi][j+1]==' ') j++;
			}
			
			else c[i++]=input_t_set[idi][j];
		}
		
		/* optional truncate */
		if (strlen(c)>64) { c[61]=c[62]=c[63]='.'; c[64]='\0'; }
	}
	
	main_menu_caption(MAIN_MENUCAPTION_PUT,idm,c);
}

void input_init(void)
{
	int i;
	
	memset(&input,0,sizeof(input));
	input_device_keyboard_cursor=input_device_mouse_cursor=input_device_joystick_cursor=input_device_keyboard_begin=input_device_mouse_begin=input_device_joystick_begin=NULL;
	input.mouse_cursor_visible=TRUE;
	
	/* std settings */
	input.joystick_isenabled=TRUE; i=settings_get_yesnoauto(SETTINGS_JOYSTICK); if (i==FALSE||i==TRUE) { input.joystick_isenabled=i; }
	input.background=FALSE; i=settings_get_yesnoauto(SETTINGS_INPBG); if (i==FALSE||i==TRUE) { input.background=i; }
	
	/* require accurate frequency: Quattro Sports/Super Sports Challenge (BMX Simulator) */
	input.updfreq=4; if (SETTINGS_GET_INT(settings_info(SETTINGS_UPDFREQUENCY),&i)) { CLAMP(i,1,1000); input.updfreq=i; }
	
	/* create dinput */
	if ((DirectInputCreate(MAIN->module,DIRECTINPUT_VERSION,&input.dinput,NULL))!=DI_OK) {
		LOG(LOG_MISC|LOG_ERROR,"can't create DirectInput!\n");
		input_clean(); exit(1);
	}
	
	/* enumerate devices */
	enumerate_devices();
	
	input.num_triggers=(input.num_keyboards*T_KEYBOARD_SIZE_TOTAL)+(input.num_mouses*T_MOUSE_SIZE)+(input.num_joysticks*T_JOYSTICK_SIZE)+4; /* +4: -4=foreground -3=active, -2=nothing, -1=empty */
	if (input.num_triggers>0x7fff) { LOG(LOG_MISC|LOG_ERROR,"input trigger overflow!\n"); input_clean(); exit(1); }
	MEM_CREATE(input.trigger,input.num_triggers*sizeof(int));
	MEM_CREATE(input.trigger_old,input.num_triggers*sizeof(int));
	MEM_CREATE(input.trigger_info,input.num_triggers*sizeof(char*));
	
	input.num_axes=(input.num_keyboards*A_KEYBOARD_SIZE)+(input.num_mouses*A_MOUSE_SIZE)+(input.num_joysticks*A_JOYSTICK_SIZE)+4;
	MEM_CREATE(input.axis,input.num_axes*sizeof(long));
	MEM_CREATE(input.axis_old,input.num_axes*sizeof(long));
	MEM_CREATE(input.axis_info,input.num_axes*sizeof(char*));
	
	MEM_CREATE(input.emshort,INPUT_T_END*sizeof(int));
	MEM_CREATE(input.emshort_old,INPUT_T_END*sizeof(int));
	
	input_fill_info();
	
	/* map settings */
	#define IGET_MAP(x,y,z) \
	input.skey_set[x]=0; \
	for (i=y;i<z;i++) { \
		input_t_set[i]=NULL; \
		SETTINGS_GET_STRING(cont_get_trigger_info(i),&input_t_set[i]); \
		if (!input_set_trigger(i,input_t_set[i],x)) { \
			/* set setting to default */ \
			MEM_CLEAN(input_t_set[i]); \
			MEM_CREATE(input_t_set[i],strlen(input_t_default[i])+1); \
			memcpy(input_t_set[i],input_t_default[i],strlen(input_t_default[i])); \
		} \
	}
	
	IGET_MAP(CONT_DA_MIC,CONT_T_MIC_BEGIN,CONT_T_MIC_END)
	IGET_MAP(CONT_DA_J1,CONT_T_J1_BEGIN,CONT_T_J1_END)
	IGET_MAP(CONT_DA_J2,CONT_T_J2_BEGIN,CONT_T_J2_END)
	IGET_MAP(CONT_DA_J3,CONT_T_J3_BEGIN,CONT_T_J3_END)
	IGET_MAP(CONT_DA_J4,CONT_T_J4_BEGIN,CONT_T_J4_END)
	IGET_MAP(CONT_DA_PPFT,CONT_T_PP_BEGIN,CONT_T_PP_END)
	IGET_MAP(CONT_DA_HS,CONT_T_HS_BEGIN,CONT_T_HS_END)
	IGET_MAP(CONT_DA_KEY,CONT_T_K_BEGIN,CONT_T_K_END)
	IGET_MAP(INPUT_DA_EM,(INPUT_T_MAX-INPUT_T_END),INPUT_T_MAX)
	
	input_affirm_ports();
	
	/* set menu keys */
	input_menucaption(IDM_RELOADGAME,INPUT_T_RELOAD);
	input_menucaption(IDM_RESETNES,INPUT_T_SOFTRESET);
	input_menucaption(IDM_OPENGAME,INPUT_T_OPENGAME);
	input_menucaption(IDM_EXITEMU,INPUT_T_EXIT);
	input_menucaption(IDM_FULLSCREEN,INPUT_T_FULLSCREEN);
	input_menucaption(IDM_PAUSEEMU,INPUT_T_PAUSE);
	input_menucaption(IDM_VIDEOFORMAT,INPUT_T_VIDEOFORMAT);
	input_menucaption(IDM_HELP,INPUT_T_HELP);
	input_menucaption(IDM_CLOSEGAME,INPUT_T_CLOSE);
	
	LOG(LOG_VERBOSE,"input initialised\n");
}

static void enumerate_devices(void)
{
	HRESULT result;
	
	/* on winxp, dinput only enumerates 1 keyboard and 1 mouse */
	
	/* keyboard */
	result=IDirectInput_EnumDevices(input.dinput,DIDEVTYPE_KEYBOARD,keyboard_create,0,DIEDFL_ATTACHEDONLY);
	if ((result!=DI_OK)|(input.num_keyboards==0)) {
		LOG(LOG_MISC|LOG_ERROR,"can't create DirectInput keyboard!\n"); input_clean(); exit(1);
	}
	
	/* mouse */
	result=IDirectInput_EnumDevices(input.dinput,DIDEVTYPE_MOUSE,mouse_create,0,DIEDFL_ATTACHEDONLY);
	if ((result!=DI_OK)|(input.num_mouses==0)) {
		LOG(LOG_MISC|LOG_ERROR,"can't create DirectInput mouse!\n"); input_clean(); exit(1);
	}
	
	/* joystick */
	if (input.joystick_isenabled) {
		IDirectInput_EnumDevices(input.dinput,DIDEVTYPE_JOYSTICK,joystick_create,0,DIEDFL_ATTACHEDONLY);
		if (result!=DI_OK) { LOG(LOG_MISC|LOG_WARNING,"can't create DirectInput joystick\n"); joystick_clean(); } /* not mandatory */
	}
}

/* create device */
#define DEVICE_WARNING(nam,war,str) \
LOG(LOG_MISC|LOG_WARNING,"%s for DI %s%d",str,war,input.num_##nam##s+1); \
	if (pd) { \
		IDirectInputDevice_Unacquire(pd); \
		IDirectInputDevice_Release(pd); \
	} return DIENUM_CONTINUE

#define _CREATE_DEVICE(nam,war,dfm) \
static BOOL CALLBACK nam##_create(LPCDIDEVICEINSTANCE instance,LPVOID v) \
{ \
	int i; \
	LPDIRECTINPUTDEVICE pd; \
	input_device* prev=input_device_##nam##_cursor; \
	\
	/* create it */ \
	if ((IDirectInput_CreateDevice(input.dinput,&instance->guidInstance,&pd,NULL))!=DI_OK) { DEVICE_WARNING(nam,war,"can't create device"); } \
	/* set dataformat */ \
	if ((IDirectInputDevice_SetDataFormat(pd,&dfm))!=DI_OK) { DEVICE_WARNING(nam,war,"can't set dataformat"); } \
	/* set cooperative level */ \
	if ((IDirectInputDevice_SetCooperativeLevel(pd,MAIN->window,DISCL_NONEXCLUSIVE|DISCL_BACKGROUND))!=DI_OK) { DEVICE_WARNING(nam,war,"can't set cooperative level"); } \
	/* acquire */ \
	for (i=10;(IDirectInputDevice_Acquire(pd)!=DI_OK)&(i>0);i--) Sleep(5); \
	if (i==0) { DEVICE_WARNING(nam,war,"can't acquire"); } \
	\
	/* root */ \
	if (input_device_##nam##_begin==NULL) { \
		MEM_CREATE_T(input_device_##nam##_begin,sizeof(input_device),input_device*); \
		input_device_##nam##_cursor=input_device_##nam##_begin; \
	} \
	/* next */ \
	else { \
		MEM_CREATE_T(input_device_##nam##_cursor->next,sizeof(input_device),input_device*); \
		input_device_##nam##_cursor=input_device_##nam##_cursor->next; \
		input_device_##nam##_cursor->prev=prev; \
	} \
	\
	input_device_##nam##_cursor->pd=pd; \
	if (instance->tszInstanceName&&strlen(instance->tszInstanceName)) { \
		i=strlen(instance->tszInstanceName); \
		MEM_CREATE(input_device_##nam##_cursor->name,i+1); \
		strcpy(input_device_##nam##_cursor->name,instance->tszInstanceName); \
	} \
	else { /* nameless, dunno if that can happen */ \
		MEM_CREATE(input_device_##nam##_cursor->name,strlen("nameless device")+1); \
		sprintf(input_device_##nam##_cursor->name,"nameless device"); \
	} \
	input.num_##nam##s++; \
	LOG(LOG_VERBOSE,"enumerated [%s] as [%s%d]\n",input_device_##nam##_cursor->name,war,input.num_##nam##s); \
	return DIENUM_CONTINUE; \
}
_CREATE_DEVICE(keyboard,"K",c_dfDIKeyboard)
_CREATE_DEVICE(mouse,"M",c_dfDIMouse)
_CREATE_DEVICE(joystick,"J",c_dfDIJoystick)


void input_affirm_ports(void)
{
	int a=input.skey_set[INPUT_DA_EM];
	
	if (nes_is_running()) {
		u32 da;
		
		#define s(p) da=cont_get_da(p); a=(a)|(input.skey_set[da&0xff])|(input.skey_set[da>>8&0xff])|(input.skey_set[da>>16&0xff])|(input.skey_set[da>>24&0xff])
		s(CONT_P1); s(CONT_P2); s(CONT_PEXT);
		#undef s
		
		/* 2 famicom pads: microphone enabled */
		if (cont_get_port(CONT_P1)==CONT_D_FJ&&cont_get_port(CONT_P2)==CONT_D_FJ) a|=input.skey_set[CONT_DA_MIC];
	}
	
	input.menukey=a&3;
	
	if (a&4) sound_create_capture();
	else sound_clean_capture();
}



/* clean */
/* clean std */
void input_clean(void)
{
	int i;
	for (i=0;i<INPUT_T_MAX;i++) {
		MEM_CLEAN(input_t_set[i]);
		MEM_CLEAN(input.trigger_rule[i]);
	}
	
	if (input.num_triggers) while (input.num_triggers--) { MEM_CLEAN(input.trigger_info[input.num_triggers]); }
	MEM_CLEAN(input.trigger); MEM_CLEAN(input.trigger_old); MEM_CLEAN(input.trigger_info);
	
	if (input.num_axes) while (input.num_axes--) { MEM_CLEAN(input.axis_info[input.num_axes]); }
	MEM_CLEAN(input.axis); MEM_CLEAN(input.axis_old); MEM_CLEAN(input.axis_info);
	
	MEM_CLEAN(input.emshort); MEM_CLEAN(input.emshort_old);
	
	mouse_clean();
	keyboard_clean();
	joystick_clean();
	
	if (input.dinput) { IDirectInput_Release(input.dinput); input.dinput=NULL; }
	
	LOG(LOG_VERBOSE,"input cleaned\n");
}

/* clean device */
#define _CLEAN_DEVICE(nam) static void nam##_clean(void) \
{ \
	if (input.num_##nam##s) { \
		for (;;) { \
			if (input_device_##nam##_cursor->pd) { \
				IDirectInputDevice_Unacquire(input_device_##nam##_cursor->pd); \
				IDirectInputDevice_Release(input_device_##nam##_cursor->pd); \
				MEM_CLEAN(input_device_##nam##_cursor->name); \
			} \
			if (input_device_##nam##_cursor->prev) { \
				input_device_##nam##_cursor=input_device_##nam##_cursor->prev; \
				MEM_CLEAN(input_device_##nam##_cursor->next); \
			} \
			else break; \
		} \
		MEM_CLEAN(input_device_##nam##_begin); \
		input_device_##nam##_cursor=NULL; \
		input.num_##nam##s=0; \
	} \
}
_CLEAN_DEVICE(keyboard)
_CLEAN_DEVICE(mouse)
_CLEAN_DEVICE(joystick)

/* save settings */
void input_settings_save(void)
{
	int i;
	input_device* d;
	
	/* std */
	SETTINGS_PUT_STRING(NULL,"; *** Input settings ***\n\n");
	SETTINGS_PUT_INT(settings_info(SETTINGS_UPDFREQUENCY),input.updfreq);
	settings_put_yesnoauto(SETTINGS_INPBG,input.background);
	settings_put_yesnoauto(SETTINGS_JOYSTICK,input.joystick_isenabled);
	
	cont_settings_save();
	
	/* devices */
	SETTINGS_PUT_STRING(NULL,"\n; Previously detected devices:");
	
	#define ISET_DETDEV(x,y) \
	if (input.num_##x##s) { \
		d=input_device_##x##_begin; \
		for (i=1;;i++) { \
			SETTINGS_PUT_STRING(NULL,"\n; "); \
			SETTINGS_PUT_STRING(NULL,y); \
			SETTINGS_PUT_INT(NULL,i); \
			SETTINGS_PUT_STRING(NULL," as "); \
			SETTINGS_PUT_STRING(NULL,d->name); \
			if (d->next) d=d->next; else break; \
		} \
	}
	
	ISET_DETDEV(keyboard,"K")
	ISET_DETDEV(mouse,"M")
	ISET_DETDEV(joystick,"J")
	else SETTINGS_PUT_STRING(NULL,"\n; (no joysticks)");
	SETTINGS_PUT_STRING(NULL,"\n");
	
	/* mappings */
	#define ISET_MAP(x,y,z) \
	SETTINGS_PUT_STRING(NULL,"\n; * "); \
	SETTINGS_PUT_STRING(NULL,cont_get_a_name(x)); \
	SETTINGS_PUT_STRING(NULL," *\n"); \
	for (i=y;i<z;i++) SETTINGS_PUT_STRING(cont_get_trigger_info(i),input_t_set[i])
	
	ISET_MAP(CONT_DA_J1,CONT_T_J1_BEGIN,CONT_T_J1_END);
	ISET_MAP(CONT_DA_J2,CONT_T_J2_BEGIN,CONT_T_J2_END);
	SETTINGS_PUT_STRING(cont_get_trigger_info(CONT_T_MIC),input_t_set[CONT_T_MIC]);
	ISET_MAP(CONT_DA_J3,CONT_T_J3_BEGIN,CONT_T_J3_END);
	ISET_MAP(CONT_DA_J4,CONT_T_J4_BEGIN,CONT_T_J4_END);
	ISET_MAP(CONT_DA_PPFT,CONT_T_PP_BEGIN,CONT_T_PP_END);
	ISET_MAP(CONT_DA_HS,CONT_T_HS_BEGIN,CONT_T_HS_END);
	ISET_MAP(CONT_DA_KEY,CONT_T_K_BEGIN,CONT_T_K_END);
	
	SETTINGS_PUT_STRING(NULL,"\n; * emulator shortcuts *\n");
	#define SETEM(x) SETTINGS_PUT_STRING(input_get_trigger_info(x),input_t_set[(INPUT_T_MAX-INPUT_T_END)+x]);
	SETEM(INPUT_T_OPENGAME);
	SETEM(INPUT_T_CLOSE);
	SETEM(INPUT_T_EXIT);
	SETEM(INPUT_T_CPUSPEEDMIN);
	SETEM(INPUT_T_CPUSPEEDPLUS);
	SETEM(INPUT_T_RELOAD);
	SETEM(INPUT_T_SOFTRESET);
	SETEM(INPUT_T_PAUSE);
	SETEM(INPUT_T_CARTSWITCHUP);
	SETEM(INPUT_T_CARTSWITCHDOWN);
	SETEM(INPUT_T_FULLSCREEN);
	SETEM(INPUT_T_VIDEOFORMAT);
	SETEM(INPUT_T_HELP);
	SETEM(INPUT_T_DISABLE);
	#undef SETEM
}




/* update device */
/* std */

void input_reset_updframe(void) { input.updcycles=input.updframe=input.vblank_updated=0; }

void input_need_update(void)
{
	if (input.updfreq==1) return;
	
	/* adjust cycles */
	if (crystal->fc>input.updframe) {
		int m=crystal->fc-input.updframe; if (m>2) m=2;
		input.updcycles+=(crystal->frame*m);
		input.updframe=crystal->fc;
	}
	
	/* update */
	if ((input.updcycles-cpu_get_cycles())>=(crystal->frame/input.updfreq)) {
		input.updcycles=cpu_get_cycles();
		
		/* minor delay when possible, can be made more accurate but that could interfere with other timing */
		if ((input.updcycles-cpu_get_cycles())>=(crystal->frame/5)&&(GetTickCount()-input.updticks)<sound_get_sleeptime_raw()&&sound_get_lostframes()==0&&sound_get_slept()<sound_get_slept_prev()) sound_sleep(1);
		
		if (!input.vblank_updated) { input_read(); input_read_shortcuts(); }
		input.vblank_updated=0;
		cont_update();
		
		input.updticks=GetTickCount();
	}
}

void input_vblank(void)
{
	if (input.updfreq==1) {
		input_read();
		input_read_shortcuts();
		cont_update();
	}
	
	else if (crystal->fc>(input.updframe+1)) {
		input.vblank_updated=TRUE;
		input_read();
		input_read_shortcuts();
	}
	
	/* every 20 seconds or so, fool possible screensaver mousecursor moved */
	if ((crystal->fc&0x7ff)==0x400&&((MAIN->foreground&(MAIN->dialog^1))|input.background)) {
		mouse_event(MOUSEEVENTF_MOVE,0,0,0,0);
	}
}

void input_read(void)
{
	int* p=input.trigger; long* a=input.axis;
	int fg=(MAIN->foreground&(MAIN->dialog^1));
	
	if (input.semaphore) return;
	input.semaphore=TRUE;
	
	/* remember previous state (swap addresses) */
	input.trigger=input.trigger_old; input.trigger_old=p;
	input.axis=input.axis_old; input.axis_old=a;
	
	if (fg|input.background) {
		sound_capt_read();
		keyboard_update();
		mouse_update();
		joystick_update();
		input.trigger[INPUT_TT_ACTIVE]=input.axis[INPUT_TA_ACTIVE]=0x80;
		input.trigger[INPUT_TT_FOREGROUND]=input.axis[INPUT_TA_FOREGROUND]=fg<<7;
	}
	else {
		memset(input.trigger,0,input.num_triggers*sizeof(int));
		memset(input.axis,0,input.num_axes*sizeof(long));
		input_mouse_cursor_reset_ticks();
	}
	
	/* debug: show held triggers */
	/*
	int ii;
	printf("\r                                                               \r");
	for (ii=0;ii<input.num_triggers;ii++) if (input.trigger[ii]) printf("%s ",input.trigger_info[ii]);
	*/
	
	/* debug: hack to do whatever if key pressed */
	/*
	if (input.trigger[DIK_RCONTROL]) {
		printf("LOG START\n");
		log_set_frame_start(0);
	}
	else if (input.trigger[DIK_RSHIFT]) {
		log_set_frame_start(~0);
		printf("LOG END\n");
	}
	*/
	
	input.semaphore=FALSE;
}

int input_menukey(int key)
{
	switch (key) {
		case VK_MENU: return input.menukey&1;
		case -1: return input.menukey<<7&input.trigger[T_KEYBOARD_A_ALT];
		case VK_F10: return input.menukey&2;
		
		default: break;
	}
	
	return FALSE;
}

/* shortcuts */
void input_cpuspeed_ticks_stall(void) { input.cpuspeed_ticks_next=input.cpuspeed_ticks+400; }
void input_cartswitch_ticks_stall(void) { input.cartswitch_ticks_next=input.cartswitch_ticks+400; }

#define pm(x,y) PostMessage(MAIN->window,WM_COMMAND,x,y)

static void input_shortcut_cpuspeed(void)
{
	/* cpuspeed: hold key */
	#define p input.emshort[INPUT_T_CPUSPEEDPLUS]
	#define m input.emshort[INPUT_T_CPUSPEEDMIN]
	if (crystal->overclock==100&&p&&m) input_cpuspeed_ticks_stall();
	else if (input.cpuspeed_ticks>input.cpuspeed_ticks_next) {
		input.cpuspeed_ticks_next=input.cpuspeed_ticks+40;
		if (p&&m) pm(IDM_CPUSPEEDNORMAL,0); /* both keys held for a while: reset to 100% */
		else if (m) pm(IDM_CPUSPEEDMIN,0);
		else pm(IDM_CPUSPEEDPLUS,0);
	}
	#undef p
	#undef m
}

#define o(x)	input.emshort_old[x]
#define p(x)	input.emshort[x]
static void input_shortcut_reload(void) { if (!o(INPUT_T_RELOAD)) pm(IDM_RELOADGAME,0); }
static void input_shortcut_softreset(void) { pm(IDM_RESETNES,0); }
static void input_shortcut_opengame(void) { if (!o(INPUT_T_OPENGAME)) pm(IDM_OPENGAME,0); }
static void input_shortcut_exit(void) { if (!o(INPUT_T_EXIT)) pm(IDM_EXITEMU,0); }
static void input_shortcut_fullscreen(void) { if (!o(INPUT_T_FULLSCREEN)) pm(IDM_FULLSCREEN,0); }
static void input_shortcut_pause(void) { if (!o(INPUT_T_PAUSE)) pm(IDM_PAUSEEMU,0); }
static void input_shortcut_videoformat(void) { if (!o(INPUT_T_VIDEOFORMAT)) pm(IDM_VIDEOFORMAT,0); }
static void input_shortcut_cpuspeedmin(void) { if (!o(INPUT_T_CPUSPEEDMIN)) { pm(IDM_CPUSPEEDMIN,0); input_cpuspeed_ticks_stall(); } else input_shortcut_cpuspeed(); }
static void input_shortcut_cpuspeedplus(void) { if (!o(INPUT_T_CPUSPEEDPLUS)) { if (!p(INPUT_T_CPUSPEEDMIN)) pm(IDM_CPUSPEEDPLUS,0); input_cpuspeed_ticks_stall(); } else { if (!o(INPUT_T_CPUSPEEDMIN)&&p(INPUT_T_CPUSPEEDMIN)) input_cpuspeed_ticks_stall(); else input_shortcut_cpuspeed(); } }
static void input_shortcut_cartswitchup(void) { if (!o(INPUT_T_CARTSWITCHUP)) { pm(IDM_CARTSWITCHUP,0); input_cartswitch_ticks_stall(); } else { if (input.cartswitch_ticks>input.cartswitch_ticks_next) { input.cartswitch_ticks_next=input.cartswitch_ticks+80; pm(IDM_CARTSWITCHUP,0); } } }
static void input_shortcut_cartswitchdown(void) { if (!o(INPUT_T_CARTSWITCHDOWN)) { if (!p(INPUT_T_CARTSWITCHUP)) pm(IDM_CARTSWITCHDOWN,0); input_cartswitch_ticks_stall(); } else { if (p(INPUT_T_CARTSWITCHUP)) input_cartswitch_ticks_stall(); else if (input.cartswitch_ticks>input.cartswitch_ticks_next) { input.cartswitch_ticks_next=input.cartswitch_ticks+80; pm(IDM_CARTSWITCHDOWN,0); } } }
static void input_shortcut_help(void) { if (!o(INPUT_T_HELP)) pm(IDM_HELP,0); }
static void input_shortcut_close(void) { if (!o(INPUT_T_CLOSE)) pm(IDM_CLOSEGAME,0); }
static void input_shortcut_disable(void) { return; } /* handled below */
#undef pm
#undef o
#undef p

typedef void(*fpt_input_shortcut)(void);
static const fpt_input_shortcut input_shortcut[INPUT_T_END]= {
	input_shortcut_reload,
	input_shortcut_softreset,
	input_shortcut_opengame,
	input_shortcut_exit,
	input_shortcut_fullscreen,
	input_shortcut_pause,
	input_shortcut_videoformat,
	input_shortcut_cpuspeedmin,
	input_shortcut_cpuspeedplus,
	input_shortcut_cartswitchup,
	input_shortcut_cartswitchdown,
	input_shortcut_help,
	input_shortcut_close,
	input_shortcut_disable
};

void input_read_shortcuts(void)
{
	int i;
	
	/* remember previous state */
	int* swap=input.emshort_old; input.emshort_old=input.emshort; input.emshort=swap;
	
	input.cartswitch_ticks=input.cpuspeed_ticks=GetTickCount();
	
	/* read */
	i=INPUT_T_END; while (i--) input.emshort[i]=input_trigger_held((INPUT_T_MAX-INPUT_T_END)+i);
	
	/* pause is a special case: if you manage to 'press' the trigger while input is inactive (eg. !IsActive, or a rapid function with !nothing), it will be interpreted as secondary hold key instead of the standard toggle */
	if (input.emshort[INPUT_T_PAUSE]&&!input.trigger[INPUT_TT_ACTIVE]) nes_set_paused(nes_get_paused()|2); else nes_set_paused(nes_get_paused()&1);
	
	if (!input.trigger_old[INPUT_TT_ACTIVE]||input.emshort[INPUT_T_DISABLE]) {
		input_cpuspeed_ticks_stall();
		input_cartswitch_ticks_stall();
		return;
	}
	
	/* run shortcut (if it's there, bottom to top) */
	i=INPUT_T_END; while (i--) { if (input.emshort[i]) { input_shortcut[i](); break; } }
}


#define GETDEVICESTATE(s) IDirectInputDevice_GetDeviceState(d->pd,s,&state)
#define DEVICE_GETSTATE_ERROR(ts,as) t+=ts; a+=as; d=d->next; continue /* error, use previous state */
#define DEVICE_GETSTATE(s,ts,as) \
	switch (GETDEVICESTATE(s)) { \
		case DI_OK: break; \
		case DIERR_NOTACQUIRED: case DIERR_INPUTLOST: \
			IDirectInputDevice_Acquire(d->pd); \
			if (GETDEVICESTATE(s)!=DI_OK) { DEVICE_GETSTATE_ERROR(ts,as); } \
			else break; \
		default: DEVICE_GETSTATE_ERROR(ts,as); \
	}

/* keyboard */
static void keyboard_update(void)
{
	int t=0,a=0;
	UCHAR state[T_KEYBOARD_SIZE];
	input_device* d=input_device_keyboard_begin;
	
	while (d) {
		/* get keyboard state */
		DEVICE_GETSTATE(T_KEYBOARD_SIZE,T_KEYBOARD_SIZE,A_KEYBOARD_SIZE)
		
		/* copy state */
		for (a=0;a<T_KEYBOARD_SIZE;a++) input.trigger[t++]=state[a]&0x80;
		input.trigger[t++]=(state[DIK_LMENU]|state[DIK_RMENU])&0x80; /* alt=lalt|ralt */
		input.trigger[t++]=(state[DIK_LCONTROL]|state[DIK_RCONTROL])&0x80; /* ctrl=lctrl|rctrl */
		input.trigger[t++]=(state[DIK_LSHIFT]|state[DIK_RSHIFT])&0x80; /* shift=lshift|rshift */
		input.trigger[t++]=(state[DIK_RETURN]|state[DIK_NUMPADENTER])&0x80; /* enter=main enter|numpad enter */
		input.trigger[t++]=state[DIK_PREVTRACK]&0x80; /* circumflex=previous track */
		
		d=d->next;
	}
}

/* mouse */
static void mouse_update(void)
{
	int t=input.num_keyboards*T_KEYBOARD_SIZE_TOTAL;
	int a=0,c=TRUE,ao,to;
	DIMOUSESTATE state;
	input_device* d=input_device_mouse_begin;
	
	while (d) {
		to=t; ao=a;
		
		/* get mouse state */
		DEVICE_GETSTATE(sizeof(DIMOUSESTATE),T_MOUSE_SIZE,A_MOUSE_SIZE)
		
		/* copy state */
		/* axes, relative */
		#define GET_M_AXIS(x) input.trigger[t++]=(state.x<0)<<7; input.trigger[t++]=(state.x>0)<<7; input.axis[a++]=state.x
		GET_M_AXIS(lX);		GET_M_AXIS(lY);		GET_M_AXIS(lZ);
		
		/* buttons */
		#define GET_M_BUT(x) input.trigger[t++]=state.rgbButtons[x]&0x80
		GET_M_BUT(0);	GET_M_BUT(1);	GET_M_BUT(2);	GET_M_BUT(3);
		
		/* cursor movement? */
		if (c) {
			if ((MAIN->foreground&(MAIN->dialog^1))&&(memcmp(&input.axis[ao],&input.axis_old[ao],A_MOUSE_SIZE*sizeof(long))==0)&&(memcmp(&input.trigger[to],&input.trigger_old[to],T_MOUSE_SIZE*sizeof(int))==0)) {
				if (input.mouse_cursor_ticks<GetTickCount()) { input.mouse_cursor_visible=FALSE; }
			}
			else { c=FALSE; input_mouse_cursor_reset_ticks(); }
		}
		
		d=d->next;
	}
}

/* joystick */
static void joystick_update(void)
{
	int i,t=(input.num_keyboards*T_KEYBOARD_SIZE_TOTAL)+(input.num_mouses*T_MOUSE_SIZE);
	int a=input.num_mouses*A_MOUSE_SIZE;
	DIJOYSTATE state;
	input_device* d=input_device_joystick_begin;
	
	if (input.num_joysticks==0) return;
	
	while (d) {
		/* get joystick state */
		DEVICE_GETSTATE(sizeof(DIJOYSTATE),T_JOYSTICK_SIZE,A_JOYSTICK_SIZE)
		
		/* copy state */
		/* axes, absolute; 0..65535 ($FFFF) */
		#define GET_J_AXIS(x) input.trigger[t++]=(state.x<0x6000)<<7; input.trigger[t++]=(state.x>=0xa000)<<7; input.axis[a++]=state.x
		GET_J_AXIS(lX);				GET_J_AXIS(lY);			GET_J_AXIS(lZ);
		GET_J_AXIS(lRx);			GET_J_AXIS(lRy);		GET_J_AXIS(lRz);
		GET_J_AXIS(rglSlider[0]);	GET_J_AXIS(rglSlider[1]);
		
		/* pov, in degrees*100.. digital interpretation only */
		for (i=0;i<4;i++) {
			input.trigger[t++]=((state.rgdwPOV[i]<=31500)&(state.rgdwPOV[i]>=22500))<<7;
			input.trigger[t++]=((state.rgdwPOV[i]<=13500)&(state.rgdwPOV[i]>=4500))<<7;
			input.trigger[t++]=((state.rgdwPOV[i]<=4500)|((state.rgdwPOV[i]>=31500)&(state.rgdwPOV[i]<36000)))<<7;
			input.trigger[t++]=((state.rgdwPOV[i]<=22500)&(state.rgdwPOV[i]>=13500))<<7;
		}
		
		/* buttons */
		for (i=0;i<32;i++) input.trigger[t++]=state.rgbButtons[i]&0x80;
		
		d=d->next;
	}
}


/* mouse cursor */
void input_mouse_cursor_update(void)
{
	POINT p;
	
	/* visibility, idea from VirtuaNES (NES emulator) */
	/* known bug (don't know how to fix it): ShowCursor doesn't work properly when opening a ROM by holding the left mouse
		button after doubleclicking, or very quick doubleclicking, from Windows explorer/desktop, however, the cursor
		will hide for a split second if you move it afterwards (so it did have effect, but didn't 'show' it).
	*/
	/* show cursor */
	if (input.mouse_cursor_visible) while (input.mouse_cursor_dc<0) input.mouse_cursor_dc=ShowCursor(TRUE);
	
	/* possibly hide cursor */
	else {
		if (nes_is_running()) {
			while (input.mouse_cursor_dc>=0) {
				if (input.mouse_cursor_dc==0) {
					if (!input_mouse_in_client(MAIN->window,NULL,&p)) {
						input.mouse_cursor_dc--;
						break;
					}
				}
				input.mouse_cursor_dc=ShowCursor(FALSE);
			}
		}
		else input.mouse_cursor_visible=TRUE;
	}
	
	/* full screen menu */
	if (draw_get_fullscreen()) {
		if (!GetCursorPos(&p)) p.y=GetSystemMetrics(SM_CYMENU);
		if (!MAIN->dialog) {
			if (p.y<GetSystemMetrics(SM_CYMENU)) draw_show_menu(TRUE);
			else draw_show_menu(FALSE);
		}
	}
}

void input_mouse_cursor_reset_ticks(void) { input.mouse_cursor_visible=TRUE; input.mouse_cursor_ticks=GetTickCount()+2500; }

/* doubleclick */
void input_doubleclick(void) { input.doubleclick=TRUE; input.doubleclick_ticks=GetTickCount(); }

/* mouse position in client area? */
int input_mouse_in_client(HWND wnd,HWND parent,POINT* p)
{
	RECT r;
	
	p->x=p->y=0;
	ClientToScreen(wnd,p);
	GetClientRect(wnd,&r);
	OffsetRect(&r,p->x,p->y);
	
	/* in client area? */
	if (GetCursorPos(p)&&PtInRect(&r,*p)) {
		HWND wnd_prev;
		RECT rc;
		
		if (parent) wnd=parent;
		
		/* check windows above this one */
		for (;;) {
			wnd_prev=GetNextWindow(wnd,GW_HWNDPREV);
			if (wnd_prev==NULL) break; /* done */
			wnd=wnd_prev;
			
			GetWindowRect(wnd,&rc);
			if (IsWindowVisible(wnd)&&PtInRect(&rc,*p)) return FALSE;
		}
		
		p->x-=r.left; p->y-=r.top;
		
		return TRUE;
	}
	
	return FALSE;
}



/* set/get trigger (rules) */

int input_set_trigger(const u32 trig,const char* set,u32 type)
{
	char *c=NULL,*t,*to,f[20];
	int i,def=TRUE,don=TRUE,off,sep,n,cur,len,end,rule_size,trc,ret=TRUE,rc=0,*trule,*rule=NULL;
	
	if (trig>=INPUT_T_MAX) return -1;
	
	if ((set!=NULL)&&(strlen(set)>0)) {
		
		MEM_CREATE(c,strlen(set)+1);
		memcpy(c,set,strlen(set));
retry:
		off=sep=def=trc=rule_size=0;
		
		/* find number of separators */
		for (i=0;;i++) {
			if (c[i]=='\0') break;
			if ((c[i]=='|')|(c[i]=='&')) rule_size++;
		}
		rule_size+=2; rule_size*=2;
		MEM_CREATE(trule,rule_size*sizeof(int));
		
		/* convert "key1 & key2 | key3" to -1 1 -2 2 -1 3 -3, a [not](!) is keynum|$8000 */
		trule[trc++]=-1;
		for (;;) {
			/* find new */
			for (;;off++) {
				if (c[off]=='\0') { off=-1; break; }
				else if ((c[off]!=' ')&(c[off]!='\t')&(c[off]!='&')&(c[off]!='|')) { break; }
			}
			if (off==-1) { trule[trc++]=-3; break; }
			
			/* find separator */
			if (sep<0) trule[trc++]=sep;
			sep=0; cur=off;
			for (;;cur++) {
				switch (c[cur]) {
					case '\0': break;
					case '|': sep=-1; break;
					case '&': sep=-2; break;
					default: continue;
				}
				break;
			}
			
			/* backtrack to calculate length */
			end=cur;
			for (;;end--) if ((c[end-1]!=' ')&(c[end-1]!='\t')) break;
			len=end-off;
			
			/* convert trigger string to int */
			/*
				bit 15: is not/invert
				
				key:
				bits 0-14: key number
				
				toggle:
				bit 16: is toggle
				bit 17: toggle is valid
				+key
				
				doubleclick:
				bit 17: is doubleclick
				
				rapid:
				bit 30: rapid is valid
				bits 18-23: rapid count
				bits 24-29: rapid max/is rapid
				+key
				
				sound:
				bit 30: is sound
				bits 0-7: sound threshold
				
				bit 31: unused (int sign bit)
			*/
			t=to=NULL;
			MEM_CREATE(t,len+2);
			memcpy(t,&c[off],len);
			to=t; n=0;
			
			#define T_FIND() \
				if (t[0]=='!') { \
					t++; n^=0x8000; /* not */ \
					while ((t[0]==' ')|(t[0]=='\t')) t++; /* ignore optional whitespace */ \
				} \
				if (strlen(t)>3) { \
					memset(f,0,10); \
					memcpy(f,t,3); \
					if (stricmp(f,"K1 ")==0) t+=3; /* ignore optional "K1 " */ \
				} \
				for (i=0;i<(input.num_triggers-1);i++) \
					if (stricmp(t,input.trigger_info[i])==0) break
			
			T_FIND();
			
			/* not found? check if it's a function */
			if (i==(input.num_triggers-1)) {
				
				int notfound=TRUE;
				
				len=strlen(t);
				
				/* doubleclick? */
				if (notfound&&len==10) {
					memset(f,0,20);
					memcpy(f,t,10);
					if (stricmp(f,"dblclick()")==0) {
						n|=0x20000; i=0;
						notfound=FALSE;
					}
				}
				
				/* rapid? */
				if (notfound&&len>9) {
					memset(f,0,20);
					memcpy(f,t,5);
					if ((stricmp(f,"rapid")==0)||(stricmp(f,"turbo")==0)) {
						int rapid=-1; t+=5;
						while ((t[0]=='(')|(t[0]==' ')|(t[0]=='\t')) t++;
						len=strlen(t);
						if (len) {
							sscanf(t,"%d",&rapid);
							while ((t[0]!='\0')&(t[0]!=',')) t++;
							t++; len=strlen(t);
							if (len) while (len--) if (t[len]==')') { t[len]='\0'; break; }
							if (len) while (len--) if ((t[len]!=' ')&(t[len]!='\t')) { t[++len]='\0'; break; }
						}
						if (len&&rapid!=-1) {
							CLAMP(rapid,1,0x3f);
							rapid<<=24;
							T_FIND(); n|=rapid;
							notfound=FALSE;
						}
					}
				}
				
				/* toggle? */
				if (notfound&&len>8) {
					memset(f,0,20);
					memcpy(f,t,6);
					if (stricmp(f,"toggle")==0) {
						t+=6;
						while ((t[0]=='(')|(t[0]==' ')|(t[0]=='\t')) t++;
						len=strlen(t);
						if (len) while (len--) if (t[len]==')') { t[len]='\0'; break; }
						if (len) while (len--) if ((t[len]!=' ')&(t[len]!='\t')) { t[++len]='\0'; break; }
						if (len) { T_FIND(); n|=0x10000; notfound=FALSE; }
					}
				}
				
				/* sound? */
				if (notfound&&len>6) {
					memset(f,0,20);
					memcpy(f,t,5);
					if (stricmp(f,"sound")==0) {
						int threshold=75; t+=5;
						while ((t[0]=='(')|(t[0]==' ')|(t[0]=='\t')) t++;
						len=strlen(t);
						if (len) {
							sscanf(t,"%d",&threshold);
							CLAMP(threshold,0,100);
							n|=0x40000000;
							n|=threshold;
							i=0;
							input.skey_set[type]|=4;
							notfound=FALSE;
						}
					}
				}
			}
			
			MEM_CLEAN(to);
			if (i==(input.num_triggers-1)) n=0;
			trule[trc++]=i|n;
			
			off=cur;
		}
		
		/* convert to [or_numkey(s)] [and_numkey(s)] [key] ([key], etc) [and_numkey(s)] [key] ([key], etc) etc */
		MEM_CREATE(rule,rule_size*sizeof(int));
		trc=end=0; rc=1;
		for (;;) {
			/* find next or */
			for (;;trc++) {
				if (trule[trc]==-1) break;
				else if (trule[trc]==-3) { end=TRUE; break; }
			}
			if (end) break;
			
			/* look if next key(s) = invalid */
			trc++; cur=trc; i=FALSE;
			for (;;trc++) {
				if ((trule[trc]&0x7fff)==(input.num_triggers-1)) { i=TRUE; }
				else if ((trule[trc]==-1)|(trule[trc]==-3)) break;
			}
			if (i) continue;
			
			/* copy next key(s) */
			off=0; rule[0]++;
			for (;;cur++) {
				if (trule[cur]<0) { if ((trule[cur]==-1)|(trule[cur]==-3)) break; }
				else { rule[rc]++; off++; rule[rc+off]=trule[cur]; }
			}
			rc=rc+off+1;
		}
		
		def=rule[0]==0;
		
		MEM_CLEAN(trule);
		MEM_CLEAN(c);
	}
	
	/* invalid; use default */
	if (def&don) {
		don=ret=FALSE;
		
		MEM_CLEAN(rule);
		MEM_CREATE(c,strlen(input_t_default[trig])+1);
		memcpy(c,input_t_default[trig],strlen(input_t_default[trig]));
		goto retry;
	}
	
	/* valid */
	if (rc) {
		int orc=rule[0];
		int andc; i=0;
		
		/* find single ALT or F10 (lalt&ralt situation would still pass though). "sound()" is handled above. */
		/* I'd want to trap LWIN and RWIN too, but that's only possible with a global hook, it seems (SetWindowsHookEx(WH_KEYBOARD,x,x,0)) */
		while (orc--) {
			i++; andc=rule[i];
			if (andc==1) {
				i++;
				switch (rule[i]&0x40007fff) { /* bit 30=sound, and uses lower bits */
					case DIK_LMENU: case DIK_RMENU: case T_KEYBOARD_A_ALT: input.skey_set[type]|=1; break;
					case DIK_F10: input.skey_set[type]|=2; break;
					default: break;
				}
			}
			else i+=andc;
		}
		
		/* copy */
		MEM_CREATE(input.trigger_rule[trig],rc*sizeof(int));
		memcpy(input.trigger_rule[trig],rule,rc*sizeof(int));
	}
	
	MEM_CLEAN(rule);
	
	return ret;
}


int __fastcall input_trigger_held(int t)
{
	int orc=input.trigger_rule[t][0];
	int andc,i=0,or=0,and=0;
	
	while (orc--) {
		i++; andc=input.trigger_rule[t][i]; and|=0x80;
		while (andc--) {
			i++;
			
			/* function */
			if (input.trigger_rule[t][i]&0x7f030000) {
				
				/* rapid */
				if (input.trigger_rule[t][i]&0x3f000000) {
					input.trigger_rule[t][i]&=0xbfffffff;
					input.trigger_rule[t][i]+=0x40000;
					if (input.trigger_old[input.trigger_rule[t][i]&0x7fff]^input.trigger[input.trigger_rule[t][i]&0x7fff]||(input.trigger_rule[t][i]&0xfc0000)==(input.trigger_rule[t][i]>>6&0xfc0000)) input.trigger_rule[t][i]&=0xbf03ffff;
					if ((input.trigger_rule[t][i]&0xfc0000)<(input.trigger_rule[t][i]>>7&0xfc0000)) input.trigger_rule[t][i]|=0x40000000;
					
					and&=(((input.trigger[input.trigger_rule[t][i]&0x7fff])^(input.trigger_rule[t][i]>>8))&(input.trigger_rule[t][i]>>23));
				}
				
				/* toggle */
				else if (input.trigger_rule[t][i]&0x10000) {
					input.trigger_rule[t][i]^=((((input.trigger_old[input.trigger_rule[t][i]&0x7fff])^0x80)&((input.trigger[input.trigger_rule[t][i]&0x7fff])))<<10);
					and&=(input.trigger_rule[t][i]>>10^(input.trigger_rule[t][i]>>8));
				}
				
				/* doubleclick */
				else if (input.trigger_rule[t][i]&0x20000) {
					and&=((input.doubleclick&((input.doubleclick_ticks+200)>GetTickCount()))<<7^(input.trigger_rule[t][i]>>8));
					input.doubleclick=FALSE;
				}
				
				/* sound */
				else if (input.trigger_rule[t][i]&0x40000000) {
					and&=((sound_get_capt_threshold()>(input.trigger_rule[t][i]&0x7f))<<7^(input.trigger_rule[t][i]>>8));
				}
			}
			
			/* normal */
			else and&=((input.trigger[input.trigger_rule[t][i]&0x7fff])^(input.trigger_rule[t][i]>>8));
		}
		or|=and;
	}
	
	return or>>7&1;
}
