#include <math.h>

#include "global.h"
#include "cartridge.h"
#include "cpu_memorymap.h"
#include "mapper.h"
#include "ppu.h"
#include "cpu.h"
#include "crystal.h"
#include "file.h"
#include "sound.h"
#include "apu.h"
#include "sample.h"
#include "settings.h"
#include "input.h"

#include "apu_components.c"


static struct {
	/* ptr */
	u8** ppu_memmap_cur_name_ptr;
	u8** ppu_memmap_name_ram_ptr;
	u8* ppu_memmap_palette_ptr;
	int* ppu_valid_scanline_ptr;
	int* ppu_enabled_ptr;
	u8* ppu_control1_ptr;
	u8* ppu_control2_ptr;
	int* ppu_reading_sprites_ptr;
	int* ppu_rendering_ptr;
	int* ppu_sc_ptr;
	int* ppu_scc_ptr;
	u16* ppu_address_ptr;
	int* cpu_cycles_ptr;
	int* ppu_cycles_ptr;
	
	/* std */
	int mirroring_initial;
	u16 ppu_address_old;
	int cpu_cycles_old;
	int ppu_cycles_old;
	int a12_rise_wait;
	int sc;
} mapper;

/* mapper info */
_MAPINIT(nothing) { return; }

static struct {
	fp_init_mapper init;
	char* name;
	char* shortname;
	int uid;
	int ines_number;
	int mapmir; /* true: mapper determines mirroring, false: info in header */
} mapper_info[MAPTYPE_MAX];

int mapperinfo_supported(u32 m) { return mapper_info[m%MAPTYPE_MAX].init!=MAPINIT_FUN(nothing); }
int mapperinfo_maptype2ines(u32 m) { return mapper_info[m%MAPTYPE_MAX].ines_number; }
const char* mapperinfo_get_name(u32 m) { return mapper_info[m%MAPTYPE_MAX].name; }
const char* mapperinfo_get_shortname(u32 m) { return mapper_info[m%MAPTYPE_MAX].shortname; }
int mapperinfo_get_mapmir(u32 m) { return mapper_info[m%MAPTYPE_MAX].mapmir; }

int mapperinfo_ines2maptype(u32 m)
{
	int r=MAPTYPE_UNKNOWN;
	
	switch (m) {
		case 0:		r=MAPTYPE_NINTENDO_NROM;						break;
		case 1:		r=MAPTYPE_NINTENDO_MMC1_SxROM;					break;
		case 2:		r=MAPTYPE_NINTENDO_UXROM;						break;
		case 3:		r=MAPTYPE_NINTENDO_CNROM;						break;
		case 4:		r=MAPTYPE_NINTENDO_MMC3_TXROM;					break;
		case 5:		r=MAPTYPE_NINTENDO_MMC5_EXROM;					break;
		case 6:		r=MAPTYPE_FRONT_FAREAST_F4XXX;					break;
		case 7:		r=MAPTYPE_NINTENDO_AXROM;						break;
		case 8:		r=MAPTYPE_FRONT_FAREAST_F3XXX;					break;
		case 9:		r=MAPTYPE_NINTENDO_MMC2_PXROM;					break;
		case 10:	r=MAPTYPE_NINTENDO_MMC4_FXROM;					break;
		case 11:	r=MAPTYPE_COLOR_DREAMS;							break;
		case 13:	r=MAPTYPE_NINTENDO_CPROM;						break;
		case 15:	r=MAPTYPE_PIRATE_MULTI_CONTRA_FUNCTION_16;		break;
		case 16:	r=MAPTYPE_BANDAI;								break;
		case 17:	r=MAPTYPE_FRONT_FAREAST_F8XXX;					break;
		case 18:	r=MAPTYPE_JALECO_SS_88006;						break;
		case 19:	r=MAPTYPE_NAMCOT_106;							break;
		case 21:	r=MAPTYPE_KONAMI_VRC4A_VRC4C;					break;
		case 22:	r=MAPTYPE_KONAMI_VRC2A;							break;
		case 23:	r=MAPTYPE_KONAMI_VRC2B_VRC4E;					break;
		case 24:	r=MAPTYPE_KONAMI_VRC6A;							break;
		case 25:	r=MAPTYPE_KONAMI_VRC4B_VRC4D;					break;
		case 26:	r=MAPTYPE_KONAMI_VRC6B;							break;
		case 32:	r=MAPTYPE_IREM_G_101;							break;
		case 33:	r=MAPTYPE_TAITO_TC;								break;
		case 34:	r=MAPTYPE_NINTENDO_BXROM;						break;
		case 37:	r=MAPTYPE_NINTENDO_MMC3_SMB_TETRIS_WORLDCUP;	break;
		case 40:	r=MAPTYPE_PIRATE_SMB2J_FDS_ALTERNATE;			break;
		case 41:	r=MAPTYPE_CALTRON_6_IN_1;						break;
		case 42:	r=MAPTYPE_PIRATE_MARIO_BABY;					break;
		case 44:	r=MAPTYPE_PIRATE_MULTI_MMC3_SUPER_BIG_7_IN_1;	break;
		case 46:	r=MAPTYPE_RUMBLESTATION;						break;
		case 47:	r=MAPTYPE_NINTENDO_MMC3_VBALL_WORLDCUP;			break;
		case 48:	r=MAPTYPE_TAITO_TC;								break;
		case 50:	r=MAPTYPE_PIRATE_SMB2J_FDS;						break;
		case 52:	r=MAPTYPE_PIRATE_MULTI_MMC3_MARIO_7_IN_1;		break;
		case 64:	r=MAPTYPE_TENGEN_RAMBO_1;						break;
		case 65:	r=MAPTYPE_IREM_H_3001;							break;
		case 66:	r=MAPTYPE_NINTENDO_GXROM;						break;
		case 67:	r=MAPTYPE_SUNSOFT_3;							break;
		case 68:	r=MAPTYPE_SUNSOFT_4;							break;
		case 69:	r=MAPTYPE_SUNSOFT_5;							break;
		case 70:	r=MAPTYPE_BANDAI_SIMPLE_M2;						break;
		case 71:	r=MAPTYPE_CAMERICA_BF9093;						break;
		case 72:	r=MAPTYPE_JALECO_JF_SIMPLE_P16S;				break;
		case 73:	r=MAPTYPE_KONAMI_VRC3;							break;
		case 75:	r=MAPTYPE_KONAMI_VRC1_JALECO_SS_88005;			break;
		case 76:	r=MAPTYPE_NAMCOT_108;							break;
		case 77:	r=MAPTYPE_IREM_SIMPLE_M4;						break;
		case 78:	r=MAPTYPE_IREM_SIMPLE_C8;						break;
		case 79:	r=MAPTYPE_AVE_NINA_003;							break;
		case 80:	r=MAPTYPE_TAITO_X_005;							break;
		case 82:	r=MAPTYPE_TAITO_X1_017;							break;
		case 85:	r=MAPTYPE_KONAMI_VRC7A_VRC7B;					break;
		case 86:	r=MAPTYPE_JALECO_JF_SIMPLE_P32S;				break;
		case 87:	r=MAPTYPE_JALECO_JF_KONAMI_SIMPLE;				break;
		case 88:	r=MAPTYPE_NAMCOT_118;							break;
		case 89:	r=MAPTYPE_SUNSOFT_2;							break;
		case 92:	r=MAPTYPE_JALECO_JF_SIMPLE_P16US;				break;
		case 93:	r=MAPTYPE_SUNSOFT_2_HARDWIRED_MIRRORING;		break;
		case 94:	r=MAPTYPE_CAPCOM_SIMPLE;						break;
		case 95:	r=MAPTYPE_NAMCOT_118_CHR_CONTROLLED_MIRRORING;	break;
		case 97:	r=MAPTYPE_IREM_SIMPLE_P16U;						break;
		case 100:	r=MAPTYPE_NINTENDO_MMC3_TXROM;					break;
		case 101:	r=MAPTYPE_JALECO_JF_SIMPLE_P32S;				break;
		case 105:	r=MAPTYPE_NINTENDO_MMC1_WORLD_CHAMPIONSHIPS;	break;
		case 113:	r=MAPTYPE_AVE_NINA_003;							break;
		case 118:	r=MAPTYPE_NINTENDO_MMC3_TXSROM;					break;
		case 119:	r=MAPTYPE_NINTENDO_MMC3_TQROM;					break;
		case 122:	r=MAPTYPE_SUNSOFT_1;							break;
		case 140:	r=MAPTYPE_JALECO_JF_SIMPLE_GN;					break;
		case 144:	r=MAPTYPE_AGCI_50282;							break;
		case 152:	r=MAPTYPE_BANDAI_SIMPLE_M1;						break;
		case 154:	r=MAPTYPE_NAMCOT_118_ONE_SCREEN_MIRRORING;		break;
		case 155:	r=MAPTYPE_NINTENDO_MMC1A_SxROM;					break;
		case 158:	r=MAPTYPE_NINTENDO_MMC3_TXSROM;					break;
		case 180:	r=MAPTYPE_NICHIBUTSU_SIMPLE;					break;
		case 184:	r=MAPTYPE_SUNSOFT_1;							break;
		case 185:	r=MAPTYPE_NINTENDO_CNROM_WITH_CHR_DISABLE3;		break;
		case 206:	r=MAPTYPE_NAMCOT_118;							break;
		case 207:	r=MAPTYPE_TAITO_X_005_ONE_SCREEN_MIRRORING;		break;
		case 210:	r=MAPTYPE_NAMCOT_106_STRIPPED_DOWN;				break;
		case 228:	r=MAPTYPE_ACTION_52;							break;
		case 232:	r=MAPTYPE_CAMERICA_BF9096;						break;
		case 234:	r=MAPTYPE_AVE_D_1012;							break;
		case 242:	r=MAPTYPE_PIRATE_WAI_XING_ZHAN_SHI;				break;
		
		default:													break;
	}
	
	return r;
}

void mapper_set_info(u32 i,const char* name,const char* shortname,int uid,int ines_number,int mapmir,fp_init_mapper init)
{
	if (i>=MAPTYPE_MAX) return;
	
	mapper_info[i].uid=uid;
	mapper_info[i].ines_number=ines_number;
	mapper_info[i].mapmir=mapmir;
	mapper_info[i].init=init==NULL?MAPINIT_FUN(nothing):init;
	
	/* shortname */
	MEM_CLEAN(mapper_info[i].shortname);
	if (shortname&&strlen(shortname)) { MEM_CREATE(mapper_info[i].shortname,strlen(shortname)+1); strcpy(mapper_info[i].shortname,shortname); }
	
	/* name */
	MEM_CLEAN(mapper_info[i].name);
	if (name&&strlen(name)) { MEM_CREATE(mapper_info[i].name,strlen(name)+1); strcpy(mapper_info[i].name,name); }
}

void mapper_clean_info(void)
{
	int i=MAPTYPE_MAX;
	while (i--) { MEM_CLEAN(mapper_info[i].name); }
}

void mapper_build_std_info(void)
{
	memset(&mapper_info,0,sizeof(mapper_info));
	
	/* simple types and subtypes have a made up addendum relating to mapper behaviour eg. Cx=chr bankswitch type, Px=prg bankswitch type, Mx=mirroring type, S=speech/sample */
	
	/*		internal number												mapper name									short name		uid|102	iNES#	cmirror	init function 						module */
	mapper_set_info(MAPTYPE_ACTION_52,									"Action 52",								"Action 52",	1,		228,	FALSE,	MAPINIT_FUN(action52)			);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_AGCI_50282,									"AGCI 50282",								"50282",		2,		144,	FALSE,	MAPINIT_FUN(colordreams)		);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_AVE_D_1012,									"AVE D1012",								"D1012",		3,		234,	TRUE,	MAPINIT_FUN(ave_d1012)			);	/* mapper_ave.c */
	mapper_set_info(MAPTYPE_AVE_NINA_001,								"AVE Nina-001",								"Nina-1",		4,		-34,	FALSE,	MAPINIT_FUN(nina1)				);	/* mapper_ave.c */
	mapper_set_info(MAPTYPE_AVE_NINA_003,								"AVE Nina-003/Nina-006",					"Nina-3/6",		5,		79,		FALSE,	MAPINIT_FUN(nina3)				);	/* mapper_ave.c */
	mapper_set_info(MAPTYPE_BANDAI,										"Bandai LZ93D50",							"Bandai LZ",	6,		16,		TRUE,	MAPINIT_FUN(bandai)				);	/* mapper_bandai.c */
	mapper_set_info(MAPTYPE_BANDAI_PT_554,								"Bandai PT-554",							"Bandai PT",	7,		-3,		FALSE,	MAPINIT_FUN(cnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_BANDAI_SIMPLE_M1,							"Bandai simple M1",							"Bandai M1",	8,		152,	TRUE,	MAPINIT_FUN(bandai_l1)			);	/* mapper_bandai.c */
	mapper_set_info(MAPTYPE_BANDAI_SIMPLE_M2,							"Bandai simple M2",							"Bandai M2",	9,		70,		TRUE,	MAPINIT_FUN(bandai_l1)			);	/* mapper_bandai.c */
	mapper_set_info(MAPTYPE_CALTRON_6_IN_1,								"Caltron 6-in-1",							"Caltron",		10,		41,		FALSE,	MAPINIT_FUN(caltron)			);	/* mapper_misc.c */
	mapper_set_info(MAPTYPE_HES_X_IN_1,									"Caltron HES X-in-1",						"HES MC",		11,		-113,	TRUE,	MAPINIT_FUN(nina3)				);	/* mapper_ave.c */
	mapper_set_info(MAPTYPE_CAMERICA_BF9093,							"Camerica BF9093",							"BF9093",		12,		71,		FALSE,	MAPINIT_FUN(camerica)			);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_CAMERICA_BF9096,							"Camerica BF9096",							"BF9096",		13,		232,	FALSE,	MAPINIT_FUN(camerica_bf9096)	);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_CAMERICA_BF9096_ALADDIN,					"Camerica BF9096 Aladdin",					"Aladdin",		14,		-232,	FALSE,	MAPINIT_FUN(camerica_bf9096)	);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_CAMERICA_BF9097,							"Camerica BF9097",							"BF9097",		15,		-71,	TRUE,	MAPINIT_FUN(camerica)			);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_CAPCOM_SIMPLE,								"Capcom simple",							"Capcom",		16,		94,		FALSE,	MAPINIT_FUN(capcom_l1)			);	/* mapper_misc.c */
	mapper_set_info(MAPTYPE_COLOR_DREAMS,								"Color Dreams",								"Dreams",		17,		11,		FALSE,	MAPINIT_FUN(colordreams)		);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_FRONT_FAREAST_F3XXX,						"Front Fareast F3XXX",						"F3XXX",		18,		8,		FALSE,	MAPINIT_FUN(ffe)				);	/* mapper_ffe.c */
	mapper_set_info(MAPTYPE_FRONT_FAREAST_F4XXX,						"Front Fareast F4XXX",						"F4XXX",		19,		6,		FALSE,	MAPINIT_FUN(ffe)				);	/* mapper_ffe.c */
	mapper_set_info(MAPTYPE_FRONT_FAREAST_F8XXX,						"Front Fareast F8XXX",						"F8XXX",		20,		17,		FALSE,	MAPINIT_FUN(ffe)				);	/* mapper_ffe.c */
	mapper_set_info(MAPTYPE_IREM_G_101,									"Irem G101",								"G101",			21,		32,		TRUE,	MAPINIT_FUN(irem_g101)			);	/* mapper_irem.c */
	mapper_set_info(MAPTYPE_IREM_G_101_ALTERNATE_MIRRORING,				"Irem G101 alternate mirroring",			"G101a",		22,		-32,	TRUE,	MAPINIT_FUN(irem_g101)			);	/* mapper_irem.c */
	mapper_set_info(MAPTYPE_IREM_H_3001,								"Irem H3001",								"H3001",		23,		65,		TRUE,	MAPINIT_FUN(irem_h3001)			);	/* mapper_irem.c */
	mapper_set_info(MAPTYPE_IREM_SIMPLE_C8,								"Irem simple C8",							"Irem C8",		24,		78,		TRUE,	MAPINIT_FUN(irem_l3)			);	/* mapper_irem.c */
	mapper_set_info(MAPTYPE_IREM_SIMPLE_P16U,							"Irem simple P16U",							"Irem P16U",	25,		97,		TRUE,	MAPINIT_FUN(irem_l1)			);	/* mapper_irem.c */
	mapper_set_info(MAPTYPE_IREM_SIMPLE_M4,								"Irem simple M4",							"Irem M4",		26,		77,		TRUE,	MAPINIT_FUN(irem_l2)			);	/* mapper_irem.c */
	mapper_set_info(MAPTYPE_JALECO_JF_KONAMI_SIMPLE,					"Jaleco JF/Konami simple",					"JF/Konami",	27,		87,		FALSE,	MAPINIT_FUN(jaleco_l1)			);	/* mapper_jaleco.c */
	mapper_set_info(MAPTYPE_JALECO_JF_SIMPLE_GN,						"Jaleco JF simple GN",						"JF GN",		28,		140,	FALSE,	MAPINIT_FUN(jaleco_l4)			);	/* mapper_jaleco.c */
	mapper_set_info(MAPTYPE_JALECO_JF_SIMPLE_M1,						"Jaleco JF simple M1",						"JF M1",		29,		-78,	TRUE,	MAPINIT_FUN(jaleco_l5)			);	/* mapper_jaleco.c */
	mapper_set_info(MAPTYPE_JALECO_JF_SIMPLE_P16S,						"Jaleco JF simple P16S",					"JF P16S",		30,		72,		FALSE,	MAPINIT_FUN(jaleco_l3)			);	/* mapper_jaleco.c */
	mapper_set_info(MAPTYPE_JALECO_JF_SIMPLE_P16US,						"Jaleco JF simple P16US",					"JF P16US",		31,		92,		FALSE,	MAPINIT_FUN(jaleco_l3)			);	/* mapper_jaleco.c */
	mapper_set_info(MAPTYPE_JALECO_JF_SIMPLE_P32S,						"Jaleco JF simple P32S",					"JF P32S",		32,		86,		FALSE,	MAPINIT_FUN(jaleco_l2)			);	/* mapper_jaleco.c */
	mapper_set_info(MAPTYPE_JALECO_SS_88006,							"Jaleco SS 88006",							"SS 88006",		33,		18,		TRUE,	MAPINIT_FUN(jaleco_ss88006)		);	/* mapper_jaleco.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC1_JALECO_SS_88005,				"Konami VRC I/Jaleco SS 88005",				"VRC1/JSS5",	34,		75,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC2A,								"Konami VRC IIa",							"VRC2a",		35,		22,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC2B_VRC4E,							"Konami VRC IIb/VRC IVe",					"VRC2b/4e",		36,		23,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC3,								"Konami VRC III",							"VRC3",			37,		73,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC4A_VRC4C,							"Konami VRC IVa/VRC IVc",					"VRC4a/c",		38,		21,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC4B_VRC4D,							"Konami VRC IVb/VRC IVd",					"VRC4b/d",		39,		25,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC6A,								"Konami VRC VIa",							"VRC6a",		40,		24,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC6B,								"Konami VRC VIb",							"VRC6b",		41,		26,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_KONAMI_VRC7A_VRC7B,							"Konami VRC VIIa/VRC VIIb",					"VRC7a/b",		42,		85,		TRUE,	MAPINIT_FUN(konami)				);	/* mapper_konami.c */
	mapper_set_info(MAPTYPE_NAMCOT_106,									"Namcot 106",								"N-106",		43,		19,		TRUE,	MAPINIT_FUN(namcot_106)			);	/* mapper_namcot.c */
	mapper_set_info(MAPTYPE_NAMCOT_106_STRIPPED_DOWN,					"Namcot 106 stripped down",					"N-106 SD",		44,		210,	FALSE,	MAPINIT_FUN(namcot_106)			);	/* mapper_namcot.c */
	mapper_set_info(MAPTYPE_NAMCOT_108,									"Namcot 108",								"N-108",		45,		76,		FALSE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NAMCOT_118,									"Namcot 118",								"N-118",		46,		88,		FALSE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NAMCOT_118_CHR_CONTROLLED_MIRRORING,		"Namcot 118 CHR controlled mirroring",		"N-118 MC",		47,		95,		TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NAMCOT_118_ONE_SCREEN_MIRRORING,			"Namcot 118 one-screen mirroring",			"N-118 M1",		48,		154,	TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NICHIBUTSU_SIMPLE,							"Nichibutsu simple",						"Nichibutsu",	49,		180,	FALSE,	MAPINIT_FUN(nichibutsu_l1)		);	/* mapper_misc.c */
	mapper_set_info(MAPTYPE_NINTENDO_AXROM,								"Nintendo AxROM",							"AxROM",		50,		7,		TRUE,	MAPINIT_FUN(aorom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_BXROM,								"Nintendo BxROM",							"BxROM",		51,		34,		FALSE,	MAPINIT_FUN(bnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_CNROM,								"Nintendo CNROM",							"CNROM",		52,		3,		FALSE,	MAPINIT_FUN(cnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_CNROM_RELYING_ON_BUS_CONFLICTS,	"Nintendo CNROM relying on bus conflicts",	"CNROM bus",	53,		-3,		FALSE,	MAPINIT_FUN(cnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_CNROM_WITH_CHR_DISABLE0,			"Nintendo CNROM with CHR disable (0/0)",	"CNROM CD0",	54,		185,	FALSE,	MAPINIT_FUN(cnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_CNROM_WITH_CHR_DISABLE1,			"Nintendo CNROM with CHR disable (0/1)",	"CNROM CD1",	55,		185,	FALSE,	MAPINIT_FUN(cnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_CNROM_WITH_CHR_DISABLE2,			"Nintendo CNROM with CHR disable (1/0)",	"CNROM CD2",	56,		185,	FALSE,	MAPINIT_FUN(cnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_CNROM_WITH_CHR_DISABLE3,			"Nintendo CNROM with CHR disable (1/1)",	"CNROM CD3",	57,		185,	FALSE,	MAPINIT_FUN(cnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_CPROM,								"Nintendo CPROM",							"CPROM",		58,		13,		FALSE,	MAPINIT_FUN(cprom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_GXROM,								"Nintendo GxROM",							"GxROM",		59,		66,		FALSE,	MAPINIT_FUN(gnrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC1_SxROM,						"Nintendo MMC1 SxROM",						"MMC1",			60,		1,		TRUE,	MAPINIT_FUN(mmc1)				);	/* mapper_mmc1.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC1A_SxROM,						"Nintendo MMC1A SxROM",						"MMC1A",		61,		155,	TRUE,	MAPINIT_FUN(mmc1)				);	/* mapper_mmc1.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC1_SOROM,						"Nintendo MMC1 SOROM",						"SOROM",		62,		-1,		TRUE,	MAPINIT_FUN(mmc1)				);	/* mapper_mmc1.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC1_SXROM,						"Nintendo MMC1 SXROM",						"SXROM",		63,		-1,		TRUE,	MAPINIT_FUN(mmc1)				);	/* mapper_mmc1.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC1_WORLD_CHAMPIONSHIPS,			"Nintendo MMC1 World Championships",		"MMC1 NWC",		64,		105,	TRUE,	MAPINIT_FUN(mmc1)				);	/* mapper_mmc1.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC2_PXROM,						"Nintendo MMC2 PxROM",						"MMC2",			65,		9,		TRUE,	MAPINIT_FUN(mmc2)				);	/* mapper_mmc2.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_TXROM,						"Nintendo MMC3 TxROM",						"MMC3",			66,		4,		FALSE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_TXSROM,						"Nintendo MMC3 TxSROM",						"TxSROM",		67,		118,	TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_TQROM,						"Nintendo MMC3 TQROM",						"TQROM",		68,		119,	TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_TR1ROM_TVROM,					"Nintendo MMC3 TR1ROM/TVROM",				"TR1ROM",		69,		-4,		TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_ALTERNATE_IRQ,				"Nintendo MMC3 alternate IRQ",				"MMC3a",		70,		-4,		FALSE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_WITHOUT_WRAM,					"Nintendo MMC3 without WRAM",				"MMC3w",		71,		-4,		FALSE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_SMB_TETRIS_WORLDCUP,			"Nintendo MMC3 SMB+Tetris+World Cup",		"MMC3 M3",		72,		37,		TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC3_VBALL_WORLDCUP,				"Nintendo MMC3 V'Ball+World Cup",			"MMC3 M2",		73,		47,		TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC4_FXROM,						"Nintendo MMC4 FxROM",						"MMC4",			74,		10,		TRUE,	MAPINIT_FUN(mmc2)				);	/* mapper_mmc2.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC5_EXROM,						"Nintendo MMC5 ExROM",						"MMC5",			75,		5,		TRUE,	MAPINIT_FUN(mmc5)				);	/* mapper_mmc5.c */
	mapper_set_info(MAPTYPE_NINTENDO_MMC6_HXROM,						"Nintendo MMC6 HxROM",						"MMC6",			76,		-4,		TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_NINTENDO_NROM,								"Nintendo NROM",							"NROM",			77,		0,		FALSE,	MAPINIT_FUN(nrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_NROM_FAMILY_BASIC,					"Nintendo NROM Family Basic",				"NROM FB",		78,		-256,	FALSE,	MAPINIT_FUN(nrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_NROM_WITH_COPYNES,					"Nintendo NROM with CopyNES",				"NROM CN",		79,		-256,	FALSE,	MAPINIT_FUN(nrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_NINTENDO_UXROM,								"Nintendo UxROM",							"UxROM",		80,		2,		FALSE,	MAPINIT_FUN(unrom)				);	/* mapper_nintendo.c */
	mapper_set_info(MAPTYPE_PIRATE_MARIO_BABY,							"Pirate Mario Baby",						"Baby",			81,		42,		TRUE,	MAPINIT_FUN(mariobaby)			);	/* mapper_pirate.c */
	mapper_set_info(MAPTYPE_PIRATE_SMB2J_FDS,							"Pirate SMB2J FDS",							"SMB2J",		82,		50,		FALSE,	MAPINIT_FUN(smb2j)				);	/* mapper_pirate.c */
	mapper_set_info(MAPTYPE_PIRATE_SMB2J_FDS_ALTERNATE,					"Pirate SMB2J FDS alternate",				"SMB2Ja",		83,		40,		FALSE,	MAPINIT_FUN(smb2j)				);	/* mapper_pirate.c */
	mapper_set_info(MAPTYPE_PIRATE_WAI_XING_ZHAN_SHI,					"Pirate Wai Xing Zhan Shi",					"WX PS4",		84,		242,	TRUE,	MAPINIT_FUN(waixing)			);	/* mapper_pirate.c */
	mapper_set_info(MAPTYPE_PIRATE_MULTI_CONTRA_FUNCTION_16,			"Pirate multi Contra Function 16",			"BMC CF16",		85,		15,		FALSE,	MAPINIT_FUN(contraf16)			);	/* mapper_pirate.c */
	mapper_set_info(MAPTYPE_PIRATE_MULTI_MMC3_MARIO_7_IN_1,				"Pirate multi MMC3 Mario 7-in-1",			"BMC M7",		86,		52,		TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_PIRATE_MULTI_MMC3_SUPER_BIG_7_IN_1,			"Pirate multi MMC3 Super Big 7-in-1",		"BMC S7",		87,		44,		TRUE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_RUMBLESTATION,								"Rumble Station",							"Rumble",		88,		46,		FALSE,	MAPINIT_FUN(rumblestation)		);	/* mapper_unlicensed.c */
	mapper_set_info(MAPTYPE_SUNSOFT_1,									"Sunsoft 1",								"Sunsoft 1",	89,		184,	FALSE,	MAPINIT_FUN(sunsoft1)			);	/* mapper_sunsoft.c */
	mapper_set_info(MAPTYPE_SUNSOFT_2,									"Sunsoft 2",								"Sunsoft 2",	90,		89,		TRUE,	MAPINIT_FUN(sunsoft2)			);	/* mapper_sunsoft.c */
	mapper_set_info(MAPTYPE_SUNSOFT_2_HARDWIRED_MIRRORING,				"Sunsoft 2 hardwired mirroring",			"Sunsoft 2a",	91,		93,		FALSE,	MAPINIT_FUN(sunsoft2)			);	/* mapper_sunsoft.c */
	mapper_set_info(MAPTYPE_SUNSOFT_3,									"Sunsoft 3",								"Sunsoft 3",	92,		67,		TRUE,	MAPINIT_FUN(sunsoft3)			);	/* mapper_sunsoft.c */
	mapper_set_info(MAPTYPE_SUNSOFT_4,									"Sunsoft 4",								"Sunsoft 4",	93,		68,		TRUE,	MAPINIT_FUN(sunsoft4)			);	/* mapper_sunsoft.c */
	mapper_set_info(MAPTYPE_SUNSOFT_5,									"Sunsoft 5",								"Sunsoft 5",	94,		69,		TRUE,	MAPINIT_FUN(sunsoft5)			);	/* mapper_sunsoft.c */
	mapper_set_info(MAPTYPE_TENGEN_MIMIC_1,								"Tengen Mimic-1",							"Mimic-1",		95,		-4,		FALSE,	MAPINIT_FUN(mmc3)				);	/* mapper_mmc3.c */
	mapper_set_info(MAPTYPE_TENGEN_RAMBO_1,								"Tengen Rambo-1",							"Rambo-1",		96,		64,		TRUE,	MAPINIT_FUN(tengen_rambo1)		);	/* mapper_tengen.c */
	mapper_set_info(MAPTYPE_TENGEN_SIMPLE,								"Tengen simple",							"Tengen",		97,		-3,		FALSE,	MAPINIT_FUN(tengen_l1)			);	/* mapper_tengen.c */
	mapper_set_info(MAPTYPE_TAITO_TC,									"Taito TC",									"Taito TC",		98,		33,		TRUE,	MAPINIT_FUN(taito_tc)			);	/* mapper_taito.c */
	mapper_set_info(MAPTYPE_TAITO_X_005,								"Taito X-005",								"X-005",		99,		80,		TRUE,	MAPINIT_FUN(taito_x)			);	/* mapper_taito.c */
	mapper_set_info(MAPTYPE_TAITO_X_005_ONE_SCREEN_MIRRORING,			"Taito X-005 one-screen mirroring",			"X-005a",		100,	207,	TRUE,	MAPINIT_FUN(taito_x)			);	/* mapper_taito.c */
	mapper_set_info(MAPTYPE_TAITO_X1_017,								"Taito X1-017",								"X1-017",		101,	82,		TRUE,	MAPINIT_FUN(taito_x)			);	/* mapper_taito.c */
	
	mapper_set_info(MAPTYPE_UNKNOWN,									"Unknown mapper type",						"Unknown",		9999,	-1000,	FALSE,	MAPINIT_FUN(nothing)			);	/* x */
}


/* mapper switches */
static struct {
	/* family basic back up switch (mapper_nintendo.c) */
	int familybasic_backupswitch;
	
	/* nintendo world championships 4 dipswitches (mapper_mmc1.c) */
	int nwc_dips;
} mapper_switches;

void mapper_init_switches(void)
{
	MAPSWITCHINIT_FUN(nrom_fb)();
	MAPSWITCHINIT_FUN(mmc1_event)();
}

void mapper_save_switches(void)
{
	SETTINGS_PUT_STRING(NULL,"; *** Cartridge settings ***\n\n");
	MAPSWITCHSAVE_FUN(nrom_fb)();
	MAPSWITCHSAVE_FUN(mmc1_event)();
}


/* shortcuts */
#define SET_MAPIO8_ALL(x) cpu_set_write_io8000(mapper_io_##x##_w,mapper_io_##x##_w,mapper_io_##x##_w,mapper_io_##x##_w)
#define SET_MAPIO8_R_ALL(x) cpu_set_read_io8000(mapper_io_##x##_r,mapper_io_##x##_r,mapper_io_##x##_r,mapper_io_##x##_r)
#define SET_MAPIO8(f1,f2,f3,f4) cpu_set_write_io8000(mapper_io_##f1##_w,mapper_io_##f2##_w,mapper_io_##f3##_w,mapper_io_##f4##_w)
#define SET_MAPIO8_R(f1,f2,f3,f4) cpu_set_read_io8000(mapper_io_##f1##_r,mapper_io_##f2##_r,mapper_io_##f3##_r,mapper_io_##f4##_r)
#define SET_MAPIO6(x) cpu_set_write_io6000(mapper_io_##x##_w)
#define SET_MAPIO6_R(x) cpu_set_read_io6000(mapper_io_##x##_r)
#define SET_MAPIO4(x) cpu_set_write_io4020(mapper_io_##x##_w)
#define SET_MAPIO4_R(x) cpu_set_read_io4020(mapper_io_##x##_r)
_MAPIO(nothing) { return; }
_MAPIO_R(nothing) { return address>>8; }

#define SET_MAPIRQ(x) cpu_set_mapper_irq(mapper_irq_##x)
_MAPIRQ(nothing) { return; }
_MAPIRQ(updateppu) { ppu_force_update(); }

#define SET_MAPAPUCLOCK(x) apu_set_mapper_apu_clock(mapper_apu_clock_##x)
_MAPAPUCLOCK(nothing) { return; }

void __fastcall (*mapppubusoverride)(u16); /* for a12 */
#define SET_MAPPPUBUS(x) ppu_memmap_set_ppubusoverride(mapper_ppubus_##x)
_MAPPPUBUS(nothing) { return; }

#define SET_MAPPPUMCS(x) ppu_set_modechangesignal(mapper_ppumodechangesignal_##x)
_MAPPPUMCS(nothing) { return; }


void (*_mapper_change_mode)(void);
#define _MAPMODE(x) static void mapper_change_mode_##x(void)
#define SET_MAPMODE(x) _mapper_change_mode=&mapper_change_mode_##x
_MAPMODE(nothing) { return; }
void mapper_change_mode(void) { (*_mapper_change_mode)(); }

void (*_mapper_set_switches)(u32);
#define _MAPSWITCHSET(x) static void mapper_set_switches_##x(u32 set)
#define SET_MAPSWITCHSET(x) _mapper_set_switches=&mapper_set_switches_##x
_MAPSWITCHSET(nothing) { return; }
void mapper_set_switches(u32 set) { (*_mapper_set_switches)(set); }

void (*_mapper_reset)(void);
#define _MAPRESET(x) static void mapper_reset_##x(void)
#define SET_MAPRESET(x) _mapper_reset=&mapper_reset_##x
_MAPRESET(nothing) { return; }
void mapper_reset(void) { (*_mapper_reset)(); }

void (*_mapper_clean)(void);
#define _MAPCLEAN(x) static void mapper_clean_##x(void)
#define SET_MAPCLEAN(x) _mapper_clean=&mapper_clean_##x
_MAPCLEAN(nothing) { return; }
void mapper_clean_mapper(void) { (*_mapper_clean)(); }

void (*_mapper_new_frame)(void);
#define _MAPNF(x) static void mapper_new_frame_##x(void)
#define SET_MAPNF(x) _mapper_new_frame=&mapper_new_frame_##x
_MAPNF(nothing) { return; }
_MAPNF(resetsc) { mapper.sc=0; }
_MAPNF(updateoldcycles) { if (mapper.cpu_cycles_old<crystal->frame) mapper.cpu_cycles_old+=crystal->frame; if (mapper.ppu_cycles_old<crystal->frame) mapper.ppu_cycles_old+=crystal->frame; }
void mapper_new_frame(void) { (*_mapper_new_frame)(); }

void (*_mapper_end_frame)(void);
#define _MAPEF(x) static void mapper_end_frame_##x(void)
#define SET_MAPEF(x) _mapper_end_frame=&mapper_end_frame_##x
_MAPEF(nothing) { return; }
void mapper_end_frame(void) { (*_mapper_end_frame)(); }

void (*_mapper_set_sound_buffer_counters)(int);
#define _MAPSOUNDBC(x) static void mapper_set_sound_buffer_counters_##x(int bc)
#define SET_MAPSOUNDBC(x) _mapper_set_sound_buffer_counters=&mapper_set_sound_buffer_counters_##x
_MAPSOUNDBC(nothing) { return; }
void mapper_set_sound_buffer_counters(int bc) { (*_mapper_set_sound_buffer_counters)(bc); }

void (*_mapper_sound_stream)(signed short*,int);
#define _MAPSTREAM(x) static void mapper_sound_stream_##x(signed short* stream,int length)
#define SET_MAPSTREAM(x) _mapper_sound_stream=&mapper_sound_stream_##x
_MAPSTREAM(nothing) { return; }
_MAPSTREAM(sample) { sample_stream(stream,length); }
void mapper_sound_stream(signed short* stream,int length) { (*_mapper_sound_stream)(stream,length); }

void (*_mapper_kill_cycle)(void);
#define _MAPKILLCYCLE(x) static void mapper_kill_cycle_##x(void)
#define SET_MAPKILLCYCLE(x) _mapper_kill_cycle=&mapper_kill_cycle_##x
_MAPKILLCYCLE(nothing) { return; }
void mapper_kill_cycle(void) { mapper.cpu_cycles_old-=crystal->ppu_cycle; mapper.ppu_cycles_old-=crystal->ppu_cycle; (*_mapper_kill_cycle)(); }

int (*_mapper_battery_load)(void);
#define _MAPBATLOAD(x) static int mapper_battery_load_##x(void)
#define SET_MAPBATLOAD(x) _mapper_battery_load=&mapper_battery_load_##x
_MAPBATLOAD(nothing) { return FALSE; }
int mapper_battery_custom_load(void) { return (*_mapper_battery_load)(); }

int (*_mapper_battery_save)(void);
#define _MAPBATSAVE(x) static int mapper_battery_save_##x(void)
#define SET_MAPBATSAVE(x) _mapper_battery_save=&mapper_battery_save_##x
_MAPBATSAVE(nothing) { return FALSE; }
int mapper_battery_custom_save(void) { return (*_mapper_battery_save)(); }

/* bankswitch */
#define MAXPRG cartridge->prgbanks
#define MAXVROM cartridge->vrombanks

#define PRG_NUM(n,i)			cartridge->cur_prgbank_num[n]=i
#define PRG_BS(n,i)				cartridge->cur_prgbank[n]=cartridge->prg[i]
#define PRG8_BS(n,i)			PRG_NUM(n,i); PRG_BS(n,i)
#define PRG16_BS(n,i)			PRG_NUM(n,i); PRG_BS(n,i++);			PRG_NUM((n)|1,i); PRG_BS((n)|1,i)
#define PRG32_BS(i)				PRG_NUM(0,i); PRG_BS(0,i++);			PRG_NUM(1,i); PRG_BS(1,i++);			PRG_NUM(2,i); PRG_BS(2,i++);			PRG_NUM(3,i); PRG_BS(3,i++)
#define PRG_BS_6000(i)			cartridge->cur_wrambank=cartridge->prg[i]; cartridge->cur_wrambank_num=i
#define INIT_PRG_FIRSTLAST()	PRG_NUM(0,0); PRG_BS(0,0);				PRG_NUM(1,1); PRG_BS(1,1);				PRG_NUM(2,MAXPRG-1); PRG_BS(2,MAXPRG-1);PRG_NUM(3,MAXPRG); PRG_BS(3,MAXPRG)
#define INIT_PRG_FIRST()		PRG_BS(0,0); PRG_BS(0,0);				PRG_NUM(1,1); PRG_BS(1,1);				PRG_NUM(2,2&MAXPRG); PRG_BS(2,2&MAXPRG);PRG_NUM(3,3&MAXPRG); PRG_BS(3,3&MAXPRG)
#define INIT_PRG_LASTFIRST()	PRG_NUM(0,MAXPRG-1); PRG_BS(0,MAXPRG-1);PRG_NUM(1,MAXPRG); PRG_BS(1,MAXPRG);	PRG_NUM(2,0); PRG_BS(2,0);				PRG_NUM(3,1); PRG_BS(3,1)
#define INIT_PRG_LAST()			PRG_NUM(0,(MAXPRG-3)&MAXPRG); PRG_BS(0,(MAXPRG-3)&MAXPRG); PRG_NUM(1,(MAXPRG-2)&MAXPRG); PRG_BS(1,(MAXPRG-2)&MAXPRG); PRG_NUM(2,MAXPRG-1); PRG_BS(2,MAXPRG-1); PRG_NUM(3,MAXPRG); PRG_BS(3,MAXPRG)

#define CHR_NUM(n,i)	cartridge->cur_patbank_num[n]=i
#define CHR_BS(n,i,d)	cartridge->cur_patbank[n]=cartridge->d[i]
#define CHR1_BS(n,i,d)	ppu_force_update(); CHR_NUM(n,i); CHR_BS(n,i,d)
#define CHR1_BS_V(n,i)	ppu_force_update(); CHR_NUM(n,i); CHR_BS(n,i,vrom)
#define CHR2_BS(n,i,d)	ppu_force_update(); CHR_NUM(n,i); CHR_BS(n,i++,d);			CHR_NUM((n)|1,i); CHR_BS((n)|1,i,d)
#define CHR2_BS_V(n,i)	ppu_force_update(); CHR_NUM(n,i); CHR_BS(n,i++,vrom);		CHR_NUM((n)|1,i); CHR_BS((n)|1,i,vrom)
#define CHR4_BS(n,i,d)	ppu_force_update(); CHR_NUM(n,i); CHR_BS(n,i++,d);			CHR_NUM((n)|1,i); CHR_BS((n)|1,i++,d);		CHR_NUM((n)|2,i); CHR_BS((n)|2,i++,d);		CHR_NUM((n)|3,i); CHR_BS((n)|3,i,d)
#define CHR4_BS_V(n,i)	ppu_force_update(); CHR_NUM(n,i); CHR_BS(n,i++,vrom);		CHR_NUM((n)|1,i); CHR_BS((n)|1,i++,vrom);	CHR_NUM((n)|2,i); CHR_BS((n)|2,i++,vrom);	CHR_NUM((n)|3,i); CHR_BS((n)|3,i,vrom)
#define CHR8_BS(i,d)	ppu_force_update(); CHR_NUM(0,i); CHR_BS(0,i++,d);			CHR_NUM(1,i); CHR_BS(1,i++,d);				CHR_NUM(2,i); CHR_BS(2,i++,d);				CHR_NUM(3,i); CHR_BS(3,i++,d);			CHR_NUM(4,i); CHR_BS(4,i++,d);		CHR_NUM(5,i); CHR_BS(5,i++,d);		CHR_NUM(6,i); CHR_BS(6,i++,d);		CHR_NUM(7,i); CHR_BS(7,i,d)
#define CHR8_BS_V(i)	ppu_force_update(); CHR_NUM(0,i); CHR_BS(0,i++,vrom);		CHR_NUM(1,i); CHR_BS(1,i++,vrom);			CHR_NUM(2,i); CHR_BS(2,i++,vrom);			CHR_NUM(3,i); CHR_BS(3,i++,vrom);		CHR_NUM(4,i); CHR_BS(4,i++,vrom);	CHR_NUM(5,i); CHR_BS(5,i++,vrom);	CHR_NUM(6,i); CHR_BS(6,i++,vrom);	CHR_NUM(7,i); CHR_BS(7,i,vrom)
#define INIT_CHR_FIRSTVROM() ppu_force_update(); CHR_NUM(0,0); CHR_BS(0,0,vrom); CHR_NUM(1,1); CHR_BS(1,1,vrom); CHR_NUM(2,2); CHR_BS(2,2,vrom); CHR_NUM(3,3); CHR_BS(3,3,vrom); CHR_NUM(4,4); CHR_BS(4,4,vrom); CHR_NUM(5,5); CHR_BS(5,5,vrom); CHR_NUM(6,6); CHR_BS(6,6,vrom); CHR_NUM(7,7); CHR_BS(7,7,vrom); cartridge->cur_patbank_isram[0]=cartridge->cur_patbank_isram[1]=cartridge->cur_patbank_isram[2]=cartridge->cur_patbank_isram[3]=cartridge->cur_patbank_isram[4]=cartridge->cur_patbank_isram[5]=cartridge->cur_patbank_isram[6]=cartridge->cur_patbank_isram[7]=FALSE
#define INIT_CHR_FIRSTVRAM() ppu_force_update(); CHR_NUM(0,0); CHR_BS(0,0,vram); CHR_NUM(1,1); CHR_BS(1,1,vram); CHR_NUM(2,2); CHR_BS(2,2,vram); CHR_NUM(3,3); CHR_BS(3,3,vram); CHR_NUM(4,4); CHR_BS(4,4,vram); CHR_NUM(5,5); CHR_BS(5,5,vram); CHR_NUM(6,6); CHR_BS(6,6,vram); CHR_NUM(7,7); CHR_BS(7,7,vram); cartridge->cur_patbank_isram[0]=cartridge->cur_patbank_isram[1]=cartridge->cur_patbank_isram[2]=cartridge->cur_patbank_isram[3]=cartridge->cur_patbank_isram[4]=cartridge->cur_patbank_isram[5]=cartridge->cur_patbank_isram[6]=cartridge->cur_patbank_isram[7]=TRUE
#define INIT_CHR_FIRST() if (MAXVROM) { INIT_CHR_FIRSTVROM(); } else { INIT_CHR_FIRSTVRAM(); }

#define DEF_6000_PRG() cartridge->wram_isenabled=TRUE; cartridge->wram_writeenable=FALSE; cartridge->cur_wrambank_isprg=TRUE
#define DEF_6000_WRAM() cartridge->wram_isenabled=cartridge->wram_writeenable=TRUE; cartridge->cur_wrambank=cartridge->wram; cartridge->cur_wrambank_isprg=FALSE
#define WRAM_WRITE(x) cpu_write_wram(address,x)

/* mirroring shortcuts, current method mostly inspired from Martin Freij's Nestopia */
#define MIR_C(x) ppu_force_update(); cartridge->mirroring=x; mapper.ppu_memmap_cur_name_ptr[0]=mapper.ppu_memmap_name_ram_ptr[cartridge->mirroring&1]; mapper.ppu_memmap_cur_name_ptr[1]=mapper.ppu_memmap_name_ram_ptr[cartridge->mirroring>>1&1]; mapper.ppu_memmap_cur_name_ptr[2]=mapper.ppu_memmap_name_ram_ptr[cartridge->mirroring>>2&1]; mapper.ppu_memmap_cur_name_ptr[3]=mapper.ppu_memmap_name_ram_ptr[cartridge->mirroring>>3&1]
#define MIR_0(x) if (x) { MIR_C(CARTRIDGE_MIRRORING_ONE_0); }
#define MIR_VH(x) if (x) { MIR_C(CARTRIDGE_MIRRORING_HORIZONTAL); } else { MIR_C(CARTRIDGE_MIRRORING_VERTICAL); }
#define MIR_HV(x) if (x) { MIR_C(CARTRIDGE_MIRRORING_VERTICAL); } else { MIR_C(CARTRIDGE_MIRRORING_HORIZONTAL); }
#define MIR_01(x) if (x) { MIR_C(CARTRIDGE_MIRRORING_ONE_1); } else { MIR_C(CARTRIDGE_MIRRORING_ONE_0); }
#define MIR_10(x) if (x) { MIR_C(CARTRIDGE_MIRRORING_ONE_0); } else { MIR_C(CARTRIDGE_MIRRORING_ONE_1); }
#define MIR_VH01(x) \
	switch (x) { \
		case 0: MIR_C(CARTRIDGE_MIRRORING_VERTICAL); break; \
		case 1: MIR_C(CARTRIDGE_MIRRORING_HORIZONTAL); break; \
		case 2: MIR_C(CARTRIDGE_MIRRORING_ONE_0); break; \
		case 3: MIR_C(CARTRIDGE_MIRRORING_ONE_1); break; \
		default: break; \
	}
#define MIR_01VH(x) \
	switch (x) { \
		case 0: MIR_C(CARTRIDGE_MIRRORING_ONE_0); break; \
		case 1: MIR_C(CARTRIDGE_MIRRORING_ONE_1); break; \
		case 2: MIR_C(CARTRIDGE_MIRRORING_VERTICAL); break; \
		case 3: MIR_C(CARTRIDGE_MIRRORING_HORIZONTAL); break; \
		default: break; \
	}
#define INIT_MIR(x) mapper.mirroring_initial=TRUE; MIR_C(x)
#define INIT_MIR_V() mapper.mirroring_initial=TRUE; MIR_C(CARTRIDGE_MIRRORING_VERTICAL)
#define INIT_MIR_H() mapper.mirroring_initial=TRUE; MIR_C(CARTRIDGE_MIRRORING_HORIZONTAL)
#define INIT_MIR_0() mapper.mirroring_initial=TRUE; MIR_C(CARTRIDGE_MIRRORING_ONE_0)
#define INIT_MIR_1() mapper.mirroring_initial=TRUE; MIR_C(CARTRIDGE_MIRRORING_ONE_1)
#define INIT_MIR_4() mapper.mirroring_initial=TRUE; MIR_C(CARTRIDGE_MIRRORING_4)
#define INIT_MIR_C() mapper.mirroring_initial=TRUE; MIR_C(CARTRIDGE_MIRRORING_CUSTOM)
#define INIT_MIR_U() mapper.mirroring_initial=TRUE;

/* a12 handler */
void __fastcall mapper_ppu_write_memory_address_warmup_a12_override(u8 data) { ppu_write_memory_address_warmup(data); }
void __fastcall mapper_ppu_write_memory_address_a12_override(u8 data) { ppu_write_memory_address(data); (*mapppubusoverride)(*mapper.ppu_address_ptr&0x1000); }
u8 __fastcall mapper_ppu_read_memory_data_a12_override(u16 address) { int ret=ppu_read_memory_data(address); (*mapppubusoverride)(*mapper.ppu_address_ptr&0x1000); return ret; }
void __fastcall mapper_ppu_write_memory_data_a12_override(u8 data) { ppu_write_memory_data(data); (*mapppubusoverride)(*mapper.ppu_address_ptr&0x1000); }

#define MAPA12_PDIS() cpu_set_write_io2000(0x2006,ppu_get_warmup()?mapper_ppu_write_memory_address_warmup_a12_override:mapper_ppu_write_memory_address_a12_override); cpu_set_read_io2000(0x2007,mapper_ppu_read_memory_data_a12_override); cpu_set_write_io2000(0x2007,mapper_ppu_write_memory_data_a12_override)
#define MAPA12_PEN() cpu_set_write_io2000(0x2006,ppu_get_warmup()?ppu_write_memory_address_warmup:ppu_write_memory_address); cpu_set_read_io2000(0x2007,ppu_read_memory_data); cpu_set_write_io2000(0x2007,ppu_write_memory_data)

_MAPPPUMCS(a12)
{
	if (enabled) { ppu_memmap_set_ppubusoverride(mapppubusoverride); MAPA12_PEN(); }
	else { SET_MAPPPUBUS(nothing); (*mapppubusoverride)(*mapper.ppu_address_ptr&0x1000); MAPA12_PDIS(); }
}

#define SET_MAPA12(x) \
ppu_force_update(); \
ppu_set_fpt(PPU_FPT_ALL,FALSE,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL); \
SET_MAPIRQ(updateppu); SET_MAPNF(updateoldcycles); SET_MAPPPUMCS(a12); \
if (*mapper.ppu_enabled_ptr&*mapper.ppu_valid_scanline_ptr) { mapper.ppu_address_old=0; MAPA12_PEN(); } \
else { mapper.ppu_address_old=*mapper.ppu_address_ptr&0x1000; MAPA12_PDIS(); } /* uninitialized but disabled on startup */ \
mapppubusoverride=mapper_ppubus_##x

static void RESET_MAPA12(void)
{
	ppu_set_fpt(PPU_FPT_NORMAL,FALSE,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
	SET_MAPIRQ(nothing); SET_MAPNF(nothing); SET_MAPPPUMCS(nothing); MAPA12_PEN();
}



void mapper_init(void)
{
	SET_MAPCLEAN(nothing);
	SET_MAPNF(nothing);
	SET_MAPBATLOAD(nothing);
	SET_MAPBATSAVE(nothing);
	SET_MAPSWITCHSET(nothing);
	cartridge->battery_data=NULL;
}

void mapper_clean(void)
{
	(*_mapper_clean)();
}

static void map_zero_structs(void); /* bottom */

void mapper_init_mapper(u32 maptype)
{
	int i;
	char v[16],loadmsg[0x100];
	memset(v,0,16);
	memset(loadmsg,0,0x100);
	
	mapper_clean_mapper();
	map_zero_structs();
	
	mapper.cpu_cycles_ptr=cpu_get_cycles_ptr();
	mapper.ppu_cycles_ptr=ppu_get_cycles_ptr();
	mapper.ppu_memmap_cur_name_ptr=ppu_memmap_get_cur_name_ptr();
	mapper.ppu_memmap_name_ram_ptr=ppu_memmap_get_name_ram_ptr();
	mapper.ppu_valid_scanline_ptr=ppu_get_valid_scanline_ptr();
	mapper.ppu_enabled_ptr=ppu_get_enabled_ptr();
	mapper.ppu_memmap_palette_ptr=ppu_memmap_get_palette_ptr();
	mapper.ppu_control1_ptr=ppu_get_control1_ptr();
	mapper.ppu_control2_ptr=ppu_get_control2_ptr();
	mapper.ppu_reading_sprites_ptr=ppu_get_reading_sprites_ptr();
	mapper.ppu_rendering_ptr=ppu_get_rendering_ptr();
	mapper.ppu_sc_ptr=ppu_get_sc_ptr();
	mapper.ppu_scc_ptr=ppu_get_scc_ptr();
	mapper.ppu_address_ptr=ppu_get_address_ptr();
	mapper.a12_rise_wait=16*crystal->ppu_cycle;
	
	/* make rom point to dummies */
	for (i=0;i<8;i++)
	{
		cartridge->cur_prgbank[i>>1]=cartridge->dummy_rom;
		cartridge->cur_prgbank_isram[i>>1]=FALSE;
		cartridge->cur_prgbank_num[i>>1]=-1;
		cartridge->cur_patbank[i]=cartridge->dummy_vrom;
		cartridge->cur_patbank_isram[i]=FALSE;
		cartridge->cur_patbank_num[i]=-1;
	}
	
	if ((cartridge->mirroring&BIN8(11110000))==0) { MIR_C(cartridge->mirroring); }
	cartridge->cur_wrambank=cartridge->wram; cartridge->battery_size=0x2000; cartridge->wram_writeenable=TRUE;
	if (cartridge->battery|cartridge->trainer) cartridge->wram_isenabled=TRUE;
	else cartridge->wram_isenabled=FALSE;
	cartridge->battery_data=NULL;
	
	SET_MAPCLEAN(nothing);
	SET_MAPRESET(nothing);
	SET_MAPMODE(nothing);
	SET_MAPNF(nothing);
	SET_MAPEF(nothing);
	SET_MAPSOUNDBC(nothing);
	SET_MAPSTREAM(nothing);
	SET_MAPKILLCYCLE(nothing);
	SET_MAPBATLOAD(nothing);
	SET_MAPBATSAVE(nothing);
	SET_MAPPPUBUS(nothing);
	SET_MAPPPUMCS(nothing);
	SET_MAPIO4(nothing);
	SET_MAPIO4_R(nothing);
	SET_MAPIO8_ALL(nothing);
	SET_MAPIRQ(nothing);
	SET_MAPAPUCLOCK(nothing);
	SET_MAPSWITCHSET(nothing);
	
	mapper_info[maptype].init(maptype);
	
	/* showing mapper info on ROM load inspired from Quietust's Nintendulator (NES emulator) */
	if (cartridge->vrombanks_real) sprintf(v,", %dK chr",cartridge->vrombanks_real);
	sprintf(loadmsg,"%s%s, %dK prg%s%s%s",mapper_info[maptype].shortname,cartridge->trainer?"+t":"",cartridge->prgbanks_real<<3,v,cartridge->battery?", SRAM":"",!mapper.mirroring_initial?cartridge->mirroring==CARTRIDGE_MIRRORING_VERTICAL?", M-V":cartridge->mirroring==CARTRIDGE_MIRRORING_HORIZONTAL?", M-H":", M-0":"");
	memset(loadmsg+strlen(loadmsg),' ',LOG_MAX_WIDTH-strlen(loadmsg));
	LOG(LOG_MISC|LOG_TYPE(LT_MAPPERLOAD),loadmsg);
}


#include "mapper_ave.c"
#include "mapper_bandai.c"
#include "mapper_ffe.c"
#include "mapper_irem.c"
#include "mapper_jaleco.c"
#include "mapper_konami.c"
#include "mapper_misc.c"
#include "mapper_mmc1.c"
#include "mapper_mmc2.c" /* and 4 */
#include "mapper_mmc3.c" /* and 6 */
#include "mapper_mmc5.c"
#include "mapper_namcot.c"
#include "mapper_nintendo.c"
#include "mapper_pirate.c"
#include "mapper_sunsoft.c"
#include "mapper_taito.c"
#include "mapper_tengen.c"
#include "mapper_unlicensed.c"

static void map_zero_structs(void)
{
	#define MAP_ZERO_STRUCT(x) memset(&x,0,sizeof(x))
	
	MAP_ZERO_STRUCT(mapper);
	
	MAP_ZERO_STRUCT(map_ave);
	MAP_ZERO_STRUCT(map_bandai);
	MAP_ZERO_STRUCT(map_ffe);
	MAP_ZERO_STRUCT(map_irem);
	MAP_ZERO_STRUCT(map_jaleco);
	MAP_ZERO_STRUCT(map_konami);
	MAP_ZERO_STRUCT(map_misc);
	MAP_ZERO_STRUCT(map_mmc1);
	MAP_ZERO_STRUCT(map_mmc2);
	MAP_ZERO_STRUCT(map_mmc3);
	MAP_ZERO_STRUCT(map_mmc5);
	MAP_ZERO_STRUCT(map_namcot);
	MAP_ZERO_STRUCT(map_nintendo);
	MAP_ZERO_STRUCT(map_pirate);
	MAP_ZERO_STRUCT(map_sunsoft);
	MAP_ZERO_STRUCT(map_taito);
	MAP_ZERO_STRUCT(map_tengen);
	MAP_ZERO_STRUCT(map_unl);
}
