/* 'backup' copier mappers (Super Magic Drive and such) */

static struct {
	int mode;
	int reset;
	int irq_isenabled;
	int irq_counter;
	int irq_hit;
	
} map_ffe;


/* --- Front Fareast (FFE) disk backup device. These mappers emulate the machine's basic IO, and RAM(512K), VRAM(256K)
	simulated as ROM, VROM/VRAM, and WRAM. The disk drive unit itself is not emulated! (iNES: 6,8,17) ---
	
	lots of hacked games
*/

/*
	several games refuse to work, dunno whether this is due to incomplete emulation;
	I think lots of these hacks have been created in the Nesticle-era...
	
	many games look glitchy, but that's probably normal
*/

static void ffe_set_mode(int);
static void ffe_reset_irq_hit(void);

static void ffe_reset_irq_hit(void) { map_ffe.irq_hit=0x10000*crystal->rel_cycle; }
_MAPMODE(ffe) { ffe_reset_irq_hit(); }

_MAPRESET(ffe)
{
	/* first reset, BIOS most likely does some initialisation */
	if (!map_ffe.reset) {
		map_ffe.reset=TRUE;
		apu_write_fs(0x40);
		cpu_memmap_fill_ram(0);
	}
}

/* general io, is always available */
_MAPIO(ffe)
{
	switch (address) {
		
		/* control register 1 */
		case 0x42fc: case 0x42fd: case 0x42fe: case 0x42ff:
			
			/* addressbus bit 0, databus bit 4: set mirroring */
			MIR_01VH((address<<1&2)|(data>>4&1))
			
			/* bits 5,6,7: change mode */
			ffe_set_mode(data>>5);
			
			break;
		
		case 0x43fe: break; /* ? */
		case 0x43ff: break; /* ? */
		
		/* control register 2 (fds/ppu/sram modes?) */
		case 0x4500: break;
		
		/* irq */
		case 0x4100: /* fantasy zone 2 */
		case 0x4501:
			map_ffe.irq_isenabled=FALSE;
			cpu_acknowledge_interrupt(~IRQ_MAPPER);
			break;
		case 0x4101: /* fantasy zone 2 */
		case 0x4502: case 0x4503: {
			u16 count=map_ffe.irq_counter/crystal->rel_cycle;
			if (address&1) { count=(count&0xff)|(data<<8); map_ffe.irq_isenabled=TRUE; }
			else count=(count&0xff00)|data;
			map_ffe.irq_counter=(count*crystal->rel_cycle)+(map_ffe.irq_counter%crystal->rel_cycle);
			break;
		}
		
		default:
			#if DEBUG_MAPPER
			LOG(LOG_MISC,"%x %x: ",address,data); LOG_BIN(LOG_MISC,data,TRUE);
			#endif
			break;
	}
}

/* set mapper 'simulation' mode */
static void ffe_set_mode(int m)
{
	if (map_ffe.mode==m) return;
	
	switch (m) {
		/* F4xxx mode 1: 32K vram, prg: $8000=selectable, $c000=7 */
		case 1: {
			int i;
			SET_MAPIO8_ALL(ffe_f4xxx_1);
			if (cartridge->vram[0]==NULL) for (i=0;i<8;i++) { MEM_CREATE(cartridge->vram[i],0x400); }
			if (cartridge->vram[8]==NULL) for (i=8;i<0x20;i++) { MEM_CREATE(cartridge->vram[i],0x400); }
			break;
		}
		
		/* UNROM: 8K vram, prg: $8000=selectable, $c000=last */
		case 2:
			SET_MAPIO8_ALL(unrom);
			if (cartridge->vram[0]==NULL) {
				int i;
				for (i=0;i<8;i++) { MEM_CREATE(cartridge->vram[i],0x400); }
			}
			break;
		
		/* F4xxx mode 2: vrom, prg: $8000=selectable, $c000=7 */
		case 3:
			SET_MAPIO8_ALL(ffe_f4xxx_2);
			break;
		
		/* F3xxx/GNROM: vrom, prg: 32K selectable */
		case 4:
			SET_MAPIO8_ALL(gnrom);
			break;
		
		default: break;
	}
	
	map_ffe.mode=m;
}

/* F4xxx mode 1 (vram) IO */
_MAPIO(ffe_f4xxx_1)
{
	/* bits 0,1: select 8K vram */
	int bank=data<<3&BIN8(00011000);
	CHR8_BS(bank,vram);
	
	/* bits 2 and up: select 16K prg in $8000 */
	bank=data>>1&BIN8(11111110)&MAXPRG;
	PRG16_BS(0,bank);
}
/* F4xxx mode 2 (vrom) IO */
_MAPIO(ffe_f4xxx_2)
{
	/* lower nibble: select 16K prg in $8000 */
	int bank=data<<1&MAXPRG;
	PRG16_BS(0,bank);
	
	/* bits 4 and up: select 8K vrom */
	bank=data>>1&BIN8(11111000)&MAXVROM;
	CHR8_BS_V(bank);
}

/* F8xxx IO */
_MAPIO(ffe_f8xxx)
{
	switch (address) {
		/* select 8k prg */
		case 0x4504: case 0x4505: case 0x4506: case 0x4507:
			PRG8_BS(address&3,data&MAXPRG);
			break;
		
		/* select 1k chr */
		case 0x4510: case 0x4511: case 0x4512: case 0x4513: case 0x4514: case 0x4515: case 0x4516: case 0x4517:
			CHR1_BS_V(address&7,data&MAXVROM);
			break;
		
		default: MAPIO_FUN(ffe)(address,bus,data); break;
	}
}

/* irq handler, based on cpu cycles, irq generated after 16 bit counter overflow */
_MAPIRQ(ffe)
{
	if (map_ffe.irq_isenabled) {
		map_ffe.irq_counter+=(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr);
		if (map_ffe.irq_counter>=map_ffe.irq_hit) { cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(map_ffe.irq_counter==map_ffe.irq_hit?INT_EDGE_IRQ_EX:0)); map_ffe.irq_counter-=map_ffe.irq_hit; }
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

/* init */
_MAPINIT(ffe)
{
	SET_MAPRESET(ffe);
	SET_MAPIO4(ffe);
	SET_MAPIRQ(ffe);
	SET_MAPNF(updateoldcycles);
	SET_MAPMODE(ffe);
	ffe_reset_irq_hit();
	map_ffe.mode=-1;
	DEF_6000_WRAM();
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	
	switch (maptype) {
		
		/* F3xxx */
		case MAPTYPE_FRONT_FAREAST_F3XXX:
			INIT_PRG_FIRST();
			ffe_set_mode(4);
			break;
		
		/* F4xxx */
		case MAPTYPE_FRONT_FAREAST_F4XXX: {
			int bank=7<<1&MAXPRG; PRG16_BS(2,bank);
			if (MAXVROM) ffe_set_mode(3);
			else ffe_set_mode(1);
			break;
		}
		
		/* F8xxx */
		case MAPTYPE_FRONT_FAREAST_F8XXX:
			SET_MAPIO4(ffe_f8xxx);
			break;
			
		default: break;
	}
}
