/* Nintendo MMC3 and subtypes mappers */

static struct {
	/* all */
	int cmdreg;
	int irq_isenabled;
	int irq_latch;
	int irq_counter;
	int irq_counter_old; /* b1 */
	int irq_reload;
	int prg[2];
	int chr[8];
	int chr_custom[8];
	
	/* latch */
	int prg_hi;
	int chr_hi;
	int reset; /* marioparty only */
	
	/* mmc6 */
	int mmc6_wram_lh_isenabled[2];
	int mmc6_wram_lh_writeenable[2];
	
	/* 4-screen */
	u8* tvrom_nt[4];
	
	/* tqrom */
	u8* tqrom_vram[8];
	
} map_mmc3;

/*
	NINTENDO
	
   --- MMC3:MMC3A/MMC3B/MMC3C(NES-TEROM/NES-TFROM/NES-TGROM/NES-TKROM/NES-TLROM/HVC-TNROM/NES-TSROM) (iNES: 4(/100)) ---
	SMB 2,3, Megaman 3,4,5,6, Double Dragon 2,3, Ninja Gaiden 2,3, Shatterhand, River City Ransom, Crystalis,
	Mighty Final Fight, Final Fantasy 3, Gun Nac, Earthbound/Mother, Kirby, Turtles 2,3, Tiny Toon 1,2, Toki, etc.
	
   --- NES-QJ (high banks latch) (iNES: 47) ---
	Super Spike V'Ball + Nintendo World Cup
	
   --- NES-??? (latch) (iNES: 37) ---
	Super Mario Bros. + Tetris + Nintendo World Cup
	
   --- NES-TLSROM/HVC-TKSROM (chr different mirroring) (iNES: 118/158) ---
	Armadillo, Alien Syndrome (U), Ys 3, Goal 2, RPG Jinsei Game, NES Play Action Football
	
   --- NES-TQROM (mixed VROM and VRAM banking) (iNES: 119) ---
	Pin Bot, High Speed
	
   --- NES-TR1ROM/NES-TVROM (4-screen 'mirroring') (iNES: 4+4screenflag) ---
	Gauntlet, Rad Racer 2
	
   --- MMC3 alternate IRQ (standard, but using b1 IRQ handler) (iNES: no/4) ---
	I don't know of any games depending on it (one of blargg's MMC3 tests tests this)
	
   --- MMC3 without WRAM (standard, but no WRAM) (iNES: no/4) ---
	Low G Man (probably more, but I know this one depends on it)
	
   --- MMC6/MMC6B(NES-HKROM) (almost same as MMC3, different WRAM handling) (iNES: no/4) ---
	Startropics 1,2
	
	NAMCOT
	
   --- Namcot 108 (iNES: 76) ---
	Digital Devil Monogatari - Megami Tensei (Japanese)
	(there's another "Namcot 108", used by Sky Kid, and probably more, incompatible with this one though)
	
   --- Namcot 118 (iNES: 88/206) ---
	Dragon Spirit - Aratanaru Densetsu, Karnov, Namcot Mahjong 3 - Mahjong Tengoku, Quinty
	
   --- Namcot 118 CHR controlled mirroring (iNES: 95) ---
	Dragon Buster (Japanese)
	
   --- Namcot 118 one-screen mirroring (iNES: 154) ---
	Devil Man (Japanese)
	
	TENGEN
	
   --- Tengen MIMIC-1 (stripped down) (iNES: no/4) ---
	Fantasy Zone (U), Indiana Jones and the Temple of Doom, Pac-Mania, R.B.I. Baseball 1,2,3,
	Road Runner, Super Sprint, Toobin, Vindicators,
	
	(Gauntlet too, but that uses extra RAM for 4 screens, use TR1ROM for that.
	Games seem to work fine in std MMC3 under current setup, some of which could be RAMBO-1,
	since they write to places of extra registers)
	
	PIRATE
	
   --- Mario Party (latch) (iNES: 52) ---
	Mario 7-in-1 (pirate)
		Super Mario Bros 3 (J), Super Mario Bros 2 (E), Mario is Missing, Jackie Chan's Action Kung-Fu (hacked),
		Takahashi Meijin no Bouken Shima III (hacked), Kaiketsu Yanchamaru 3 - Taiketsu! Zouringen (hacked),
		Flintstones - The Rescue of Dino & Hoppy (hacked)
	
   --- Super Big 7-in-1 (latch) (iNES: 44) ---
	Super Big 7-in-1 (pirate)
		Adventure Island 2, Super Contra, Robocop 2, Jackie Chan, Silk Worm, Double Dragon III (J),
		Teenage Mutant Ninja Turtles II - The Arcade Game
*/


/* TODO:
	I'm not sure if IRQ handling is 100% correct. Problem games:
	- Armadillo: statusbar shakes in-game
	- Days of Thunder: screen shakes when entering/exiting pits. garbage scanline on the game over textbox (not in PAL version)
	- Juuouki: statusbar shakes in-game (normal? it's a crappy port overall)
	- Mickey's Safari in Letterland: bottom half of statusbar shakes in-game (can be fixed with A12 rise/fall valid timing of 36 PPU cycles, but that's not accurate)
	- Super Mario Bros 3 (J): white line above the 'checkerboard' shouldn't be visible on the Famicom (different MMC3 revision?), should be on NES though
	- Ys III: statusbar shakes at the first conversation (pending IRQs are involved, same mapper as Armadillo)
	
	I'm not satisfied with the current IRQ implementation: checking the PPU address constantly,
	while also running the handler whenever the PPU makes a memory access. Though I'm certain
	the wrong behaviour (above) would be the same if it's implemented more cleanly.
*/


static void mmc3_disable(u8);
static void __fastcall mmc3_nothing(u8 data) { return; }
static __inline__ void mmc3_set_chr_mirroring(void);
static __inline__ void mmc3_set_chr_custom(u8);

static void __fastcall mmc3_cmd(u8);				static void __fastcall mmc3_page(u8);				static void __fastcall mmc3_mirroring(u8);
static void __fastcall mmc3_wram(u8);				static void __fastcall mmc3_irq_counter(u8);		static void __fastcall mmc3_irq_latch(u8);
static void __fastcall mmc3_irq_c0(u8);				static void __fastcall mmc3_irq_c1(u8);

static void __fastcall mmc3_std_bs(void);
static void __fastcall mmc3_superbig7in1_bs(void);	static void __fastcall mmc3_superbig7in1_register(u8);
static void __fastcall mmc3_qj_bs(void);
static void __fastcall mmc3_marioparty_bs(void);
static void __fastcall mmc3_3in1_bs(void);
static void __fastcall mmc3_namcot_118c_bs(void);	static void __fastcall mmc3_namcot_118c_page(u8);
static void __fastcall mmc3_namcot_108_bs(void);	static void __fastcall mmc3_namcot_108_page(u8);
static void __fastcall mmc3_namcot_118_bs(void);	static void __fastcall mmc3_namcot_118_cmd(u8);		static void __fastcall mmc3_namcot_118_mirroring(u8);
static void __fastcall mmc3_tqrom_bs(void);			static void __fastcall mmc3_tqrom_page(u8);
static void __fastcall mmc3_tlsrom_bs(void);		static void __fastcall mmc3_tlsrom_page(u8);
static void __fastcall mmc6_cmd(u8);				static void __fastcall mmc6_wram(u8);

static __inline__ void mmc3_prg_bs(void);
typedef void __fastcall (*fpt_mmc3_io)(u8);
static fpt_mmc3_io mmc3_io[8];
void __fastcall (*mmc3_bs)(void);
static const fpt_mmc3_io mmc3_io_std[8]={ mmc3_cmd, mmc3_page, mmc3_mirroring, mmc3_wram, mmc3_irq_latch, mmc3_irq_counter, mmc3_irq_c0, mmc3_irq_c1 };

/* disable standard registers */
static void mmc3_disable(u8 r)
{
	int i;
	for (i=0;i<8;i++) {
		if ((r&0x80)==0) {
			mmc3_io[i]=mmc3_nothing;
			if (i>3) RESET_MAPA12();
		}
		r<<=1;
	}
}


/* IO shortcut */
_MAPIO(mmc3) { mmc3_io[(address>>12&6)|(address&1)](data); }

/* --- standard MMC3 calls --- */

/* irq handler, based on ppu addressbus bit 12 rises (0 to 1) */
/* b1 (above problem games stay broken) */
_MAPA12(mmc3b1)
{
	if (address&0x1000) {
		if (((*mapper.ppu_cycles_ptr+mapper.a12_rise_wait)<=mapper.ppu_cycles_old)&(mapper.ppu_address_old==0)) {
			if ((map_mmc3.irq_counter==0)|map_mmc3.irq_reload) {
				map_mmc3.irq_counter=map_mmc3.irq_latch;
				map_mmc3.irq_counter_old=0;
			}
			else map_mmc3.irq_counter_old=map_mmc3.irq_counter--;
			if (((map_mmc3.irq_counter==0)&((map_mmc3.irq_counter_old!=0)|map_mmc3.irq_reload))&map_mmc3.irq_isenabled) {
				map_mmc3.irq_counter_old=0;
				cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(*mapper.ppu_cycles_ptr-*mapper.cpu_cycles_ptr?0:INT_EDGE_IRQ_EX));
			}
			map_mmc3.irq_reload=FALSE;
		}
		mapper.ppu_cycles_old=*mapper.ppu_cycles_ptr;
	}
	mapper.ppu_address_old=address&0x1000;
}
/* b2 (default) */
_MAPA12(mmc3b2)
{
	if (address&0x1000) {
		if (((*mapper.ppu_cycles_ptr+mapper.a12_rise_wait)<=mapper.ppu_cycles_old)&(mapper.ppu_address_old==0)) {
			if ((map_mmc3.irq_counter==0)|map_mmc3.irq_reload) {
				map_mmc3.irq_reload=FALSE;
				map_mmc3.irq_counter=map_mmc3.irq_latch;
			}
			else map_mmc3.irq_counter--;
			if ((map_mmc3.irq_counter==0)&map_mmc3.irq_isenabled) cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(*mapper.ppu_cycles_ptr-*mapper.cpu_cycles_ptr?0:INT_EDGE_IRQ_EX));
		}
		mapper.ppu_cycles_old=*mapper.ppu_cycles_ptr;
	}
	mapper.ppu_address_old=address&0x1000;
}


/* registers */
/* $8000: set command register */
static void __fastcall mmc3_cmd(u8 data) { map_mmc3.cmdreg=data; (*mmc3_bs)(); }

/* $8001: set page */
static void __fastcall mmc3_page(u8 data)
{
	int chr;
	if (MAXVROM) chr=data&MAXVROM;
	else chr=data&7;
	
	switch(map_mmc3.cmdreg&7) {
		case 0: map_mmc3.chr[0]=chr&~1; map_mmc3.chr[1]=chr|1; break; /* select 2k chr bank at $0000 */
		case 1: map_mmc3.chr[2]=chr&~1; map_mmc3.chr[3]=chr|1; break; /* select 2k chr bank at $0800 */
		case 2: map_mmc3.chr[4]=chr; break; /* select 1k chr bank at $1000 */
		case 3: map_mmc3.chr[5]=chr; break; /* select 1k chr bank at $1400 */
		case 4: map_mmc3.chr[6]=chr; break; /* select 1k chr bank at $1800 */
		case 5: map_mmc3.chr[7]=chr; break; /* select 1k chr bank at $c000 */
		case 6: map_mmc3.prg[0]=data&MAXPRG; break; /* select 1st prg bank */
		case 7: map_mmc3.prg[1]=data&MAXPRG; break; /* select 2nd prg bank */
	}
	(*mmc3_bs)();
}

/* $a000: set mirroring */
static void __fastcall mmc3_mirroring(u8 data) { MIR_VH(data&1); }

/* $a001: wram settings */
static void __fastcall mmc3_wram(u8 data)
{
	/* bit 7: enable wram, bit 6: write protected wram */
	cartridge->wram_isenabled=data>>7;
	cartridge->wram_writeenable=~data>>6&1;
}

/* $c000: store irq counter */
static void __fastcall mmc3_irq_latch(u8 data) { map_mmc3.irq_latch=data; }

/* $c001: reload irq */
static void __fastcall mmc3_irq_counter(u8 data) { map_mmc3.irq_reload=TRUE; }

/* $e000: disable irq */
static void __fastcall mmc3_irq_c0(u8 data) { map_mmc3.irq_isenabled=FALSE; cpu_acknowledge_interrupt(~IRQ_MAPPER); }

/* $e001: enable irq */
static void __fastcall mmc3_irq_c1(u8 data) { map_mmc3.irq_isenabled=TRUE; }


/* bankswitch */
static void __fastcall mmc3_std_bs(void)
{
	int i=8;
	const u8 ca=map_mmc3.cmdreg>>5&4;
	if (MAXVROM) { while(i--) { CHR1_BS_V(i^ca,map_mmc3.chr[i]); } }
	else { while(i--) { CHR1_BS(i^ca,map_mmc3.chr[i],vram); } }
	mmc3_prg_bs();
}
static __inline__ void mmc3_prg_bs(void)
{
	PRG8_BS(1,map_mmc3.prg[1]);
	PRG8_BS(3,MAXPRG);
	if (map_mmc3.cmdreg&BIT(6)) { PRG8_BS(0,MAXPRG-1); PRG8_BS(2,map_mmc3.prg[0]); }
	else { PRG8_BS(0,map_mmc3.prg[0]); PRG8_BS(2,MAXPRG-1); }
}


/* --- custom shared --- */

/* set custom chr bit */
static __inline__ void mmc3_set_chr_custom(u8 b)
{
	const u8 bank=map_mmc3.cmdreg&7;
	if (bank<2) { map_mmc3.chr_custom[bank<<1]=b; map_mmc3.chr_custom[bank<<1|1]=b; }
	else if (bank<6) map_mmc3.chr_custom[bank+2]=b;
}

/* set mirroring based on chr bit */
static __inline__ void mmc3_set_chr_mirroring(void)
{
	int i=4;
	const u8 ca=map_mmc3.cmdreg>>5&4;
	
	cartridge->mirroring&=BIN8(11110000);
	while(i--) cartridge->mirroring|=(map_mmc3.chr_custom[i^ca]<<i);
	MIR_C(cartridge->mirroring);
}


/* --- MMC6 --- */

/* registers */
/* $8000: set command register */
static void __fastcall mmc6_cmd(u8 data)
{
	/* bit 5: enable wram, can't be unset */
	map_mmc3.cmdreg=data|(cartridge->wram_isenabled<<5);
	cartridge->wram_isenabled=map_mmc3.cmdreg>>5&1;
	(*mmc3_bs)();
}

/* $a001: wram settings */
static void __fastcall mmc6_wram(u8 data)
{
	if (cartridge->wram_isenabled) { /* bits 4567: enable/write protect 512byte wram bank */
		map_mmc3.mmc6_wram_lh_isenabled[0]=data>>5&1;
		map_mmc3.mmc6_wram_lh_isenabled[1]=data>>7&1;
		map_mmc3.mmc6_wram_lh_writeenable[0]=data>>4&1;
		map_mmc3.mmc6_wram_lh_writeenable[1]=data>>6&1;
	}
}

/* wram handler */
_MAPIO_R(mmc6_wram_handler)
{
	if (cartridge->wram_isenabled&&(address&0x1000)) {
		if (map_mmc3.mmc6_wram_lh_isenabled[(address&0x3ff)>0x1ff]) return cartridge->cur_wrambank[address&0x3ff];
		else if (map_mmc3.mmc6_wram_lh_isenabled[(address&0x3ff)<0x200]) return 0;
	}
	return address>>8;
}
_MAPIO(mmc6_wram_handler) { if (cartridge->wram_isenabled&&(address&0x1000)&&(map_mmc3.mmc6_wram_lh_isenabled[(address&0x3ff)>0x1ff]&map_mmc3.mmc6_wram_lh_writeenable[(address&0x3ff)>0x1ff])) cartridge->cur_wrambank[address&0x3ff]=data; }


/* --- Super Big 7-in-1 --- */

/* registers */
/* $a001: set high banks */
static void __fastcall mmc3_superbig7in1_register(u8 data)
{
	/* bits 0,1,2: select game (0-5=128/128, 6=7=256/256) */
	data&=7; if (data>5) data=6;
	
	map_mmc3.chr_hi=data<<7;
	map_mmc3.prg_hi=data<<4;
	MAXPRG=0xf|map_mmc3.prg_hi|((map_mmc3.prg_hi==0x60)<<4);
	
	(*mmc3_bs)();
}

/* bankswitch */
static void __fastcall mmc3_superbig7in1_bs(void)
{
	int i=map_mmc3.prg_hi==0x60;
	const int prgmask=0xf|(i<<4);
	const int chrmask=0x7f|(i<<7);
	
	i=8; while(i--) map_mmc3.chr[i]=(map_mmc3.chr[i]&chrmask)|map_mmc3.chr_hi;
	i=2; while(i--) map_mmc3.prg[i]=(map_mmc3.prg[i]&prgmask)|map_mmc3.prg_hi;
	
	mmc3_std_bs();
}


/* --- MMC3 NES-QJ --- */

/* registers */
/* $6000: set high banks */
_MAPIO(mmc3_qj)
{
	/* 2 128K banks, bit 0 selects for both chr and prg */
	map_mmc3.chr_hi=data<<7&BIT(7);
	map_mmc3.prg_hi=data<<4&BIT(4);
	MAXPRG=0xf|map_mmc3.prg_hi;
	
	(*mmc3_bs)();
}

/* bankswitch */
static void __fastcall mmc3_qj_bs(void)
{
	int i;
	i=8; while(i--) map_mmc3.chr[i]=(map_mmc3.chr[i]&0x7f)|map_mmc3.chr_hi;
	i=2; while(i--) map_mmc3.prg[i]=(map_mmc3.prg[i]&0xf)|map_mmc3.prg_hi;
	
	mmc3_std_bs();
}


/* --- Mario Party --- */

/* registers */

static __inline__ void mmc3_marioparty_register(int data)
{
	/* chr: bits 5,2: select 256k bank. bit 6: bank size (256 or 128). bit 4: 128k bank (out of 256k) */
	map_mmc3.chr_hi=(((data<<1)&(data<<3))&BIT(7))|(data<<4&BIT(9))|(data<<6&BIT(8));
	
	/* prg: bits 1,2: select 256k bank. bit 3: bank size (256 or 128). bit 0: 128k bank (out of 256k) */
	map_mmc3.prg_hi=(((data<<1)&(data<<4))&BIT(4))|(data<<4&BIN8(01100000));
	MAXPRG=0xf|map_mmc3.prg_hi|((~data&BIT(3))<<1);
	
	(*mmc3_bs)();
}

/* $6000: set high banks (once) */
_MAPIO(mmc3_marioparty)
{
	WRAM_WRITE(data);
	
	if (map_mmc3.reset) {
		mmc3_marioparty_register(data);
		map_mmc3.reset=FALSE;
	}
}

/* reset: reset register */
_MAPRESET(mmc3_marioparty)
{
	mmc3_marioparty_register(0);
	map_mmc3.reset=TRUE;
}

/* bankswitch */
static void __fastcall mmc3_marioparty_bs(void)
{
	int i;
	const int prgmask=0xf|(~map_mmc3.prg_hi&0x10);
	const int chrmask=0x7f|(~map_mmc3.chr_hi&0x80);
	
	i=8; while(i--) map_mmc3.chr[i]=(map_mmc3.chr[i]&chrmask)|map_mmc3.chr_hi;
	i=2; while(i--) map_mmc3.prg[i]=(map_mmc3.prg[i]&prgmask)|map_mmc3.prg_hi;
	
	mmc3_std_bs();
}


/* --- SMB+Tetris+World Cup --- */

/* registers */
/* $6000: set high banks */
_MAPIO(mmc3_3in1)
{
	/* bit 0: select 128K chr bank */
	map_mmc3.chr_hi=data<<5&BIT(7);
	
	/* bit 1: select 64K prg bank, bit 2: select 128K prg bank */
	map_mmc3.prg_hi=data<<2&BIN8(00011000);
	MAXPRG=7|map_mmc3.prg_hi;
	if (map_mmc3.prg_hi&BIT(4)) MAXPRG|=8; /* 128K mode */
	
	(*mmc3_bs)();
}

/* bankswitch */
static void __fastcall mmc3_3in1_bs(void)
{
	int i;
	i=8; while(i--) map_mmc3.chr[i]=(map_mmc3.chr[i]&0x7f)|map_mmc3.chr_hi;
	i=2; while(i--) map_mmc3.prg[i]=(map_mmc3.prg[i]&(7|(MAXPRG&8)))|map_mmc3.prg_hi;
	
	mmc3_std_bs();
}


/* --- Namcot 108 --- */

/* registers */
/* $8001: set page */
static void __fastcall mmc3_namcot_108_page(u8 data)
{
	switch(map_mmc3.cmdreg&7) {
		/* nothing */
		case 0: case 1: break;
		
		/* select 2K chr */
		case 2: case 3: case 4: case 5:
			CHR1_BS_V(((map_mmc3.cmdreg&7)-2)<<1,data<<1&MAXVROM); CHR1_BS_V((((map_mmc3.cmdreg&7)-2)<<1)|1,(data<<1&MAXVROM)|1);
			break;
		
		/* select std prg */
		case 6: case 7: mmc3_page(data);
	}
}

/* bankswitch */
static void __fastcall mmc3_namcot_108_bs(void) { mmc3_prg_bs(); }


/* --- Namcot 118 --- */

/* registers */
/* $8000: set command register */
static void __fastcall mmc3_namcot_118_cmd(u8 data)
{
	/* $8000 also seems to affect mirroring on type 2 */
	mmc3_namcot_118_mirroring(data);
	
	mmc3_cmd(data);
}

/* $c000: type 2: set mirroring (bit 6) */
static void __fastcall mmc3_namcot_118_mirroring(u8 data) { MIR_01(data&BIT(6)) }

/* bankswitch */
static void __fastcall mmc3_namcot_118_bs(void)
{
	int i=8;
	while (i--) {
		/* ppu $0000-$0fff: use banks 0-63 */
		map_mmc3.chr[i]&=0x3f;
		
		/* ppu $1000-$1fff: use banks 63-127 */
		if (i>3) map_mmc3.chr[i]|=0x40;
		
		map_mmc3.chr[i]&=MAXVROM;
	}
	
	map_mmc3.cmdreg&=7;
	mmc3_std_bs();
}

/* --- Namcot 118 with chr mirroring --- */

/* registers */
/* $8001: set page */
static void __fastcall mmc3_namcot_118c_page(u8 data)
{
	mmc3_set_chr_custom(data>>5&1); /* bit 5: set nametable/mirroring */
	mmc3_page(data);
}

/* bankswitch */
static void __fastcall mmc3_namcot_118c_bs(void)
{
	mmc3_set_chr_mirroring();
	mmc3_std_bs();
}


/* --- MMC3 HVC-TKSROM/NES-TLSROM --- */

/* registers */
/* $8001: set page */
static void __fastcall mmc3_tlsrom_page(u8 data)
{
	mmc3_set_chr_custom(data>>7); /* bit 7: set nametable/mirroring */
	mmc3_page(data);
}

/* bankswitch */
static void __fastcall mmc3_tlsrom_bs(void)
{
	mmc3_set_chr_mirroring();
	mmc3_std_bs();
}


/* --- MMC3 NES-TQROM --- */

/* clean */
_MAPCLEAN(mmc3_tqrom) { int i=8; while(i--) { MEM_CLEAN(map_mmc3.tqrom_vram[i]); } }

/* registers */
/* $8001: set page */
static void __fastcall mmc3_tqrom_page(u8 data)
{
	mmc3_set_chr_custom(data>>6&1); /* bit 6: set vram */
	mmc3_page(data);
}

/* bankswitch */
static void __fastcall mmc3_tqrom_bs(void)
{
	int i=8;
	const u8 ca=map_mmc3.cmdreg>>5&4;
	ppu_force_update();
	
	while(i--) {
		cartridge->cur_patbank_isram[i^ca]=map_mmc3.chr_custom[i];
		if (map_mmc3.chr_custom[i]) cartridge->cur_patbank[i^ca]=map_mmc3.tqrom_vram[map_mmc3.chr[i]&7];
		else { CHR1_BS(i^ca,map_mmc3.chr[i],vrom); }
	}
	
	mmc3_prg_bs();
}


/* MMC3 NES-TR1ROM/NES-TVROM */

/* clean */
_MAPCLEAN(mmc3_tvrom) { int i=4; while(i--) { MEM_CLEAN(map_mmc3.tvrom_nt[i]); } }



/* init: for all */
_MAPINIT(mmc3)
{
	int i;
	
	/* set MMC3 standard values */
	for (i=0;i<8;i++) mmc3_io[i]=mmc3_io_std[i];
	mmc3_bs=&mmc3_std_bs;
	SET_MAPIO8_ALL(mmc3);
	
	map_mmc3.prg[0]=0; map_mmc3.prg[1]=1;
	INIT_CHR_FIRST();
	for (i=0;i<8;i++) map_mmc3.chr[i]=i;
	DEF_6000_WRAM();
	
	SET_MAPA12(mmc3b2);
	
	switch (maptype) {
		
		/* NINTENDO */
		
		/* standard MMC3 */
		case MAPTYPE_NINTENDO_MMC3_TXROM:
			if (cartridge->mirroring!=CARTRIDGE_MIRRORING_VERTICAL) cartridge->mirroring=CARTRIDGE_MIRRORING_HORIZONTAL; /* either horizontal or vertical: can be hardwired on eg. Tengen MIMIC-1. normal MMC3: $a000=0 (vertical) */
			break;
		
		/* Super Spike V'Ball + Nintendo World Cup */
		case MAPTYPE_NINTENDO_MMC3_VBALL_WORLDCUP:
			mmc3_disable(BIN8(11101111)); /* no wram settings */
			cartridge->wram_isenabled=cartridge->battery=FALSE;
			MAXPRG=0xf;
			mmc3_bs=&mmc3_qj_bs;
			SET_MAPIO6(mmc3_qj);
			INIT_MIR_V();
			break;
		
		/* Super Mario Bros. + Tetris + Nintendo World Cup */
		case MAPTYPE_NINTENDO_MMC3_SMB_TETRIS_WORLDCUP:
			mmc3_disable(BIN8(11101111)); /* no wram settings */
			cartridge->wram_isenabled=cartridge->battery=FALSE;
			MAXPRG=7;
			mmc3_bs=&mmc3_3in1_bs;
			SET_MAPIO6(mmc3_3in1);
			INIT_MIR_V();
			break;
		
		/* TxSROM */
		case MAPTYPE_NINTENDO_MMC3_TXSROM:
			mmc3_disable(BIN8(11011111)); /* no mirroring switch */
			mmc3_io[1]=mmc3_tlsrom_page;
			mmc3_bs=&mmc3_tlsrom_bs;
			INIT_MIR_C();
			break;
		
		/* TQROM */
		case MAPTYPE_NINTENDO_MMC3_TQROM:
			mmc3_io[1]=mmc3_tqrom_page;
			mmc3_bs=&mmc3_tqrom_bs;
			SET_MAPCLEAN(mmc3_tqrom);
			for (i=0;i<8;i++) { MEM_CREATE(map_mmc3.tqrom_vram[i],0x400); }
			INIT_MIR_V();
			break;
		
		/* 4-screen */
		case MAPTYPE_NINTENDO_MMC3_TR1ROM_TVROM:
			mmc3_disable(BIN8(11011111)); /* no mirroring switch */
			INIT_MIR_4();
			SET_MAPCLEAN(mmc3_tvrom);
			for (i=0;i<4;i++) {
				MEM_CREATE(map_mmc3.tvrom_nt[i],0x400);
				mapper.ppu_memmap_cur_name_ptr[i]=map_mmc3.tvrom_nt[i];
			}
			break;
		
		/* standard MMC3, but alt irq */
		case MAPTYPE_NINTENDO_MMC3_ALTERNATE_IRQ:
			if (cartridge->mirroring!=CARTRIDGE_MIRRORING_VERTICAL) cartridge->mirroring=CARTRIDGE_MIRRORING_HORIZONTAL;
			SET_MAPA12(mmc3b1);
			break;
		
		/* standard MMC3, but without wram */
		case MAPTYPE_NINTENDO_MMC3_WITHOUT_WRAM:
			mmc3_disable(BIN8(11101111)); /* no wram settings */
			cartridge->wram_isenabled=cartridge->battery=FALSE;
			if (cartridge->mirroring!=CARTRIDGE_MIRRORING_VERTICAL) cartridge->mirroring=CARTRIDGE_MIRRORING_HORIZONTAL;
			break;
		
		/* MMC6 */
		case MAPTYPE_NINTENDO_MMC6_HXROM:
			mmc3_io[0]=mmc6_cmd; mmc3_io[3]=mmc6_wram;
			cartridge->wram_isenabled=cartridge->wram_writeenable=FALSE;
			SET_MAPIO6(mmc6_wram_handler);
			SET_MAPIO6_R(mmc6_wram_handler);
			cartridge->battery_size=0x400;
			INIT_MIR_V();
			break;
		
		/* NAMCOT */
		
		/* Namcot 108 */
		case MAPTYPE_NAMCOT_108:
			mmc3_disable(BIN8(11100000)); /* not sure about the mirroring */
			cartridge->wram_isenabled=cartridge->battery=FALSE;
			mmc3_io[1]=mmc3_namcot_108_page;
			mmc3_bs=&mmc3_namcot_108_bs;
			break;
		
		/* Namcot 118 */
		case MAPTYPE_NAMCOT_118: case MAPTYPE_NAMCOT_118_ONE_SCREEN_MIRRORING:
			mmc3_disable(BIN8(11000000)); /* cmd and page only */
			cartridge->wram_isenabled=cartridge->battery=FALSE;
			for (i=4;i<8;i++) {
				map_mmc3.chr[i]=(map_mmc3.chr[i]|0x40)&MAXVROM;
				CHR1_BS_V(i,map_mmc3.chr[i]);
			}
			mmc3_bs=&mmc3_namcot_118_bs;
			if (maptype==MAPTYPE_NAMCOT_118_ONE_SCREEN_MIRRORING) { /* 118 with 1-screen mirroring */
				mmc3_io[0]=mmc3_namcot_118_cmd;
				mmc3_io[4]=mmc3_namcot_118_mirroring;
				INIT_MIR_0();
			}
			break;
		case MAPTYPE_NAMCOT_118_CHR_CONTROLLED_MIRRORING: /* 118 with chr mirroring */
			mmc3_disable(BIN8(11000000)); /* cmd and page only */
			cartridge->wram_isenabled=cartridge->battery=FALSE;
			mmc3_io[1]=mmc3_namcot_118c_page;
			mmc3_bs=&mmc3_namcot_118c_bs;
			INIT_MIR_C();
			break;
		
		/* TENGEN */
		
		/* MIMIC-1 */
		case MAPTYPE_TENGEN_MIMIC_1:
			mmc3_disable(BIN8(11000000)); /* cmd and page only */
			cartridge->wram_isenabled=cartridge->battery=FALSE;
			break;
		
		/* PIRATE */
		
		/* Mario Party */
		case MAPTYPE_PIRATE_MULTI_MMC3_MARIO_7_IN_1:
			SET_MAPRESET(mmc3_marioparty);
			mmc3_bs=&mmc3_marioparty_bs;
			SET_MAPIO6(mmc3_marioparty);
			INIT_MIR_V();
			break;
		
		/* Super Big 7-in-1 */
		case MAPTYPE_PIRATE_MULTI_MMC3_SUPER_BIG_7_IN_1:
			mmc3_io[3]=mmc3_superbig7in1_register;
			MAXPRG=0xf;
			mmc3_bs=&mmc3_superbig7in1_bs;
			INIT_MIR_V();
			break;
		
		default: break;
	}
	
	mmc3_prg_bs(); /* initial prg init here */
}
