/* Nintendo MMC5 mapper */

static struct {
	/* std */
	int wram_size;
	int wram_protection[2];
	int exram_usage;
	int fillnt_ti;
	int fillnt_at;
	int nt_mode[4];
	int nt_index;
	int prgbs_mode;
	int chrbs_mode;
	int mirroring;
	int prg[4][4];
	int irq_scanline;
	int irq_isenabled;
	int irq_status;
	int chr[2][4][8];
	int chr_use;
	int chr_high;
	int split_isenabled;
	int split_side;
	int split_pos;
	int split_yscroll;
	int split_chr;
	int split_offset;
	int addressmode_split;
	int mul[2];
	u8* wram;
	u8* exram;
	
	/* sound */
	int* dac_squares;
	int* dac_pcm;
	
	int* pcm_buffer;
	int pcm_cycles;
	int pcm_buffer_counter;
	int pcm_sample;
} map_mmc5;


/* --- MMC5/MMC5B(NES-EKROM/NES-ELROM/NES-ETROM/NES-EWROM) (iNES: 5) ---
	Bandit Kings of Ancient China/Suikoden - Tenmei no Chikai, Uncharted Waters/Daikoukai Jidai, Just Breed,
	Romance of the Three Kingdoms II/Sangokushi 2, Nobunaga's Ambition 2/Nobunaga no Yabou - Sengoku Gunyuu Den,
	Gemfire/Royal Blood, Laser Invasion/Gun Sight, Castlevania III - Dracula's Curse, Metal Slader Glory,
	Shin 4 Nin Uchi Mahjong - Yakuman Tengoku, Uchuu Keibitai SDF, Nobunaga no Yabou - Bushou Fuuun Roku,
	Ishin no Arashi, Aoki Ookami to Shiroki Mejika - Genchou Hishi
	
	PD: BAPSACRSTRKCGTF, demo3, Teenage Mutant Ninja Turtles DEMO
*/

/* TODO:
	- $5010: PCM related?
	- sound gets clocked by the APU frame sequencer?
	- Castlevania III (PAL): supposedly, part of the background should shine through during vertical scrolling
	
	(BAPSACRSTRKCGTF doesn't work (IRQ related, probably tested on old inaccurate emulators))
*/

#define MMC5_BATTERY_64K FALSE /* 64K for every battery sram file, instead of autodetection */

static __fastcall void mmc5_mirroring(void);
static __fastcall void mmc5_prgbs(void);
static __inline__ void mmc5_chrbs(void);
static __inline__ int mmc5_insplit(void);


/* sound; 2 squares, and an 8 bit PCM channel. used mostly by Just Breed and Shin 4 Nin Uchi Mahjong, the latter being the only game using PCM */

static void mmc5_sound_init(void);
static void mmc5_sound_clean(void);

static __fastcall void mmc5_square0_update(void);
static __fastcall void mmc5_square1_update(void);
static __fastcall void mmc5_pcm_update(void);

static ac_channel* mmc5_square0;
static ac_channel* mmc5_square1;

static int mmc5_pcm_volume_lookup[0x100];

/* clean */
static void mmc5_sound_clean(void)
{
	AC_CLEAN_CHANNEL(mmc5_square0);
	AC_CLEAN_CHANNEL(mmc5_square1);
}

/* soft reset */
_MAPRESET(mmc5)
{
	MAPIO_FUN(mmc5)(0x5015,0,0);
}

_MAPSOUNDBC(mmc5) { map_mmc5.pcm_buffer_counter=mmc5_square0->buffer_counter=mmc5_square1->buffer_counter=bc; }

_MAPKILLCYCLE(mmc5)
{
	map_mmc5.pcm_cycles-=crystal->ppu_cycle;
	mmc5_square0->cycles-=crystal->ppu_cycle;
	mmc5_square1->cycles-=crystal->ppu_cycle;
}

/* squares are almost the same as the ones in the APU, but without sweep */
#define MMC5_SQUARE_UPDATE(x) \
	int cpu_cycles; \
	\
	for (cpu_cycles=cpu_get_cycles();x->cycles>cpu_cycles;x->cycles-=crystal->rel_cycle) { \
		if (!--x->timer) { \
			x->timer=(x->timer_raw+1)<<1; \
			x->sequencer_counter=(x->sequencer_counter+1)&7; \
			x->sample=ac_duty_waveform[x->duty][x->sequencer_counter]; \
		} \
		if (x->sample) { \
			if ((x->length_counter==0)|(x->timer_raw<8)) x->buffer[x->buffer_counter++]=0; \
			else x->buffer[x->buffer_counter++]=x->volume; \
		} \
		else x->buffer[x->buffer_counter++]=0; \
	}

static __fastcall void mmc5_square0_update(void) { MMC5_SQUARE_UPDATE(mmc5_square0); }
static __fastcall void mmc5_square1_update(void) { MMC5_SQUARE_UPDATE(mmc5_square1); }

/* pcm */
static __fastcall void mmc5_pcm_update(void)
{
	int cpu_cycles;
	
	for (cpu_cycles=cpu_get_cycles();map_mmc5.pcm_cycles>cpu_cycles;map_mmc5.pcm_cycles-=crystal->rel_cycle) {
		map_mmc5.pcm_buffer[map_mmc5.pcm_buffer_counter++]=map_mmc5.pcm_sample;
	}
}

/* new frame */
_MAPNF(mmc5)
{
	mapper_new_frame_resetsc(); /* for irq */
	map_mmc5.pcm_cycles=mmc5_square0->cycles=mmc5_square1->cycles=cpu_get_cycles(); /* for sound */
}

/* end frame */
_MAPEF(mmc5)
{
	int i;
	
	mmc5_pcm_update();
	mmc5_square0_update();
	mmc5_square1_update();
	
	/* fill dacs */
	i=mmc5_square0->buffer_counter;
	while (i--) {
		map_mmc5.dac_squares[i]=ac_square_volume_lookup[mmc5_square0->buffer[i]+mmc5_square1->buffer[i]];
		map_mmc5.dac_pcm[i]=mmc5_pcm_volume_lookup[map_mmc5.pcm_buffer[i]];
	}
}

/* gets clocked by the APU frame sequencer */
_MAPAPUCLOCK(mmc5)
{
	AC_ENVELOPE_CLOCK(mmc5_square0);
	AC_ENVELOPE_CLOCK(mmc5_square1);
	if (c) {
		AC_LENGTH_COUNTER_CLOCK(mmc5_square0);
		AC_LENGTH_COUNTER_CLOCK(mmc5_square1);
	}
}

/* sound init */
static void mmc5_sound_init(void)
{
	int i;
	
	SET_MAPRESET(mmc5);
	SET_MAPNF(mmc5);
	SET_MAPEF(mmc5);
	SET_MAPKILLCYCLE(mmc5);
	SET_MAPSOUNDBC(mmc5);
	SET_MAPAPUCLOCK(mmc5);
	map_mmc5.pcm_buffer=sound_create_channel();
	AC_CREATE_CHANNEL(mmc5_square0);
	AC_CREATE_CHANNEL(mmc5_square1);
	mmc5_square0->buffer=sound_create_channel();
	mmc5_square1->buffer=sound_create_channel();
	map_mmc5.dac_pcm=sound_create_dac();
	map_mmc5.dac_squares=sound_create_dac();
	
	mmc5_square0->timer=mmc5_square1->timer=2;
	ac_fill_volume_lookup();
	for (i=0;i<0x100;i++) mmc5_pcm_volume_lookup[i]=i*33;
}



/* ppu memmap overrides */
/* patterntable reads */
#define MMC5_GET_PATDATA(x) \
	u8 ret=0; \
	\
	if ((map_mmc5.exram_usage==1)&!*mapper.ppu_reading_sprites_ptr) { \
		ret=cartridge->vrom[((((map_mmc5.chr_high<<6)|(map_mmc5.exram[map_mmc5.nt_index]&BIN8(00111111)))<<2)|(x&3))&MAXVROM][address&0x3ff]; \
	} \
	else { \
		if (((*mapper.ppu_control1_ptr&BIT(5))!=0)&*mapper.ppu_rendering_ptr) { /* 8*16 sprite mode */ \
			if (*mapper.ppu_reading_sprites_ptr) ret=cartridge->vrom[map_mmc5.chr[0][map_mmc5.chrbs_mode][x]][address&0x3ff]; \
			else ret=cartridge->vrom[map_mmc5.chr[1][map_mmc5.chrbs_mode][x]][address&0x3ff]; \
		} \
		else ret=cartridge->cur_patbank[x][address&0x3ff]; \
	} \
	if (mmc5_insplit()) { \
		if (map_mmc5.addressmode_split) { map_mmc5.addressmode_split=FALSE; *mapper.ppu_address_ptr=mapper.ppu_address_old; } \
		ret=cartridge->vrom[(map_mmc5.split_chr<<2|(x&3))&MAXVROM][address&0x3ff]; \
	} \
	return ret
/* PT0 */
_MAPPPU_R(mmc5,pt0a) { MMC5_GET_PATDATA(0); }
_MAPPPU_R(mmc5,pt0b) { MMC5_GET_PATDATA(1); }
_MAPPPU_R(mmc5,pt0c) { MMC5_GET_PATDATA(2); }
_MAPPPU_R(mmc5,pt0d) { MMC5_GET_PATDATA(3); }
/* PT1 */
_MAPPPU_R(mmc5,pt1a) { MMC5_GET_PATDATA(4); }
_MAPPPU_R(mmc5,pt1b) { MMC5_GET_PATDATA(5); }
_MAPPPU_R(mmc5,pt1c) { MMC5_GET_PATDATA(6); }
_MAPPPU_R(mmc5,pt1d) { MMC5_GET_PATDATA(7); }
/* nametable reads */
#define MMC5_GET_NAMEDATA(x) \
	u8 ret=0; \
	\
	if (map_mmc5.nt_mode[x]&1) return ret; \
	if (mmc5_insplit()) { \
		if (!map_mmc5.addressmode_split) { \
			map_mmc5.addressmode_split=TRUE; \
			mapper.ppu_address_old=*mapper.ppu_address_ptr; \
			*mapper.ppu_address_ptr=((((*mapper.ppu_sc_ptr+map_mmc5.split_yscroll)<<12)|((((*mapper.ppu_sc_ptr+map_mmc5.split_yscroll)<240)*(*mapper.ppu_sc_ptr+map_mmc5.split_yscroll))<<2))&BIN16(01110011,11100000))|(map_mmc5.split_offset>>3&BIN8(00011111)); /* cloak ppu address */ \
		} \
		if ((address&0x3ff)<0x3c0) ret=map_mmc5.exram[*mapper.ppu_address_ptr&BIN16(00000011,11111111)]; \
		else { \
			ret=map_mmc5.exram[address&0x3ff]; \
			if (map_mmc5.addressmode_split) { map_mmc5.addressmode_split=FALSE; *mapper.ppu_address_ptr=mapper.ppu_address_old; }; \
		} \
	} \
	else { \
		if (map_mmc5.nt_mode[x]&2) { \
			if ((address&0x3ff)<0x3c0) ret=map_mmc5.fillnt_ti; \
			else ret=map_mmc5.fillnt_at; \
		} \
		else ret=mapper.ppu_memmap_cur_name_ptr[x][address&0x3ff]; \
	} \
	if (map_mmc5.exram_usage==1) { \
		if ((address&0x3ff)<0x3c0) map_mmc5.nt_index=address&0x3ff; \
		else ret=(map_mmc5.exram[map_mmc5.nt_index]>>6)|(map_mmc5.exram[map_mmc5.nt_index]>>4&BIN8(00001100))|(map_mmc5.exram[map_mmc5.nt_index]>>2&BIN8(00110000))|(map_mmc5.exram[map_mmc5.nt_index]&BIN8(11000000)); \
	} \
	return ret

_MAPPPU_R(mmc5,name0) { MMC5_GET_NAMEDATA(0); }
_MAPPPU_R(mmc5,name1) { MMC5_GET_NAMEDATA(1); }
_MAPPPU_R(mmc5,name2) { MMC5_GET_NAMEDATA(2); }
_MAPPPU_R(mmc5,name3) { MMC5_GET_NAMEDATA(3); }
/* nametable writes */
#define MMC5_SET_NAMEDATA(x) if (!map_mmc5.nt_mode[x]) ppu_write_name##x(address,data)

_MAPPPU_W(mmc5,name0) { MMC5_SET_NAMEDATA(0); }
_MAPPPU_W(mmc5,name1) { MMC5_SET_NAMEDATA(1); }
_MAPPPU_W(mmc5,name2) { MMC5_SET_NAMEDATA(2); }
_MAPPPU_W(mmc5,name3) { MMC5_SET_NAMEDATA(3); }
/* --- */


/* IO reads */
_MAPIO_R(mmc5)
{
	u8 ret=address>>8;
	switch (address) {
		/* sound */
		case 0x5010: break; /* ? */
		case 0x5015: ret=ret|(mmc5_square0->length_counter!=0)|((mmc5_square1->length_counter!=0)<<1); break; /* length counters > 0 */
		
		/* other */
		case 0x5204: ret=map_mmc5.irq_status; map_mmc5.irq_status&=BIN8(01111111); cpu_acknowledge_interrupt(~IRQ_MAPPER); break; /* get irq status */
		case 0x5205: ret=(map_mmc5.mul[0]*map_mmc5.mul[1])&0xff; break; /* get multiplier result: low byte */
		case 0x5206: ret=(map_mmc5.mul[0]*map_mmc5.mul[1])>>8&0xff; break; /* get multiplier result: high byte */
		
		default:
			if (address>0x5bff) { /* read exram, only when enabled */
				if (map_mmc5.exram_usage&2) ret=map_mmc5.exram[address&0x3ff];
			}
			
			#if DEBUG_MAPPER
			else LOG(LOG_MAPPER,"%x\n",address);
			#endif
			break;
	}
	return ret;
}


/* IO writes */
_MAPIO(mmc5)
{
	int i,p;
	switch (address) {
		/* sound */
		/* square0 */
		case 0x5000: AC_ENVELOPE_REGISTER(mmc5_square0); AC_DUTY_REGISTER(mmc5_square0); break;
		case 0x5002: AC_TIMER_LOW_REGISTER(mmc5_square0); break;
		case 0x5003: AC_LENGTH_COUNTER_REGISTER(mmc5_square0); AC_TIMER_HIGH_REGISTER(mmc5_square0); AC_DUTY_RESET(mmc5_square0); break;
		/* square1 */
		case 0x5004: AC_ENVELOPE_REGISTER(mmc5_square1); AC_DUTY_REGISTER(mmc5_square1); break;
		case 0x5006: AC_TIMER_LOW_REGISTER(mmc5_square1); break;
		case 0x5007: AC_LENGTH_COUNTER_REGISTER(mmc5_square1); AC_TIMER_HIGH_REGISTER(mmc5_square1); AC_DUTY_RESET(mmc5_square1); break;
		/* pcm */
		case 0x5010: break; /* ? */
		case 0x5011: mmc5_pcm_update(); map_mmc5.pcm_sample=data; break; /* set 8 bit pcm sample */
		
		case 0x5015: AC_LENGTH_COUNTER_ENABLE_REGISTER(mmc5_square0,0) AC_LENGTH_COUNTER_ENABLE_REGISTER(mmc5_square1,1) break; /* sound */
		
		/* bankswitching modes */
		#define MMC5_WRAM_PROTECTION(x) map_mmc5.wram_protection[x]=data&3; cartridge->wram_writeenable=(map_mmc5.wram_protection[0]==2)&(map_mmc5.wram_protection[1]==1)
		case 0x5100: map_mmc5.prgbs_mode=data&3; mmc5_prgbs(); break; /* set prg bankswitch mode */
		case 0x5101: ppu_force_update(); map_mmc5.chrbs_mode=data&3; mmc5_chrbs(); break; /* set chr bankswitch mode */
		case 0x5102: MMC5_WRAM_PROTECTION(0); break; /* set wram protection (reg1) */
		case 0x5103: MMC5_WRAM_PROTECTION(1); break; /* set wram protection (reg2) */
		case 0x5104: ppu_force_update(); map_mmc5.exram_usage=data&3; mmc5_mirroring(); break; /* set exram usage */
		case 0x5105: map_mmc5.mirroring=data; mmc5_mirroring(); break; /* set nametable/mirroring */
		case 0x5106: ppu_force_update(); map_mmc5.fillnt_ti=data; break; /* set fill nametable tile index */
		case 0x5107: ppu_force_update(); map_mmc5.fillnt_at=((data&3)<<6)|((data&3)<<4)|((data&3)<<2)|(data&3); break; /* set fill nametable attribute data */
		
		/* prg banks */
		case 0x5113: /* select 8k ram for $6000 */
			p=data&7;
			if (p>map_mmc5.wram_size) map_mmc5.wram_size=p;
			cartridge->cur_wrambank=map_mmc5.wram+(p<<13);
			mmc5_prgbs();
			break;
		case 0x5114: /* select 8k prg bank for $8000-$9fff */
			map_mmc5.prg[3][0]=data;
			mmc5_prgbs();
			break;
		case 0x5115: /* select 8/16k prg bank for x-$bfff */
			map_mmc5.prg[3][1]=data; /* 3: 8k */
			data&=BIN8(11111110); map_mmc5.prg[1][0]=map_mmc5.prg[2][0]=data; map_mmc5.prg[1][1]=map_mmc5.prg[2][1]=data|1; /* 1/2: 16k */
			mmc5_prgbs();
			break;
		case 0x5116: /* select 8 prg bank for $c000-$dfff */
			map_mmc5.prg[2][2]=map_mmc5.prg[3][2]=data; /* 2/3: 8k */
			mmc5_prgbs();
			break;
		case 0x5117: /* select 8/16/32k prg bank for x-$ffff */
			data=data|0x80;
			map_mmc5.prg[2][3]=map_mmc5.prg[3][3]=data; /* 2/3: 8k */
			data&=BIN8(11111110); map_mmc5.prg[1][2]=data; map_mmc5.prg[1][3]=data|1; /* 1: 16k */
			data&=BIN8(11111100); for (i=3;i>=0;i--) map_mmc5.prg[0][i]=data|i; /* 0: 32k */
			mmc5_prgbs();
			break;
		
		/* chr banks */
		case 0x5120: case 0x5121: case 0x5122: case 0x5123: case 0x5124: case 0x5125: case 0x5126: case 0x5127: /* sprite banks */
			ppu_force_update();
			map_mmc5.chr[0][3][address&7]=(data|(map_mmc5.chr_high<<8))&MAXVROM; /* 1k */
			if (address&1) { p=(data<<1)|(map_mmc5.chr_high<<9); map_mmc5.chr[0][2][(address&7)-1]=p&MAXVROM; map_mmc5.chr[0][2][address&7]=(p|1)&MAXVROM; } /* 2k */
			if ((address&3)==3) { p=data<<2; for (i=(address&7)-3;i<8;i++) map_mmc5.chr[0][1][i]=(p|(i&3))&MAXVROM; } /* 4k */
			if (address==0x5127) { p=data<<3; for (i=7;i>=0;i--) map_mmc5.chr[0][0][i]=(p|i)&MAXVROM; } /* 8k */
			map_mmc5.chr_use=0; mmc5_chrbs();
			break;
		case 0x5128: case 0x5129: case 0x512a: case 0x512b: /* background banks */
			ppu_force_update();
			map_mmc5.chr[1][3][address&3]=map_mmc5.chr[1][3][(address&3)+4]=(data|(map_mmc5.chr_high<<8))&MAXVROM; /* 1+1k */
			if (address&1) { /* 2+2k */
				p=(data<<1)|(map_mmc5.chr_high<<9);
				map_mmc5.chr[1][2][(address&3)-1]=map_mmc5.chr[1][2][(address&3)+3]=p&MAXVROM;
				map_mmc5.chr[1][2][address&3]=map_mmc5.chr[1][2][(address&3)+4]=(p|1)&MAXVROM;
			}
			if (address==0x512b) {
				p=data<<2; for (i=3;i>=0;i--) map_mmc5.chr[1][1][i]=map_mmc5.chr[1][1][i+4]=(p|i)&MAXVROM; /* 4+4k */
				p=data<<3; for (i=7;i>=0;i--) map_mmc5.chr[1][0][i]=(p|i)&MAXVROM; /* 8k */
			}
			map_mmc5.chr_use=1; mmc5_chrbs();
			break;
		case 0x5130: ppu_force_update(); map_mmc5.chr_high=data&3; break; /* high bits for small high banks selection */
		
		/* other */
		case 0x5200: /* set screensplit enable, side, column position. seems to be used only in the intro of Uchuu Keibitai SDF */
			ppu_force_update();
			map_mmc5.split_isenabled=data&0x80;
			map_mmc5.split_side=data&0x40;
			map_mmc5.split_pos=data<<3&0xff;
			break;
		case 0x5201: ppu_force_update(); map_mmc5.split_yscroll=data; break; /* set screensplit y scroll value */
		case 0x5202: ppu_force_update(); map_mmc5.split_chr=data; break; /* set screensplit chrbank */
		case 0x5203: map_mmc5.irq_scanline=data; break; /* set scanline for irq */
		case 0x5204: /* enable/disable irqs */
			map_mmc5.irq_isenabled=data&0x80;
			if (map_mmc5.irq_isenabled&map_mmc5.irq_status) cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|INT_EDGE_IRQ_EX);
			else cpu_acknowledge_interrupt(~IRQ_MAPPER);
			break;
		case 0x5205: map_mmc5.mul[0]=data; break; /* multiplier value 1 */
		case 0x5206: map_mmc5.mul[1]=data; break; /* multiplier value 2 */
		
		default:
			if (address>0x5bff) {
				switch (map_mmc5.exram_usage) {
					case 0: case 1:
						ppu_force_update();
						if (*mapper.ppu_enabled_ptr&*mapper.ppu_valid_scanline_ptr) map_mmc5.exram[address&0x3ff]=data;
						else map_mmc5.exram[address&0x3ff]=0;
						break;
					case 2: map_mmc5.exram[address&0x3ff]=data; break;
					default: break;
				}
			}
			
			#if DEBUG_MAPPER
			else LOG(LOG_MAPPER,"%x %x\n",address,data);
			#endif
			break;
	}
}


/* handler for writes to programbanks (in case of wram) */
_MAPIO(mmc5_pb) { if (cartridge->wram_writeenable&&cartridge->cur_prgbank_isram[address>>13&3]) cartridge->cur_prgbank[address>>13&3][address&0x1fff]=data; }

/* irq handler, based on scanlines */
_MAPIRQ(mmc5)
{
	for (;;) {
		if (*mapper.cpu_cycles_ptr<crystal->sc[mapper.sc+1]) { /* new scanline */
			mapper.sc++;
			if (mapper.sc>241) return;
			else if (mapper.sc==241) { map_mmc5.irq_status&=0x80; return; } /* end of visible frame */
			if (*mapper.ppu_enabled_ptr) {
				if (mapper.sc==1) { map_mmc5.irq_status=0; cpu_acknowledge_interrupt(~IRQ_MAPPER); }
				else if (mapper.sc>1) {
					if ((map_mmc5.irq_scanline+1)==mapper.sc) map_mmc5.irq_status|=0x80;
					if (map_mmc5.irq_isenabled&map_mmc5.irq_status) cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER);
					map_mmc5.irq_status|=0x40;
				}
			}
		}
		else break;
	}
}

/* update mirroring */
static __fastcall void mmc5_mirroring(void)
{
	int i=4;
	ppu_force_update();
	while(i--) {
		switch ((map_mmc5.mirroring>>(i<<1))&3) {
			case 0: mapper.ppu_memmap_cur_name_ptr[i]=mapper.ppu_memmap_name_ram_ptr[0]; map_mmc5.nt_mode[i]=0; break;
			case 1: mapper.ppu_memmap_cur_name_ptr[i]=mapper.ppu_memmap_name_ram_ptr[1]; map_mmc5.nt_mode[i]=0; break;
			case 2:
				if (map_mmc5.exram_usage&2) map_mmc5.nt_mode[i]=1;
				else { mapper.ppu_memmap_cur_name_ptr[i]=map_mmc5.exram; map_mmc5.nt_mode[i]=0; }
				break;
			case 3: map_mmc5.nt_mode[i]=2; break;
			default: break;
		}
	}
}

/* prg bankswitch */
static __fastcall void mmc5_prgbs(void)
{
	int i=4;
	while(i--) {
		if(map_mmc5.prg[map_mmc5.prgbs_mode][i]&0x80) {
			PRG8_BS(i,(map_mmc5.prg[map_mmc5.prgbs_mode][i]&BIN8(01111111))&MAXPRG);
			cartridge->cur_prgbank_isram[i]=FALSE;
		}
		else {
			int j=map_mmc5.prg[map_mmc5.prgbs_mode][i]&7;
			if (j>map_mmc5.wram_size) map_mmc5.wram_size=j;
			cartridge->cur_prgbank[i]=map_mmc5.wram+(j<<13);
			cartridge->cur_prgbank_isram[i]=TRUE;
		}
	}
}

/* chr bankswitch */
static __inline__ void mmc5_chrbs(void)
{
	#define MMC5_CHR1_BS(x) CHR1_BS(x,map_mmc5.chr[map_mmc5.chr_use][map_mmc5.chrbs_mode][x],vrom)
	MMC5_CHR1_BS(0); MMC5_CHR1_BS(1); MMC5_CHR1_BS(2); MMC5_CHR1_BS(3); MMC5_CHR1_BS(4); MMC5_CHR1_BS(5); MMC5_CHR1_BS(6); MMC5_CHR1_BS(7);
}

/* see if it's in split */
static __inline__ int mmc5_insplit(void)
{
	if ((map_mmc5.split_isenabled)&&(*mapper.ppu_rendering_ptr&!*mapper.ppu_reading_sprites_ptr&(((~map_mmc5.exram_usage)&2)>>1))&&
	((*mapper.ppu_scc_ptr<257)|((*mapper.ppu_scc_ptr>320)&(*mapper.ppu_scc_ptr<337)))) { /* ignore bogus nametables */
		
		if (*mapper.ppu_scc_ptr>256) map_mmc5.split_offset=*mapper.ppu_scc_ptr&8; /* hblank */
		else map_mmc5.split_offset=(*mapper.ppu_scc_ptr&BIN8(11111000))+16; /* +16 due to tile fetch cache */
		if (map_mmc5.split_side) return map_mmc5.split_offset>=map_mmc5.split_pos;
		else return map_mmc5.split_offset<=map_mmc5.split_pos;
	}
	else return FALSE;
}



/* clean */
_MAPCLEAN(mmc5)
{
	MEM_CLEAN(map_mmc5.exram);
	MEM_CLEAN(map_mmc5.wram);
	
	mmc5_sound_clean();
}

/* battery */
_MAPBATLOAD(mmc5)
{
	map_mmc5.wram_size=(file->size-(file->size>0))>>13&7;
	return FALSE;
}
_MAPBATSAVE(mmc5)
{
	if (!MMC5_BATTERY_64K) cartridge->battery_size=(map_mmc5.wram_size+1)<<13;
	cartridge->cur_wrambank=map_mmc5.wram;
	return FALSE;
}


/* main init */
_MAPINIT(mmc5)
{
	int i,j,k;
	SET_MAPCLEAN(mmc5);
	SET_MAPIO8(mmc5_pb,mmc5_pb,mmc5_pb,nothing);
	SET_MAPIO4_R(mmc5);
	SET_MAPIO4(mmc5);
	ppu_set_fpt(PPU_FPT_PT0,FALSE,mapper_mmc5_ppu_read_pt0a,mapper_mmc5_ppu_read_pt0b,mapper_mmc5_ppu_read_pt0c,mapper_mmc5_ppu_read_pt0d,NULL,NULL,NULL,NULL);
	ppu_set_fpt(PPU_FPT_PT1,FALSE,mapper_mmc5_ppu_read_pt1a,mapper_mmc5_ppu_read_pt1b,mapper_mmc5_ppu_read_pt1c,mapper_mmc5_ppu_read_pt1d,NULL,NULL,NULL,NULL);
	ppu_set_fpt(PPU_FPT_NT,FALSE,mapper_mmc5_ppu_read_name0,mapper_mmc5_ppu_read_name1,mapper_mmc5_ppu_read_name2,mapper_mmc5_ppu_read_name3,mapper_mmc5_ppu_write_name0,mapper_mmc5_ppu_write_name1,mapper_mmc5_ppu_write_name2,mapper_mmc5_ppu_write_name3);
	SET_MAPIRQ(mmc5);
	
	MEM_CREATE(map_mmc5.exram,0x400);
	MEM_CREATE(map_mmc5.wram,0x10000);
	DEF_6000_WRAM();
	cartridge->cur_wrambank=map_mmc5.wram;
	if (cartridge->battery) {
		cartridge->battery_size=0x10000;
		SET_MAPBATLOAD(mmc5);
		SET_MAPBATSAVE(mmc5);
	}
	map_mmc5.chr_use=1;
	map_mmc5.prgbs_mode=3;
	for (j=0;j<4;j++) for (i=0;i<4;i++) { map_mmc5.prg[j][i]=MAXPRG|0x80; PRG8_BS(i,MAXPRG); } /* prg banks: last 8k everywhere */
	for (k=0;k<2;k++) for (j=0;j<4;j++) for (i=0;i<8;i++) map_mmc5.chr[k][j][i]=i;
	INIT_CHR_FIRST();
	INIT_MIR_C();
	
	mmc5_sound_init();
}
