#include <windows.h>
#include <shlobj.h>

#include "global.h"
#include "crystal.h"
#include "mapper.h"
#include "cartridge.h"
#include "cpu_memorymap.h"
#include "cpu.h"
#include "apu.h"
#include "ppu.h"
#include "draw.h"
#include "input.h"
#include "sound.h"
#include "file.h"
#include "main.h"
#include "nes.h"
#include "cont.h"
#include "settings.h"
#include "sample.h"
#include "resource.h"


static const char* nes_priority_name[NES_PRIORITY_MAX]={"normal","above normal","high","auto"};
const char* nes_get_priority_name(u32 p) { if (p>=NES_PRIORITY_MAX) return NULL; else return nes_priority_name[p]; }


static struct {
	int wait;
	int wait_paused;
	int waiting;
	
	int warm_boot;
	int soft_reset;
	int main_interrupt;
	int is_running;
	int paused;
	DWORD pauselog_ticks;
	DWORD pauselog_ticks_next;
	int pauselog_visible;
	void* pause_sample;
	int semaphore;
	DWORD tid;
	HANDLE thread;
	int priority;
	int priority_auto;
	int thread_priority_prev;
	DWORD app_priority_prev;
	
	int filterindex;
	char openfn_prev[STRING_SIZE];
	int openview;
	
	struct {
		u32 crc;
		u32 crc_p;
		int mir;
		int maptype;
		int battery;
	} mapinfo[4];
	
	u8 header[4][0x10];
	HWND template;
} nes;


void nes_set_soft_reset(void) { nes.soft_reset=TRUE; }
void nes_reset(int hard)
{
	if (nes_get_paused()&1) SendMessage(MAIN->window,WM_COMMAND,IDM_PAUSEEMU,0); /* force unpause */
	
	ppu_clear_nmi_enable();
	apu_read_status(0x4015);
	apu_write_status(0);
	apu_rewrite_fs();
	mapper_reset();
	cont_reset();
	cpu_reset(hard);
	ppu_reset(hard,nes.warm_boot);
	
	nes.warm_boot=nes.soft_reset=FALSE;
}

void nes_videoformat(void)
{
	crystal->mode^=1;
	if (draw_get_vidformat()!=DRAW_VIDFORMAT_AUTO) draw_set_vidformat(crystal->mode?DRAW_VIDFORMAT_PAL:DRAW_VIDFORMAT_NTSC);
	crystal_set_mode(crystal->mode);
	LOG(LOG_MISC|LOG_COLOUR(LC_GREEN)|LOG_TYPE(LT_VIDFORMAT),"%s video format ",crystal->mode?"PAL":"NTSC");
	
	if (nes.is_running) {
		apu_fill_fs_update_frequency_lookup();
		apu_fill_dmc_update_frequency_lookup();
		apu_fill_noise_frequency_lookup();
		mapper_change_mode();
	}
}



int __fastcall nes_is_running(void) { return nes.is_running|(nes.semaphore&1); }

int __fastcall nes_get_paused(void) { return nes.paused; }
int nes_get_filterindex(void) { return nes.filterindex; }
int nes_get_priority(void) { return nes.priority; }
int nes_get_priority_auto(void) { return nes.priority_auto; }

void __fastcall nes_set_main_interrupt(void)
{
	nes.main_interrupt=TRUE;
	if (nes_is_running()) return;
	
	input_read(); input_read_shortcuts();
	draw_draw();
}

void __fastcall nes_set_paused(int p)
{
	if (nes_is_running()) {
		int po=nes.paused;
		nes.paused=p;
		
		/* play SMB3 pause sound, thanks to Samor for this funny idea */
		if ((po^1)&p&1) sound_play_sample(nes.pause_sample,FALSE);
	}
	else nes.paused=FALSE;
}

void __fastcall nes_set_priority(int p)
{
	int tp;
	DWORD ap;
	
	nes.priority=p;
	
	switch (nes.priority) {
		case 1: tp=THREAD_PRIORITY_ABOVE_NORMAL; ap=ABOVE_NORMAL_PRIORITY_CLASS; break;
		case 2: tp=THREAD_PRIORITY_HIGHEST; ap=HIGH_PRIORITY_CLASS; break;
		default: tp=THREAD_PRIORITY_NORMAL; ap=NORMAL_PRIORITY_CLASS; break;
	}
	
	/* app */
	if (nes.priority_auto) {
		if (MAIN->foreground&&nes.is_running&&!MAIN->dialog&&sound_get_slept()) ap=HIGH_PRIORITY_CLASS;
		else ap=NORMAL_PRIORITY_CLASS;
	}
	
	if (ap!=nes.app_priority_prev) {
		nes.app_priority_prev=ap;
		if (!SetPriorityClass(GetCurrentProcess(),ap)&&ap==ABOVE_NORMAL_PRIORITY_CLASS) SetPriorityClass(GetCurrentProcess(),HIGH_PRIORITY_CLASS); /* 'above normal' not supported in winnt/95/98/me */
	}
	
	/* thread */
	if (nes.is_running) {
		if (nes.priority_auto) {
			if (MAIN->foreground&&!MAIN->dialog&&sound_get_slept()) tp=THREAD_PRIORITY_ABOVE_NORMAL;
			else tp=THREAD_PRIORITY_NORMAL;
		}
		
		if (tp!=nes.thread_priority_prev) {
			nes.thread_priority_prev=tp;
			SetThreadPriority(nes.thread,tp);
		}
	}
}

void nes_init(void)
{
	char* c=NULL;
	int p,size;
	u8* s;
	
	memset(&nes,0,sizeof(nes));
	nes.thread_priority_prev=-1;
	nes.app_priority_prev=~0;
	
	/* settings */
	/* priority */
	/* unstable on not-auto: seems to cause very busy app on anything else but 'normal' if app is taking 99% cpu (eg. with sleep off) */
	p=NES_PRIORITY_AUTO;
	SETTINGS_GET_STRING(settings_info(SETTINGS_PRIORITY),&c);
	if (c&&strlen(c)) {
		for (p=0;p<NES_PRIORITY_MAX;p++) if (stricmp(c,nes_get_priority_name(p))==0) break;
		if (p==NES_PRIORITY_MAX) p=NES_PRIORITY_AUTO;
	}
	MEM_CLEAN(c);
	if (p==NES_PRIORITY_AUTO) { nes.priority_auto=TRUE; p=NES_PRIORITY_NORMAL; }
	nes_set_priority(p);
	
	/* pause sample */
	s=file_from_resource(ID_SAMPLE_PAUSE,&size);
	nes.pause_sample=sound_create_sample((void*)s,size);
	file_from_resource_close();
	
	/* file filter index */
	nes.filterindex=1;
	SETTINGS_GET_INT(settings_info(SETTINGS_FILTERINDEX),&nes.filterindex);
	CLAMP(nes.filterindex,1,3);
	
	cartridge_init();
	mapper_build_std_info();
	mapper_init();
	mapper_init_switches();
	nes_stop(FALSE);
	main_menu_enable(IDM_RELOADGAME,FALSE);
}

void nes_clean(void)
{
	memset(&nes,0,sizeof(nes));
	
	cartridge_clean();
}

static DWORD WINAPI nes_thread(void)
{
	while (nes.semaphore&2) { Sleep(1); } /* not done yet opening game */
	
	if ((CoInitializeEx(NULL,COINIT_APARTMENTTHREADED))!=S_OK) {
		LOG(LOG_MISC|LOG_ERROR,"can't initialize COM!\n");
		exit(1);
	}
	
	/* init */
	ppu_set_update_nothing();
	sound_create_ll();
	input_reset_updframe();
	crystal->fc=0;
	cont_init_cont();
	input_affirm_ports();
	
	draw_init_palette(TRUE);
	draw_reset_burst_phase();
	ppu_memmap_init();
	cpu_memmap_init();
	sample_init();
	mapper_init();
	mapper_init_mapper(cartridge->type);
	cartridge_battery_load();
	
	cpu_init();
	apu_init();
	ppu_init();
	
	input_mouse_cursor_reset_ticks();
	
	nes_reset(TRUE);
	
	sound_reset_write_cursor();
	sound_reset_capt_cursor();
	sound_reset_lostframes();
	sound_reset_slept();
	
	/* run */
	while (nes.is_running) {
		
		/* new frame */
		sound_new_frame();
		ppu_new_frame(); /* crystal_new_frame done there */
		cpu_new_frame();
		apu_new_frame();
		mapper_new_frame();
		
		cpu_sprite_dma_transfer_continue();
		
		if (nes.soft_reset) {
			cpu_sync_reset_time();
			nes_reset(FALSE);
		}
		
		if (crystal->odd_frame) {
			cpu_execute(crystal->sc[1]); /* kill cycle */
			ppu_force_update();
		}
		
		if (cpu_execute(crystal->vblank_trigger)) cpu_execute_halted(crystal->vblank_trigger);
		ppu_force_update();
		cpu_do_vblank_nmi();
		cpu_set_nmi(0);
		
		if (draw_get_source_height()==242) {
			if (cpu_execute(crystal->vblank_trigger-crystal->full_scanline)) cpu_execute_halted(crystal->vblank_trigger-crystal->full_scanline);
			ppu_force_update();
		}
		draw_draw();
		
		/* vblank */
		input_vblank();
		if (cpu_execute(0)) cpu_execute_halted(0);
		/*ppu_force_update();*/ /* no need */
		ppu_clear_status();
		
		apu_end_frame();
		mapper_end_frame();
		sound_compress();
		
		/* wait request? (time for external changes) */
		while (nes.is_running&(nes.waiting=nes.wait)) Sleep(1);
		
		
		if (nes.paused) {
			
			sound_silence(TRUE); sound_reset_lostframes();
			nes.pauselog_visible=TRUE; nes.pauselog_ticks_next=0;
			nes.main_interrupt=FALSE;
			
			while (nes.is_running&&nes.paused) {
				nes.pauselog_ticks=GetTickCount();
				if (nes.pauselog_ticks>nes.pauselog_ticks_next) { nes.pauselog_visible^=1; nes.pauselog_ticks_next=nes.pauselog_ticks+500; }
				
				if (!nes.wait_paused) {
					LOG(LOG_MISC|LOG_COLOUR(LC_ORANGE)|LOG_TYPE(LT_PAUSE),nes.pauselog_visible?"= paused =":"          ");
					
					/* main sends "interrupt" */
					if (nes.main_interrupt) {
						nes.main_interrupt=FALSE;
						
						input_read(); input_read_shortcuts();
						draw_draw();
					}
				}
				Sleep(3);
			}
			
			while (nes.is_running&nes.wait_paused) Sleep(1);
			WaitForSingleObject(MAIN->mutex,INFINITE); ReleaseMutex(MAIN->mutex);
			
			LOG(LOG_MISC|LOG_COLOUR(LC_ORANGE)|LOG_TYPE(LT_PAUSE|LT_IGNORE),"          ");
			
			sound_reset_write_cursor();
			sound_stop_sample(nes.pause_sample,TRUE);
		}
		
		sound_write();
		
		if (nes.priority_auto) nes_set_priority(nes.priority);
		
		ppu_warm();
	}
	
	/* clean */
	
	sound_clean_ll();
	sound_silence(TRUE);
	sound_reset_lostframes();
	sound_reset_slept();
	
	cartridge_battery_save(cpu_get_halt());
	apu_clean();
	cpu_clean();
	cpu_memmap_clean();
	ppu_clean();
	mapper_clean();
	cont_clean_cont();
	sample_clean();
	
	CoUninitialize();
	
	return 0;
}

int nes_start(void)
{
	if (cartridge_load(CARTRIDGE_LOAD_PUT,&nes.mapinfo[1].maptype,&nes.mapinfo[1].battery,&nes.mapinfo[1].mir,&nes.mapinfo[1].crc,&nes.mapinfo[1].crc_p)) {
		char fn[STRING_SIZE]={0};
		
		nes_stop(FALSE);
		cartridge_unload(cartridge);
		cartridge_set_cartridge();
		
		/* remember file/dir for open dialog */
		file_setfile(NULL,cartridge->filename,NULL,"nes");
		if (file->name) strcpy(fn,file->name);
		if (file->ext) { strcat(fn,"."); strcat(fn,file->ext); }
		strcpy(nes.openfn_prev,fn);
		file_setdirectory(&file->romdir);
		
		if (nes.mapinfo[3].crc==cartridge->crc32&&nes.mapinfo[3].crc_p==cartridge->crc32_patch) cartridge->mode=crystal->mode; /* same game: keep previous video format setting */
		crystal_set_mode(cartridge->mode);
		nes.mapinfo[3].crc=cartridge->crc32; nes.mapinfo[3].crc_p=cartridge->crc32_patch; /* rest is unused */
		sound_create_secondary_buffer(FALSE);
		nes.is_running=TRUE;
		nes_wait_done();
		WaitForSingleObject(MAIN->mutex,INFINITE); ReleaseMutex(MAIN->mutex);
		
		draw_set_repaint(1);
		main_titlebar(cartridge->name);
		nes.thread=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)nes_thread,NULL,0,&nes.tid);
		SetThreadAffinityMask(nes.thread,1);
		
		nes.thread_priority_prev=-1;
		nes.app_priority_prev=~0;
		nes_set_priority(nes.priority);
		
		main_menu_enable(IDM_RELOADGAME,TRUE);
		main_menu_enable(IDM_CLOSEGAME,TRUE);
		main_menu_enable(IDM_RESETNES,TRUE);
		main_menu_enable(IDM_PAUSEEMU,TRUE);
		
		return TRUE;
	}
	
	else cartridge_unload(cartridge_temp);
	
	return FALSE;
}

void nes_stop(int pal)
{
	if (!nes_is_running()) return;
	
	main_menu_enable(IDM_CLOSEGAME,FALSE);
	main_menu_enable(IDM_RESETNES,FALSE);
	main_menu_enable(IDM_PAUSEEMU,FALSE);
	
	nes.semaphore=TRUE;
	nes.is_running=nes.paused=FALSE;
	main_menu_check(IDM_PAUSEEMU,FALSE);
	
	if (nes.thread!=NULL) {
		WaitForSingleObject(nes.thread,INFINITE);
		CloseHandle(nes.thread);
	}
	
	if (pal) draw_init_palette(FALSE);
	main_titlebar(NULL);
	
	nes.semaphore=FALSE;
	nes_set_priority(nes.priority);
	
	input_affirm_ports();
}

void nes_wait(void)
{
	if (!nes_is_running()||nes_get_paused()) {
		nes.wait_paused=TRUE;
		return;
	}
	
	while (nes.waiting|nes.wait|nes.wait_paused) Sleep(1);
	nes.wait=TRUE;
	while (!nes.waiting) Sleep(1);
}

void nes_wait_done(void)
{
	nes.wait_paused=nes.wait=FALSE;
}


void nes_drop(WPARAM p)
{
	char fn[STRING_SIZE]={0};
	HDROP drop=(HDROP)p;
	UINT numfiles=DragQueryFile(drop,~0,NULL,0);
	
	DragQueryFile(drop,0,fn,STRING_SIZE);
	DragFinish(drop);
	
	if (numfiles!=1) { LOG(LOG_MISC|LOG_WARNING,"can't load multiple files\n"); return; }
	
	file_setfile(NULL,fn,NULL,"nes");
	nes_start();
	
	SetForegroundWindow(MAIN->window);
}

void nes_reload(void)
{
	if (cartridge&&cartridge->filename&&strlen(cartridge->filename)) {
		file_setfile(NULL,cartridge->filename,NULL,"nes");
		nes.warm_boot=nes_is_running();
		if (!nes_start()) nes.warm_boot=FALSE;
	}
}


/* edit mapper settings dialogs */

/* save ines header */
static int nes_inesedit_save(HWND dialog)
{
	if (!file_open()||!file_read(nes.header[3],0x10)||memcmp(nes.header[0],nes.header[3],0x10)!=0) {
		file_close();
		log_set_window(dialog);
		LOG(LOG_MISC|LOG_ERROR,"couldn't check file for saving!\n");
		return 0;
	}
	file_close();
	
	if (memcmp(nes.header[0],nes.header[1],0x10)!=0) { /* identicle: no need to save */
		if (!file_open_rw()) {
			file_close();
			log_set_window(dialog);
			LOG(LOG_MISC|LOG_ERROR,"couldn't open file for saving!\n");
			return 0;
		}
		
		if (!file_write(nes.header[1],0x10)) {
			file_close();
			log_set_window(dialog);
			LOG(LOG_MISC|LOG_ERROR,"couldn't write iNES header to file!\n");
			return 0;
		}
		file_close();
	}
	
	memcpy(nes.header[0],nes.header[1],0x10);
	
	return 1;
}

static void nes_inesedit_checkapply(HWND dialog) { EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_APPLY),memcmp(nes.header[1],nes.header[2],0x10)!=0); }

#define OH_UPDATE_INES(x) sprintf(c,"(......)"); if (x&&mapperinfo_maptype2ines(nes.mapinfo[0].maptype)>=0) sprintf(c,"(%03d)",mapperinfo_maptype2ines(nes.mapinfo[0].maptype)); SetDlgItemText(dialog,IDC_OPENGAME_MAPTYPETEXT,c)

/* check new file */
static void nes_open_hook_newfile(HWND dialog)
{
	char fn[STRING_SIZE]={0},c[256];
	memset(c,0,256);
	SendMessage(GetParent(dialog),CDM_GETFILEPATH,STRING_SIZE,(LPARAM)fn);
	file_setfile(NULL,fn,NULL,"nes");
	
	log_disable(); /* disable warning messages from cartridge_load */
	
	if (cartridge_load(CARTRIDGE_LOAD_GET,&nes.mapinfo[0].maptype,&nes.mapinfo[0].battery,&nes.mapinfo[0].mir,&nes.mapinfo[0].crc,&nes.mapinfo[0].crc_p)) {
		if (nes.mapinfo[0].crc==nes.mapinfo[1].crc&&nes.mapinfo[0].crc_p==nes.mapinfo[1].crc_p) memcpy(&nes.mapinfo[0],&nes.mapinfo[1],sizeof(nes.mapinfo[0]));
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_PROPT),TRUE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_MAPTYPE),TRUE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_BATTERY),TRUE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_HMIR),mapperinfo_get_mapmir(nes.mapinfo[0].maptype)?FALSE:TRUE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_VMIR),mapperinfo_get_mapmir(nes.mapinfo[0].maptype)?FALSE:TRUE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_EDIT),file->is_zip?FALSE:TRUE);
	
		CheckDlgButton(dialog,IDC_OPENGAME_BATTERY,nes.mapinfo[0].battery?BST_CHECKED:BST_UNCHECKED);
		SendMessage(GetDlgItem(dialog,IDC_OPENGAME_MAPTYPE),CB_SETCURSEL,nes.mapinfo[0].maptype,0);
		CheckRadioButton(dialog,IDC_OPENGAME_HMIR,IDC_OPENGAME_VMIR,nes.mapinfo[0].mir?IDC_OPENGAME_VMIR:IDC_OPENGAME_HMIR);
		OH_UPDATE_INES(TRUE);
	}
	else {
		SendMessage(GetDlgItem(dialog,IDC_OPENGAME_MAPTYPE),CB_SETCURSEL,MAPTYPE_UNKNOWN,0);
		CheckDlgButton(dialog,IDC_OPENGAME_HMIR,BST_UNCHECKED);
		CheckDlgButton(dialog,IDC_OPENGAME_VMIR,BST_UNCHECKED);
		CheckDlgButton(dialog,IDC_OPENGAME_BATTERY,BST_UNCHECKED);
		OH_UPDATE_INES(FALSE);
		
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_PROPT),FALSE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_MAPTYPE),FALSE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_BATTERY),FALSE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_HMIR),FALSE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_VMIR),FALSE);
		EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_EDIT),FALSE);
	}
	
	log_enable();
}

/* edit ines header dialog */
static BOOL CALLBACK nes_inesedit(HWND dialog,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch (msg) {
		
		case WM_INITDIALOG:
			memcpy(nes.header[1],nes.header[0],0x10); /* edit */
			memcpy(nes.header[2],nes.header[0],0x10); /* last applied */
			EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_APPLY),FALSE);
			EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_REMOVEGARBAGE),(((nes.header[1][7]&BIN8(00001110))!=0)|(nes.header[1][9]>1)|(memcmp(nes.header[1]+10,"\0\0\0\0\0\0",6)!=0))?TRUE:FALSE);
			
			/* init editboxes */
			#define s(x,y) SendDlgItemMessage(dialog,x,EM_LIMITTEXT,3,0); SetDlgItemInt(dialog,x,y,FALSE)
			s(IDC_INESEDIT_MAPNUM,(nes.header[1][7]&0xf0)|(nes.header[1][6]>>4));
			s(IDC_INESEDIT_PRG,nes.header[1][4]);
			s(IDC_INESEDIT_CHR,nes.header[1][5]);
			s(IDC_INESEDIT_RAM,nes.header[1][8]);
			#undef s
			
			/* init videoformat */
			SendDlgItemMessage(dialog,IDC_INESEDIT_VIDEOFORMAT,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)"NTSC"));
			SendDlgItemMessage(dialog,IDC_INESEDIT_VIDEOFORMAT,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)"PAL"));
			SendDlgItemMessage(dialog,IDC_INESEDIT_VIDEOFORMAT,CB_SETCURSEL,nes.header[1][9]&1,0);
			
			/* init checkboxes/radiobuttons */
			CheckDlgButton(dialog,IDC_INESEDIT_BATTERY,(nes.header[1][6]&2)?BST_CHECKED:BST_UNCHECKED);
			CheckDlgButton(dialog,IDC_INESEDIT_TRAINER,(nes.header[1][6]&4)?BST_CHECKED:BST_UNCHECKED);
			CheckDlgButton(dialog,IDC_INESEDIT_4SCREEN,(nes.header[1][6]&8)?BST_CHECKED:BST_UNCHECKED);
			EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_VMIR),(nes.header[1][6]&8)==0); EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_HMIR),(nes.header[1][6]&8)==0);
			CheckDlgButton(dialog,IDC_INESEDIT_VS,(nes.header[1][7]&1)?BST_CHECKED:BST_UNCHECKED);
			CheckRadioButton(dialog,IDC_INESEDIT_HMIR,IDC_INESEDIT_VMIR,(nes.header[1][6]&1)?IDC_INESEDIT_VMIR:IDC_INESEDIT_HMIR);
			
			return TRUE;
			break;
		
		case WM_COMMAND: {
			
			u32 item;
			char c[256];
			memset(c,0,256);
			
			switch (LOWORD(wParam)) {
				
				/* remove garbage. if used bits, start clearing at high mapper number bits */
				case IDC_INESEDIT_REMOVEGARBAGE:
					if ((nes.header[1][7]&BIN8(00001110))|(nes.header[1][8]>8)|(nes.header[1][9]>1)) {
						memset(nes.header[1]+7,0,9);
						SetDlgItemInt(dialog,IDC_INESEDIT_MAPNUM,nes.header[1][6]>>4,FALSE);
						SetDlgItemInt(dialog,IDC_INESEDIT_RAM,0,FALSE);
						CheckDlgButton(dialog,IDC_INESEDIT_VS,BST_UNCHECKED);
						SendDlgItemMessage(dialog,IDC_INESEDIT_VIDEOFORMAT,CB_SETCURSEL,0,0);
					}
					else memset(nes.header[1]+10,0,6);
					EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_REMOVEGARBAGE),FALSE);
					nes_inesedit_checkapply(dialog);
					break;
				
				
				/* editboxes */
				#define READ() if ((item=GetDlgItemInt(dialog,LOWORD(wParam),NULL,FALSE))>255) item=255 /* read editbox */
				#define UPD() if ((HIWORD(wParam))==EN_KILLFOCUS) SetDlgItemInt(dialog,LOWORD(wParam),item,FALSE); /* update information bar */ \
					nes_inesedit_checkapply(dialog)
				
				/* mapper number */
				case IDC_INESEDIT_MAPNUM:
					READ(); nes.header[1][7]=(nes.header[1][7]&0xf)|(item&0xf0); nes.header[1][6]=(nes.header[1][6]&0xf)|(item<<4&0xf0);
					sprintf(c,"(%s)",mapperinfo_get_shortname(mapperinfo_ines2maptype(item))); SetDlgItemText(dialog,IDC_INESEDIT_MAPNUMEXPL,c); UPD();
					break;
				
				/* prg banks (0 is impossible) */
				case IDC_INESEDIT_PRG:
					item=GetDlgItemInt(dialog,IDC_INESEDIT_PRG,NULL,FALSE); if (item>255) item=255; else if (item<1) item=1;
					nes.header[1][4]=item; sprintf(c,"(%dK ROM)",item*16); SetDlgItemText(dialog,IDC_INESEDIT_PRGEXPL,c); UPD();
					break;
				
				/* chr banks */
				case IDC_INESEDIT_CHR:
					READ(); nes.header[1][5]=item; sprintf(c,"(%dK R%cM)",(item|(item==0))*8,(item==0)?'A':'O'); SetDlgItemText(dialog,IDC_INESEDIT_CHREXPL,c); UPD();
					break;
				
				/* ram banks */
				case IDC_INESEDIT_RAM:
					READ(); nes.header[1][8]=item; sprintf(c,"(%dK RAM)",(item|(item==0))*8); SetDlgItemText(dialog,IDC_INESEDIT_RAMEXPL,c); UPD();
					break;
				
				#undef READ
				#undef UPD
				/****/
				
				
				/* videoformat */
				case IDC_INESEDIT_VIDEOFORMAT:
					if ((HIWORD(wParam))==CBN_SELCHANGE) nes.header[1][9]=(nes.header[1][9]&0xfe)|(SendMessage(GetDlgItem(dialog,IDC_INESEDIT_VIDEOFORMAT),CB_GETCURSEL,0,0)&1);
					nes_inesedit_checkapply(dialog);
					break;
				
				/* checkboxes/radiobuttons */
				case IDC_INESEDIT_4SCREEN: case IDC_INESEDIT_BATTERY: case IDC_INESEDIT_TRAINER: case IDC_INESEDIT_VS: case IDC_INESEDIT_HMIR: case IDC_INESEDIT_VMIR:
					nes.header[1][6]=(nes.header[1][6]&0xf0)|(IsDlgButtonChecked(dialog,IDC_INESEDIT_VMIR)==BST_CHECKED)|((IsDlgButtonChecked(dialog,IDC_INESEDIT_BATTERY)==BST_CHECKED)<<1)|((IsDlgButtonChecked(dialog,IDC_INESEDIT_TRAINER)==BST_CHECKED)<<2)|((IsDlgButtonChecked(dialog,IDC_INESEDIT_4SCREEN)==BST_CHECKED)<<3);
					nes.header[1][7]=(nes.header[1][7]&0xfe)|(IsDlgButtonChecked(dialog,IDC_INESEDIT_VS)==BST_CHECKED);
					EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_VMIR),(nes.header[1][6]&8)==0); EnableWindow(GetDlgItem(dialog,IDC_INESEDIT_HMIR),(nes.header[1][6]&8)==0);
					nes_inesedit_checkapply(dialog);
					break;
				
				
				case IDOK: case IDC_INESEDIT_OK:
					EndDialog(dialog,nes_inesedit_save(dialog));
					break;
				
				case IDCANCEL: case IDC_INESEDIT_CANCEL:
					EndDialog(dialog,0);
					break;
				
				case IDC_INESEDIT_APPLY:
					memcpy(nes.header[2],nes.header[1],0x10);
					if (!nes_inesedit_save(dialog)) EndDialog(dialog,0);
					else {
						nes_inesedit_checkapply(dialog);
						nes_open_hook_newfile(nes.template); /* getparent or getancestor won't retrieve it */
					}
					break;
				
				default: break;
			}
			break;
		}
		
		default: break;
	}
	
	return 0;
}

/* file open template dialog */
static UINT_PTR CALLBACK nes_open_hook(HWND dialog,UINT msg,WPARAM wParam,LPARAM lParam)
{
	char c[256];
	memset(c,0,256);
	
	switch (msg) {
		
		case WM_INITDIALOG: {
			int i;
			HWND combo=GetDlgItem(dialog,IDC_OPENGAME_MAPTYPE);
			
			nes.openview&=BIN8(00000101);
			
			/* fill combobox */
			for (i=0;i<MAPTYPE_MAX;i++) SendMessage(combo,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)mapperinfo_get_name(i)));
			SendMessage(combo,CB_SETCURSEL,MAPTYPE_UNKNOWN,0);
			
			CheckDlgButton(dialog,IDC_OPENGAME_HMIR,BST_CHECKED);
			EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_EDIT),FALSE);
			OH_UPDATE_INES(FALSE);
			
			return TRUE;
			break;
		}
		
		case WM_NOTIFY:
			
			switch (((LPOFNOTIFY)lParam)->hdr.code) {
				
				/* file listbox selection was changed */
				case CDN_SELCHANGE:
					nes_open_hook_newfile(dialog);
					break;
				
				/* exit dialog affirmatively */
				case CDN_FILEOK:
					nes.mapinfo[0].mir=IsDlgButtonChecked(dialog,IDC_OPENGAME_VMIR)==BST_CHECKED;
					nes.mapinfo[0].battery=IsDlgButtonChecked(dialog,IDC_OPENGAME_BATTERY)==BST_CHECKED;
					break;
					
				
				default: break;
			}
			
			break;
		
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				
				case IDC_OPENGAME_EDIT:
					if (!file_open()||!file_read(nes.header[0],0x10)) {
						file_close();
						log_set_window(dialog);
						LOG(LOG_MISC|LOG_ERROR,"couldn't open file!\n");
					}
					else {
						file_close();
						nes.template=dialog;
						if (DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_INESEDIT),dialog,nes_inesedit)) nes_open_hook_newfile(dialog);
					}
					break;
				
				case IDC_OPENGAME_MAPTYPE:
					if ((HIWORD(wParam))==CBN_SELCHANGE&&(nes.mapinfo[0].maptype=SendMessage(GetDlgItem(dialog,IDC_OPENGAME_MAPTYPE),CB_GETCURSEL,0,0))!=CB_ERR) {
						/* changing ROM type on load inspired from dvik's blueMSX (MSX emulator) */
						/* user changed combobox selection */
						EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_HMIR),mapperinfo_get_mapmir(nes.mapinfo[0].maptype)?FALSE:TRUE);
						EnableWindow(GetDlgItem(dialog,IDC_OPENGAME_VMIR),mapperinfo_get_mapmir(nes.mapinfo[0].maptype)?FALSE:TRUE);
						OH_UPDATE_INES(TRUE);
					}
					break;
				
				default: break;
			}
			
			break;
		
		case WM_SIZE: {
			RECT rect;
			
			/* reposition edit button */
			GetClientRect(GetParent(dialog),&rect);
			SetWindowPos(GetDlgItem(dialog,IDC_OPENGAME_EDIT),NULL,rect.right-87,rect.bottom-74,0,0,SWP_NOSIZE|SWP_NOZORDER);
			break;
		}
		
		case WM_PAINT: {
			HWND parent=GetParent(dialog);
			
			/* set height (only once) */
			if (~nes.openview&1) {
				RECT rect;
				nes.openview|=1;
				GetWindowRect(parent,&rect);
				SetWindowPos(parent,NULL,0,0,rect.right-rect.left,604,SWP_NOMOVE|SWP_NOZORDER);
			}
			
			/* change default view mode to details */
			if (~nes.openview&2) {
				nes.openview|=2;
				SendMessage(parent,WM_COMMAND,FCIDM_BROWSERFIRST|FVM_DETAILS,0);
				SendMessage(parent,WM_COMMAND,0x702c,0); /* on older OS? */
			}
			
			/* I don't know the ID of the listbox control (to be able to make it select the last opened file when reopening the dialog), it's not lst1 */
			
			break;
		}
		
		default: break;
	}
	
	return 0;
}

/* common open dialog */
void nes_open(void)
{
	OPENFILENAME of;
	char fn[STRING_SIZE]={0};
	
	if (nes.openview&4) return;
	nes.openview|=4;
	
	memset(&of,0,sizeof(OPENFILENAME));
	
	memcpy(fn,nes.openfn_prev,strlen(nes.openfn_prev));
	
	of.lStructSize=sizeof(OPENFILENAME);
	of.hwndOwner=MAIN->window;
	of.hInstance=MAIN->module;
	of.lpstrFile=fn;
	of.nMaxFile=STRING_SIZE;
	of.lpstrTitle="Open Cartridge";
	of.lpstrFilter="All Supported Media\0*.nes;*.zip\0iNES ROMs (*.nes, *.zip)\0*.nes;*.zip\0All Files\0*.*\0\0";
	of.nFilterIndex=nes.filterindex;
	of.Flags=OFN_ENABLESIZING|OFN_EXPLORER|OFN_HIDEREADONLY|OFN_FILEMUSTEXIST|OFN_ENABLETEMPLATE|OFN_ENABLEHOOK;
	of.lpstrInitialDir=file->romdir;
	of.lpTemplateName=MAKEINTRESOURCE(IDD_OPENGAME_TEMPLATE);
	of.lpfnHook=nes_open_hook;
	
	MAIN->dialog=TRUE;
	if (!GetOpenFileName(&of)) { MAIN->dialog=FALSE; nes.openview&=~4; return; }
	MAIN->dialog=FALSE;
	nes.filterindex=of.nFilterIndex;
	
	file_setfile(NULL,fn,NULL,"nes");
	
	memcpy(&nes.mapinfo[1],&nes.mapinfo[0],sizeof(nes.mapinfo[0])); nes.openview&=~4;
	nes.semaphore|=2;
	if (nes_start()) memcpy(&nes.mapinfo[2],&nes.mapinfo[1],sizeof(nes.mapinfo[0]));
	else memcpy(&nes.mapinfo[1],&nes.mapinfo[2],sizeof(nes.mapinfo[0]));
	nes.semaphore&=1;
}
