#include "global.h"
#include "draw.h"
#include "crystal.h"
#include "cpu.h"
#include "cpu_memorymap.h"
#include "cartridge.h"
#include "ppu.h"
#include "mapper.h"
#include "mt19937ar.h"

/*
2c02, PPU

Games/programs with problems:
	- Games that change fine x mid-scanline: this can be fixed by resetting fine x the same way as normal x. Though, this behaviour is accurate! confirmed on PAL/NTSC NES and also Famicom (SMB3 bonusround test) (hack: enable HACK_FINE_X below)
		- Castlevania 3 (at mist effect, Japanese version at intro mist scroll too)
		- Double Dragon 2 (in-game above statusbar)
		- Final Lap (2nd player: bottom screen half)
		- Kaijuu Monogatari (scrolling titlescreen)
		- Super Mario Bros 3 (at rotating mushroom/star/flower bonus round)
		- World Grand Prix (in-game road)
		- etc.
	
	- games/programs relying on NMI/vblank flag being stable when reading $2002 around the start of vblank, causing skipped frames ('game' bug?) (hack: enable HACK_NO_SUPPRESS below)
		- Capcom OST (PD)
		- Portopia Renzoku Satsujin Jiken (not noticeable)
		- Sprite by Damian Yerrick (PD) (not noticeable)
		- Visual Player by Memblers (PD)
		- Who's Cuter by Damian Yerrick (PD)
		- Warehouse No. 18 (Russian) (Pirate)
		- etc?
	
	- Viva! Las Vegas (J): garbage on screen at blackjack (bad dump?, confirmed on NES to not be a game bug)
	- Megaman 3: garbage scanline at titlescreen and boss intro (game bug! confirmed on Famicom and USA NES)
	- Rad Racer: yellow scanline on the road when you steer to the far left (game bug! confirmed on USA and PAL NES)
	- Commando, Trojan (Japanese versions too): very glitchy (game bug?)
	- Knight Rider (U) (Japanese version is fine): garbage scanline at title screen (game bug?)

Unknown:
	can never be completely certain about these:
	- odd frame cycle deduction timing: verified to be at end: 340 (Quietust), and also tested to be around scc 328-339 (blargg).. currently using 336
	- reset x/y timing: verified/tested to be at x:scc 257, y:scc 304 (Quietust)
	- behaviour of sprite evaluation, based on tests done by Quietust (seems OK, and blargg's tests pass)
	- open bus reads from $2004 (sprite data when ppu is rendering) (seems OK: read2004.nes (1st version))
	
	not certain about at all:
	- open bus reads from $2007 (background data when ppu is rendering), same for writes and to sprite address/data
	- behaviour of sprite address during rendering (it's known that $2003 fluctuates during rendering if the PPU is on, details are not)
		- read2004 (2nd version) by Quietust
		- Akira relies on it being reset every once in a while (it currently works OK)
		- Tatakai no Banka (Japanese version of Trojan) writes $80 to the sprite address every other frame
		  just before doing a sprite dma transfer (currently works OK)
*/

#define HACK_FINE_X			FALSE	/* hacked fine x scroll behaviour to make some games look better */
#define HACK_NO_SUPPRESS	FALSE	/* disable nmi/vblank flag suppression */


static void __fastcall ppu_execute(void);
static void __fastcall ppu_execute_disabled(void);
static void __fastcall ppu_execute_vblank(void);
static void __fastcall ppu_execute_nothing(void);
void __fastcall (*ppuupdate)(void);
void __fastcall (*ppubusoverride)(u16);
void __fastcall (*ppumodecs)(int);

static fp_ppu_read_memorymap ppu_read_memorymap[0x140]; /* last mirror in case of overflow */
static fp_ppu_write_memorymap ppu_write_memorymap[0x140];

/* attribute lookup tables */
static int ppu_atshift_table[0x400];
static int ppu_atloc_table[0x400];

static struct {
	/* p*=padding for 32-bit alignment */
	
	u8 control1,p0[3];
	u8 control2,p1[3];
	u8 status,p2[3];
	u8 latch,p3[3];
	u8 read_buffer,p4[3];
	
	int table_sp;
	int table_tp;
	int address_inc;
	
	int emphasis;
	int mono_mask;
	int left_tc_disable_mask;
	int left_sc_disable_mask;
	int enabled;
	int tiles_enabled;
	int sprites_enabled;
	
	int warmup;
	int frame_started;
	int vblank;
	int overscan;
	int overscan_bottom;
	int overscan_enabled;
	int overscan_bottom_sc;
	int skipped_cycle;
	int skip_cycle;
	
	int cycles;
	int sc;
	int scc;
	int visible_scanline;
	int valid_scanline;
	
	int nt;
	int pt1;
	int pt2;
	int at;
	
	int reading_sprites;
	int rendering;
	
	u16 address,p10[1];
	u16 temp_address,p11[1];
	int fine_x;
	
#if HACK_FINE_X
	int fine_x_render;
#endif
	
	int tile_buffer[256+16];
	int sprite_buffer[256+8];
	int sprite_ram[0x100];
	int sprite_ram_secondary[32+1];
	
	int sprite_found;
	int sr_offset;
	int srs_offset;
	int sprite_0;
	int sprbus;
	
	u16 sprite_pa,p6[1];
	u8 sprite_y_max,p7[3];
	
	u16* screen_offset;
	
	u8 databus,p8[3];
	u8 sprite_address,p9[3];
	
	int vblank_offset[5]; /* +2, +1, 0, -1, -2 */
	int vblank_flag_disallowed;
	int vblank_nmi_disallowed[3]; /* +1, 0, -1 */
	
	u8 palette[0x20];
	u8* name_ram[2];
	u8* cur_name[4];
} ppu;

/* get/set */
int* ppu_get_enabled_ptr(void) { return &ppu.enabled; }
int* ppu_get_valid_scanline_ptr(void) { return &ppu.valid_scanline; }
u8* ppu_get_control1_ptr(void) { return &ppu.control1; }
u8* ppu_get_control2_ptr(void) { return &ppu.control2; }
int* ppu_get_reading_sprites_ptr(void) { return &ppu.reading_sprites; }
int* ppu_get_rendering_ptr(void) { return &ppu.rendering; }
int* ppu_get_scc_ptr(void) { return &ppu.scc; }
int* ppu_get_sc_ptr(void) { return &ppu.sc; }
u16* ppu_get_address_ptr(void) { return &ppu.address; }
int* ppu_get_cycles_ptr(void) { return &ppu.cycles; }
int ppu_get_cycles(void) { return ppu.cycles; }
int __fastcall ppu_get_warmup(void) { return ppu.warmup; }
void ppu_set_update_nothing(void) { ppuupdate=&ppu_execute_nothing; }
void ppu_set_modechangesignal(fp_ppu_modechangesignal fun) { if (fun==NULL) fun=mapper_ppumodechangesignal_nothing; ppumodecs=fun; }
void ppu_set_overscan(void) { ppu.overscan_enabled=draw_get_source_width()!=256; ppu.overscan_bottom_sc=draw_get_source_height(); }
void ppu_give_screen_ptr(u16* screen) { ppu.screen_offset=screen; }
void __fastcall ppu_force_update(void) { if (!ppu.rendering) (*ppuupdate)(); } /* prevent reentry */
void __fastcall ppu_clear_status(void) { ppu.status=0; }
int __fastcall ppu_get_vblank_nmi_disallowed(void) { return ppu.vblank_nmi_disallowed[0]|ppu.vblank_nmi_disallowed[1]|ppu.vblank_nmi_disallowed[2]; }
int __fastcall ppu_get_skipped_cycle(void) { return ppu.skipped_cycle; }
u8* ppu_memmap_get_palette_ptr(void) { return ppu.palette; }
u8** ppu_memmap_get_cur_name_ptr(void) { return ppu.cur_name; }
u8** ppu_memmap_get_name_ram_ptr(void) { return ppu.name_ram; }
void ppu_memmap_set_ppubusoverride(fp_ppubusoverride fun) { ppubusoverride=fun; }


/* memmap */

/* --- pattern tables */
#define GET_PATDATA(x)		return cartridge->cur_patbank[x][address&0x3ff]
#define SET_PATDATA(x)		if (cartridge->cur_patbank_isram[x]) cartridge->cur_patbank[x][address&0x3ff]=data

/* read */
/* PT0 */
u8 __fastcall ppu_read_pt0a(u16 address) { GET_PATDATA(0); } /* $0000-$03ff */
u8 __fastcall ppu_read_pt0b(u16 address) { GET_PATDATA(1); } /* $0400-$07ff */
u8 __fastcall ppu_read_pt0c(u16 address) { GET_PATDATA(2); } /* $0800-$0bff */
u8 __fastcall ppu_read_pt0d(u16 address) { GET_PATDATA(3); } /* $0c00-$0fff */
/* PT1 */
u8 __fastcall ppu_read_pt1a(u16 address) { GET_PATDATA(4); } /* $1000-$13ff */
u8 __fastcall ppu_read_pt1b(u16 address) { GET_PATDATA(5); } /* $1400-$17ff */
u8 __fastcall ppu_read_pt1c(u16 address) { GET_PATDATA(6); } /* $1800-$1bff */
u8 __fastcall ppu_read_pt1d(u16 address) { GET_PATDATA(7); } /* $1c00-$1fff */

/* write */
/* PT0 */
void __fastcall ppu_write_pt0a(u16 address,u8 data) { SET_PATDATA(0); } /* $0000-$03ff */
void __fastcall ppu_write_pt0b(u16 address,u8 data) { SET_PATDATA(1); } /* $0400-$07ff */
void __fastcall ppu_write_pt0c(u16 address,u8 data) { SET_PATDATA(2); } /* $0800-$0bff */
void __fastcall ppu_write_pt0d(u16 address,u8 data) { SET_PATDATA(3); } /* $0c00-$0fff */
/* PT1 */
void __fastcall ppu_write_pt1a(u16 address,u8 data) { SET_PATDATA(4); } /* $1000-$13ff */
void __fastcall ppu_write_pt1b(u16 address,u8 data) { SET_PATDATA(5); } /* $1400-$17ff */
void __fastcall ppu_write_pt1c(u16 address,u8 data) { SET_PATDATA(6); } /* $1800-$1bff */
void __fastcall ppu_write_pt1d(u16 address,u8 data) { SET_PATDATA(7); } /* $1c00-$1fff */

/* overrides */
u8 __fastcall ppu_read_pt0a_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(0); }
u8 __fastcall ppu_read_pt0b_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(1); }
u8 __fastcall ppu_read_pt0c_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(2); }
u8 __fastcall ppu_read_pt0d_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(3); }
u8 __fastcall ppu_read_pt1a_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(4); }
u8 __fastcall ppu_read_pt1b_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(5); }
u8 __fastcall ppu_read_pt1c_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(6); }
u8 __fastcall ppu_read_pt1d_override(u16 address) { (*ppubusoverride)(address); GET_PATDATA(7); }
void __fastcall ppu_write_pt0a_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(0); }
void __fastcall ppu_write_pt0b_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(1); }
void __fastcall ppu_write_pt0c_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(2); }
void __fastcall ppu_write_pt0d_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(3); }
void __fastcall ppu_write_pt1a_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(4); }
void __fastcall ppu_write_pt1b_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(5); }
void __fastcall ppu_write_pt1c_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(6); }
void __fastcall ppu_write_pt1d_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(7); }

/* --- nametables --- */
#define GET_NAMEDATA(x)		return ppu.cur_name[x][address&0x3ff]
#define SET_NAMEDATA(x)		ppu.cur_name[x][address&0x3ff]=data

/* read */
u8 __fastcall ppu_read_name0(u16 address) { GET_NAMEDATA(0); } /* nametable 0: $2000-$23ff */
u8 __fastcall ppu_read_name1(u16 address) { GET_NAMEDATA(1); } /* nametable 1: $2400-$27ff */
u8 __fastcall ppu_read_name2(u16 address) { GET_NAMEDATA(2); } /* nametable 2: $2800-$2bff */
u8 __fastcall ppu_read_name3(u16 address) { GET_NAMEDATA(3); } /* nametable 3: $2c00-$2fff */

/* write */
void __fastcall ppu_write_name0(u16 address,u8 data) { SET_NAMEDATA(0); } /* nametable 0: $2000-$23ff */
void __fastcall ppu_write_name1(u16 address,u8 data) { SET_NAMEDATA(1); } /* nametable 1: $2400-$27ff */
void __fastcall ppu_write_name2(u16 address,u8 data) { SET_NAMEDATA(2); } /* nametable 2: $2800-$2bff */
void __fastcall ppu_write_name3(u16 address,u8 data) { SET_NAMEDATA(3); } /* nametable 3: $2c00-$2fff */

/* overrides */
u8 __fastcall ppu_read_name0_override(u16 address) { (*ppubusoverride)(address); GET_NAMEDATA(0); }
u8 __fastcall ppu_read_name1_override(u16 address) { (*ppubusoverride)(address); GET_NAMEDATA(1); }
u8 __fastcall ppu_read_name2_override(u16 address) { (*ppubusoverride)(address); GET_NAMEDATA(2); }
u8 __fastcall ppu_read_name3_override(u16 address) { (*ppubusoverride)(address); GET_NAMEDATA(3); }
void __fastcall ppu_write_name0_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(0); }
void __fastcall ppu_write_name1_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(1); }
void __fastcall ppu_write_name2_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(2); }
void __fastcall ppu_write_name3_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(3); }

/* --- palette --- */
#define GET_PALETTE() \
	if (ppu.control2&1) return ppu.palette[address&0x1f]&BIN8(11110000); /* mono */ \
	else return ppu.palette[address&0x1f]
#define SET_PALETTE() \
	if((address&BIN8(00000011))==0) ppu.palette[(address&0x1f)^0x10]=data&BIN8(00111111); \
	ppu.palette[address&0x1f]=data&BIN8(00111111)

static int __inline__ ppu_read_palette_inl(int address) { GET_PALETTE(); }
u8 __fastcall ppu_read_palette(u16 address) { GET_PALETTE(); }
void __fastcall ppu_write_palette(u16 address,u8 data) { SET_PALETTE(); }

/* overrides */
u8 __fastcall ppu_read_palette_override(u16 address) { (*ppubusoverride)(address); GET_PALETTE(); }
void __fastcall ppu_write_palette_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PALETTE(); }


/* shortcuts */

#define READ_BYTE(x) ppu_read_memorymap[(x)>>8](x)
#define WRITE_BYTE(x,y) ppu_write_memorymap[(x)>>8](x,y)


/* registers */

/* nmi suppression */
#define PPU_CHECK_NMI_DISALLOWED() \
	if ((ppu.cycles>ppu.vblank_offset[0])&(ppu.cycles<ppu.vblank_offset[4])) { \
		ppu.vblank_nmi_disallowed[0]|=(ppu.cycles<ppu.vblank_offset[2]); \
		ppu.vblank_nmi_disallowed[1]|=((ppu.cycles>ppu.vblank_offset[1])&(ppu.cycles<ppu.vblank_offset[3])); \
		ppu.vblank_nmi_disallowed[2]|=(ppu.cycles>ppu.vblank_offset[2]); \
	}


u8 __fastcall ppu_read_openbus(u16 address) { return ppu.databus; }

/* $2000, control register 1
	7	enable nmi on vblank
	6	ppu layer (always 0 for NES, unused)
	5	8x16 sprite mode
	4	tile table address (0 or $1000)
	3	sprite table address (0 or $1000)
	2	address increment value (1 or 32)
	1,0	base nametable address ($2000, $2400, $2800, $2c00 */
void __fastcall ppu_write_control1(u8 data)
{
	(*ppuupdate)();
	
	if (data&0x80) { if (ppu.status&~ppu.control1&ppu.vblank&&ppu.cycles>crystal->ppu_cycle) cpu_set_interrupt(INTERRUPT_NMI|INT_EDGE_NMI_EX); }
#if !HACK_NO_SUPPRESS
	else { PPU_CHECK_NMI_DISALLOWED() }
#endif
	
	ppu.databus=ppu.control1=data;
	ppu.table_sp=data<<9&0x1000;
	ppu.table_tp=data<<8&0x1000;
	ppu.address_inc=data&BIT(2)?32:1;
	ppu.sprite_y_max=(data>>2&8)|7;
	ppu.temp_address=(ppu.temp_address&BIN16(11110011,11111111))|(data<<10&BIN16(00001100,00000000));
}
void __fastcall ppu_write_control1_warmup(u8 data)
{
	/* ignore */
	ppu.databus=data;
}
void __fastcall ppu_clear_nmi_enable(void) { ppu.control1&=BIN8(01111111); }

/* $2001, control register 2
	7	more blue
	6	more green
	5	more red
	4	enable sprites
	3	enable tiles
	2	enable left sprite column (8 pixels)
	1	enable left tiles column (8 pixels)
	0	monochrome */
void __fastcall ppu_write_control2(u8 data)
{
	(*ppuupdate)();
	ppu.databus=ppu.control2=data;
	ppu.emphasis=(data&BIN8(11100000))<<1;
	ppu.mono_mask=data&1?BIN8(11110000):0xff;
	ppu.left_tc_disable_mask=data&2?0xff:0;
	ppu.left_sc_disable_mask=data&4?0xff:0;
	if (((data&BIN8(00011000))!=0)^ppu.enabled) {
		ppu.enabled^=1;
		if (ppuupdate!=&ppu_execute_vblank) {
			ppuupdate=ppu.valid_scanline&ppu.enabled?&ppu_execute:&ppu_execute_disabled;
			if (ppu.valid_scanline) (*ppumodecs)(ppu.enabled);
		}
	}
	ppu.tiles_enabled=data&BIT(3);
	ppu.sprites_enabled=data<<3&0x80;
}
void __fastcall ppu_write_control2_warmup(u8 data)
{
	if (ppu.warmup&PPU_WARMUP_HARD) {
		/* accept bits 0,5,6,7 at powerup */
		(*ppuupdate)();
		
		ppu.emphasis=(data&BIN8(11100000))<<1;
		ppu.mono_mask=data&1?BIN8(11110000):0xff;
	}
	
	ppu.databus=data;
}

/* $2002, status
	7	vblank
	6	sprite 0 hit
	5	sprite overflow */
u8 __fastcall ppu_read_status(u16 address)
{
	(*ppuupdate)();
#if !HACK_NO_SUPPRESS
	PPU_CHECK_NMI_DISALLOWED()
	ppu.vblank_flag_disallowed=ppu.vblank_nmi_disallowed[2];
#endif
	ppu.databus=ppu.status|(ppu.databus&BIN8(00011111));
	ppu.latch=FALSE;
	ppu.status&=~BIT(7); /* vblank flag resets on read */
	return ppu.databus;
}

/* $2003, sprite memory address */
void __fastcall ppu_write_sprite_memory_address(u8 data)
{
	(*ppuupdate)();
	ppu.databus=ppu.sprite_address=data;
}

/* $2004, sprite memory data */
u8 __fastcall ppu_read_sprite_memory_data(u16 address)
{
	(*ppuupdate)();
	
	if (ppu.valid_scanline&ppu.enabled) {
		/* sprite ram open bus */
		if (ppu.scc<66) ppu.databus=0xff;
		else if (ppu.scc<258) { /* sprite evaluation starts at cycle 65 (and scc++ at end) */
			if (ppu.sr_offset<0x100) ppu.databus=ppu.sprbus;
			else ppu.databus=ppu.scc&1?ppu.sprite_ram[0xfc]:ppu.sprbus;
		}
		else if (ppu.scc<322) ppu.databus=ppu.sprite_ram_secondary[(ppu.scc>>1&BIN8(00011111))|(ppu.scc&BIN8(00000011))];
		else ppu.databus=ppu.srs_offset?ppu.sprite_ram_secondary[0]:ppu.sprite_ram[0xfc];
	}
	else ppu.databus=(ppu.sprite_address&3)==2?ppu.sprite_ram[ppu.sprite_address]&BIN8(11100011):ppu.sprite_ram[ppu.sprite_address];
	
	return ppu.sprbus=ppu.databus;
}
void __fastcall ppu_write_sprite_memory_data(u8 data)
{
	(*ppuupdate)();
	ppu.databus=ppu.sprite_ram[ppu.sprite_address++]=data;
}

/* $2005, memory address(scroll) */
void __fastcall ppu_write_scroll(u8 data)
{
	(*ppuupdate)();
	ppu.databus=data;
	if (ppu.latch) ppu.temp_address=(ppu.temp_address&BIN16(00001100,00011111))|(data<<12&BIN16(01110000,00000000))|(data<<2&BIN16(00000011,11100000));
	else { ppu.fine_x=data&BIN8(00000111); ppu.temp_address=(ppu.temp_address&BIN16(01111111,11100000))|data>>3; }
	ppu.latch^=1;
}
void __fastcall ppu_write_scroll_warmup(u8 data)
{
	/* ignore */
	ppu.databus=data;
}

/* $2006, memory address */
void __fastcall ppu_write_memory_address(u8 data)
{
	(*ppuupdate)();
	ppu.databus=data;
	if (ppu.latch) ppu.address=ppu.temp_address=(ppu.temp_address&BIN16(01111111,00000000))|data;
	else ppu.temp_address=(ppu.temp_address&0x00ff)|(data<<8&BIN16(00111111,00000000));
	ppu.latch^=1;
}
void __fastcall ppu_write_memory_address_warmup(u8 data)
{
	/* ignore */
	ppu.databus=data;
}

/* $2007, memory data */
u8 __fastcall ppu_read_memory_data(u16 address)
{
	(*ppuupdate)();
	
	if ((ppu.address&0x3f00)==0x3f00) {
		ppu.databus=ppu_read_palette_inl(ppu.address);
		ppu.read_buffer=READ_BYTE(ppu.address);
	}
	else {
		ppu.databus=ppu.read_buffer;
		ppu.read_buffer=READ_BYTE(ppu.address);
	}
	ppu.address+=ppu.address_inc;
	
	return ppu.databus;
}
void __fastcall ppu_write_memory_data(u8 data)
{
	(*ppuupdate)();
	WRITE_BYTE(ppu.address,data);
	ppu.databus=data;
	ppu.address+=ppu.address_inc;
}



/* render */

/* reset x/y */
#define RESET_X() ppu.address=(ppu.address&BIN16(11111011,11100000))|(ppu.temp_address&BIN16(00000100,00011111))
#define RESET_Y() ppu.address=ppu.temp_address

/* fetch nametable byte */
#define FETCH_NT() ppu.nt=READ_BYTE((ppu.address&0xfff)|0x2000)<<4|(ppu.address>>12&7)|ppu.table_tp

/* fetch attribute byte */
#define FETCH_AT() int a=ppu.address&0x3ff; ppu.at=(READ_BYTE((ppu.address&0xc00)|ppu_atloc_table[a])>>ppu_atshift_table[a]&3)<<2|0x80

/* increase x */
#define INC_X() \
a=(ppu.address+1)&BIN8(00011111); \
ppu.address=(ppu.address&BIN16(11111111,11100000))|a; \
if (!a) ppu.address^=BIT(10)

/* increase y */
#define INC_Y() \
a=(ppu.address+BIT(12))&BIN16(01110000,00000000); \
ppu.address=(ppu.address&BIN16(10001111,11111111))|a; \
if (!a) { \
	a=(ppu.address+BIT(5))&BIN16(00000011,11100000); \
	if (a==BIN16(00000011,11000000)) { a^=a; ppu.address^=BIT(11); } \
	ppu.address=(ppu.address&BIN16(11111100,00011111))|a; \
}

/* fetch tile bytes */
#define FETCH_T1() ppu.pt1=READ_BYTE(ppu.nt)
#define FETCH_T2() ppu.pt2=READ_BYTE(ppu.nt|8)

/* set 8 tile pixels */
#define TSET t=(p1&1)|(p2&2); ppu.tile_buffer[i]=t?t|ppu.at:0
#define TSHM p1>>=1; p2>>=1; i--
#define TSHP p1>>=1; p2>>=1; i++
#define TINI(o) int p1=ppu.pt1; int p2=ppu.pt2<<1; int i=o
#define SET_TLINE(o) int t; TINI(o); TSET; TSHM; TSET; TSHM; TSET; TSHM; TSET; TSHM; TSET; TSHM; TSET; TSHM; TSET; TSHM; TSET

/* set 8 sprite pixels */
#define SET_SBIT_OR() \
if (ppu.sprite_ram_secondary[n]&0x20) n=(ppu.sprite_ram_secondary[n]<<2&BIN8(00001100))|BIN8(10010000); \
else n=(ppu.sprite_ram_secondary[n]<<2&BIN8(00001100))|BIN8(11010000); \
if (ppu.sprite_0) { \
	if (ppu.scc==263) n|=0x20; \
	else ppu.sprite_0=FALSE; \
}
#define SSET \
if (!(ppu.sprite_buffer[i]&BIN8(00011111))) { \
	ppu.sprite_buffer[i]|=(p1&1)|(p2&2); \
	if (ppu.sprite_buffer[i]&BIN8(00011111)) ppu.sprite_buffer[i]|=n; \
}
#define SET_SLINE(x) \
int n=x; \
if (n<ppu.srs_offset) { \
	if (ppu.sprite_ram_secondary[n]&0x40) { /* hflip */ \
		TINI(ppu.sprite_ram_secondary[n+1]); SET_SBIT_OR() \
		SSET TSHP; SSET TSHP; SSET TSHP; SSET TSHP; SSET TSHP; SSET TSHP; SSET TSHP; SSET \
	} \
	else { \
		TINI(ppu.sprite_ram_secondary[n+1]+7); SET_SBIT_OR() \
		SSET TSHM; SSET TSHM; SSET TSHM; SSET TSHM; SSET TSHM; SSET TSHM; SSET TSHM; SSET \
	} \
}

/* fetch sprite tile bytes */
#define FETCH_S1() ppu.reading_sprites^=1; ppu.pt1=READ_BYTE(ppu.sprite_pa); ppu.reading_sprites^=1
#define FETCH_S2() ppu.reading_sprites^=1; ppu.pt2=READ_BYTE(ppu.sprite_pa|8); ppu.reading_sprites^=1

/* set sprite */
#define SET_SPRITE(x) \
int n=x; u8 y=ppu.sc-1-ppu.sprite_ram_secondary[n++]; \
if (ppu.sprite_ram_secondary[n+1]&0x80) y^=ppu.sprite_y_max; /* vflip */ \
if (ppu.sprite_y_max&8) { \
	int i=ppu.sprite_ram_secondary[n]<<4; \
	if (y>7) y=(y&7)|0x10; \
	ppu.sprite_pa=(i<<8&0x1000)|(i&BIN16(00001111,11100000))|y; \
} \
else ppu.sprite_pa=ppu.sprite_ram_secondary[n]<<4|y|ppu.table_sp

/* sprite evaluation */
#define EVAL_SPR_EMP() ppu.sprite_ram_secondary[ppu.scc>>1]=0xff
#define EVAL_SPR_Y() \
if (ppu.sr_offset<0x100) { \
	int o=ppu.sr_offset&BIN8(11111000)?ppu.sr_offset:(ppu.sprite_address&BIN8(11111000))|ppu.sr_offset; \
	/* copy sprite data */ \
	if (ppu.sprite_found) { \
		ppu.sprbus=ppu.sprite_ram_secondary[ppu.srs_offset++]=ppu.sprite_ram[o]; \
		ppu.sprite_found=++ppu.sr_offset&3; \
	} \
	\
	else { \
		u8 ys=ppu.sc-1-ppu.sprite_ram[o]; \
		ppu.sprbus=ppu.sprite_ram_secondary[ppu.srs_offset]=ppu.sprite_ram[o]; \
		if ((ppu.sprbus<0xf0)&(ys<=ppu.sprite_y_max)) { \
			/* found valid sprite */ \
			if (ppu.srs_offset++<0x20) { \
				ppu.sprite_0|=(ppu.sr_offset++==0); /* =sprite0? */ \
				ppu.sprite_found^=1; \
			} \
			/* overflow */ \
			else { \
				ppu.srs_offset&=0x20; \
				if (++ppu.sr_offset&3) ppu.sr_offset+=4; \
				ppu.status|=BIT(5); \
			} \
		} \
		/* y of new sprite, optional diagonal reads */ \
		else { \
			if (ppu.srs_offset<0x20) ppu.sr_offset+=4; \
			else if (++ppu.sr_offset&3) ppu.sr_offset+=4; \
		} \
	} \
} \
else { \
	/* thrash reads */ \
	ppu.sprbus=ppu.sprite_ram[ppu.sr_offset&0xfc]; \
	ppu.sr_offset+=4; \
}

/* empty sprite/tile buffers */
#define EMPTY_BUFFERS() \
	int* xo; int i; \
	ppu.sprite_pa=ppu.table_sp; \
	xo=ppu.sprite_buffer; i=0x21; while (i--) { *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; } \
	xo=ppu.tile_buffer; i=0x22; while (i--) { *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; *(xo++)=0; }

/* overscan */
#define LEFT_OVERSCAN_START() \
ppu.overscan_bottom=(ppu.sc>239)&(ppu.sc<ppu.overscan_bottom_sc); \
ppu.overscan|=(ppu.overscan_enabled&(ppu.overscan_bottom|(ppu.sc<240)))

/* new frame */
#define FRAME_START() \
crystal_new_frame(); ppu.skipped_cycle=FALSE; ppu.skip_cycle=crystal->odd_frame; \
ppu.vblank_flag_disallowed=ppu.vblank_nmi_disallowed[0]=ppu.vblank_nmi_disallowed[1]=ppu.vblank_nmi_disallowed[2]=FALSE; \
ppu.visible_scanline=ppu.vblank=ppu.sc=ppu.scc=ppu.overscan=ppu.overscan_bottom=0; \
ppu.valid_scanline=TRUE; \
ppuupdate=ppu.enabled?&ppu_execute:&ppu_execute_disabled; \
ppu.screen_offset=draw_get_screen_ptr(); \
if (ppu.enabled) (*ppumodecs)(TRUE)



void __fastcall ppu_new_frame(void)
{
	if (ppu.frame_started) {
		ppu.cycles+=crystal->frame;
		ppu.frame_started=FALSE;
	}
	else {
		ppu.cycles=crystal->frame;
		FRAME_START();
	}
}


static void __fastcall ppu_execute(void)
{
	int cpu_cycles=cpu_get_cycles();
	int line=ppu.scc<256;
	ppu.rendering=TRUE;
	
 while (ppu.cycles>cpu_cycles) {
	
	ppu.cycles-=crystal->ppu_cycle;
	
 switch (ppu.scc) {
	/* visible area */
	/* memory fetch 1-128: NT, AT, T1, T2 */
	/* cycle 0-63: empty secondary sprite buffer in parallel */
	/* cycle 64-255: evaluate sprite y coordinates in parallel */
	/* NT (double cases are unallowed :( (duh) ) */
	case 1:  case 9:  case 17: case 25: case 33: case 41: case 49: case 57:
		FETCH_NT(); EVAL_SPR_EMP(); break;
	case 65: case 73: case 81: case 89: case 97: case 105:case 113:case 121:
	case 129:case 137:case 145:case 153:case 161:case 169:case 177:case 185:
	case 193:case 201:case 209:case 217:case 225:case 233:case 241:case 249:
		FETCH_NT(); EVAL_SPR_Y() break;
	/* AT + INC X ( + INC Y ) */
	case 3:  case 11: case 19: case 27: case 35: case 43: case 51: case 59:
		{ FETCH_AT(); INC_X(); EVAL_SPR_EMP(); break; }
	case 67: case 75: case 83: case 91: case 99: case 107:case 115:case 123:
	case 131:case 139:case 147:case 155:case 163:case 171:case 179:case 187:
	case 195:case 203:case 211:case 219:case 227:case 235:case 243:/*case 251:*/
		{ FETCH_AT(); INC_X(); EVAL_SPR_Y() break; }
	case 251: { FETCH_AT(); INC_X(); INC_Y() EVAL_SPR_Y() break; }
	/* T1 */
	case 5:  case 13: case 21: case 29: case 37: case 45: case 53: case 61:
		FETCH_T1(); EVAL_SPR_EMP(); break;
	case 69: case 77: case 85: case 93: case 101:case 109:case 117:case 125:
	case 133:case 141:case 149:case 157:case 165:case 173:case 181:case 189:
	case 197:case 205:case 213:case 221:case 229:case 237:case 245:case 253:
		FETCH_T1(); EVAL_SPR_Y() break;
	/* T2 + fill tile buffer */
	case 7:  case 15: case 23: case 31: case 39: case 47: case 55: case 63:
		FETCH_T2(); EVAL_SPR_EMP(); if (ppu.visible_scanline) { SET_TLINE(ppu.scc+16); } break;
	case 71: case 79: case 87: case 95: case 103:case 111:case 119:case 127:
	case 135:case 143:case 151:case 159:case 167:case 175:case 183:case 191:
	case 199:case 207:case 215:case 223:case 231:case 239:case 247:case 255:
		FETCH_T2(); EVAL_SPR_Y() if (ppu.visible_scanline) { SET_TLINE(ppu.scc+16); } break;
	
	/* hblank */
	case 256: {
		EMPTY_BUFFERS()
		line=FALSE;
		
		/* right overscan start */
		ppu.overscan=ppu.visible_scanline&ppu.overscan_enabled;
		
		#if DEBUG_PPU
		LOG(LOG_PPU,"------- hblank ------- sc:%d cc:%d\n",ppu.sc,ppu.cycles);
		#endif
		
		break;
	}
	
	/* memory fetch 129-160: NT (unused), NT (unused), S1, S2, sprites in parallel */
	/* NT */
	/*case 257:*/case 265:case 273:case 281:case 289:case 297:case 305:case 313:
		FETCH_NT(); break;
	/* NT ( + RESET X ) */
	case 257:
#if HACK_FINE_X
		ppu.fine_x_render=ppu.fine_x;
#endif
		RESET_X(); FETCH_NT(); break;
	
	/* RESET Y */
	case 304:
		if (ppu.sc==0) { RESET_Y(); } break;
	/* NT + sprites */
	case 259:/*case 267:*/case 275:case 283:case 291:case 299:case 307:case 315:
		{ SET_SPRITE(ppu.scc>>1&BIN8(00011100)); FETCH_NT(); break; }
	case 267: /* + right overscan end */
		{ SET_SPRITE(ppu.scc>>1&BIN8(00011100)); FETCH_NT(); ppu.overscan=FALSE; break; }
	
	/* S1 */
	case 261:case 269:case 277:case 285:case 293:case 301:case 309:case 317:
		FETCH_S1(); break;
	
	/* S2 + fill sprite buffer */
	case 263:case 271:case 279:case 287:case 295:case 303:case 311:case 319:
		FETCH_S2(); if (ppu.sc<240) { SET_SLINE(ppu.scc>>1&BIN8(00011110)) } break;
	
	/* memory fetch 161-168: NT, AT, T1, T2 (prefetch) */
	/* NT */
	case 321:case 329: FETCH_NT(); break;
	
	/* AT + INC X */
	case 323:case 331: { FETCH_AT(); INC_X(); break; }
	
	/* T1 */
	case 325: case 333: FETCH_T1(); break;
	
	/* left overscan start */
	case 326: LEFT_OVERSCAN_START(); break;
	
	/* T2 + fill tile buffer */
	case 327:case 335: FETCH_T2(); if (ppu.sc<240) { SET_TLINE(ppu.scc-320); } break;
	
	/* optional cycle skip */
	case 336:
		if (ppu.skip_cycle) {
			ppu.skipped_cycle=TRUE;
			ppu.scc++;
			ppu.cycles-=crystal->ppu_cycle;
			cpu_cycles-=crystal->ppu_cycle;
			cpu_kill_cycle();
			
			/* extra overscan pixel */
			if (ppu.overscan) *ppu.screen_offset++=(ppu.palette[ppu.overscan_bottom&((ppu.address&0x3f00)==0x3f00)?ppu.address&0x1f:0]&ppu.mono_mask)|ppu.emphasis;
		}
		else break;
	/* memory fetch 169-170: NT (unused), NT (unused) */
	case 337:case 339: FETCH_NT(); break;
	
	/* end */
	case 340:
		ppu.skip_cycle=ppu.sprite_0=ppu.sprite_found=ppu.sr_offset=ppu.srs_offset=0;
		ppu.visible_scanline=ppu.valid_scanline=++ppu.sc<241;
		line=TRUE; ppu.scc=-1;
		
		/* sprite 0 hit doesn't work on rightmost pixel */
		ppu.sprite_buffer[0xff]&=BIN8(11011111);
		
		#if DEBUG_PPU
		LOG(LOG_PPU,"------- new scanline ------- sc:%d cc:%d\n",ppu.sc,ppu.cycles);
		#endif
		
		/* end of visible scanlines */
		if (ppu.sc==241) {
			ppu.scc++;
			if (ppu.overscan) *ppu.screen_offset++=(ppu.palette[(ppu.address&0x3f00)==0x3f00?ppu.address&0x1f:0]&ppu.mono_mask)|ppu.emphasis;
			ppu.overscan|=ppu.overscan_bottom;
			ppuupdate=&ppu_execute_disabled;
			ppu.rendering=FALSE;
			(*ppumodecs)(FALSE);
			(*ppuupdate)();
			return;
		}
		
		if (ppu.overscan_enabled) goto skip_pixel;
		else goto skip_overscan; /* line=1 */
		
		break;
	
	default: break;
} /* switch */
	
	/* draw 1 pixel */
	if (ppu.visible_scanline&line) {
		/* bg pixel */
#if HACK_FINE_X
		int p=ppu.tiles_enabled?ppu.tile_buffer[ppu.scc+ppu.fine_x_render]:0;
#else
		int p=ppu.tiles_enabled?ppu.tile_buffer[ppu.scc+ppu.fine_x]:0;
#endif
		/* show/hide left column */
		if (ppu.scc<8) {
			p&=ppu.left_tc_disable_mask;
			ppu.sprite_buffer[ppu.scc]&=ppu.left_sc_disable_mask;
		}
		
		/* sprite pixel */
		if (ppu.sprites_enabled&ppu.sprite_buffer[ppu.scc]) { /* $80 */
			
			/* non-transparent bg pixel */
			if (p&0x80) {
				#if DEBUG_PPU
				if (((ppu.status&BIT(6))==0)&((ppu.sprite_buffer[ppu.scc]&0x20)!=0)) LOG(LOG_PPU,"------- sprite 0 hit ------- sc:%d cc:%d\n",ppu.sc,ppu.cycles);
				#endif
				
				/* sprite 0 hit? */
				ppu.status|=(ppu.sprite_buffer[ppu.scc]<<1&BIT(6));
				
				/* sprite has priority over bg? */
				if (ppu.sprite_buffer[ppu.scc]&0x40) p=ppu.sprite_buffer[ppu.scc];
			}
			
			else p=ppu.sprite_buffer[ppu.scc];
		}
		
		/* put pixel */
		*ppu.screen_offset++=(ppu.palette[p&BIN8(00011111)]&ppu.mono_mask)|ppu.emphasis;
	}
	
	else
	
skip_pixel:
	
	if (ppu.overscan) *ppu.screen_offset++=(ppu.palette[ppu.overscan_bottom&((ppu.address&0x3f00)==0x3f00)?ppu.address&0x1f:0]&ppu.mono_mask)|ppu.emphasis; /* just the background colour */
	
skip_overscan:
	
	ppu.scc++;
	
} /* main loop */

	ppu.rendering=FALSE;
}

static void __fastcall ppu_execute_disabled(void)
{
	int cpu_cycles=cpu_get_cycles();
	int line=ppu.scc<256;
	
 while (ppu.cycles>cpu_cycles) {
	
	ppu.cycles-=crystal->ppu_cycle;
	
 switch (ppu.scc) {
	
	/* hblank */
	case 256:
		if (ppu.valid_scanline) { EMPTY_BUFFERS() }
		line=FALSE;
		
		/* right overscan start */
		ppu.overscan=(ppu.visible_scanline|ppu.overscan_bottom)&ppu.overscan_enabled;
		
		#if DEBUG_PPU
		LOG(LOG_PPU,"------- hblank ------- sc:%d cc:%d\n",ppu.sc,ppu.cycles);
		#endif
		
		break;
	
	/* right overscan end */
	case 267: ppu.overscan=FALSE; break;
	
	/* left overscan start */
	case 326: LEFT_OVERSCAN_START(); break;
	
	/* end */
	case 340:
		ppu.skip_cycle=ppu.sprite_0=ppu.sprite_found=ppu.sr_offset=ppu.srs_offset=0;
		ppu.visible_scanline=ppu.valid_scanline=++ppu.sc<241;
		ppu.overscan|=ppu.overscan_bottom;
		line=TRUE; ppu.scc=-1;
		
		#if DEBUG_PPU
		LOG(LOG_PPU,"------- new scanline ------- sc:%d cc:%d\n",ppu.sc,ppu.cycles);
		#endif
		
		if (ppu.sc>241) {
			
			/* last overscan line ends */
			#if !DEBUG_PPU
			if (ppu.sc==243) {
				ppu.scc++;
				ppuupdate=&ppu_execute_vblank;
				(*ppuupdate)();
				return;
			}
			#else
			if (ppu.sc==242) {
			#endif
			
			/* enter vblank */
			if (ppu.enabled) ppu.sprite_address=0;
			ppu.vblank=BIT(7);
			if (!ppu.vblank_flag_disallowed) ppu.status|=BIT(7);
			cpu_set_nmi(ppu.control1&BIT(7));
			
			#if DEBUG_PPU
			} LOG(LOG_PPU,"------- vblank ------- sc:%d cc:%d\n",ppu.sc,ppu.cycles);
			#endif
			
			/* no overscan: frame ends */
			#if !DEBUG_PPU
			if (!ppu.overscan) {
				ppu.scc++;
				ppuupdate=&ppu_execute_vblank;
				(*ppuupdate)();
				return;
			}
			#endif
		}
		
		/* end of frame */
		#if DEBUG_PPU
		else if (ppu.sc==crystal->lines) {
			ppuupdate=&ppu_execute_vblank; /* possibly restart frame there */
			(*ppuupdate)();
			return;
		}
		#endif
		
		if (ppu.overscan_enabled) goto dis_skip_pixel;
		else goto dis_skip_overscan; /* line=1 */
		
		break;
	
	default: break;
} /* switch */
	
	if (ppu.visible_scanline&line) *ppu.screen_offset++=(ppu.palette[(ppu.address&0x3f00)==0x3f00?ppu.address&0x1f:0]&ppu.mono_mask)|ppu.emphasis; /* draw 1 pixel, just the background colour */
	
	else
	
dis_skip_pixel:
	
	if (ppu.overscan) *ppu.screen_offset++=(ppu.palette[(~ppu.enabled|ppu.overscan_bottom)&((ppu.address&0x3f00)==0x3f00)?ppu.address&0x1f:0]&ppu.mono_mask)|ppu.emphasis; /* just the background colour */
	
dis_skip_overscan:
	
	ppu.scc++;
	
} /* main loop */

}

static void __fastcall ppu_execute_vblank(void)
{
	int cpu_cycles=cpu_get_cycles();
	
	/* new frame (if the ppu was forced to execute the whole frame) */
	if (cpu_cycles<=0) {
		ppu.cycles=0;
		ppu.frame_started=TRUE;
		
		FRAME_START();
		ppu.status=0;
		
		(*ppuupdate)();
	}
	else ppu.cycles=cpu_cycles;
}

static void __fastcall ppu_execute_nothing(void) { return; }



/************/

void ppu_memmap_printf_internal_nametable(int n)
{
	/* debug related */
	int i;
	for (i=0;i<0x400;i++) {
		printf("0x%02X",ppu.name_ram[0][i]);
		if (i!=0x3ff) printf(", ");
	}
	printf("\n");
}

static void ppu_fill_tables(void)
{
	int i,j=5;
	
	/* fill vblank offset table */
	i=crystal->vblank_trigger+(2*crystal->ppu_cycle);
	while (j--) {
		ppu.vblank_offset[j]=i;
		i-=crystal->ppu_cycle;
	}
	
	/* fill attribute lookup tables: same as attribute calculation was before, during rendering */
	for (i=0;i<0x400;i++) {
		ppu_atshift_table[i]=(i&2)|(i>>4&4);
		ppu_atloc_table[i]=0x23c0|((i>>2&7)|(i>>4&BIN8(00111000)));
	}
}

void ppu_memmap_init(void)
{
	/* palette values on startup: probably different on each NES, this is blargg's NES */
	const u8 palette_startup[]={
		0x09,0x01,0x00,0x01,0x00,0x02,0x02,0x0d,0x08,0x10,0x08,0x24,0x00,0x00,0x04,0x2c,
		0x09,0x01,0x34,0x03,0x00,0x04,0x00,0x14,0x08,0x3a,0x00,0x02,0x00,0x20,0x2c,0x08
	};
	
	memset(&ppu,0,sizeof(ppu));
	
	memcpy(ppu.palette,palette_startup,0x20);
	
	/* or black startup palette */
	/* memset(ppu.palette,BIN8(00111111),0x20); */
	
	MEM_CREATE_N(ppu.name_ram[0],0x400); memset(ppu.name_ram[0],0xff,0x400);
	MEM_CREATE_N(ppu.name_ram[1],0x400); memset(ppu.name_ram[1],0xff,0x400);
	
	ppu_set_fpt(PPU_FPT_NORMAL,FALSE,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
	
	LOG(LOG_VERBOSE,"PPU memorymap initialised\n");
}

void ppu_init(void)
{
	ppu_fill_tables();
	ppu_set_overscan();
	
	LOG(LOG_VERBOSE,"PPU initialised\n");
}

void ppu_reset(int hard,int warm)
{
	int i=0x100;
	while (i--) ppu.sprite_ram[i]=0x55;
	
	if (hard) {
		
		ppu.read_buffer=genrand_int32()&0xff;
		
		if (warm) {
			ppu.address=1;
			ppu.sprite_address=1;
			ppu.status=0xa0;
			ppu.warmup=PPU_WARMUP_HARD;
		}
	}
	else {
		ppu.read_buffer=0;
		ppu.latch=0;
		ppu.warmup=PPU_WARMUP_SOFT;
	}
	
	ppu_write_control1(0);
	ppu_write_control2(0);
	ppu_write_scroll(0); ppu_write_scroll(0);
	
	if (ppu.warmup) {
		cpu_set_write_io2000(0x2000,ppu_write_control1_warmup);
		cpu_set_write_io2000(0x2001,ppu_write_control2_warmup);
		cpu_set_write_io2000(0x2005,ppu_write_scroll_warmup);
		cpu_set_write_io2000(0x2006,((cpu_get_write_io2000(0x2006)==mapper_ppu_write_memory_address_a12_override)||(cpu_get_write_io2000(0x2006)==mapper_ppu_write_memory_address_warmup_a12_override))?mapper_ppu_write_memory_address_warmup_a12_override:ppu_write_memory_address_warmup);
	}
}

void ppu_warm(void)
{
	if (!ppu.warmup) return;
	
	ppu.warmup=0;
	
	cpu_set_write_io2000(0x2000,ppu_write_control1);
	cpu_set_write_io2000(0x2001,ppu_write_control2);
	cpu_set_write_io2000(0x2005,ppu_write_scroll);
	cpu_set_write_io2000(0x2006,((cpu_get_write_io2000(0x2006)==mapper_ppu_write_memory_address_warmup_a12_override)||(cpu_get_write_io2000(0x2006)==mapper_ppu_write_memory_address_a12_override))?mapper_ppu_write_memory_address_a12_override:ppu_write_memory_address);
}

void ppu_clean(void)
{
	MEM_CLEAN(ppu.name_ram[0]);
	MEM_CLEAN(ppu.name_ram[1]);
	
	memset(&ppu,0,sizeof(ppu));
	
	LOG(LOG_VERBOSE,"PPU cleaned\n");
}

void ppu_set_fpt (int type,int custom,fp_ppu_read_memorymap funr1,fp_ppu_read_memorymap funr2,fp_ppu_read_memorymap funr3,fp_ppu_read_memorymap funr4,fp_ppu_write_memorymap funw1,fp_ppu_write_memorymap funw2,fp_ppu_write_memorymap funw3,fp_ppu_write_memorymap funw4)
{
	int i;
	int all=type&8;
	type&=7;
	
	#define _PSR(r) ppu_read_memorymap[i]=r
	#define _PSW(w) ppu_write_memorymap[i]=w
	#define _PSRW(r,w) _PSR(r); _PSW(w)
	
	for (i=0;i<0x140;i++) {
		switch (i>>2&0xf) {
			/* patterntables $0000-$0fff */
			case 0x0:			if (type==0) { _PSRW(ppu_read_pt0a,ppu_write_pt0a); }	if (all) { _PSRW(ppu_read_pt0a_override,ppu_write_pt0a_override); }	if (type==1) { if (funr1) _PSR(funr1); if (funw1) _PSW(funw1); } break;
			case 0x1:			if (type==0) { _PSRW(ppu_read_pt0b,ppu_write_pt0b); }	if (all) { _PSRW(ppu_read_pt0b_override,ppu_write_pt0b_override); }	if (type==1) { if (funr2) _PSR(funr2); if (funw2) _PSW(funw2); } break;
			case 0x2:			if (type==0) { _PSRW(ppu_read_pt0c,ppu_write_pt0c); }	if (all) { _PSRW(ppu_read_pt0c_override,ppu_write_pt0c_override); }	if (type==1) { if (funr3) _PSR(funr3); if (funw3) _PSW(funw3); } break;
			case 0x3:			if (type==0) { _PSRW(ppu_read_pt0d,ppu_write_pt0d); }	if (all) { _PSRW(ppu_read_pt0d_override,ppu_write_pt0d_override); }	if (type==1) { if (funr4) _PSR(funr4); if (funw4) _PSW(funw4); } break;
			/* patterntables $1000-$1fff */
			case 0x4:			if (type==0) { _PSRW(ppu_read_pt1a,ppu_write_pt1a); }	if (all) { _PSRW(ppu_read_pt1a_override,ppu_write_pt1a_override); }	if (type==2) { if (funr1) _PSR(funr1); if (funw1) _PSW(funw1); } break;
			case 0x5:			if (type==0) { _PSRW(ppu_read_pt1b,ppu_write_pt1b); }	if (all) { _PSRW(ppu_read_pt1b_override,ppu_write_pt1b_override); }	if (type==2) { if (funr2) _PSR(funr2); if (funw2) _PSW(funw2); } break;
			case 0x6:			if (type==0) { _PSRW(ppu_read_pt1c,ppu_write_pt1c); }	if (all) { _PSRW(ppu_read_pt1c_override,ppu_write_pt1c_override); }	if (type==2) { if (funr3) _PSR(funr3); if (funw3) _PSW(funw3); } break;
			case 0x7:			if (type==0) { _PSRW(ppu_read_pt1d,ppu_write_pt1d); }	if (all) { _PSRW(ppu_read_pt1d_override,ppu_write_pt1d_override); }	if (type==2) { if (funr4) _PSR(funr4); if (funw4) _PSW(funw4); } break;
			/* nametables */
			case 0x8: case 0xc:	if (type==0) { _PSRW(ppu_read_name0,ppu_write_name0); }	if (all) { _PSRW(ppu_read_name0_override,ppu_write_name0_override); }	if (type==3) { if (funr1) _PSR(funr1); if (funw1) _PSW(funw1); } break;
			case 0x9: case 0xd:	if (type==0) { _PSRW(ppu_read_name1,ppu_write_name1); }	if (all) { _PSRW(ppu_read_name1_override,ppu_write_name1_override); }	if (type==3) { if (funr2) _PSR(funr2); if (funw2) _PSW(funw2); } break;
			case 0xa: case 0xe:	if (type==0) { _PSRW(ppu_read_name2,ppu_write_name2); }	if (all) { _PSRW(ppu_read_name2_override,ppu_write_name2_override); }	if (type==3) { if (funr3) _PSR(funr3); if (funw3) _PSW(funw3); } break;
			case 0xb: case 0xf:	if (type==0) { _PSRW(ppu_read_name3,ppu_write_name3); }	if (all) { _PSRW(ppu_read_name3_override,ppu_write_name3_override); }	if (type==3) { if (funr4) _PSR(funr4); if (funw4) _PSW(funw4); } break;
			
			default: break;
		}
		if ((i&0x3f)==0x3f) {
			_PSW(ppu_write_palette);
			if (all) _PSW(ppu_write_palette_override);
		}
		
		if (type==4) { if ((i&0x3f)==custom) { if (funr1) _PSR(funr1); if (funw1) _PSW(funw1); } }
	}
}
