#include "global.h"
#include "file.h"
#include "cartridge.h"
#include "sample.h"

/*
games using samples, simulated to play one on first request:
(correct behaviour wouldn't work with samples that have unexpected length/
different length than the original ones)

mapper 3:

Family Trainer - Aerobics Studio:
	0	hello, let's go
	1	last
	2	good
	3	hey
	4	four
	5	three
	6	two
	7	one

mapper 18:

Moe Pro! '90 - Kandou Hen (USA version is Bases Loaded 3), Moe Pro! - Saikyou Hen (USA version is Bases Loaded 4),
Shin Moero!! Pro Yakyuu: use same samples as Moero!! Pro Yakyuu '88 (see below)

Terao no Dosukoi Oozumou:
	0	tree beating sound
	1	Hakkyoyoi
	2	Nokotta
	3	Matta Nashi
	4	Nokotta Nokotta
	5	Matta Arimasen

mapper 72/92:

Moero!! Pro Tennis (USA version is Racket Attack):
	0	"no" (lady)
	1	"no" (man)
	2	0="love" (points)
	3	15 (points)
	4	30 (points)
	5	40 (points)
	6	advantage
	7	(advantage) server
	8	(advantage) receiver
	9	all (equal points)
	10	deuce
	11	game (match over)
	12	racket hits ball
	13	crowd applauds
	14	fault
	15	let (serve ball touched the net and went over)
	16	out
	17	hit ball at close range (rare)
	18	ball hits player

Moero!! Pro Yakyuu '88 - Ketteiban (USA version is Bases Loaded II):
	0	out
	1	safe
	2	foul
	3	fair
	4	strike
	5	ball
	6	time
	7	batter out
	8	ball four
	9	home run
	10	play ball
	11	new pitcher
	12	pinch-hit
	13	groan (batter gets hit by pitcher)
	14	bat hits ball
	15	bunt
	16	caught the ball (catcher 2)
	17	caught the ball (catcher)
	18	caught the ball (pitcher/field)
	19	crowd cheers

mapper 86:

Moero!! Pro Yakyuu (Black/Red) (USA version is Bases Loaded):
	0	strike
	1	ball
	2	time (ask for time out)
	3	out
	4	safe
	5	foul
	6	"yah" (field player catching a fast ball)
	7	batter out (batter out after 3 strikes)
	8	play ball
	9	ball four
	10	home run
	11	new pitcher (choosing new pitcher in time out)
	12	ouch (batter gets hit by pitcher)
	13	aho ('fool' by supervisor)
	14	bat hits ball
	15	crowd cheers

*/

/*
sample loading errors:

00: wrong filename name: last 1 or 2 characters must be a valid (and unused before) number
01: general error reading file
02: unexpected end of file (file size too small)
03: not a RIFF WAVE file
04: chunk 1 (format) size too small
05: wrong sound format: only PCM (uncompressed) is allowed
06: wrong number of channels: should be 1 or 2 (mono/stereo)
07: wrong samplerate: should be 11025,22050,or 44100hz
08: wrong bits: should be 8 or 16
09: no "data" field
10: wrong amount of bytes in data (0, or uneven with 16 bit/stereo)

*/


#define SAMPLE_MAX 0x20

static struct {
	s16* sam[SAMPLE_MAX];
	u32 len[SAMPLE_MAX];
	
	int cur;
	int bc;
	int playing;
	int invalid;
} sample;

void sample_init(void)
{
	memset(&sample,0,sizeof(sample));
}

void sample_clean(void)
{
	int i;
	for (i=0;i<SAMPLE_MAX;i++) { MEM_CLEAN(sample.sam[i]); }
	memset(&sample,0,sizeof(sample));
}

void sample_load(void)
{
	/* can't use mmio* */
	
	char* fn;
	int i=strlen(cartridge->name);
	
	sample.invalid=TRUE;
	
	if (i==0) return;
	
	MEM_CREATE(fn,i+1);
	memcpy(fn,cartridge->name,i);
	
	while (i--) {
		file_setfile(&file->sampledir,fn,"zip","wav");
		if (file_accessible()) { i=-2; break; }
		fn[i]=0;
	}
	
	MEM_CLEAN(fn);
	
	if (i!=-2) return;
	
	/* open samples */
	for (;;) {
		
		if (file_open()) {
			int len=strlen(file->filename_inzip);
			
			if (file_get_zn()==-1) break; /* end of zipfile */
			
			/* .wav file? */
			if ((len>4)&&(stricmp((file->filename_inzip+len-4),".wav")==0)) {
				
				#define SAMPLE_WARN_N(x) LOG(LOG_MISC|LOG_WARNING,"error %02d reading",x); LOG(LOG_MISC|LOG_WARNING,"%s\n",file->filename_inzip)
				#define SAMPLE_WARN(x) SAMPLE_WARN_N(x); MEM_CLEAN(sf); continue
				#define SAMPLE_SEEK(x) o+=(x); if (o>=file->size) { SAMPLE_WARN(2); }
				
				/* get sample number */
				int sn,inv;
				int o=len-6; if (o<0) o=0;
				o+=((file->filename_inzip[o]<48)|(file->filename_inzip[o]>57)); /* 0-9 */
				
				sn=0; sscanf(file->filename_inzip+o,"%d",&sn); inv=sn==0;
				sn=1; sscanf(file->filename_inzip+o,"%d",&sn); inv&=(sn==1);
				inv|=(sn>=SAMPLE_MAX);
				if (!inv) inv|=(sample.len[sn]>0);
				
				/* found number */
				if (!inv) {
					u8* sf;
					o=0;
					
					MEM_CREATE(sf,file->size+32);
					if (file_read(sf,file->size)) {
						int oo,step;
						u16* sb;
						u32 l,sr;
						u16 chans,bits;
						o=0;
						
						/* examine header */
						memset(sf+4,'x',4); if (memcmp(sf,"RIFFxxxxWAVEfmt ",16)!=0) { SAMPLE_WARN(3); } SAMPLE_SEEK(16) /* header */
						memcpy(&l,sf+o,4); if (l<16) { SAMPLE_WARN(4); } SAMPLE_SEEK(4) /* chunk 1 size */
						if (memcmp(sf+o,"\1\0",2)!=0) { SAMPLE_WARN(5); } /* PCM (uncompressed) */
						memcpy(&chans,sf+o+2,2); if (chans!=1&&chans!=2) { SAMPLE_WARN(6); } /* mono/stereo */
						memcpy(&sr,sf+o+4,4); if (sr!=44100&&sr!=22050&&sr!=11025) { SAMPLE_WARN(7); } /* samplerate */
						memcpy(&bits,sf+o+14,2); if (bits!=8&&bits!=16) { SAMPLE_WARN(8); } /* 8/16 bit */
						SAMPLE_SEEK(l)
						
						if (memcmp(sf+o,"data",4)!=0) { SAMPLE_WARN(9); } SAMPLE_SEEK(4)
						memcpy(&l,sf+o,4); if ((l<((bits>>3)*chans))||(((bits==16)|(chans==2))&l)||((l>>1)&(bits==16)&(chans==2))) { SAMPLE_WARN(10); } SAMPLE_SEEK(4) /* data size */
						
						
						/* copy data */
						
						oo=o;
						step=44100/sr;
						
						SAMPLE_SEEK(l-1);
						
						MEM_CREATE(sb,l+2);
						memcpy(sb,sf+oo,l);
						l>>=1;
						
						/* convert 8 to 16 bits */
						if (bits==8) {
							u16* sb2;
							l<<=1; i=l;
							MEM_CREATE(sb2,(l<<1)+2);
							while (i--) sb2[i]=(sb[i>>1]>>(i<<3&8)^0x80)<<8&0xff00;
							MEM_CLEAN(sb); sb=sb2;
						}
						
						/* convert stereo to mono */
						if (chans==2) {
							l>>=1;
							for (i=0;i<l;i++) sb[i]=((s16)sb[i<<1]+(s16)sb[i<<1|1])/2;
						}
						
						MEM_CREATE(sample.sam[sn],l*step*2); sample.len[sn]=l*step;
						
						oo=0;
						
						/* convert samplerate (interpolate) */
						if (sr!=44100) {
							int sc,sd;
							
							for (i=0;i<l;i++) {
								sample.sam[sn][oo++]=sb[i];
								
								sc=step-1; sd=((s16)sb[i+1]-(s16)sb[i])/step;
								while (sc--) {
									sample.sam[sn][oo]=sample.sam[sn][oo-1]+sd;
									oo++;
								}
							}
						}
						
						else memcpy(sample.sam[sn],sb,l<<1);
						
						MEM_CLEAN(sb);
						
						
						/* adjust volume */
						i=sample.len[sn];
						while (i--) { sample.sam[sn][i]/=4; }
						
					}
					else { SAMPLE_WARN(1); }
					MEM_CLEAN(sf);
				}
				else { SAMPLE_WARN_N(0); }
			}
		}
		
		else break; /* fopen error */
		
		file_close();
		file_zn_next();
	}
	
	file_close();
	file_zn_reset();
	
	for (i=0;i<SAMPLE_MAX;i++) sample.invalid&=(sample.len[i]==0);
}

void sample_stream(signed short* stream,int len)
{
	if ((len!=0)&sample.playing&(sample.len[sample.cur]>0)&(sample.bc<sample.len[sample.cur])) {
		int i;
		for (i=0;(i<len)&(sample.bc<sample.len[sample.cur]);i++) { stream[i]+=sample.sam[sample.cur][sample.bc++]; }
		
		sample.playing=sample.bc!=sample.len[sample.cur];
	}
}

void __fastcall sample_play(int sn)
{
	sample.playing=TRUE;
	sample.cur=sn;
	sample.bc=0;
	
	if (!sample.invalid&&sample.len[sample.cur]==0) {
		LOG(LOG_MISC|LOG_WARNING,"sample %d unavailable\n",sample.cur);
	}
}

void __fastcall sample_stop(void)
{
	sample.playing=FALSE;
}
